#include "stdafx.h"
//#include "types.h"
#include "CSBuildDoc.h"
#include "Misc.h"
#include "Actuator.h"

CString TranslateObjectName(const char *name);

bool IsActuatorClassAllowed(ACTUATORCLASS ac, LOCATIONREL locr)
{
  ROOMTYPE rt;
  rt = RoomType(locr);
  switch (ac)
  {
  case AC_PRESSUREPAD:
    return    (rt == roomOPEN)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomNONBLOCKING)  //|| (rt == roomFALSEWALL)
           || (rt == roomSTONE)
           ;
  case  AC_DISABLED:
    return    (rt == roomOPEN)
           || (rt == roomSTONE)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomNONBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  case  AC_MONSTERGENERATOR:
    return    (rt == roomOPEN)
           || (rt == roomPIT)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomNONBLOCKING)  //|| (rt == roomFALSEWALL)
           ;
  case  AC_PUSHBUTTON:
    return    (rt == roomSTONE) || (rt == roomBLOCKING)
           ;
  case  AC_ANDOR:
    return    (rt == roomSTONE)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomNONBLOCKING) || (rt == roomBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  case  AC_COUNTER:
    return    (rt == roomOPEN)
           || (rt == roomSTONE)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomBLOCKING) || (rt == roomNONBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  case  AC_ENDGAME:
    return    (rt == roomOPEN)
           || (rt == roomSTONE)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomBLOCKING) || (rt == roomNONBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  case  AC_LAUNCHER:
    return    (rt == roomSTONE)
           ;
  case  AC_PORTRAIT:
    return    (rt == roomSTONE)
           || (rt == roomBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  case  AC_OBJECTHOLDER:   //Torch Holder, eg.
    return    (rt == roomSTONE)
           ;
  case  AC_SWAPOBJECT:     //Swap ZoKathRa for Firestaff, eg.
    return    (rt == roomSTONE)
           ;
  case  AC_DSA:
    return    (rt == roomOPEN)
           || (rt == roomSTONE)
           || (rt == roomPIT)
           || (rt == roomSTAIRS)
           || (rt == roomDOOR)
           || (rt == roomTELEPORTER)
           || (rt == roomBLOCKING) || (rt == roomNONBLOCKING) //|| (rt == roomFALSEWALL)
           ;
  default:
    Strange("Unknown Actuator Type");
    return false;
  };
}

bool IsValidTarget(LOCATIONREL target)
{
  ROOMTYPE rt;
  rt = RoomType(pDoc->GetCellFlag(target));
  if (rt == roomPIT) return true;;
  if (rt == roomTELEPORTER) return true;
  //if (rt == roomFALSEWALL) return true;
  if (rt == roomBLOCKING) return true;
  if (rt == roomNONBLOCKING) return true;
  if (rt == roomDOOR) return true;
  if (rt == roomOPEN)
  {
    RN obj;
    for (obj = pDoc->FindFirstObject(target);
        obj != RNeof;
        obj = pDoc->GetDBRecordLink(obj)
        )
    {
      if (obj.dbType() == dbTEXT) return true;
      if (obj.dbType() == dbACTUATOR) return true;
    };
  };
  if (rt == roomSTONE)
  {
    RN obj;
    for (obj = pDoc->FindFirstObject(target);
        obj != RNeof;
        obj = pDoc->GetDBRecordLink(obj)
        )
    {
      if (obj.dbType() == dbTEXT) return true;
      if (obj.dbType() == dbACTUATOR) return true;
    };
  };
  return false;
}

void DB3::operator = (const ACTUATOR_PressurePad& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_Portrait& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_MonsterGenerator& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_Launcher& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_PushButton& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_Counter& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_AndOr& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_ObjectHolder& )
{
  InternalError("DB3::Operator =");
}

void DB3::operator = (const ACTUATOR_Disabled& )
{
  InternalError("DB3::Operator =");
}

CString DB3::Name(ROOMTYPE rt)
{
  ACTUATORCLASS ac;
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DISABLED:
    return Cast2AC_Disabled(rt)->Name();
  case AC_PUSHBUTTON:
    return Cast2AC_PushButton(rt)->Name();
  case AC_PRESSUREPAD:
    return Cast2AC_PressurePad(rt)->Name();
  case AC_ANDOR:
    return Cast2AC_AndOr(rt)->Name();
  case AC_COUNTER:
    return Cast2AC_Counter(rt)->Name();
  case AC_MONSTERGENERATOR:
    return Cast2AC_MonsterGenerator(rt)->Name();
  case AC_LAUNCHER:
    return Cast2AC_Launcher(rt)->Name();
  case AC_OBJECTHOLDER:
    return Cast2AC_ObjectHolder(rt)->Name();
  case AC_SWAPOBJECT:
    return Cast2AC_SwapObject(rt)->Name();
  case AC_PORTRAIT:
    return Cast2AC_Portrait(rt)->Name();
  case AC_ENDGAME:
    return Cast2AC_EndGame(rt)->Name();
  case AC_DSA:
    return Cast2AC_DSA(rt)->Name();
  };
  return CString("Unknown");
}

CString DB3::ShortName(ROOMTYPE rt)
{
  ACTUATORCLASS ac;
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DISABLED:
    return Cast2AC_Disabled(rt)->Name();
  case AC_PUSHBUTTON:
    return Cast2AC_PushButton(rt)->Name();
  case AC_PRESSUREPAD:
    return Cast2AC_PressurePad(rt)->ShortName();
  case AC_ANDOR:
    return Cast2AC_AndOr(rt)->Name();
  case AC_COUNTER:
    return Cast2AC_Counter(rt)->Name();
  case AC_MONSTERGENERATOR:
    return Cast2AC_MonsterGenerator(rt)->Name();
  case AC_LAUNCHER:
    return Cast2AC_Launcher(rt)->Name();
  case AC_OBJECTHOLDER:
    return Cast2AC_ObjectHolder(rt)->Name();
  case AC_SWAPOBJECT:
    return Cast2AC_SwapObject(rt)->Name();
  case AC_PORTRAIT:
    return Cast2AC_Portrait(rt)->Name();
  case AC_ENDGAME:
    return Cast2AC_EndGame(rt)->Name();
  case AC_DSA:
    return Cast2AC_DSA(rt)->ShortName();
  };
  return CString("Unknown");
}

bool DB3::OnceOnly(ROOMTYPE rt)
{
  ACTUATORCLASS ac;
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DISABLED:
    return Cast2AC_Disabled(rt)->onceOnly();
  case AC_PUSHBUTTON:
    return Cast2AC_PushButton(rt)->onceOnly();
  case AC_PRESSUREPAD:
    return Cast2AC_PressurePad(rt)->onceOnly();
  case AC_ANDOR:
    return Cast2AC_AndOr(rt)->onceOnly();
  case AC_COUNTER:
    return Cast2AC_Counter(rt)->onceOnly();
  case AC_MONSTERGENERATOR:
    return Cast2AC_MonsterGenerator(rt)->onceOnly();
  case AC_LAUNCHER:
    return Cast2AC_Launcher(rt)->onceOnly();
  case AC_OBJECTHOLDER:
    return Cast2AC_ObjectHolder(rt)->onceOnly();
  case AC_SWAPOBJECT:
    return Cast2AC_SwapObject(rt)->onceOnly();
  case AC_PORTRAIT:
    return Cast2AC_Portrait(rt)->onceOnly();
  case AC_ENDGAME:
    return Cast2AC_EndGame(rt)->onceOnly();
  case AC_DSA:
    return Cast2AC_DSA(rt)->onceOnly();
  };
  return false;
}

CString DB3::GraphicName(LOCATIONREL locr)
{
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  rt = RoomType(locr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DISABLED:
    return Cast2AC_Disabled(rt)->GraphicName(locr);
  case AC_PUSHBUTTON:
    return Cast2AC_PushButton(rt)->GraphicName(locr);
  case AC_PRESSUREPAD:
    return Cast2AC_PressurePad(rt)->GraphicName(locr);
  case AC_ANDOR:
    return Cast2AC_AndOr(rt)->GraphicName(locr);
  case AC_COUNTER:
    return Cast2AC_Counter(rt)->GraphicName(locr);
  case AC_MONSTERGENERATOR:
    return Cast2AC_MonsterGenerator(rt)->GraphicName(locr);
  case AC_LAUNCHER:
    return Cast2AC_Launcher(rt)->GraphicName(locr);
  case AC_OBJECTHOLDER:
    return Cast2AC_ObjectHolder(rt)->GraphicName(locr);
  case AC_SWAPOBJECT:
    return Cast2AC_SwapObject(rt)->GraphicName(locr);
  case AC_PORTRAIT:
    return Cast2AC_Portrait(rt)->GraphicName(locr);
  case AC_ENDGAME:
    return Cast2AC_EndGame(rt)->GraphicName(locr);
  case AC_DSA:
    return Cast2AC_DSA(rt)->GraphicName(locr);
  };
  return CString("Unknown");
}

CString DB3::StandardGraphicName(int ordinal, LOCATIONREL locr)
{
  ROOMTYPE rt;
  if (ordinal == 0) return CString("none");
  rt = RoomType(locr);
  if (rt == roomSTONE)
  {
    WALLTYPE wt;
    if (ordinal > pDoc->NumWallDecorations(locr.l())) return CString("illegal");
    if (!pDoc->IsLegalWallDecorationIndex(locr.l(), ordinal)) return CString("illegal");
    wt = pDoc->WallDecorationType(locr.l(), ordinal-1);
    return CString(wt.Name());
  }
  else
  {
    FLOORTYPE ft;
    if (ordinal > pDoc->NumFloorDecorations(locr.l())) return CString("illegal");
    ft = pDoc->FloorDecorationType(locr.l(), ordinal);
    return CString(ft.Name());
  };
}

MODIFIEDOBJECTTYPE DB3::TriggerObject(LOCATIONREL locr)
{
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  rt = RoomType(locr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DISABLED:
    return Cast2AC_Disabled(rt)->neededObject(locr);
  case AC_PUSHBUTTON:
    return Cast2AC_PushButton(rt)->neededObject(locr);
  case AC_PRESSUREPAD:
    return Cast2AC_PressurePad(rt)->neededObject(locr);
  case AC_ANDOR:
    return Cast2AC_AndOr(rt)->neededObject(locr);
  case AC_COUNTER:
    return Cast2AC_Counter(rt)->neededObject(locr);
  case AC_MONSTERGENERATOR:
    return Cast2AC_MonsterGenerator(rt)->neededObject(locr);
  case AC_LAUNCHER:
    return Cast2AC_Launcher(rt)->neededObject(locr);
  case AC_OBJECTHOLDER:
    return Cast2AC_ObjectHolder(rt)->neededObject(locr);
  case AC_SWAPOBJECT:
    return Cast2AC_SwapObject(rt)->neededObject(locr);
  case AC_PORTRAIT:
    return Cast2AC_Portrait(rt)->neededObject(locr);
  case AC_ENDGAME:
    return Cast2AC_EndGame(rt)->neededObject(locr);
  case AC_DSA:
    return Cast2AC_DSA(rt)->neededObject(locr);
  };
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void DB3::SwapTargetLevel(int L1, int L2, LOCATIONREL *pLocr)
{
  ROOMTYPE rt;
  ACTUATORCLASS ac;
  rt = RoomType(*pLocr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_DSA:
    {
      ACTUATOR_DSA *pDSA;
      pDSA = Cast2AC_DSA(rt);
      pDSA->SwapTargetLevel(L1, L2);
    };
    break;
  default:
    break;
  };
}

void DB3::MakeBig(void)
{
  if (actuatorType() != 47) return; //DSA
  m_word8 = (ui16)(((m_word6>>12)&0x0c) | ((m_word4>>14)&0x03));
  m_word8 &= 0x3fff;
  m_word6 &= 0x3fff;
}

ACTUATOR_MonsterGenerator::ACTUATOR_MonsterGenerator(void)
{
  InternalError();
}

void ACTUATOR_MonsterGenerator::Clear(void)
{
  MONSTERTYPE mt;
  DB3::Clear();
  link(RNeof);
  localAction(true);
  actuatorType(6);
  numMonster(1);
  hitPointMultiplier(1);
  mt = "Screamer";
  monsterType(mt);
  onceOnly(false);
  audible(false);
  disableTime(50);
  floorGraphicOrdinal(0);
  invisible(false);
}

CString ACTUATOR_MonsterGenerator::Name(void)
{
  return CString("Generator");
}

CString ACTUATOR_MonsterGenerator::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(floorGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_MonsterGenerator::GetNthTarget(LOCATIONREL locr, int /*n*/)
{
  locr.l(-1); 
  return locr;
}

ACTUATOR_MonsterGenerator::~ACTUATOR_MonsterGenerator(void)
{
}

void ACTUATOR_MonsterGenerator::Dump(FILE *f)
{
  fprintf(f,"\n");
}

ACTUATOR_PressurePad::ACTUATOR_PressurePad(void)
{
  InternalError();
}

void ACTUATOR_PressurePad::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  localAction(false);
  actuatorType(1); //Operated by anything
  action(AA_SET);
  delay(0);
  value(0);
  onceOnly(false);
  normallyClosed(false);
  audible(true);
  targetX(0);
  targetY(0);
  targetPos(0);
  graphic(0);
}

ACTUATOR_PressurePad::~ACTUATOR_PressurePad(void)
{
}

CString ACTUATOR_PressurePad::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(graphic(), locr);
}

CString ACTUATOR_PressurePad::Name(void)
{
  return CString("Pressure Pad");
}

CString ACTUATOR_PressurePad::ShortName(void)
{
  return CString("Pad");
}

LOCATIONREL ACTUATOR_PressurePad::GetNthTarget(LOCATIONREL locr, int n)
{
  LOCATIONREL target;
  if ((n==0) && !localAction()) 
  {
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
  }
  else
  {
    target.l(-1);
  };
  return target;
}



CString ACTUATOR_PressurePad::CheckTarget(LOCATIONREL locr, 
                                          char isTarget[16][32][32])
{
  return StandardCheckTarget(locr, isTarget, Name());
}

CString ACTUATOR_PressurePad::CheckTrigger(LOCATIONREL /*locr*/, 
                                          char /*isTarget*/[16][32][32])
{
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_PressurePad::neededObject(LOCATIONREL /*locr*/)
{
  int acType;
  MODIFIEDOBJECTTYPE mot;
  acType = actuatorType();
  switch (acType)
  {
  case 0:
    mot = "NONE"; //Nothing
    break;
  case 1:
    mot = "ANY"; //Anything
    break;
  case 2:
    mot = "BEING"; //Party or Monster
    break;
  case 3:
    if (value() == 0) //any direction
    {
      mot = "STEP"; //party moves on or off
    }
    else
    {
      mot = "MOVE"; //Party moves or turns (facing)
    };
    break;
  case 4:
    if (mot.IsValid(value())) mot = value(); //Particular object
    else mot = "ERROR";
    break;
  case 7:
    mot = "MONSTER"; //Monster
    break;
  case 8:
    if (mot.IsValid(value())) mot = value(); //Party possesses object
    else mot = "ERROR";
    break;
  case 9:
    mot = "PARTY"; //Party steps on pad
    break;
  default:
    mot = "NONE";
  };
  return mot;
}

void ACTUATOR_PressurePad::Dump(FILE *f)
{
  CString opby, target, graphic, oo, sound, delay, nc, msg;
  int acType;
  acType = actuatorType();
  switch (acType)
  {
  case 0:
    opby = "Nothing"; //Nothing
    break;
  case 1:
    opby = "anything"; //Anything
    break;
  case 2:
    opby = "Party/Monster"; //Party or Monster
    break;
  case 3:
    if (value() == 0) //any direction
    {
      opby = "PartyMovesOnOrOff"; //party moves on or off
    }
    else
    {
      opby = "PartyMovesOrTurns"; //Party moves or turns (facing)
      switch (value()) 
      {
        case 1: opby += "North"; break;
        case 2: opby += "East"; break;
        case 3: opby += "South"; break;
        case 4: opby += "West"; break;
      };
    };
    break;
  case 4:
    opby.Format("Object=%d",value()); //Particular object
    break;
  case 7:
    opby = "Monster"; //Monster
    break;
  case 8:
    opby.Format("PartyPossesses=%d", value()); //Party possesses object
    break;
  case 9:
    opby = "PartyStepsOn"; //Party steps on pad
    break;
  default:
    opby = "????";
  };
  target.Format("(%02d,%02d)%d",targetX(), targetY(), targetPos());
  graphic.Format("%d",this->graphic());
  oo = onceOnly()?"Yes":"No";
  sound = audible()?"Yes":"No";
  delay.Format("%d",this->delay());
  nc = normallyClosed()?"Yes":"No";
  if (localAction())
  {
    msg = "LocalAction";
  }
  else
  {
    switch (action())
    {
    case AA_SET:    msg="Set";    break;
    case AA_CLEAR:  msg="Clear";  break;
    case AA_TOGGLE: msg="Toggle"; break;
    case AA_HOLD:   msg="Hold";   break;
    };
  };
  fprintf(f,"OpBy:%s Target:%s Graphic:%s OO:%s Sound:%s Delay:%s NC:%s Msg:%s\n",
            opby, target, graphic, oo, sound, delay, nc, msg);
}

ACTUATOR_Launcher::ACTUATOR_Launcher(void)
{
  InternalError();
}

void ACTUATOR_Launcher::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  localAction(true);
  actuatorType(8); //Operated by anything
  onceOnly(false);
  audible(true);
  wallGraphicOrdinal(0);
}

ACTUATOR_Launcher::~ACTUATOR_Launcher(void)
{
}

CString ACTUATOR_Launcher::Name(void)
{
  return CString("Shooter");
}

CString ACTUATOR_Launcher::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_Launcher::GetNthTarget(LOCATIONREL locr, int /*n*/)
{
  locr.l(-1); 
  return locr;
}



CString ACTUATOR_Launcher::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (!StandardCheckTrigger(locr, isTarget))
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_Launcher::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_Launcher::Dump(FILE *f)
{
  fprintf(f, "\n");
}

ACTUATOR_Disabled::ACTUATOR_Disabled(void)
{
  InternalError();
}

void ACTUATOR_Disabled::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  actuatorType(0); //Disabled
  wallGraphicOrdinal(0);
}

ACTUATOR_Disabled::~ACTUATOR_Disabled(void)
{
}

CString ACTUATOR_Disabled::Name(void)
{
  return CString("Disabled");
}

CString ACTUATOR_Disabled::GraphicName(LOCATIONREL locr)
{
  ROOMTYPE rt;
  rt = RoomType(locr);
  if (rt == roomSTONE)
      return StandardGraphicName(floorGraphicOrdinal(), locr);
  else
      return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_Disabled::GetNthTarget(LOCATIONREL locr,int /*n*/)
{
  locr.l(-1); 
  return locr;
}



CString ACTUATOR_Disabled::CheckTrigger(LOCATIONREL /*locr*/,
                                        char [16][32][32])
{
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_Disabled::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_Disabled::Dump(FILE *f)
{
  fprintf(f,"Graphic:%d\n",DB3::graphicOrdinal());
}

ACTUATOR_PushButton::ACTUATOR_PushButton(ACTUATORCLASS /*ac*/)
{
  InternalError();
}

void ACTUATOR_PushButton::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  localAction(false);
  actuatorType(1); //Operated by anything
  action(AA_SET);
  delay(0);
  value(0);
  onceOnly(false);
  normallyClosed(false);
  audible(true);
  targetX(0);
  targetY(0);
  targetPos(0);
  wallGraphicOrdinal(0);
}

CString ACTUATOR_PushButton::Name(void)
{
  return CString("PushButton");
}

CString ACTUATOR_PushButton::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

CString ACTUATOR_PushButton::CheckTarget(LOCATIONREL locr,
                                         char isTarget[16][32][32])
{
  return StandardCheckTarget(locr, isTarget, Name());
}

static bool Alcove(LOCATIONREL locr)
{
  RN actuator;
  DB3 *pActuator;
  int k;
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  WALLTYPE wt;
  rt = RoomType(pDoc->GetCellFlag(locr));
  for (actuator = pDoc->FindFirstActuator(locr);
       actuator != RNeof;
       actuator = pDoc->FindNextActuator(actuator))
  {
    if (actuator.pos() != locr.p()) continue;
    pActuator = pDoc->GetRecordAddressDB3(actuator);
    ac = pActuator->actuatorClass(rt);
    switch (ac)
    {
    case AC_PRESSUREPAD: 
      continue;
    case AC_ANDOR:
      k = pActuator->Cast2AC_AndOr(rt)->wallGraphicOrdinal();
      break;
    case AC_PUSHBUTTON:
      k =pActuator->Cast2AC_PushButton(rt)->wallGraphicOrdinal();
      break;
    case AC_COUNTER:
      k =pActuator->Cast2AC_Counter(rt)->wallGraphicOrdinal();
      break;
    case AC_DISABLED:
      k = pActuator->Cast2AC_Disabled(rt)->wallGraphicOrdinal();
      break;
    case AC_PORTRAIT:
      k = pActuator->Cast2AC_Portrait(rt)->graphic();
      break;
    case AC_OBJECTHOLDER:
      k = pActuator->Cast2AC_ObjectHolder(rt)->wallGraphicOrdinal();
      break;
    case AC_LAUNCHER:
      k = pActuator->Cast2AC_Launcher(rt)->wallGraphicOrdinal();
      break;
    case AC_DSA:
      k = 0;
      break;
    default:NotImplemented();
      k = 0;
      break;
    };
    if (k == 0) continue;
    wt = pDoc->WallDecorationType(locr.l(),k-1);
    if (wt.IsAlcove()) return true;
  };
  return false;
}



MODIFIEDOBJECTTYPE ACTUATOR_PushButton::neededObject(LOCATIONREL locr)
{
  MODIFIEDOBJECTTYPE mot;
  PBTYPE PBtype;
  switch (actuatorType())
  {
  case 1:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveAnything;
    }
    else
    {
      PBtype = PB_SwitchAnything;
    };
    break;
  case 2:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveObject;
    }
    else
    {
      PBtype = PB_SwitchHand;
    };
    break;
  case 3:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveNotObject;
    }
    else
    {
      PBtype = PB_LockKeep;;
    };
    break;
  case 4:
    PBtype = PB_LockRemove;
    break;
  case 11:
    PBtype = PB_LastLockRotate;
    break;
  case 12:
    PBtype = PB_GiftFromGod;
    break;
  case 17:
    PBtype = PB_LastLockDelete;
    break;
  default:
    Strange("Unknown Pushbutton Type");
    mot = "NONE";
    return mot;
  };
  switch (PBtype)
  {
  case PB_SwitchAnything:
  case PB_SwitchHand:
  case PB_AlcoveAnything:
  case PB_GiftFromGod:
    mot = "NONE";
    return mot;
  case PB_LockKeep:
  case PB_LastLockRotate:
  case PB_LockRemove:
  case PB_AlcoveObject:
  case PB_AlcoveNotObject:
  case PB_LastLockDelete:
    break;
  };
  mot = value();
  return mot;
}


/*
  switch (m_pPushButton->actuatorType())
  {
  case 1:
    if (IsAlcove())
    {
      m_PBtype = PB_AlcoveAnything;
    }
    else
    {
      m_PBtype = PB_SwitchAnything;
    };
    m_remove = false;
    localAction = m_pPushButton->localAction();
    if (!localAction)
    {
      m_column = m_pPushButton->targetX() + m_levelInfo.left;
      m_row = m_pPushButton->targetY() + m_levelInfo.top;
      m_positionIndex = m_pPushButton->targetPos();
      m_effect = 0; //targeted
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    }
    else
    {
      m_column = m_levelInfo.left;
      m_row = m_levelInfo.top;
      m_positionIndex = 0;
      if (m_pPushButton->localActionType() == 10)
      {
        m_effect = 2; //skills
      }
      else
      {
        m_effect = 1; //rotate
      };
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    };
    break;
  case 2:
    if (IsAlcove())
    {
      m_PBtype = PB_AlcoveObject;
    }
    else
    {
      m_PBtype = PB_SwitchHand;
    };
    m_remove = true;
    localAction = m_pPushButton->localAction();
    if (!localAction)
    {
      m_column = m_pPushButton->targetX() + m_levelInfo.left;
      m_row = m_pPushButton->targetY() + m_levelInfo.top;
      m_positionIndex = m_pPushButton->targetPos();
      m_effect = 0; //targeted
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    }
    else
    {
      m_column = m_levelInfo.left;
      m_row = m_levelInfo.top;
      m_positionIndex = 0;
      if (m_pPushButton->localActionType() == 10)
      {
        m_effect = 2; //skills
      }
      else
      {
        m_effect = 1; //rotate
      };
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    };
    break;
  case 3:
    if (IsAlcove())
    {
      m_PBtype = PB_AlcoveNotObject;
    }
    else
    {
      m_PBtype = PB_LockKeep;;
    };
    m_remove = false;
    localAction = m_pPushButton->localAction();
    if (!localAction)
    {
      m_column = m_pPushButton->targetX() + m_levelInfo.left;
      m_row = m_pPushButton->targetY() + m_levelInfo.top;
      m_positionIndex = m_pPushButton->targetPos();
      m_effect = 0; //targeted
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    }
    else
    {
      m_column = m_levelInfo.left;
      m_row = m_levelInfo.top;
      m_positionIndex = 0;
      if (m_pPushButton->localActionType() == 10)
      {
        m_effect = 2; //skills
      }
      else
      {
        m_effect = 1; //rotate
      };
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    };
    break;
  case 4:
    m_PBtype = PB_LockRemove;
    m_remove = true;
    localAction = m_pPushButton->localAction();
    if (!localAction)
    {
      m_column = m_pPushButton->targetX() + m_levelInfo.left;
      m_row = m_pPushButton->targetY() + m_levelInfo.top;
      m_positionIndex = m_pPushButton->targetPos();
      m_effect = 0; //targeted
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    }
    else
    {
      m_column = m_levelInfo.left;
      m_row = m_levelInfo.top;
      m_positionIndex = 0;
      if (m_pPushButton->localActionType() == 10)
      {
        m_effect = 2; //skills
      }
      else
      {
        m_effect = 1; //rotate
      };
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    };
    break;
  case 11:
    m_PBtype = PB_LastLockRotate;
    m_remove = true;
    m_action = m_pPushButton->action();
    m_NC = m_pPushButton->normallyClosed();
    localAction = m_pPushButton->localAction();
    if (!localAction)
    {
      m_column = m_pPushButton->targetX() + m_levelInfo.left;
      m_row = m_pPushButton->targetY() + m_levelInfo.top;
      m_positionIndex = m_pPushButton->targetPos();
      m_effect = 0; //targeted
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    }
    else
    {
      m_column = m_levelInfo.left;
      m_row = m_levelInfo.top;
      m_positionIndex = 0;
      if (m_pPushButton->localActionType() == 10)
      {
        m_effect = 2; //skills
      }
      else
      {
        m_effect = 1; //rotate
      };
      m_action = m_pPushButton->action();
      m_NC = m_pPushButton->normallyClosed();
    };
    break;
case 17:
  m_PBtype = PB_LastLockDelete;
  m_action = m_pPushButton->action();
  m_NC = m_pPushButton->normallyClosed();
  m_remove = true;
  localAction = m_pPushButton->localAction();
  if (!localAction)
  {
    m_column = m_pPushButton->targetX() + m_levelInfo.left;
    m_row = m_pPushButton->targetY() + m_levelInfo.top;
    m_positionIndex = m_pPushButton->targetPos();
    m_effect = 0; //targeted
    m_action = m_pPushButton->action();
    m_NC = m_pPushButton->normallyClosed();
  }
  else
  {
    m_column = m_levelInfo.left;
    m_row = m_levelInfo.top;
    m_positionIndex = 0;
    if (m_pPushButton->localActionType() == 10)
    {
      m_effect = 2; //skills
    }
    else
    {
      m_effect = 1; //rotate
    };
    m_action = m_pPushButton->action();
    m_NC = m_pPushButton->normallyClosed();
  };
  break;
default:
  NotImplemented();
};

*/
void ACTUATOR_PushButton::Dump(FILE * /*f*/)
{
  /*
  CString opby, target, nc, oo, sound, msg, delay, graphic;
  PBTYPE PBtype;
  switch (actuatorType())
  {
  case 1:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveAnything;
    }
    else
    {
      PBtype = PB_SwitchAnything;
    };
    break;
  case 2:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveObject;
    }
    else
    {
      PBtype = PB_SwitchHand;
    };
    break;
  case 3:
    if (Alcove(locr))
    {
      PBtype = PB_AlcoveNotObject;
    }
    else
    {
      PBtype = PB_LockKeep;;
    };
    break;
  case 4:
    PBtype = PB_LockRemove;
    break;
  case 11:
    PBtype = PB_LastLockRotate;
    break;
  case 17:
    PBtype = PB_LastLockDelete;
    break;
  default:
    Strange("Unknown Pushbutton Type");
    mot = "NONE";
    return mot;
  };
  switch (
  fprintf(f,"%s", opby);
  fprintf(f,"%s", target);
  fprintf(f,"%s", nc);
  fprintf(f,"%s", oo);
  fprintf(f,"%s", sound);
  fprintf(f,"%s", msg);
  fprintf(f,"%s", delay);
  fprintf(f,"%s", graphic);
  fprintf(f,"\n");
  */
}


ACTUATOR_ObjectHolder::ACTUATOR_ObjectHolder(void)
{
  InternalError();
}

void ACTUATOR_ObjectHolder::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  localAction(false);
  actuatorType(13); //Operated by anything
  action(AA_SET);
  delay(0);
  value(0);
  onceOnly(false);
  normallyClosed(false);
  audible(true);
  targetX(0);
  targetY(0);
  targetPos(0);
  wallGraphicOrdinal(0);
}

ACTUATOR_ObjectHolder::~ACTUATOR_ObjectHolder(void)
{
}

CString ACTUATOR_ObjectHolder::Name(void)
{
  return CString("Object Holder");
}

CString ACTUATOR_ObjectHolder::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_ObjectHolder::GetNthTarget(LOCATIONREL locr, int n)
{
  LOCATIONREL target;
  if ((n==0) && !localAction()) 
  {
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
  }
  else
  {
    target.l(-1);
  };
  return target;
}


CString ACTUATOR_ObjectHolder::CheckTarget(LOCATIONREL locr,
                                           char isTarget[16][32][32])
{
  if (localAction()) return CString("");
  return StandardCheckTarget(locr, isTarget,Name());
}

CString ACTUATOR_ObjectHolder::CheckTrigger(LOCATIONREL /*locr*/, 
                                          char /*isTarget*/[16][32][32])
{
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_ObjectHolder::neededObject(LOCATIONREL /*locr*/)
{
  return value();
}

void ACTUATOR_ObjectHolder::Dump(FILE *f)
{
  fprintf(f, "\n");
}

ACTUATOR_SwapObject::ACTUATOR_SwapObject(void)
{
  InternalError();
}

ACTUATOR_SwapObject::~ACTUATOR_SwapObject(void)
{
}

CString ACTUATOR_SwapObject::Name(void)
{
  return CString("SwapObject");
}

CString ACTUATOR_SwapObject::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_SwapObject::GetNthTarget(LOCATIONREL locr, int /*n*/)
{
  locr.l(-1); 
  return locr;
}



CString ACTUATOR_SwapObject::CheckTarget(LOCATIONREL locr,
                                         char isTarget[16][32][32])
{
  if (localAction()) return CString("");
  return StandardCheckTarget(locr, isTarget,Name());
}


void ACTUATOR_SwapObject::Clear(void)
{
  MODIFIEDOBJECTTYPE mot;
  DB3::Clear();
  type(16);
  mot = TranslateObjectName("APPLE");
  value(mot.Integer());
}

CString ACTUATOR_SwapObject::CheckTrigger(LOCATIONREL /*locr*/,
                                        char [16][32][32])
{
  return CString("");
}

void ACTUATOR_SwapObject::Dump(FILE *f)
{
  fprintf(f,"\n");
}

ACTUATOR_PushButton::~ACTUATOR_PushButton(void)
{
}

CString ACTUATOR_PushButton::CheckTrigger(LOCATIONREL /*locr*/, 
                                          char /*isTarget*/[16][32][32])
{
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_SwapObject::neededObject(LOCATIONREL /*locr*/)
{
  return value();
}

LOCATIONREL ACTUATOR_PushButton::GetNthTarget(LOCATIONREL locr, int n)
{
  LOCATIONREL target;
  if ((n==0) && !localAction()) 
  {
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
  }
  else
  {
    target.l(-1);
  };
  return target;
}



ACTUATOR_Counter::ACTUATOR_Counter()
{
  InternalError();
}

void ACTUATOR_Counter::Clear()
{
  DB3::Clear();
  link(RNeof);
  localAction(false);
  //actuatorType(6); //Originally 6 but 6 is shared with
  //                 //monster generators.  So any new counters
                     //will have a unique type.  Now you can
                     //put a counter in an open room.
  actuatorType(48);
  delay(0);
  value(0);
  onceOnly(false);
  normallyClosed(false);
  targetX(0);
  targetY(0);
  targetPos(0);
  wallGraphicOrdinal(0);
}

ACTUATOR_Counter::~ACTUATOR_Counter(void)
{
}

CString ACTUATOR_Counter::Name(void)
{
  return CString("Counter");
}

CString ACTUATOR_Counter::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_Counter::GetNthTarget(LOCATIONREL locr, int n)
{
  LOCATIONREL target;
  if ((n==0) && !localAction()) 
  {
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
  }
  else
  {
    target.l(-1);
  };
  return target;
}


CString ACTUATOR_Counter::CheckTarget(LOCATIONREL locr,
                                      char isTarget[16][32][32])
{
  return StandardCheckTarget(locr, isTarget,Name());
}

CString ACTUATOR_Counter::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (isTarget[locr.l()][locr.x()][locr.y()] == 0)
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_Counter::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_Counter::Dump(FILE *f)
{
  fprintf(f,"\n");
}

void ACTUATOR_EndGame::Clear(void)
{
  DB3::Clear();
  actuatorType(18);
}

CString ACTUATOR_EndGame::Name(void)
{
  return CString("EndGame");
}

CString ACTUATOR_EndGame::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(graphic(), locr);
}

LOCATIONREL ACTUATOR_EndGame::GetNthTarget(LOCATIONREL locr, int /*n*/)
{
  locr.l(-1); 
  return locr;
}



CString ACTUATOR_EndGame::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (isTarget[locr.l()][locr.x()][locr.y()] == 0)
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_EndGame::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_EndGame::Dump(FILE *f)
{
  fprintf(f,"<EndGame>\n");
}

CString ACTUATOR_EndGame::CheckTarget(LOCATIONREL locr,
                                    char isTarget[16][32][32])
{
  return StandardCheckTarget(locr, isTarget,Name());
}

ACTUATOR_DSA::ACTUATOR_DSA(void)
{
  InternalError();
}

ACTUATOR_DSA::~ACTUATOR_DSA(void)
{
}

CString ACTUATOR_DSA::Name(void)
{
  return CString("DSA");
}

CString ACTUATOR_DSA::ShortName(void)
{
  return CString("DSA");
}

CString ACTUATOR_DSA::GraphicName(LOCATIONREL /*locr*/)
{
  return CString("none");
}

CString ACTUATOR_DSA::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (isTarget[locr.l()][locr.x()][locr.y()] == 0)
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_DSA::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

CString ACTUATOR_DSA::CheckTarget(LOCATIONREL locr,
                                  char isTarget[16][32][32])
{
  LOCATIONREL target;
  CString result;
  bool validTarget = true;
  target.Integer(ParameterA());
  isTarget[target.l()][target.x()][target.y()] |= 1<<target.p();
  validTarget &= IsValidTarget(locr);
  target.Integer(ParameterB());
  isTarget[target.l()][target.x()][target.y()] |= 1<<target.p();
  validTarget &= IsValidTarget(locr);
  if (!validTarget)
  {
    result.Format("Warning %s has no target", Name());
  };
  return result;
}


LOCATIONREL ACTUATOR_DSA::GetNthTarget(LOCATIONREL locr, int n)
{
  i32 p = -1;
  switch (n)
  {
  case 0: 
    if (!AisInteger()) {p = ParameterA(); break;};
    if (!BisInteger()) {p = ParameterB(); break;};
    break;
  case 1: 
    if (!AisInteger() && !BisInteger()) p = ParameterB();
    break;
  }; //switch
  if (p == -1) locr.l(-1);
  else locr.Integer(p);
  return locr;
}

void ACTUATOR_DSA::SwapTargetLevel(int /*L1*/, int /*L2*/)
{
  NotImplemented();
}

void ACTUATOR_DSA::Clear(void)
{
  link(RNeof);
  actuatorType(47);
  State(0);
  DSAselector(0);
  ParameterA(0);
  ParameterB(0);
}

void ACTUATOR_DSA::Dump(FILE *f)
{
  fprintf(f,"Designer Specified Actuator\n");
}

ACTUATOR_AndOr::ACTUATOR_AndOr()
{
  InternalError();
}

void ACTUATOR_AndOr::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  localAction(false);
  actuatorType(5);
  delay(0);
  initialValue(0);
  finalValue(0);
  onceOnly(false);
  normallyClosed(false);
  action(AA_SET);
  targetX(0);
  targetY(0);
  targetPos(0);
  wallGraphicOrdinal(0);
  Edge(0);
}

ACTUATOR_AndOr::~ACTUATOR_AndOr(void)
{
}

CString ACTUATOR_AndOr::Name(void)
{
  return CString("And/Or");
}

CString ACTUATOR_AndOr::GraphicName(LOCATIONREL locr)
{
  return StandardGraphicName(wallGraphicOrdinal(), locr);
}

LOCATIONREL ACTUATOR_AndOr::GetNthTarget(LOCATIONREL locr, int n)
{
  LOCATIONREL target;
  if ((n==0) && !localAction()) 
  {
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
  }
  else
  {
    target.l(-1);
  };
  return target;
}

CString ACTUATOR_AndOr::CheckTarget(LOCATIONREL locr,
                                    char isTarget[16][32][32])
{
  return StandardCheckTarget(locr, isTarget,Name());
}

CString ACTUATOR_AndOr::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (isTarget[locr.l()][locr.x()][locr.y()] == 0)
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_AndOr::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_AndOr::Dump(FILE *f)
{
  CString graphic, delay, oo, nc, local, target, initial, final, msg;
  graphic.Format("%d",graphicOrdinal());
  delay.Format("%d",this->delay());
  oo = onceOnly()?"Yes":"No";
  nc = normallyClosed()?"Yes":"No";
  if (!localAction())
  {
    target.Format("Target:(%02d,%02d)%d",targetX(),targetY(),targetPos());
  }
  else
  {
    local = "Local:Yes";
  };
  initial.Format("%d",initialValue());
  final.Format("%d",finalValue());
  switch (action())
  {
  case AA_SET:    msg = "Set";    break;
  case AA_CLEAR:  msg = "Clear";  break;
  case AA_TOGGLE: msg = "Toggle"; break;
  case AA_HOLD:   msg = "Hold";   break;
  };
  fprintf(f,"Graphic:%s Delay:%s OO:%s %s NC:%s %s Initial:%s Final:%s Msg:%s\n",
              graphic,  delay,   oo, local, nc, target, initial, final,   msg);
}

ACTUATOR_Portrait::ACTUATOR_Portrait(void)
{
  InternalError();
}

void ACTUATOR_Portrait::Clear(void)
{
  DB3::Clear();
  link(RNeof);
  actuatorType(127); //Operated by anything
  value(0);
  graphic(0);
  BlockDisables(false);
  SetResponses(0);
  ClearResponses(0);
  ToggleResponses(0);
  DisableMessage(0);
  State(0);
  Position(0);
}

ACTUATOR_Portrait::~ACTUATOR_Portrait(void)
{
}

CString ACTUATOR_Portrait::Name(void)
{
  return CString("Portrait");
}

CString ACTUATOR_Portrait::GraphicName(LOCATIONREL locr)
{
  WALLTYPE wt;
  int ordinal;
  ordinal = graphic();
  if (ordinal == 0) return CString("none");
  wt = pDoc->WallDecorationType(locr.l(), ordinal-1);
  return CString(wt.Name());
}

LOCATIONREL ACTUATOR_Portrait::GetNthTarget(LOCATIONREL locr, int /*n*/)
{
  locr.l(-1); 
  return locr;
}

CString ACTUATOR_Portrait::CheckTarget(LOCATIONREL locr, 
                                       char isTarget[16][32][32])
{
  if (!MightBeActiveAtRuntime()) return CString("");
  if (DisableMessage() != 0)
  {
    isTarget[locr.l()][locr.x()][locr.y()] |= 1 << Position();
  };
  return CString("");
}

bool ACTUATOR_Portrait::MightBeActiveAtRuntime(void)
{
  int response;
  if ((State() & PORTRAIT_InActive) == 0) return true;
  response = (SetResponses() >> PORTRAIT_ActiveResponse) & 3;
  if (response == PORTRAIT_ResponseSet) return true;
  if (response == PORTRAIT_ResponseToggle) return true;
  response = (ClearResponses() >> PORTRAIT_ActiveResponse) & 3;
  if (response == PORTRAIT_ResponseSet) return true;
  if (response == PORTRAIT_ResponseToggle) return true;
  response = (ToggleResponses() >> PORTRAIT_ActiveResponse) & 3;
  if (response == PORTRAIT_ResponseSet) return true;
  if (response == PORTRAIT_ResponseToggle) return true;
  return false;
}

MODIFIEDOBJECTTYPE ACTUATOR_Portrait::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}

void ACTUATOR_Portrait::Dump(FILE *f)
{
  fprintf(f,"\n");
}

CString ACTUATOR_MonsterGenerator::CheckTrigger(LOCATIONREL locr,
                                        char isTarget[16][32][32])
{
  if (!StandardCheckTrigger(locr, isTarget))
    return Name() + CString(" has no trigger");
  return CString("");
}

MODIFIEDOBJECTTYPE ACTUATOR_MonsterGenerator::neededObject(LOCATIONREL /*locr*/)
{
  MODIFIEDOBJECTTYPE mot;
  mot = "NONE";
  return mot;
}


ACTUATOR_PressurePad *DB3::Cast2AC_PressurePad(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_PRESSUREPAD);
  return (ACTUATOR_PressurePad *)this;
}

ACTUATOR_DSA *DB3::Cast2AC_DSA(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_DSA);
  return (ACTUATOR_DSA *)this;
}

ACTUATOR_EndGame *DB3::Cast2AC_EndGame(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_ENDGAME);
  return (ACTUATOR_EndGame *)this;
}

ACTUATOR_Portrait *DB3::Cast2AC_Portrait(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_PORTRAIT);
  return (ACTUATOR_Portrait *)this;
}

ACTUATOR_MonsterGenerator *DB3::Cast2AC_MonsterGenerator(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_MONSTERGENERATOR);
  return (ACTUATOR_MonsterGenerator *)this;
}


ACTUATOR_PushButton *DB3::Cast2AC_PushButton(ROOMTYPE rt)
{
  ASSERT (actuatorClass(rt) == AC_PUSHBUTTON);
  return (ACTUATOR_PushButton *)this;
}

ACTUATOR_ObjectHolder *DB3::Cast2AC_ObjectHolder(ROOMTYPE rt)
{
  ASSERT (actuatorClass(rt) == AC_OBJECTHOLDER);
  return (ACTUATOR_ObjectHolder *)this;
}

ACTUATOR_SwapObject *DB3::Cast2AC_SwapObject(ROOMTYPE rt)
{
  ASSERT (actuatorClass(rt) == AC_SWAPOBJECT);
  return (ACTUATOR_SwapObject *)this;
}


ACTUATOR_AndOr *DB3::Cast2AC_AndOr(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_ANDOR);
  return (ACTUATOR_AndOr *)this;
}

ACTUATOR_Disabled *DB3::Cast2AC_Disabled(ROOMTYPE rt)
{
  ASSERT(  (actuatorClass(rt) == AC_DISABLED)
         ||(actuatorClass(rt) == AC_UNKNOWN));
  return (ACTUATOR_Disabled *)this;
}

ACTUATOR_Launcher *DB3::Cast2AC_Launcher(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_LAUNCHER);
  return (ACTUATOR_Launcher *)this;
}


ACTUATOR_Counter *DB3::Cast2AC_Counter(ROOMTYPE rt)
{
  ASSERT(actuatorClass(rt) == AC_COUNTER);
  return (ACTUATOR_Counter *)this;
}

/*
ACTUATORCLASS DB3::actuatorClass(LOCATIONREL locr)
{
  CELLFLAG cf;
  ROOMTYPE rt;
  cf = pDoc->GetCellFlag(locr);
  rt = RoomType(cf);
  if (rt == roomFALSEWALL)
  {
    bool invisibleTrickWall, nonBlockingTrickWall;
    invisibleTrickWall = (cf & 0x04) != 0;
    nonBlockingTrickWall = (cf & 0x01) != 0;
    if (invisibleTrickWall || nonBlockingTrickWall)
    {
      rt = roomOPEN;
    }
    else
    {
      rt = roomSTONE;
    };
  };
  return actuatorClass(rt);
}
*/

ACTUATORCLASS DB3::actuatorClass(ROOMTYPE rt)
{
  if (localAction())
  {//Some sort of local action
    switch (actuatorType())
    {
    case 0:
      return AC_DISABLED;
    case 1:
    case 2:
    case 3:
    case 4:
    case 11:
    case 12:
    case 17:
      return AC_PUSHBUTTON;
    case 5: 
      return AC_ANDOR;
    case 6:
      if (rt == roomSTONE) return AC_COUNTER;
      else return AC_MONSTERGENERATOR;
    case 7:
    case 8:
    case 9:
    case 10:
    case 14:
    case 15:
      return AC_LAUNCHER;
    case 13:
      return AC_OBJECTHOLDER;
    case 16:
      return AC_SWAPOBJECT;
    case 48:
      return AC_COUNTER;
    case 0x7f:
      return AC_PORTRAIT;
    default:
      return AC_UNKNOWN;
    };
  }
  else
  {
    if ((rt == roomSTONE) || (rt == roomBLOCKING) )
    {
      switch (actuatorType())
      {
      case 0:
        return AC_DISABLED;
      case 1:
      case 2:
      case 3:
      case 4:
      case 11:
      case 12:
      case 17:
        return AC_PUSHBUTTON;
      case 5:
        return AC_ANDOR;
      case 6:
        return AC_COUNTER;
      case 13:
        return AC_OBJECTHOLDER;
      case 7:
      case 8:
      case 9:
      case 10:
      case 14:
      case 15:
        return AC_LAUNCHER;
      case 16:
        return AC_SWAPOBJECT;
      case 18:
        return AC_ENDGAME;
      case 48:
        return AC_COUNTER;
      case 0x7f:
        return AC_PORTRAIT;
      case 47:
        return AC_DSA;
      default:
        actuatorType();
        return AC_UNKNOWN;
      }; // switch (Actuator type)
    }
    else if (rt == roomNONBLOCKING) //(rt == roomFALSEWALL) 
    { 
      switch(actuatorType())
      {
      case 0:
        return AC_DISABLED;
      case 1:
      case 2:
      case 3:
      case 4:
      case 7:
      case 8:
      case 9:
        return AC_PRESSUREPAD;
      case 5:
        return AC_ANDOR;
      case 6:
        return AC_MONSTERGENERATOR;
      case 47:
        return AC_DSA;
      case 48:
        return AC_COUNTER;
      case 0x7f:
        return AC_PORTRAIT;
      default:
        return AC_UNKNOWN;
      }
    }
    else
    { // open room; a presure pad
      switch(actuatorType())
      {
      case 0:
       return AC_DISABLED;
      case 1:
      case 2:
      case 3:
      case 4:
      case 7:
      case 8:
      case 9:
        return AC_PRESSUREPAD;
      case 5:
        return AC_ANDOR;
      case 6:
        return AC_MONSTERGENERATOR;
      case 18:
        return AC_ENDGAME;
      case 47:
        return AC_DSA;
      case 48:
        return AC_COUNTER;
      default:
        return AC_UNKNOWN;
      };
    };
  };
}

bool DB3::HasTarget(LOCATIONREL locr)
{
  ACTUATORCLASS ac;
  ac = actuatorClass(RoomType(locr));
  switch (ac)
  {
  case AC_PRESSUREPAD:
  case AC_PUSHBUTTON:
  case AC_OBJECTHOLDER:
  case AC_ANDOR:
  case AC_COUNTER:
  case AC_SWAPOBJECT:
  case AC_ENDGAME:
    return !localAction();
  case AC_PORTRAIT:
  case AC_DISABLED:
  case AC_LAUNCHER:
  case AC_MONSTERGENERATOR:
    return false;
  case AC_DSA:
    return true;
  default:
    NotImplemented();
    return true;
  };
}

LOCATIONREL DB3::GetNthTarget(LOCATIONREL locr, int n)
{
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  rt = RoomType(locr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_PRESSUREPAD:
    return this->Cast2AC_PressurePad(rt)->GetNthTarget(locr, n);
  case AC_PUSHBUTTON:
    return this->Cast2AC_PushButton(rt)->GetNthTarget(locr, n);
  case AC_OBJECTHOLDER:
    return this->Cast2AC_ObjectHolder(rt)->GetNthTarget(locr, n);
  case AC_ANDOR:
    return this->Cast2AC_AndOr(rt)->GetNthTarget(locr, n);
  case AC_COUNTER:
    return this->Cast2AC_Counter(rt)->GetNthTarget(locr, n);
  case AC_SWAPOBJECT:
    return this->Cast2AC_SwapObject(rt)->GetNthTarget(locr, n);
  case AC_ENDGAME:
    return this->Cast2AC_EndGame(rt)->GetNthTarget(locr, n);
  case AC_PORTRAIT:
    return this->Cast2AC_Portrait(rt)->GetNthTarget(locr, n);
  case AC_DISABLED:
    return this->Cast2AC_Disabled(rt)->GetNthTarget(locr, n);
  case AC_LAUNCHER:
    return this->Cast2AC_Launcher(rt)->GetNthTarget(locr, n);
  case AC_MONSTERGENERATOR:
    return this->Cast2AC_MonsterGenerator(rt)->GetNthTarget(locr, n);
  case AC_DSA:
    return this->Cast2AC_DSA(rt)->GetNthTarget(locr, n);
  default:
    locr.l(-1);
    return locr;
  };
}


CString DB3::StandardCheckTarget(LOCATIONREL locr, 
                                 char isTarget[16][32][32],
                                 const char *name)
{
  LOCATIONREL target;
  CString result;
  for(;;)  //To make exit easy
  {
    if (!HasTarget(locr)) return result;
    target.l(locr.l());
    target.x(targetX());
    target.y(targetY());
    target.p(targetPos());
    isTarget[target.l()][target.x()][target.y()] |= 1<<target.p();
    if (IsValidTarget(target)) break;
    result.Format("%s has no target",name);
    return result;
  };
  if (   (RoomType(target) != roomSTONE)  
      //&& (RoomType(target) != roomFALSEWALL)
      && (RoomType(target) != roomBLOCKING)
      && (RoomType(target) != roomNONBLOCKING)
      && (target.p() != 0))
  {
    result.Format("Warning %s targets non-zero position",name);
  };
  return result;
}

bool DB3::StandardCheckTrigger(LOCATIONREL locr, 
                              char isTarget[16][32][32])
{
  if (isTarget[locr.l()][locr.x()][locr.y()] & (1<<locr.p()))
                return true;
  return false;
}

int DB3::Edit(LOCATIONREL locr, RN RNactuator)  
{
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  int result;
  rt = RoomType(pDoc->GetCellFlag(locr));
  OBJECT_CHECKPOINT checkpoint(RNactuator, rt, &locr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_PRESSUREPAD:
    {
      result = Cast2AC_PressurePad(rt)->Edit(locr, RNactuator);
    };
    break;
  case AC_ANDOR:
    {
      result = Cast2AC_AndOr(rt)->Edit(locr, RNactuator);
    };
    break;
  case AC_MONSTERGENERATOR:
    {
      result = Cast2AC_MonsterGenerator(rt)->Edit(locr, RNactuator);
    };
    break;
  case AC_PUSHBUTTON:
    {
      result = Cast2AC_PushButton(rt)->Edit(locr);
    };
    break;
  case AC_COUNTER:
    {
      result = Cast2AC_Counter(rt)->Edit(locr);
    };
    break;
  case AC_OBJECTHOLDER:
    {
      result = Cast2AC_ObjectHolder(rt)->Edit(locr);
    };
    break;
  case AC_DISABLED:
    {
      result = Cast2AC_Disabled(rt)->Edit(locr);
    };
    break;
  case AC_LAUNCHER:
    {
      result = Cast2AC_Launcher(rt)->Edit(locr);
    };
    break;
  case AC_SWAPOBJECT:
    {
      result = Cast2AC_SwapObject(rt)->Edit(locr);
    };
    break;
  case AC_PORTRAIT:
    {
      result = Cast2AC_Portrait(rt)->Edit(locr, RNactuator);
    };
    break;
  case AC_DSA:
    {
      result = Cast2AC_DSA(rt)->Edit(locr, RNactuator);
    };
    break;
  default:
    {
      result = Cast2AC_EndGame(rt)->Edit(locr, RNactuator);
    };
    break;
  };
  if (result != IDOK)
  {
    checkpoint.RebuildObject(RNactuator);
  }
  else
  {
    OBJECT_CHECKPOINT newCheckpoint(RNactuator, rt, &locr);
    if (!(newCheckpoint == checkpoint)) MODIFY
  };
  return result;
}

i16 DB3::GraphicOrdinal(LOCATIONREL locr)  
{
  ACTUATORCLASS ac;
  ROOMTYPE rt;
  rt = RoomType(locr);
  ac = actuatorClass(rt);
  switch (ac)
  {
  case AC_ENDGAME:
    return (i16)Cast2AC_EndGame(rt)->graphic();
  case AC_PRESSUREPAD:
    return (i16)Cast2AC_PressurePad(rt)->graphic();
  case AC_ANDOR:
    return (i16)Cast2AC_AndOr(rt)->wallGraphicOrdinal();
  case AC_MONSTERGENERATOR:
    return (i16)Cast2AC_MonsterGenerator(rt)->floorGraphicOrdinal();
  case AC_PUSHBUTTON:
    return (i16)Cast2AC_PushButton(rt)->wallGraphicOrdinal();
  case AC_COUNTER:
    return (i16)Cast2AC_Counter(rt)->wallGraphicOrdinal();
  case AC_OBJECTHOLDER:
    return (i16)Cast2AC_ObjectHolder(rt)->wallGraphicOrdinal();
  case AC_DISABLED:
    return (i16)Cast2AC_Disabled(rt)->wallGraphicOrdinal();
  case AC_LAUNCHER:
    return (i16)Cast2AC_Launcher(rt)->wallGraphicOrdinal();
  case AC_SWAPOBJECT:
    return (i16)Cast2AC_SwapObject(rt)->wallGraphicOrdinal();
  case AC_PORTRAIT:
    return (i16)Cast2AC_Portrait(rt)->graphic();
  case AC_DSA:
    return Cast2AC_DSA(rt)->graphicOrdinal();
  default:
    NotImplemented();
    return 0;
  };
}
