// CSBuildView.h : interface of the CCSBuildView class
//
/////////////////////////////////////////////////////////////////////////////

#define MAXACLINE 500


#if !defined(AFX_CSBUILDVIEW_H__7F7EED2C_C6A0_11D6_A264_8CD155B1182C__INCLUDED_)
#define AFX_CSBUILDVIEW_H__7F7EED2C_C6A0_11D6_A264_8CD155B1182C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class CELL_CHECKPOINT;
class dlg_InsertObject;
class SAFELIST;
class EditText;

class CCSBuildView : public CView
{
  class TARGETS;
  class TARGET
  {
    friend class TARGETS;
    TARGET *m_next;
    LOCATIONREL m_locr;
  public:
    TARGET(void);
    ~TARGET(void);
    LOCATIONREL Locr(void){return m_locr;};
  };

  class TARGETS
  {
    TARGET *m_first;
  public:
    TARGETS(void);
    ~TARGETS(void);
    void AddTarget(LOCATIONREL locr);
    void Clear(void);
    TARGET *First(void){return m_first;};
    TARGET *Next(TARGET *pT){return pT->m_next;};
  };
protected: // create from serialization only
	CCSBuildView();
	DECLARE_DYNCREATE(CCSBuildView)

  int m_mouseCol, m_mouseRow, m_mousePos, m_mouseLevel;
  TARGETS m_targets;
// Attributes
public:
	CCSBuildDoc* GetDocument();
  BITMAPINFO m_bmi;
  CBitmap m_tileBitmap;
  bool m_tileBitmapValid;
  int  m_tileSize;
  bool m_bitmapDump;
  int m_currentTileTool;
  bool m_LButton;
  CBrush *m_savedBackgroundBrush;
  CRect  m_teleportLine[MAXACLINE];
  CRect  m_actuatorLine[MAXACLINE];
// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CCSBuildView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnDragLeave();
	virtual BOOL Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext = NULL);
	virtual void OnInitialUpdate();
	virtual DROPEFFECT OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CCSBuildView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif
  CString PitError(LOCATIONREL locr);
  CString StairsError(LOCATIONREL locr);
  void ClearNotesArea(int *top, int *left);
//  int ListTargets(int top, int left); //Returns vertical space used
  int ListCellObjects(int top, int left,
                      int col, int row);//Returns vertical space used.
  int ListTriggers(int top, int left, 
                   int col, int row); //Returns vertical space used.
  int ListCellType(int top, int left,
                   int col, int row);
  void EraseLine(CRect *rect);
  void DrawLine(CDC *pDC, CRect *rect, COLORREF color);
  bool GetTargets(const RN objActuator,
                  const LOCATIONREL& locr,
                  const int n,
                  LOCATIONREL& locrt);
  void HandleLines(int col, int row, ROOMTYPE rt);
  void DrawCell(CDC *pDC, CDC *memDC, 
                int level, int col, int row,
                int markCell = 0);
  void ReadBitmaps(CDC *pDC);
  bool CheckClickOnTileTools(int x, int y);
  bool CheckLClickOnDungeonMap(int x, int y, bool drag);
  bool CheckRClickOnDungeonMap(int x, int y);
  bool IsWall(LOCATIONREL locr);
  bool ActuatorWouldChange(
                    LOCATIONREL locr,
                    ROOMTYPE oldRT,
                    ROOMTYPE newRT,
                    bool printError);
  bool PlaceOpenRoom(LOCATIONREL locr, 
                     bool drag,
                     CELL_CHECKPOINT **ppCCBefore);
  bool PlaceStoneRoom(LOCATIONREL locr, bool drag);
  void PlaceStairsRoom(LOCATIONREL locr, 
                       int direction,
                       CELL_CHECKPOINT **ppCCBefore);
  void PlaceDoorRoom(LOCATIONREL locr, int type,
                     CELL_CHECKPOINT **ppCCBefore);
  void PlaceTeleporterRoom(LOCATIONREL locr,
                           CELL_CHECKPOINT **ppCCBefore);
  bool PlaceFalseWallRoom(LOCATIONREL locr, 
                          bool drag,
                          CELL_CHECKPOINT **ppCCBefore);
  CString IsPitOK(LOCATIONREL locr);
  bool IsDoorOK(LOCATIONREL locr, int type);
  bool IsStairsOK(LOCATIONREL locr, int direction);
  bool PlacePitRoom(LOCATIONREL locr, 
                    bool drag,
                    CELL_CHECKPOINT **ppCCBefore);
  void SetPitOpen(LOCATIONREL locr);
  void SetPitEasyToSee(LOCATIONREL locr);
  void SetPitReal(LOCATIONREL locr);
  bool AnyAttachedDoors(LOCATIONREL locr, bool msg);
  int  CountAttachedWalls(LOCATIONREL locr);
  int  FindStairExit(LOCATIONREL locr);
  int  Position(int x, int y);
  void EditRoomNonSTONE(int col, int row, ROOMTYPE rt);
  void EditRoomSTONE(int col, int row);
  bool StairwayWouldBeMadeInvalid(LOCATIONREL locr);
  void CheckTargetsOfObjectList(
                    RN obj,
                    LOCATIONREL locr,
                    char isTarget[64][32][32],
                    CHKERRMSG *result);
  void CheckTargetsOfActuatorAt(LOCATIONREL locr, 
                                    char t[16][32][32],
                                    CHKERRMSG *result);
  void CheckOneFilterTarget(
            EXPOOL_DATA_TYPE filterType,
            char isTarget[64][32][32],
            char *name,
            CHKERRMSG *result,
            LOCATIONREL *filterLoc);
  void CheckTargetsOfFilters(
                     char isTarget[64][32][32],
                     CHKERRMSG  *  /* result */);
  void CheckTriggerForActuatorAt(LOCATIONREL locr, 
                                    char t[16][32][32],
                                    CHKERRMSG *result);
  void CheckActuators(LOCATIONREL locr, CHKERRMSG *result);
  CString DecorationError(char *w, LOCATIONREL locr); 
  void DrawTeleportorLines(int level, int col, int row);
  void DrawActuatorLines(const LOCATIONREL& locr);
  void DrawTeleporteeLines(const LOCATIONREL& locr);
  void DrawPartyCircle(CDC *pDC);
  void DrawActuateeLines(const LOCATIONREL& locr);
  bool SearchList(RN first, 
                  dlg_InsertObject *dlg,
                  LOCATIONREL& locr) ;
  void VerifyObjectList(RN obj, 
                        CHKERRMSG *result,
                        LOCATIONREL locr,
                        int containerMask);
  void VerifyCharacterObjects(CHKERRMSG *result);
  void VerifyTimerObjects(CHKERRMSG *result);
  void VerifyAllObjects(CHKERRMSG *result);
  void CheckOnePortrait(RN objPort, 
                        CHKERRMSG *result, 
                        LOCATIONREL locr,
                        EditText *util);
  void AddPortraitError(LOCATIONREL locr,
                        CHKERRMSG *result,
                        char *msg);
  void DumpText(RN obj, FILE *f, LOCATIONREL locr);
  bool CheckDB2Indices(void) const;
  bool CheckCompressedTextIndices(void) const;
  bool ReleaseUnusedCompressedText(SAFELIST& textList);
  void InvalidateTargets(LOCATIONREL locr);
  void SwapAllReferences(int level1, int level2);


protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CCSBuildView)
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnUpdateDownLevel(CCmdUI* pCmdUI);
	afx_msg void OnUpdateUpLevel(CCmdUI* pCmdUI);
	afx_msg void OnDownLevel();
	afx_msg void OnUpLevel();
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnViewObjectCounts();
	afx_msg void OnKillFocus(CWnd* pNewWnd);
	afx_msg void OnCaptureChanged(CWnd *pWnd);
	afx_msg void OnLevelInfo();
	afx_msg void OnGlobalInfo();
	afx_msg void OnVERIFYTargetOfActuators();
	afx_msg void OnVerifyObjects();
	afx_msg void OnEditActiveTimers();
	afx_msg void OnVerifyMonsters();
	afx_msg void OnCheckDecorations();
	afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnEditUndo();
	afx_msg void OnEncrypt();
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnEditRedo();
	afx_msg void OnPitsAndStairs();
	afx_msg void OnEditOptions();
	afx_msg void OnFindObject();
	afx_msg void OnCheckAll();
	afx_msg void OnEditDSA();
	afx_msg void OnEditLevelInfo();
	afx_msg void OnWritePortraits();
	afx_msg void OnReadPortraits();
	afx_msg void OnTargetLines();
	afx_msg void OnCheckText();
	afx_msg void OnDumpText();
	afx_msg void OnLocateText();
	afx_msg void OnWriteAscii();
	afx_msg void OnVerifyTimers();
	afx_msg void OnSetFocus(CWnd* pOldWnd);
	afx_msg void OnCheckCharacterPortraits();
	afx_msg void OnUpdateTargetLines(CCmdUI* pCmdUI);
	afx_msg void OnSwapLevels();
	afx_msg void OnLocateDecoration();
	afx_msg void OnPortraitFiles();
	afx_msg void OnSkins();
	afx_msg void OnHints();
	afx_msg void OnAdvancedSearch();
	afx_msg void OnImportCharacterData();
	afx_msg void OnWriteCharacterData();
	afx_msg void OnBitmapDump();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  // debug version in CSBuildView.cpp
inline CCSBuildDoc* CCSBuildView::GetDocument()
   { return (CCSBuildDoc*)m_pDocument; }
#endif

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_CSBUILDVIEW_H__7F7EED2C_C6A0_11D6_A264_8CD155B1182C__INCLUDED_)
