#include "stdafx.h"

#include "CSBuild.h"
#include "Misc.h"
#include "CSBuildDoc.h"
#include "WarningDialog.h"

extern i32 dbEntrySizes[16];

void OBJECT_CHECKPOINT::AddToPossessions(RN obj, ROOMTYPE rt)
{
  RN objPoss= obj;
  OBJECT_CHECKPOINT *pOC, *p;
  while (objPoss != RNeof)
  {
    pOC = new OBJECT_CHECKPOINT(objPoss, rt, NULL);
    if (m_firstPossession == NULL) m_firstPossession = pOC;
    else
    {
      p = m_firstPossession;
      while (p->m_nextObject != NULL) p = p->m_nextObject;
      p->m_nextObject = pOC;
    };
    objPoss = pDoc->GetDBRecordLink(objPoss);
  };
}

void OBJECT_CHECKPOINT::Dump(void)
{
  WarningDialog dlg;
  CString line;
  CHKERRMSG result;
  line.Format("OBJECT_CHECKPOINT at %08x\n",this);
  result.NewError(line);
  line.Format(" next = %08x\n",m_nextObject);
  result += line;
  line.Format(" first = %08x\n",m_firstPossession);
  result += line;
  line.Format(" dbType = %08x\n",m_dbType);
  result += line;
  line.Format(" position = %08x\n",m_position);
  result += line;
  line.Format(" object data = \n");
  result += line;
  line.Format(" text = %08x\n",m_text);
  result += line;
  dlg.m_chkErrMsg = &result;
  dlg.DoModal();
}


void OBJECT_CHECKPOINT::ConstructDB0(RN obj, ROOMTYPE /*rt*/)
{
  DB0 *pDB;
  pDB = pDoc->GetRecordAddressDB0(obj);
  memcpy(m_objectData, pDB, sizeof (DB0));
}

void OBJECT_CHECKPOINT::ConstructDB1(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB1 *pDB;
  pDB = pDoc->GetRecordAddressDB1(obj);
  memcpy(m_objectData, pDB, sizeof (DB1));
}

void OBJECT_CHECKPOINT::ConstructDB2(RN obj, ROOMTYPE /*rt*/)
{
  DB2 *pDB;
  int textIndex, textLength;
  ui16 *pWord;
  pDB = pDoc->GetRecordAddressDB2(obj);
  memcpy(m_objectData, pDB, sizeof (DB2));
  textIndex = pDB->index();
  textLength = pDoc->ComputeTextLength(textIndex);
  pWord = pDoc->FindRawText(textIndex);
  m_text = (ui16 *)malloc((textLength+1)*2, MALLOC062);
  m_text[0] = (ui16)textLength;
  memcpy(m_text+1, pWord, textLength*2);
}

void OBJECT_CHECKPOINT::ConstructDB3(RN obj, LOCATIONREL *plocr, ROOMTYPE rt)
{
  int i;
  TIMER *pTimer;
  //ACTUATORCLASS ac;
  DB3 *pDB;
  pDB = pDoc->GetRecordAddressDB3(obj);
  memcpy(m_objectData, pDB, sizeof (DB3));
  m_acClass = pDB->actuatorClass(rt);
  switch (m_acClass)
  {
    case AC_PRESSUREPAD:
    case AC_MONSTERGENERATOR:
    case AC_PUSHBUTTON:
    case AC_ANDOR:
    case AC_COUNTER:
    case AC_ENDGAME:
    case AC_LAUNCHER:
    case AC_PORTRAIT:
    case AC_OBJECTHOLDER:   //Torch Holder, eg.
    case AC_SWAPOBJECT:     //Swap ZoKathRa for Firestaff, eg.
    case AC_DSA:
      break;
    case AC_DISABLED:
      // See if there is a timer associated with this.
      if (plocr != NULL)
      {
        for (i=0; i<pDoc->NumTimers(); i++)
        {
          pTimer = pDoc->Timer(i);
          if (pTimer->timerFunction != timerFunction65) continue;
          if ((unsigned)plocr->l() != ((pTimer->time >> 24) & 0xff)) continue;
          if (plocr->x != pTimer->uByte6) continue;
          if (plocr->y != pTimer->uByte7) continue;
          //locr.p = pTimer->uByte8;
          m_text = (ui16 *)malloc(sizeof(TIMER)+2,MALLOC063);
          m_text[0] = sizeof(TIMER);
          memcpy(m_text+1,pTimer,sizeof(TIMER));
        };
      };
      break;
    default:
      NotImplemented();
      break;
  };
}

void OBJECT_CHECKPOINT::ConstructDB4(RN obj, ROOMTYPE rt)
{
  DB4 *pDB;
  pDB = pDoc->GetRecordAddressDB4(obj);
  memcpy(m_objectData, pDB, sizeof (DB4));
  AddToPossessions(pDB->possession(), rt);
}

void OBJECT_CHECKPOINT::ConstructDB5(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB5 *pDB;
  pDB = pDoc->GetRecordAddressDB5(obj);
  memcpy(m_objectData, pDB, sizeof (DB5));
}

void OBJECT_CHECKPOINT::ConstructDB6(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB6 *pDB;
  pDB = pDoc->GetRecordAddressDB6(obj);
  memcpy(m_objectData, pDB, sizeof (DB6));
}

void OBJECT_CHECKPOINT::ConstructDB7(RN obj, ROOMTYPE rt)
{
  OBJECT_CHECKPOINT *pOC;
  RN objText;
  DB7 *pDB;
  pDB = pDoc->GetRecordAddressDB7(obj);
  memcpy(m_objectData, pDB, sizeof (DB6));
  objText = pDB->text();
  pOC = new OBJECT_CHECKPOINT(objText, rt, NULL);
  m_firstPossession = pOC;
}

void OBJECT_CHECKPOINT::ConstructDB8(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB8 *pDB;
  pDB = pDoc->GetRecordAddressDB8(obj);
  memcpy(m_objectData, pDB, sizeof (DB8));
}

void OBJECT_CHECKPOINT::ConstructDB9(RN obj, ROOMTYPE rt)
{
  DB9 *pDB;
  pDB = pDoc->GetRecordAddressDB9(obj);
  memcpy(m_objectData, pDB, sizeof(DB9));
  AddToPossessions(pDB->contents(), rt);
}

void OBJECT_CHECKPOINT::ConstructDB10(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB10 *pDB;
  pDB = pDoc->GetRecordAddressDB10(obj);
  memcpy(m_objectData, pDB, sizeof (DB10));
}

void OBJECT_CHECKPOINT::ConstructDB14(RN obj, ROOMTYPE rt)
{
  DB14 *pDB;
  pDB = pDoc->GetRecordAddressDB14(obj);
  memcpy(m_objectData, pDB, sizeof(DB14));
  if (!pDB->flyingObject().IsMagicSpell()) 
  {
    AddToPossessions(pDB->flyingObject(), rt);
  };
}

void OBJECT_CHECKPOINT::ConstructDB15(RN obj, ROOMTYPE /*rt*/)
{
  //Only needs data;
  DB15 *pDB;
  pDB = pDoc->GetRecordAddressDB15(obj);
  memcpy(m_objectData, pDB, sizeof (DB15));
}

OBJECT_CHECKPOINT::OBJECT_CHECKPOINT(RN obj, 
                                     ROOMTYPE rt,
                                     LOCATIONREL *plocr)
{
  m_nextObject = NULL;
  m_firstPossession = NULL;
  m_text = NULL;
  m_dbType = obj.dbType();
  m_position = obj.pos();
  memset(m_objectData, 0, sizeof(m_objectData));
  switch (m_dbType)
  {
  case dbDOOR:       ConstructDB0(obj, rt); break;
  case dbTELEPORTER: ConstructDB1(obj, rt); break;
  case dbTEXT:       ConstructDB2(obj, rt); break;
  case dbACTUATOR:   ConstructDB3(obj, plocr, rt); break;
  case dbMONSTER:    ConstructDB4(obj, rt); break;
  case dbWEAPON:     ConstructDB5(obj, rt); break;
  case dbCLOTHING:   ConstructDB6(obj, rt); break;
  case dbSCROLL:     ConstructDB7(obj, rt); break;
  case dbPOTION:     ConstructDB8(obj, rt); break;
  case dbCHEST:      ConstructDB9(obj, rt); break;
  case dbMISC:       ConstructDB10(obj, rt);break;
  case dbMISSILE:    ConstructDB14(obj, rt);break;
  case dbCLOUD:      ConstructDB15(obj, rt);break;
  default: InternalError("Illegal object type");
  };
}


void OBJECT_CHECKPOINT::RebuildPossessions(RN *pRN) const
{
  RN poss;
  OBJECT_CHECKPOINT *pOC;
  *pRN = RNeof;
  for (pOC=m_firstPossession; pOC!=NULL; pOC=pOC->m_nextObject)
  {
    poss = pOC->RebuildObject(); 
    pDoc->AddObjectToList(poss, pRN, RNeof);
  };
}

RN OBJECT_CHECKPOINT::RebuildDB0(RN obj) const
{
  DB0 *pDB0;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB0Entry();
  };
  pDB0 = pDoc->GetRecordAddressDB0(obj);
  memcpy((char *)pDB0+2,m_objectData+2,sizeof (DB0)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB1(RN obj) const
{
  DB1 *pDB1;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB1Entry();
  };
  pDB1 = pDoc->GetRecordAddressDB1(obj);
  memcpy((char *)pDB1+2,m_objectData+2,sizeof (DB1)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB2(RN obj) const
{
  int textIndex;
  DB2 *pDB2;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB2Entry(TEXTTYPE_WALL);
  };
  pDB2 = pDoc->GetRecordAddressDB2(obj);
  memcpy((char *)pDB2+2,m_objectData+2,sizeof (DB2)-2);
  textIndex = pDoc->AddText(m_text+1, m_text[0]);
  pDB2->index(textIndex);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB3(RN obj) const
{
  //ACTUATORCLASS ac;
  DB3 *pActuator, *pDB3;
  pActuator = (DB3 *)m_objectData;
  //ac = pActuator->actuatorClass(RoomType(*plocr));
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB3Entry(m_acClass);
  };
  pDB3 = pDoc->GetRecordAddressDB3(obj);
  memcpy((char *)pDB3+2,m_objectData+2,sizeof (DB3)-2);
  if (m_text != NULL) // A timer entry
  {
    pDoc->SetTimer((TIMER *)(m_text+1));
  };
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB4(RN obj) const
{
  DB4 *pDB4;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB4Entry();
  };
  pDB4 = pDoc->GetRecordAddressDB4(obj);
  while (pDB4->possession() != RNeof)
  {
    pDoc->DeleteObjectFromList(pDB4->possession(), pDB4->pPossession(), true);
  };
  memcpy((char *)pDB4+2,m_objectData+2,sizeof (DB4)-2);
  pDB4->possession(RNeof);
  RebuildPossessions(pDB4->pPossession());
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB5(RN obj) const
{
  DB5 *pDB5;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB5Entry();
  };
  pDB5 = pDoc->GetRecordAddressDB5(obj);
  memcpy((char *)pDB5+2,m_objectData+2,sizeof (DB5)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB6(RN obj) const
{
  DB6 *pDB6;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB6Entry();
  };
  pDB6 = pDoc->GetRecordAddressDB6(obj);
  memcpy((char *)pDB6+2,m_objectData+2,sizeof (DB6)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB7(RN obj) const
{
  RN objText;
  DB7 *pDB7;
  objText = m_firstPossession->RebuildObject();
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB7Entry();
  };
  pDB7 = pDoc->GetRecordAddressDB7(obj);
  memcpy((char *)pDB7+2,m_objectData+2,sizeof (DB7)-2);
  pDB7->text(objText);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB8(RN obj) const
{
  DB8 *pDB8;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB8Entry();
  };
  pDB8 = pDoc->GetRecordAddressDB8(obj);
  memcpy((char *)pDB8+2,m_objectData+2,sizeof (DB8)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB9(RN obj) const
{
  DB9 *pDB9;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB9Entry();
  };
  pDB9 = pDoc->GetRecordAddressDB9(obj);
  while (pDB9->contents() != RNeof)
  {
    pDoc->DeleteObjectFromList(pDB9->contents(), pDB9->pContents(), true);
  };
  memcpy((char *)pDB9+2,m_objectData+2,sizeof (DB9)-2);
  pDB9->contents(RNeof);
  RebuildPossessions(pDB9->pContents());
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB10(RN obj) const
{
  DB10 *pDB10;
  if (obj == RNnul)
  {
    obj = pDoc->FindEmptyDB10Entry();
  };
  pDB10 = pDoc->GetRecordAddressDB10(obj);
  memcpy((char *)pDB10+2,m_objectData+2,sizeof (DB10)-2);
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB14(RN obj) const
{
  //obj = pDoc->FindEmptyDB14Entry();
  NotImplemented();
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildDB15(RN obj) const
{
  //obj = pDoc->FindEmptyDB15Entry();
  NotImplemented();
  return obj;
}

RN OBJECT_CHECKPOINT::RebuildObject(RN obj)
{
  RN newobj;
  switch (m_dbType)
  {
  case dbDOOR:       newobj = RebuildDB0(obj); break;
  case dbTELEPORTER: newobj = RebuildDB1(obj); break;
  case dbTEXT:       newobj = RebuildDB2(obj); break;
  case dbACTUATOR:   newobj = RebuildDB3(obj); break;
  case dbMONSTER:    newobj = RebuildDB4(obj); break;
  case dbWEAPON:     newobj = RebuildDB5(obj); break;
  case dbCLOTHING:   newobj = RebuildDB6(obj); break;
  case dbSCROLL:     newobj = RebuildDB7(obj); break;
  case dbPOTION:     newobj = RebuildDB8(obj); break;
  case dbCHEST:      newobj = RebuildDB9(obj); break;
  case dbMISC:       newobj = RebuildDB10(obj);break;
  case dbMISSILE:    newobj = RebuildDB14(obj);break;
  case dbCLOUD:      newobj = RebuildDB15(obj);break;
  default: 
    {
      CString msg;
      msg.Format("Illegal Object Type=%d", m_dbType);
      DIE((LPCSTR)msg);
    };
  };
  return newobj;
}

OBJECT_CHECKPOINT::~OBJECT_CHECKPOINT(void)
{
  OBJECT_CHECKPOINT *pOC;
  ASSERT(m_nextObject == NULL);
  m_nextObject = NULL;
  while (m_firstPossession != NULL)
  {
    pOC = m_firstPossession;
    m_firstPossession = pOC->m_nextObject;
    pOC->m_nextObject = NULL;
    delete pOC;
  };
  if (m_text != NULL) free(m_text);
  m_text = NULL;
}

bool OBJECT_CHECKPOINT::operator == (const OBJECT_CHECKPOINT& oc) const
{
  OBJECT_CHECKPOINT *pOC1, *pOC2;
  int entrySize;
  if (m_dbType != oc.m_dbType) return false;
  if (m_position != oc.m_position) return false;
  if (m_text == NULL)
  {
    if (oc.m_text != NULL) return false;
  }
  else
  {
    if (memcmp(m_text, oc.m_text, (m_text[0]+1)*2) != 0) return false;
  };
  for (pOC1=m_firstPossession, pOC2=oc.m_firstPossession;
       pOC1 != NULL;
       pOC1=pOC1->m_nextObject, pOC2=pOC2->m_nextObject)
  {
    if (pOC2 == NULL) return false;
    if (!(*pOC1 == *pOC2)) return false;
  };
  if (pOC2 != NULL) return false;
  entrySize = dbEntrySizes[m_dbType];
  if (memcmp(m_objectData+2, oc.m_objectData+2, entrySize-2) != 0) return false;
  return true;
}

CELL_CHECKPOINT::CELL_CHECKPOINT(void)
{
  m_nextCell = NULL;
  m_prevCell = NULL;
  m_firstObject = NULL;
  m_cellFlag = 0;
  m_locr.l(0);
  m_locr.x(0);
  m_locr.y(0);
  m_locr.p(0);
  m_flags = 0;
}

CELL_CHECKPOINT::~CELL_CHECKPOINT(void)
{
  OBJECT_CHECKPOINT *pOC;
  ASSERT(m_nextCell == NULL);
  ASSERT(m_prevCell == NULL);
  m_nextCell = NULL;
  m_cellFlag = 0;
  while (m_firstObject != NULL)
  {
    pOC = m_firstObject;
    m_firstObject = pOC->m_nextObject;
    pOC->m_nextObject = NULL;
    delete pOC;
  };
}

void CELL_CHECKPOINT::SetFlag(CELL_CHECKPOINT_FLAGS f)
{
  m_flags |= f;
}

bool CELL_CHECKPOINT::Flags(CELL_CHECKPOINT_FLAGS f)
{
  return (f & m_flags) != 0;
}

void CELL_CHECKPOINT::Dump(void)
{
  WarningDialog dlg;
  CHKERRMSG result;
  CString line;
  line.Format("CELL_CHECKPOINT at %08x\n",this);
  result.NewError(line);
  line.Format(" next = %08x\n",m_nextCell);
  result.NewError(line);
  line.Format(" prev = %08x\n",m_prevCell);
  result.NewError(line);
  line.Format(" first = %08x\n", m_firstObject);
  result.NewError(line);
  line.Format(" %d(%02d,%02d)%d\n",
                  m_locr.l(), m_locr.x(), m_locr.y(), m_locr.p());
  result.NewError(line);
  line.Format(" Cellflag = %08x\n", m_cellFlag);
  result.NewError(line);
  dlg.m_chkErrMsg = &result;
  dlg.DoModal();
}

bool CELL_CHECKPOINT::operator == (const CELL_CHECKPOINT& cc) const
{
  OBJECT_CHECKPOINT *pOC1, *pOC2;
  if (m_cellFlag != cc.m_cellFlag) return false;
  if (!(m_locr == cc.m_locr)) return false;
  for (pOC1=m_firstObject, pOC2=cc.m_firstObject; 
       pOC1!=NULL; 
       pOC1=pOC1->m_nextObject, pOC2=pOC2->m_nextObject)
  {
    if (pOC2 == NULL) return false;
    if (!(*pOC1 == *pOC2)) return false;
  };
  if (pOC2 != NULL) return false;
  return true;
}

void CELL_CHECKPOINT::AddObject(OBJECT_CHECKPOINT *pOC)
{
  OBJECT_CHECKPOINT *p;
  ASSERT(pOC->m_nextObject == NULL);
  pOC->m_nextObject = NULL;
  if (m_firstObject == NULL) m_firstObject = pOC;
  else
  {
    p = m_firstObject;
    while (p->m_nextObject != NULL) p = p->m_nextObject;
    p->m_nextObject = pOC;
  };
}


CELL_CHECKPOINT_LIST::CELL_CHECKPOINT_LIST(void)
{
  m_firstCell = NULL;
  m_lastCell = NULL;
  m_currentCursor = NULL;
  m_maxLength = 1000;
  m_curLength = 0;
}

CELL_CHECKPOINT_LIST::~CELL_CHECKPOINT_LIST(void)
{
  Clear();
}

void CELL_CHECKPOINT_LIST::Clear(void)
{
  CELL_CHECKPOINT *pCC;
  while (m_firstCell != NULL)
  {
    pCC = m_firstCell;
    m_firstCell = pCC->m_nextCell;
    if (m_firstCell != NULL) m_firstCell->m_prevCell = NULL;
    pCC->m_nextCell = NULL;
    pCC->m_prevCell = NULL;
    delete pCC;
  };
  m_currentCursor = NULL;
  m_lastCell = NULL;
  m_maxLength = 1000;
  m_curLength = 0;
}

bool CELL_CHECKPOINT_LIST::Backspace()
{
  if (m_currentCursor == NULL)
  {
    if (m_lastCell == NULL) return false;
    m_currentCursor = m_lastCell;
    return true;
  };
  if (m_currentCursor->m_prevCell==NULL) return false;
  m_currentCursor = m_currentCursor->m_prevCell;
  return true;
}

CELL_CHECKPOINT *CELL_CHECKPOINT_LIST::ReadCellAtCursor(void)
{
  return m_currentCursor;
}

CELL_CHECKPOINT *CELL_CHECKPOINT_LIST::ForwardSpace(void)
{
  CELL_CHECKPOINT *result;
  result = m_currentCursor;
  if (result != NULL)
  {
    m_currentCursor = m_currentCursor->m_nextCell;
  };
  return result;
}

bool CELL_CHECKPOINT_LIST::IsRedoAvailable(void)
{
  return m_currentCursor != NULL;
}

void CELL_CHECKPOINT_LIST::WriteAtCursor(CELL_CHECKPOINT *pCC)
{
  CELL_CHECKPOINT *pCCprev, *pCCnext;
  while (m_currentCursor != NULL)
  {
    pCCprev = m_currentCursor->m_prevCell;
    pCCnext = m_currentCursor->m_nextCell;
    if (pCCprev != NULL) pCCprev->m_nextCell = pCCnext;
    else m_firstCell = pCCnext;
    if (pCCnext != NULL) pCCnext->m_prevCell = pCCprev;
    else m_lastCell = pCCprev;
    m_currentCursor->m_nextCell = NULL;
    m_currentCursor->m_prevCell = NULL;
    delete m_currentCursor;
    m_curLength--;
    m_currentCursor = pCCnext;
  };
  m_curLength++;
  if (m_firstCell == NULL)
  {
    ASSERT(m_lastCell == NULL);
    m_firstCell = m_lastCell = pCC;
    pCC->m_prevCell = pCC->m_nextCell = NULL;
  }
  else
  {
    pCC->m_nextCell = NULL;
    pCC->m_prevCell = m_lastCell;
    m_lastCell->m_nextCell = pCC;
    m_lastCell = pCC;
  };
  if (  (m_curLength > m_maxLength)
      &&(m_maxLength > 2) )
  {
    CELL_CHECKPOINT *temp;
    temp = m_firstCell;
    m_firstCell = temp->m_nextCell;
    m_firstCell->m_prevCell = NULL;
    temp->m_nextCell = NULL;
    temp->m_prevCell = NULL;
    delete temp;
    m_curLength--;
  };
}

CELL_CHECKPOINT *CELL_CHECKPOINT_LIST::Replace(CELL_CHECKPOINT *pCC)
{
  CELL_CHECKPOINT *result;
  result = m_currentCursor;
  if (result == NULL) DIE("NULL CHECKPOINT REPLACEMENT");
  pCC->m_prevCell = result->m_prevCell;
  pCC->m_nextCell = result->m_nextCell;
  if (result->m_prevCell != NULL) result->m_prevCell->m_nextCell = pCC;
  else m_firstCell = pCC;
  if (result->m_nextCell != NULL) result->m_nextCell->m_prevCell = pCC;
  else m_lastCell = pCC;
  m_currentCursor = pCC;
  result->m_nextCell = result->m_prevCell = NULL;
  return result;
}

void CCSBuildDoc::WriteCellUndo(CELL_CHECKPOINT *pCC)
{
  m_cellUndoList.WriteAtCursor(pCC);
}

OBJECT_CHECKPOINT *CCSBuildDoc::CheckpointObject(RN obj, LOCATIONREL *plocr)
{
  OBJECT_CHECKPOINT *pOC;
  pOC = new OBJECT_CHECKPOINT(obj, RoomType(*plocr), plocr);
  return pOC;
}

CELL_CHECKPOINT *CCSBuildDoc::CheckpointCell(LOCATIONREL locr)
{
  RN obj;
  CELL_CHECKPOINT *pCC;
  OBJECT_CHECKPOINT *pOC;
  pCC = new CELL_CHECKPOINT;
  pCC->CellFlag(GetCellFlag(locr));
  pCC->Location(locr);
  // Each object in this cell must be copied to a
  // OBJECT_CHECKPOINT entry and added to the end of
  // pCC's m_firstObject list.
  for (obj = FindFirstObject(locr);
       obj != RNeof;
       obj = GetDBRecordLink(obj))
  {
    pOC = CheckpointObject(obj, &locr);
    pCC->AddObject(pOC);
  };
  return pCC;
}

void CCSBuildDoc::ClearCellCheckpoints(void)
{
  m_cellUndoList.Clear();
}

CELL_CHECKPOINT *CCSBuildDoc::ReplaceCellUndo(CELL_CHECKPOINT *pCCnew)
{
  return m_cellUndoList.Replace(pCCnew);
}

bool CCSBuildDoc::BackspaceCellUndo(void)
{
  return m_cellUndoList.Backspace();
}

CELL_CHECKPOINT *CCSBuildDoc::ReadCellUndoAtCursor(void)
{
  return m_cellUndoList.ReadCellAtCursor();
}

CELL_CHECKPOINT *CCSBuildDoc::ForwardSpaceCellUndo(void)
{
  return m_cellUndoList.ForwardSpace();
}

bool CCSBuildDoc::IsCellUndoAvailable(void)
{
  if (m_cellUndoList.Backspace() == NULL) return false;
  m_cellUndoList.ForwardSpace();
  return true;
}

bool CCSBuildDoc::IsCellRedoAvailable(void)
{
  return m_cellUndoList.IsRedoAvailable();
}

void CCSBuildDoc::RemoveAndDeleteAllObjectsInRoom(LOCATIONREL locr)
{
  while (FindFirstObject(locr) != RNeof)
  {
    RemoveObjectFromRoom(locr, FindFirstObject(locr), true, true);
  };
}

void CCSBuildDoc::ReplaceOneObjectInRoom(LOCATIONREL locr, OBJECT_CHECKPOINT *pOC)
{
  RN obj;
  obj = pOC->RebuildObject();
  locr.p(pOC->Position());
  AddObjectToRoom(obj, locr);
}

void CCSBuildDoc::ReplaceAllObjectsInRoom(CELL_CHECKPOINT *pCC)
{
  //This is a bit of a tough job.
  //All the objects and their attachments...things like
  // movement timers....must be put back into the game.
  //Oh, well.  Let us get started.  We can do them
  // one at a time.
  OBJECT_CHECKPOINT *pOC;
  for (pOC=pCC->FirstObject(); pOC!=NULL; pOC=pCC->NextObject(pOC))
  {
    ReplaceOneObjectInRoom(pCC->Location(), pOC);  // Makes it look easy!
  };
}

void CCSBuildDoc::UndoCell(CELL_CHECKPOINT *pCC)
{
  CELLFLAG newCellFlag;
  RemoveAndDeleteAllObjectsInRoom(pCC->Location());
  ReplaceAllObjectsInRoom(pCC);
  if ((pCC->CellFlag()&0x10) != (GetCellFlag(pCC->Location()) &0x10))
  {
    InternalError("Undo Failed..Bit 4 changed");
  };
  newCellFlag = (CELLFLAG)(GetCellFlag(pCC->Location()) & 0x10);
  newCellFlag |= pCC->CellFlag() & 0xef;
  SetCellFlag(pCC->Location(), newCellFlag);
  SetModifiedFlag();
}
