#include "stdafx.h"
#include "CSBuildDoc.h"
#include "Clipboard.h"
#include "Misc.h"

// Version 01
// Represents a single binary object in hexadecimal.
// Line 1:
//   "CSBuild01"
// Line 2:
//   4-digit DataType from enum CLIPBOARDDATATYPE
//     example: "0003"
// Line 3:
//   Data title
//     example: Actuator
// Line 4:
//   Binary data represented as two hex digits per byte.
//   Counted string; 4 decimal digits followed by string"
// example:  "00140123fffe3333ca"


CLIPBOARD::CLIPBOARD(void)
{
}

CLIPBOARD::~CLIPBOARD(void)
{
}

void CLIPBOARD::PutBinaryData(
                      void *data,
                      int numbytes,
                      CString dataTitle,
                      CLIPBOARDDATATYPE cdt)
{
  char textByte[3];
  int i;
  m_text.Format("CSBuild01\n%04d\n%s\n%04d",
                    cdt,dataTitle,numbytes);
  for (i=0; i<numbytes; i++)
  {
    sprintf(textByte,"%02x", ((unsigned char *)data)[i]);
    m_text += textByte;
  };
}

bool CLIPBOARD::SendToClipboard(void)
{
  if ( !OpenClipboard(NULL) )  
  {    
    AfxMessageBox( "Cannot open the Clipboard" );
    return false;  
  };  // Remove the current Clipboard contents
  if( !EmptyClipboard() )  
  {    
    AfxMessageBox( "Cannot empty the Clipboard" );
    return false;  
  }  // ...  // Get the currently selected data  // ...
  // For the appropriate data formats...
  if ( ::SetClipboardData(CF_TEXT, (void *)(LPCSTR)m_text ) == NULL )  
  {
    AfxMessageBox( "Unable to set Clipboard data" );    CloseClipboard();
    return false;  
  }  // ...  CloseClipboard();}
  return true;
}

bool CLIPBOARD::FetchFromClipboard(void)
{
  //PLABELBOX pbox;     
  HGLOBAL   hglb; 
  LPTSTR    lptstr;     
  //PLABELBOX pboxCopy;     
  //int cx, cy;     
  //HWND hwnd;  
  //pbox = hwndSelected == NULL ? NULL : (PLABELBOX) GetWindowLong(hwndSelected, 0);  
  // If the application is in edit mode,     
  // get the clipboard text.  
  //if (pbox != NULL && pbox->fEdit)     
  //{ 
  if (!IsClipboardFormatAvailable(CF_TEXT)) return false; 
  if (!OpenClipboard(NULL)) return false;  
  hglb = GetClipboardData(CF_TEXT);
  if (hglb == NULL)         
  { 
    CloseClipboard();
    return false;
  };
  lptstr = (LPTSTR)GlobalLock(hglb);             
  if (lptstr == NULL) 
  {
    GlobalUnlock(hglb);
    CloseClipboard();
    return false;
  };
  if (strlen(lptstr) < 9)
  {
    GlobalUnlock(hglb);
    CloseClipboard();
    return false;
  };    
  if (memcmp("CSBuild01",lptstr,9) != 0)
  {
    GlobalUnlock(hglb);
    CloseClipboard();
    return false;
  };
  m_text = lptstr;
  GlobalUnlock(hglb);
  CloseClipboard();
  return true;
}

int CLIPBOARD::findLine(int linenum)
{
  int i, len;
  len = m_text.GetLength();
  for (i=0; i<len; i++)
  {
    if (linenum == 0) return i;
    if (m_text[i] == '\n') linenum--;
  };
  return -1;
}

CLIPBOARDDATATYPE CLIPBOARD::GetDataType(void)
{
  int i, col, len, result;
  char c;
  if (m_text == "") return CDT_Unknown;
  col = findLine(1);
  if (col < 0) return CDT_Unknown;
  len = m_text.GetLength();
  if (col + 4 > len) return CDT_Unknown;
  result = 0;
  for (i=0; i<4; i++)
  {
    c = m_text[col+i];
    if ( (c<'0')||(c>'9') )return CDT_Unknown;
    result = 10*result + c - '0';
  };
  if (result >= CDT_TooBig) return CDT_Unknown;
  return (CLIPBOARDDATATYPE)result;
}

bool CLIPBOARD::GetBinaryData(void *data, int numBytes)
{
  int i, col, len, size;
  unsigned char c;
  unsigned int n;
  col = findLine(3);
  if (col < 0) return false;
  len = m_text.GetLength();
  if (col+4 > len) return false;
  size = 0;
  for (i=0; i<4; i++)
  {
    c = m_text[col+i];
    if ( (c<'0')||(c>'9') ) return false;
    size = 10*size + c - '0';
  };
  col += 4;
  if (size != numBytes) return false;
  if (col + size > len) return false;
  for (i=0; i<numBytes; col+=2, i++)
  {
    c = m_text[col];
    if ( (c>='0') && (c<='9')) n=c-'0';
    else if ( (c>='a') && (c<='f') ) n=c-'a'+10;
    else if ( (c>='A') && (c<='F') ) n=c-'A'+10;
    else return false;
    n <<= 4;
    c = m_text[col+1];
    if ( (c>='0') && (c<='9')) n|=c-'0';
    else if ( (c>='a') && (c<='f') ) n|=c-'a'+10;
    else if ( (c>='A') && (c<='F') ) n|=c-'A'+10;
    else return false;
    ((unsigned char *)data)[i] = (unsigned char)n;
  };
  return true;
}

bool RN::CopyToClipboard(CLIPBOARD *clipboard)
{
  DBTYPE dbtype;
  dbtype = dbType();
  switch (dbtype)
  {
  case dbACTUATOR:
    clipboard->PutBinaryData(
                 pDoc->GetRecordAddressDB3(*this),
                 sizeof (DB3),
                 CString ("Actuator"),
                 CDT_Actuator);
    break;
  default: 
    pDoc->Sorry("Copy of this type of object not implemented");
    return false;
  };
  return true;
}