// EditDSA.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "EditDSA.h"
#include "Actuator.h"
#include "SelectDSA.h"
#include "Misc.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static bool aIsTarget = true;
static bool bIsTarget = true;

/////////////////////////////////////////////////////////////////////////////
// EditDSA dialog


EditDSA::EditDSA(LOCATIONREL locr, 
                 RN RNdsa,
                 CWnd* pParent /*=NULL*/)
	: CDialog(EditDSA::IDD, pParent)
{
  m_locr = locr;
  m_RNdsa = RNdsa;
	//{{AFX_DATA_INIT(EditDSA)
	m_aIsTarget = 0;
	m_bIsTarget = 0;
	m_columnA = 0;
	m_columnB = 0;
	m_levelA = 0;
	m_levelB = 0;
	m_positionA = 0;
	m_positionB = 0;
	m_rowA = 0;
	m_rowB = 0;
	m_valueA = 0;
	m_valueB = 0;
	m_state = 0;
	m_dsaDescription = _T("");
  m_dsaLevelIndex = 0;
	//}}AFX_DATA_INIT
}

void EditDSA::Show(int id, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(id);
  if (pWnd != NULL) pWnd->ShowWindow(show);
}

bool EditDSA::IsLegalTarget(int l, int x, int y, int p)
{
  LOCATIONREL locr;
  LOCATIONABS loca;
  if (l < 0) return false;
  if (p < 0) return false;
  if (pDoc->NumLevel() <= l) return false;
  loca.l = l;
  loca.x = x;
  loca.y = y;
  loca.p = p;
  locr = loca;
  if (pDoc->LastColumn(l) < locr.x()) return false;
  if (pDoc->LastRow(l) < locr.y()) return false;
  if (p > 3) return false;
  if (x < pDoc->OffsetX(l)) return false;
  if (y < pDoc->OffsetY(l)) return false;
  return true;
}


void EditDSA::ExtractDataFromRecord(void)
{
  LOCATIONREL locr;
  LOCATIONABS loca;
  m_state = m_pDSA->State();
  m_valueA = m_pDSA->ParameterA();
  m_aIsTarget = m_pDSA->AisInteger()?1:0;
  locr.Integer(m_valueA);
  loca = locr;
  m_levelA = loca.l;
  m_columnA = loca.x;
  m_rowA = loca.y;
  m_positionA = loca.p;
  m_valueB = m_pDSA->ParameterB();
  m_bIsTarget = m_pDSA->BisInteger()?1:0;
  locr.Integer(m_valueB);
  loca = locr;
  m_levelB = loca.l;
  m_columnB = loca.x;
  m_rowB = loca.y;
  m_positionB = loca.p;
  m_dsaLevelIndex = m_pDSA->DSAselector();
}

bool EditDSA::EncodeDataIntoRecord(void)
{ //true if something changes.
  LOCATIONREL locr;
  LOCATIONABS loca;
  bool change = false;
  if (m_pDSA->State() != m_state) change = true;
  m_pDSA->State((ui8)m_state);
  if (aIsTarget)
  {
    loca.l = m_levelA;
    loca.x = m_columnA;
    loca.y = m_rowA;
    loca.p = m_positionA;
    locr = loca;
    m_valueA = locr.Integer();
  };
  if (bIsTarget)
  {
    loca.l = m_levelB;
    loca.x = m_columnB;
    loca.y = m_rowB;
    loca.p = m_positionB;
    locr = loca;
    m_valueB = locr.Integer();
  };
  if (m_pDSA->ParameterA() != m_valueA) change = true;
  m_pDSA->ParameterA(m_valueA);
  if (m_pDSA->AisInteger() == (m_aIsTarget!=0)) change = true;
  m_pDSA->AisInteger(m_aIsTarget!=0);
  if (m_pDSA->ParameterB() != m_valueB) change = true;
  m_pDSA->ParameterB(m_valueB);
  if (m_pDSA->BisInteger() == (m_bIsTarget!=0)) change = true;
  m_pDSA->BisInteger(m_bIsTarget!=0);
  if (m_pDSA->DSAselector() != m_dsaLevelIndex) change = true;
  m_pDSA->DSAselector((ui8)m_dsaLevelIndex);
  return change;
}

void EditDSA::DoDataExchange(CDataExchange* pDX)
{
  if (!pDX->m_bSaveAndValidate)
  {
    DSA *pDSA = NULL;
    ui16 absIndex;
    absIndex = pDoc->getDSA(m_locr.l(), m_dsaLevelIndex);
    m_dsaDescription = "";
    if (absIndex != 0xffff)
    {
      pDSA = pDoc->getDSA(absIndex);
    };
    if (pDSA != NULL)
    {
      m_dsaDescription = pDSA->Description();
    };
  };
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditDSA)
	DDX_Control(pDX, IDC_ValueLabelB, m_valueLabelB);
	DDX_Control(pDX, IDC_ValueLabelA, m_valueLabelA);
	DDX_Control(pDX, IDC_RowLabelB, m_rowLabelB);
	DDX_Control(pDX, IDC_RowLabelA, m_rowLabelA);
	DDX_Control(pDX, IDC_PositionLabelB, m_positionLabelB);
	DDX_Control(pDX, IDC_PositionLabelA, m_positionLabelA);
	DDX_Control(pDX, IDC_LevelLabelB, m_levelLabelB);
	DDX_Control(pDX, IDC_LevelLabelA, m_levelLabelA);
	DDX_Control(pDX, IDC_ColumnLabelB, m_columnLabelB);
	DDX_Control(pDX, IDC_ColumnLabelA, m_columnLabelA);
	DDX_Radio(pDX, IDC_AisTarget, m_aIsTarget);
	DDX_Radio(pDX, IDC_BisTarget, m_bIsTarget);
	DDX_Text(pDX, IDC_ColumnA, m_columnA);
	DDX_Text(pDX, IDC_ColumnB, m_columnB);
	DDX_Text(pDX, IDC_LevelA, m_levelA);
	DDX_Text(pDX, IDC_LevelB, m_levelB);
	DDX_CBIndex(pDX, IDC_PositionA, m_positionA);
	DDX_CBIndex(pDX, IDC_PositionB, m_positionB);
	DDX_Text(pDX, IDC_RowA, m_rowA);
	DDX_Text(pDX, IDC_RowB, m_rowB);
	DDX_Text(pDX, IDC_ValueA, m_valueA);
	DDV_MinMaxUInt(pDX, m_valueA, 0, 262143);
	DDX_Text(pDX, IDC_ValueB, m_valueB);
	DDV_MinMaxUInt(pDX, m_valueB, 0, 262143);
	DDX_Text(pDX, IDC_State, m_state);
	DDV_MinMaxUInt(pDX, m_state, 0, 15);
	DDX_Text(pDX, IDC_DSADescription, m_dsaDescription);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    Show(IDC_ValueLabelA   , m_aIsTarget != 0);
    Show(IDC_ValueA        , m_aIsTarget != 0);
    Show(IDC_ValueLabelB   , m_bIsTarget != 0);
    Show(IDC_ValueB        , m_bIsTarget != 0);
    Show(IDC_LevelLabelA   , m_aIsTarget == 0);
    Show(IDC_LevelA        , m_aIsTarget == 0);
    Show(IDC_ColumnLabelA  , m_aIsTarget == 0);
    Show(IDC_ColumnA       , m_aIsTarget == 0);
    Show(IDC_RowLabelA     , m_aIsTarget == 0);
    Show(IDC_RowA          , m_aIsTarget == 0);
    Show(IDC_PositionLabelA, m_aIsTarget == 0);
    Show(IDC_PositionA     , m_aIsTarget == 0);
    Show(IDC_LevelLabelB   , m_bIsTarget == 0);
    Show(IDC_LevelB        , m_bIsTarget == 0);
    Show(IDC_ColumnLabelB  , m_bIsTarget == 0);
    Show(IDC_ColumnB       , m_bIsTarget == 0);
    Show(IDC_RowLabelB     , m_bIsTarget == 0);
    Show(IDC_RowB          , m_bIsTarget == 0);
    Show(IDC_PositionLabelB, m_bIsTarget == 0);
    Show(IDC_PositionB     , m_bIsTarget == 0);
  };
}


BEGIN_MESSAGE_MAP(EditDSA, CDialog)
	//{{AFX_MSG_MAP(EditDSA)
	ON_BN_CLICKED(IDC_AisInteger, OnAisInteger)
	ON_BN_CLICKED(IDC_AisTarget, OnAisTarget)
	ON_BN_CLICKED(IDC_BisInteger, OnBisInteger)
	ON_BN_CLICKED(IDC_BisTarget, OnBisTarget)
	ON_BN_CLICKED(IDC_SelectNewDSAType, OnSelectNewDSAType)
	ON_BN_CLICKED(IDC_BrowseA, OnBrowseA)
	ON_BN_CLICKED(IDC_BrowseB, OnBrowseB)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditDSA message handlers

int ACTUATOR_DSA::Edit(LOCATIONREL locr, RN RNDSA)
{
  EditDSA dlg(locr, RNDSA);
  dlg.m_pDSA = this;
  return dlg.DoModal();
}



void EditDSA::OnAisInteger() 
{
	// TODO: Add your control notification handler code here
  if (m_aIsTarget == 1) return; //Already set
	UpdateData(true);
  LOCATIONREL locr;
  LOCATIONABS loca;
  loca.l = m_levelA;
  loca.x = m_columnA;
  loca.y = m_rowA;
  loca.p = m_positionA;
  locr = loca;
  m_valueA = locr.Integer();
  UpdateData(false);
	
}

void EditDSA::OnAisTarget() 
{
	// TODO: Add your control notification handler code here
  if (m_aIsTarget == 0) return; // Already set
	UpdateData(true);
  LOCATIONREL locr;
  LOCATIONABS loca;
  locr.Integer(m_valueA);
  loca = locr;
  m_levelA    = loca.l;
  m_columnA   = loca.x;
  m_rowA      = loca.y;
  m_positionA = loca.p;
  UpdateData(false);
}

void EditDSA::OnBisInteger() 
{
	// TODO: Add your control notification handler code here
  if (m_bIsTarget == 1) return; //Already set
	UpdateData(true);
  LOCATIONREL locr;
  LOCATIONABS loca;
  loca.l = m_levelB;
  loca.x = m_columnB;
  loca.y = m_rowB;
  loca.p = m_positionB;
  locr = loca;
  m_valueB = locr.Integer();
  UpdateData(false);
	
}

void EditDSA::OnBisTarget() 
{
	// TODO: Add your control notification handler code here
  if (m_bIsTarget == 0) return; //Already set
	UpdateData(true);
  LOCATIONREL locr;
  LOCATIONABS loca;
  locr.Integer(m_valueB);
  loca = locr;
  m_levelB    = loca.l;
  m_columnB   = loca.x;
  m_rowB      = loca.y;
  m_positionB = loca.p;
  UpdateData(false);
	
}

void EditDSA::OnSelectNewDSAType() 
{
	// TODO: Add your control notification handler code here
	SelectDSA dlg(-1);
  UpdateData(true);
  dlg.m_level = m_locr.l();
  if (dlg.DoModal() == IDOK)
  {
    m_dsaLevelIndex = dlg.m_DSAindex;
  };
  UpdateData(false);
}

BOOL EditDSA::OnInitDialog() 
{
  CString text;
  LOCATIONABS loca;
  ExtractDataFromRecord();
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
  loca = m_locr;
  text.Format("Edit DSA Actuator at %d(%02d,%02d)%d",
               loca.l, loca.x, loca.y, loca.p);
	SetWindowText(text);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditDSA::OnOK() 
{
	// TODO: Add extra validation here
	
  UpdateData(true);
	aIsTarget = m_aIsTarget==0;
	bIsTarget = m_bIsTarget==0;
  if (bIsTarget)
  {
    if (!IsLegalTarget(m_levelB, m_columnB, m_rowB, m_positionB))
    {
      Error("Illegal Target B");
      return;
    };
  };
  if (aIsTarget)
  {
    if (!IsLegalTarget(m_levelA, m_columnA, m_rowA, m_positionA))
    {
      Error("Illegal Target A");
      return;
    };
  };
	CDialog::OnOK();
  if (EncodeDataIntoRecord())
  {
    setmodifiedflag();
  };
}

void EditDSA::OnCancel() 
{
	// TODO: Add extra cleanup here
	aIsTarget = m_aIsTarget == 0;
	bIsTarget = m_bIsTarget == 0;
	CDialog::OnCancel();
}

void EditDSA::OnBrowseA() 
{
	// TODO: Add your control notification handler code here
	
  LOCATIONABS loca, locaDSA;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_levelA;
  loca.x = m_columnA;
  loca.y = m_rowA;
  locrTarget = loca;
  locaDSA = m_locr;
	EditTarget dlg(locrTarget,"Designer Specified Actuator",locaDSA,true);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_levelA = loca.l;
    m_columnA = loca.x;
    m_rowA = loca.y;
    UpdateData(false);
  };
}

void EditDSA::OnBrowseB() 
{
	// TODO: Add your control notification handler code here
	
  LOCATIONABS loca, locaDSA;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_levelB;
  loca.x = m_columnB;
  loca.y = m_rowB;
  locrTarget = loca;
  locaDSA = m_locr;
	EditTarget dlg(locrTarget,"Designer Specified Actuator",locaDSA,true);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_levelB = loca.l;
    m_columnB = loca.x;
    m_rowB = loca.y;
    UpdateData(false);
  };
}
