// EditLauncher.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "Actuator.h"
#include "EditLauncher.h"
#include "Misc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// EditLauncher dialog


EditLauncher::EditLauncher(LOCATIONREL locr,
                           CWnd* pParent /*=NULL*/)
	: CDialog(EditLauncher::IDD, pParent)
{
  m_locr = locr;
  m_levelInfo = m_locr;
	//{{AFX_DATA_INIT(EditLauncher)
	m_availableObject = -1;
	m_energy = 0;
	m_one = -1;
	//}}AFX_DATA_INIT
}


void EditLauncher::DoDataExchange(CDataExchange* pDX)
{
  WALLTYPE wall;
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditLauncher)
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Control(pDX, IDC_ObjectList, m_objectList);
	DDX_Control(pDX, IDC_MagicList, m_magicList);
	DDX_Radio(pDX, IDC_AvailableObject, m_availableObject);
	DDX_Text(pDX, IDC_Energy, m_energy);
	DDV_MinMaxUInt(pDX, m_energy, 0, 255);
	DDX_Radio(pDX, IDC_One, m_one);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    int i;
    FillObjectList();
    FillMagicList();
    m_magicList.SetCurSel(0);
    m_objectList.SetCurSel(0);
    for (i=0; i<m_magicList.GetCount(); i++)
    {
      if (m_magicList.GetItemData(i) == (unsigned)m_magicType)
      {
        m_magicList.SetCurSel(i);
      };
    };
    for (i=0; i<m_objectList.GetCount(); i++)
    {
      if (m_objectList.GetItemData(i) == (unsigned)m_objectType)
      {
        m_objectList.SetCurSel(i);
      };
    };
    Enable(IDC_ObjectList, m_availableObject==1);
    Enable(IDC_MagicList,  m_availableObject==2);
    m_graphic.ResetContent();
    m_graphic.AddString("none");
    for (i=0; i<pDoc->NumWallDecorations(m_locr.l()); i++)   
    {
      if (pDoc->IsLegalWallDecorationIndex(m_locr.l(), i+1))
      {
        wall = pDoc->WallDecorationType(m_locr.l(),i);
        m_graphic.AddString(wall.Name());
      }
      else
      {
        m_graphic.AddString("illegal");
      };
    };
    m_graphic.SetCurSel(m_graphicListIndex);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_magicType = m_magicList.GetItemData(m_magicList.GetCurSel());
    m_objectType = m_objectList.GetItemData(m_objectList.GetCurSel());
    m_graphicListIndex = m_graphic.GetCurSel();
  };
}


BEGIN_MESSAGE_MAP(EditLauncher, CDialog)
	//{{AFX_MSG_MAP(EditLauncher)
	ON_BN_CLICKED(IDC_AvailableObject, OnAvailableObject)
	ON_BN_CLICKED(IDC_PredefindObject, OnAvailableObject)
	ON_BN_CLICKED(IDC_MagicSpell, OnAvailableObject)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditLauncher message handlers

void EditLauncher::Enable(int ID, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(ID);
  if (pWnd != NULL) pWnd->ShowWindow(show);
}

void EditLauncher::AddMagic(int n, char *name)
{
  int index;
  index = m_magicList.AddString(name);
  m_magicList.SetItemData(index, n);
}

void EditLauncher::FillMagicList(void)
{
  m_magicList.ResetContent();
  AddMagic(0,"Fireball");
  AddMagic(1,"Poison");
  AddMagic(2,"Lightning");
  AddMagic(3,"Dispell");
  AddMagic(6,"Poison Bolt");
  AddMagic(7,"Poison Cloud");
}

void EditLauncher::AddObject(int n, char *name)
{
  int index;
  index = m_objectList.AddString(name);
  m_objectList.SetItemData(index, n);
}

void EditLauncher::FillObjectList(void)
{
  m_objectList.ResetContent();
  AddObject(  0,"nothing");
  AddObject(  4,"Torch");
  AddObject( 54,"Rock");
  AddObject(128,"Boulder");
  AddObject( 51,"Arrow");
  AddObject( 52,"Slayer");
  AddObject( 55,"Poison Dart");
  AddObject( 56,"Throwing Star");
  AddObject( 32,"Dagger");
}

void EditLauncher::ExtractDataFromRecord(void)
{
  int at;
  at = m_pLauncher->launcherType();
  switch (at)
  {
  case 7:
    m_availableObject = 1;
    m_one = 0;
    break;
  case 8:
    m_availableObject = 2;
    m_one = 0;
    break;
  case 9:
    m_availableObject = 1;
    m_one = 1;
    break;
  case 10:
    m_availableObject = 2;
    m_one = 1;
    break;
  case 14:
    m_availableObject = 0;
    m_one = 0;
    break;
  case 15:
    m_availableObject = 0;
    m_one = 1;
    break;
  default: InternalError();
  };
  m_energy = m_pLauncher->energyRemaining();
  m_objectType = m_pLauncher->missileType();
  m_magicType = m_pLauncher->missileType();
  m_graphicListIndex = m_pLauncher->wallGraphicOrdinal();
}

bool EditLauncher::EncodeDataIntoRecord(void)
{
  int at = 0;
  switch (m_availableObject)
  {
  case 0:  //An available Object
    at = m_one == 0 ? 14 : 15;
    break;
  case 1:  //A predefined object
    at = m_one == 0 ? 7 : 9;
    m_pLauncher->missileType(m_objectType);
    break;
  case 2:  //A magic spell
    at = m_one == 0 ? 8 : 10;
    m_pLauncher->missileType(m_magicType);
    break;
  default: InternalError();
  };
  m_pLauncher->launcherType(at);
  m_pLauncher->energyRemaining(m_energy);
  m_pLauncher->onceOnly(false);
  m_pLauncher->wallGraphicOrdinal(m_graphicListIndex);
  return true;
}

BOOL EditLauncher::OnInitDialog() 
{
  ExtractDataFromRecord();
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditLauncher::OnOK() 
{
	// TODO: Add extra validation here
  if (!UpdateData(true)) return;
	if (!EncodeDataIntoRecord()) return;
	CDialog::OnOK();
}

void EditLauncher::OnAvailableObject() 
{
	// TODO: Add your control notification handler code here
  UpdateData(true);
  UpdateData(false);
}

int ACTUATOR_Launcher::Edit(LOCATIONREL locr)
{
  EditLauncher dlg(locr);
  dlg.m_pLauncher = this;
  return dlg.DoModal();
}

