// EditMonsterGenerator.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "EditMonsterGenerator.h"
#include "Actuator.h"
#include "misc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// EditMonsterGenerator dialog


EditMonsterGenerator::EditMonsterGenerator(LOCATIONREL locr,
                                           RN RNmonsterGenerator,
                                           CWnd* pParent /*=NULL*/)
	: CDialog(EditMonsterGenerator::IDD, pParent)
{
  m_locr = locr;
  m_RNmonsterGenerator = RNmonsterGenerator;
	//{{AFX_DATA_INIT(EditMonsterGenerator)
	m_fixed = -1;
	m_numberOfMonsters = 0;
	m_audible = FALSE;
	m_onceOnly = FALSE;
	m_disableTime = 0;
	m_hitPoints = 0;
	m_invisibleMonster = FALSE;
	m_drawAsSize4Monster = FALSE;
	m_alternateGraphic = 0;
	//}}AFX_DATA_INIT
}

void EditMonsterGenerator::ExtractDataFromRecord(void)
{
  m_numberOfMonsters    = m_pMonsterGenerator->numMonster();
  if (m_numberOfMonsters & 8)
  {
    m_numberOfMonsters = m_numberOfMonsters & 7;
    m_fixed = 1;
  }
  else
  {
    m_fixed = 0;
  }
  m_hitPoints           = m_pMonsterGenerator->hitPointMultiplier();
  m_monsterType         = m_pMonsterGenerator->monsterType();
  m_alternateGraphic    = m_pMonsterGenerator->alternateGraphics();
  m_audible             = m_pMonsterGenerator->audible();
  m_onceOnly            = m_pMonsterGenerator->onceOnly();
  m_disableTime         = m_pMonsterGenerator->disableTime();
  if (m_disableTime >= 128)
  {
    m_disableTime = 64*(m_disableTime-126);
  };
  m_floorGraphicOrdinal = m_pMonsterGenerator->floorGraphicOrdinal();
  m_invisibleMonster = m_pMonsterGenerator->invisible();
  m_drawAsSize4Monster = m_pMonsterGenerator->drawAsSize4();
}

bool EditMonsterGenerator::EncodeDataIntoRecord(void)
{
  if (m_fixed == 0)
  {
    m_pMonsterGenerator->numMonster(m_numberOfMonsters);
  }
  else
  {
    m_pMonsterGenerator->numMonster(m_numberOfMonsters | 8);
  };
  m_pMonsterGenerator->hitPointMultiplier(m_hitPoints);
  m_pMonsterGenerator->monsterType(m_monsterType);
  m_pMonsterGenerator->alternateGraphics(m_alternateGraphic);
  m_pMonsterGenerator->audible(m_audible?true:false);
  m_pMonsterGenerator->onceOnly(m_onceOnly?true:false);
  m_pMonsterGenerator->invisible(m_invisibleMonster?true:false);
  m_pMonsterGenerator->drawAsSize4(m_drawAsSize4Monster?true:false);
  if (m_disableTime >= 128)
  {
    m_pMonsterGenerator->disableTime(m_disableTime/64 + 126);
  }
  else
  {
    m_pMonsterGenerator->disableTime(m_disableTime);
  };
  m_pMonsterGenerator->floorGraphicOrdinal(m_floorGraphicOrdinal);
  return true;
}

void EditMonsterGenerator::DoDataExchange(CDataExchange* pDX)
{
  int i, j, index;
  FLOORTYPE ft;
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditMonsterGenerator)
	DDX_Control(pDX, IDC_FloorGraphic, m_floorGraphic);
	DDX_Control(pDX, IDC_MonsterType, m_monsterList);
	DDX_Radio(pDX, IDC_NumberFixed, m_fixed);
	DDX_Text(pDX, IDC_NumberOfMonsters, m_numberOfMonsters);
	DDX_Check(pDX, IDC_Audible, m_audible);
	DDX_Check(pDX, IDC_OnceOnly, m_onceOnly);
	DDX_Text(pDX, IDC_DisableTime, m_disableTime);
	DDV_MinMaxUInt(pDX, m_disableTime, 0, 8256);
	DDX_Text(pDX, IDC_HitPoints, m_hitPoints);
	DDV_MinMaxUInt(pDX, m_hitPoints, 0, 15);
	DDX_Check(pDX, IDC_InvisibleMonster, m_invisibleMonster);
	DDX_Check(pDX, IDC_DrawAsSize4Monster, m_drawAsSize4Monster);
	DDX_CBIndex(pDX, IDC_AlternateGraphic, m_alternateGraphic);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    index = m_monsterType.FillComboBox(&m_monsterList);
    m_monsterList.SetCurSel(index);
    m_floorGraphic.ResetContent();
    j = m_floorGraphic.AddString("none");
    m_floorGraphic.SetItemData(j, 0);
    for (i=0; i<pDoc->NumFloorDecorations(m_locr.l()); i++)   
    {
      ft = pDoc->FloorDecorationType(m_locr.l(),i+1);
      j = m_floorGraphic.AddString(ft.Name());
      m_floorGraphic.SetItemData(j, ft.Integer()); 
    };
    m_floorGraphic.SetCurSel(m_floorGraphicOrdinal);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_floorGraphicOrdinal = m_floorGraphic.GetCurSel();
    index = m_monsterList.GetCurSel();
    m_monsterType = m_monsterList.GetItemData(index);
    //m_fixed
    //m_numberOfMonsters
    //m_audible
    //m_onceOnly
    //m_disableTime
  };
}


BEGIN_MESSAGE_MAP(EditMonsterGenerator, CDialog)
	//{{AFX_MSG_MAP(EditMonsterGenerator)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditMonsterGenerator message handlers

int ACTUATOR_MonsterGenerator::Edit(LOCATIONREL locr,
                                    RN RNmonsterGenerator)
{
  EditMonsterGenerator dlg(locr,RNmonsterGenerator);
  dlg.m_pMonsterGenerator = this;
  return dlg.DoModal();
}

BOOL EditMonsterGenerator::OnInitDialog() 
{
  CString title;
  LOCATIONABS loca;
  ExtractDataFromRecord();

	CDialog::OnInitDialog();

  loca = m_locr;
  title.Format("Monster Generator [%04X] %02d(%02d,%02d)%d",
               m_RNmonsterGenerator.ConvertToInteger(),
               loca.l, loca.x, loca.y, loca.p);
  SetWindowText(title);
	
	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditMonsterGenerator::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	if (!EncodeDataIntoRecord()) return;;
	CDialog::OnOK();
}
