// EditPortrait.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "Actuator.h"

#include "EditPortrait.h"
#include "Misc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

char *dmNames[24] = {
  "ELIJA",
  "HALK",
  "SYRA",
  "HISSSSA",
  "ZED",
  "CHANI",
  "HAWK",
  "BORIS",
  "MOPHUS",
  "LEIF",
  "WUTSE",
  "ALEX",
  "LINFLAS",
  "AZIZI",
  "IAIDO",
  "GANDO",
  "STAMM",
  "LEYLA",
  "TIGGY",
  "SONJA",
  "NABI",
  "GOTHMOG",
  "WUUF",
  "DAROOU"
};

char *csbNames[24] = {
  "AIRWING",
  "AROC",
  "TALON",
  "LETA",
  "DEMA",
  "ALGOR",
  "TOADROT",
  "VEN",
  "MANTIA",
  "GNATU",
  "SLOGAR",
  "STING",
  "SKELAR",
  "DETH",
  "NECRO",
  "PLAGUE",
  "TUNDA",
  "LANA",
  "BUZZZZZ",
  "PETAL",
  "ITZA",
  "TULA",
  "KAZAI",
  "LOR"
};

int savedDMCharacter = 0;

/////////////////////////////////////////////////////////////////////////////
// EditPortrait dialog


EditPortrait::EditPortrait(LOCATIONREL locr, 
                           RN RNportrait,
                           CWnd* pParent /*=NULL*/) 
	                : CDialog(EditPortrait::IDD, pParent)
{
  m_locr = locr;
  m_RNportrait = RNportrait;
	//{{AFX_DATA_INIT(EditPortrait)
	m_characterNumber = 0;
	m_dmCharacter = -1;
	m_external = FALSE;
	m_portraitNumber = 0;
	m_portraitClearIgnore = -1;
	m_portraitSetIgnore = -1;
	m_graphicClearIgnore = -1;
	m_graphicSetIgnore = -1;
	m_graphicToggleIgnore = -1;
	m_activeClearIgnore = -1;
	m_activeSetIgnore = -1;
	m_activeToggleIgnore = -1;
	m_blockDisables = FALSE;
	m_disableMessageNone = -1;
	m_portraitToggleIgnore = -1;
	m_targetPos = -1;
	m_advanced = FALSE;
	m_inactive = FALSE;
	m_hidePortrait = FALSE;
	m_hideGraphic = FALSE;
	//}}AFX_DATA_INIT
}

bool EditPortrait::AnyAdvancedFeatures(void)
{
  return    m_inactive
         || (m_activeSetIgnore      != 0)
         || (m_activeClearIgnore    != 0)
         || (m_activeToggleIgnore   != 0)
         || (m_portraitSetIgnore    != 0)
         || (m_portraitClearIgnore  != 0)
         || (m_portraitToggleIgnore != 0)
         || (m_graphicSetIgnore     != 0)
         || (m_graphicClearIgnore   != 0)
         || (m_graphicToggleIgnore  != 0)
         || (m_disableMessageNone   != 0)
         || m_hidePortrait
         || m_hideGraphic
         || m_blockDisables
         || m_external;
}

void EditPortrait::ExtractDataFromRecord(void)
{
  int actions, state;
  m_characterNumber = m_pPortrait->characterNumber();
  if (m_characterNumber >= 24)
  {
    m_external = true;
    m_portraitNumber = m_characterNumber;
    m_characterNumber = 0;
  }
  else
  {
    m_external = false;
    m_portraitNumber = 24;
  };
  m_graphicIndex = m_pPortrait->graphic();
  m_dmCharacter = savedDMCharacter;
  m_blockDisables = m_pPortrait->BlockDisables();
  actions = m_pPortrait->SetResponses();
  m_activeSetIgnore      = (actions >> 0) & 3;
  m_portraitSetIgnore    = (actions >> 2) & 3;
  m_graphicSetIgnore     = (actions >> 4) & 3;
  actions = m_pPortrait->ClearResponses();
  m_activeClearIgnore    = (actions >> 0) & 3;
  m_portraitClearIgnore  = (actions >> 2) & 3;
  m_graphicClearIgnore   = (actions >> 4) & 3;
  actions = m_pPortrait->ToggleResponses();
  m_activeToggleIgnore   = (actions >> 0) & 3;
  m_portraitToggleIgnore = (actions >> 2) & 3;
  m_graphicToggleIgnore  = (actions >> 4) & 3;
  m_disableMessageNone = m_pPortrait->DisableMessage();
  state = m_pPortrait->State();
  m_inactive     = (state & PORTRAIT_InActive) != 0;
  m_hidePortrait = (state & PORTRAIT_HidePortrait) != 0;
  m_hideGraphic  = (state & PORTRAIT_HideGraphic) != 0;
  m_targetPos = m_pPortrait->Position();
  m_advanced = AnyAdvancedFeatures();
}

void EditPortrait::EncodeDataIntoRecord(void)
{
  ui16 actions, state;
  if (m_external)
  {
    m_pPortrait->characterNumber(m_portraitNumber);
  }
  else
  {
    m_pPortrait->characterNumber(m_characterNumber);
  };
  m_pPortrait->graphic(m_graphicIndex);
  savedDMCharacter = m_dmCharacter;
  m_pPortrait->BlockDisables((ui16)m_blockDisables);
  actions = 0;
  actions |= (m_activeSetIgnore      & 3) << 0;
  actions |= (m_portraitSetIgnore    & 3) << 2;
  actions |= (m_graphicSetIgnore     & 3) << 4;
  m_pPortrait->SetResponses(actions);
  actions = 0;
  actions |= (m_activeClearIgnore    & 3) << 0;
  actions |= (m_portraitClearIgnore  & 3) << 2;
  actions |= (m_graphicClearIgnore   & 3) << 4;
  m_pPortrait->ClearResponses(actions);
  actions = 0;
  actions |= (m_activeToggleIgnore   & 3) << 0;
  actions |= (m_portraitToggleIgnore & 3) << 2;
  actions |= (m_graphicToggleIgnore  & 3) << 4;
  m_pPortrait->ToggleResponses(actions);
  m_pPortrait->DisableMessage((ui16)m_disableMessageNone);
  state = 0;
  if (m_inactive)     state |= PORTRAIT_InActive;
  if (m_hidePortrait) state |= PORTRAIT_HidePortrait;
  if (m_hideGraphic)  state |= PORTRAIT_HideGraphic;
  m_pPortrait->State(state);
  m_pPortrait->Position((ui16)m_targetPos);
}

void EditPortrait::Show(int id, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(id);
  if (pWnd == NULL) return;
  pWnd->ShowWindow(show);
}

void EditPortrait::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditPortrait)
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Control(pDX, IDC_CharacterName, m_characterName);
	DDX_Radio(pDX, IDC_DMcharacters, m_dmCharacter);
	DDX_Check(pDX, IDC_External, m_external);
	DDX_Text(pDX, IDC_PortraitNumber, m_portraitNumber);
	DDV_MinMaxUInt(pDX, m_portraitNumber, 24, 511);
	DDX_Radio(pDX, IDC_PortraitClearIgnore, m_portraitClearIgnore);
	DDX_Radio(pDX, IDC_PortraitSetIgnore, m_portraitSetIgnore);
	DDX_Radio(pDX, IDC_GraphicClearIgnore, m_graphicClearIgnore);
	DDX_Radio(pDX, IDC_GraphicSetIgnore, m_graphicSetIgnore);
	DDX_Radio(pDX, IDC_GraphicToggleIgnore, m_graphicToggleIgnore);
	DDX_Radio(pDX, IDC_ActiveClearIgnore, m_activeClearIgnore);
	DDX_Radio(pDX, IDC_ActiveSetIgnore, m_activeSetIgnore);
	DDX_Radio(pDX, IDC_ActiveToggleIgnore, m_activeToggleIgnore);
	DDX_Check(pDX, IDC_BlockDisables, m_blockDisables);
	DDX_Radio(pDX, IDC_DisableMessageNone, m_disableMessageNone);
	DDX_Radio(pDX, IDC_PortraitToggleIgnore, m_portraitToggleIgnore);
	DDX_CBIndex(pDX, IDC_TargetPos, m_targetPos);
	DDX_Check(pDX, IDC_Advanced, m_advanced);
	DDX_Check(pDX, IDC_Inactive, m_inactive);
	DDX_Check(pDX, IDC_HidePortrait, m_hidePortrait);
	DDX_Check(pDX, IDC_HideGraphic, m_hideGraphic);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    CString n;
    int i;
    m_characterName.ResetContent();
    for (i=0; i<24; i++)
    {
      if (m_dmCharacter == 0) n.Format("%02d - %s",i,dmNames[i]);
      else n.Format("%02d - %s",i,csbNames[i]);
      m_characterName.AddString(n);
    };
    m_characterName.SetCurSel(m_characterNumber);
    FillWallDecorationsComboBox();
    m_graphic.SetCurSel(m_graphicIndex);
    Show(IDC_PortraitNumber,   m_external != 0);
    Show(IDC_nnn,              m_external != 0);
    Show(IDC_Filename,         m_external != 0);
    Show(IDC_CharacterName,    m_external == 0);
    Show(IDC_Champion,         m_external == 0);
    Show(IDC_BlockDisables,    m_advanced != 0);
    Show(IDC_Inactive,         m_advanced != 0);
    Show(IDC_External,         m_advanced != 0);
    Show(IDC_SetTitle,         m_advanced != 0);
    Show(IDC_ClearTitle,       m_advanced != 0);
    Show(IDC_ToggleTitle,      m_advanced != 0);
    Show(IDC_ObituaryBox,      m_advanced != 0);
    Show(IDC_ActiveSetIgnore,  m_advanced != 0);
    Show(IDC_ActiveSetActive,  m_advanced != 0);
    Show(IDC_ActiveSetInactive,m_advanced != 0);
    Show(IDC_ActiveSetToggle,  m_advanced != 0);
    Show(IDC_PortraitSetIgnore,m_advanced != 0);
    Show(IDC_PortraitSetShow,  m_advanced != 0);
    Show(IDC_PortraitSetHide,  m_advanced != 0);
    Show(IDC_PortraitSetToggle,m_advanced != 0);
    Show(IDC_GraphicSetIgnore, m_advanced != 0);
    Show(IDC_GraphicSetShow,   m_advanced != 0);
    Show(IDC_GraphicSetHide,   m_advanced != 0);
    Show(IDC_GraphicSetToggle, m_advanced != 0);

    Show(IDC_ActiveClearIgnore,  m_advanced != 0);
    Show(IDC_ActiveClearActive,  m_advanced != 0);
    Show(IDC_ActiveClearInactive,m_advanced != 0);
    Show(IDC_ActiveClearToggle,  m_advanced != 0);
    Show(IDC_PortraitClearIgnore,m_advanced != 0);
    Show(IDC_PortraitClearShow,  m_advanced != 0);
    Show(IDC_PortraitClearHide,  m_advanced != 0);
    Show(IDC_PortraitClearToggle,m_advanced != 0);
    Show(IDC_GraphicClearIgnore, m_advanced != 0);
    Show(IDC_GraphicClearShow,   m_advanced != 0);
    Show(IDC_GraphicClearHide,   m_advanced != 0);
    Show(IDC_GraphicClearToggle, m_advanced != 0);

    Show(IDC_ActiveToggleIgnore,  m_advanced != 0);
    Show(IDC_ActiveToggleActive,  m_advanced != 0);
    Show(IDC_ActiveToggleInactive,m_advanced != 0);
    Show(IDC_ActiveToggleToggle,  m_advanced != 0);
    Show(IDC_PortraitToggleIgnore,m_advanced != 0);
    Show(IDC_PortraitToggleShow,  m_advanced != 0);
    Show(IDC_PortraitToggleHide,  m_advanced != 0);
    Show(IDC_PortraitToggleToggle,m_advanced != 0);
    Show(IDC_GraphicToggleIgnore, m_advanced != 0);
    Show(IDC_GraphicToggleShow,   m_advanced != 0);
    Show(IDC_GraphicToggleHide,   m_advanced != 0);
    Show(IDC_GraphicToggleToggle, m_advanced != 0);

    
    Show(IDC_TargetPos,              m_advanced != 0);
    Show(IDC_DisableMessagePosTitle, m_advanced != 0);
    Show(IDC_DisableMessageNone,     m_advanced != 0);
    Show(IDC_DisableMessageSet,      m_advanced != 0);
    Show(IDC_DisableMessageClear,    m_advanced != 0);
    Show(IDC_DisableMessageToggle,   m_advanced != 0);

    Show(IDC_HideGraphic,            m_advanced != 0);
    Show(IDC_HidePortrait,           m_advanced != 0);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_characterNumber = m_characterName.GetCurSel();
    m_graphicIndex = m_graphic.GetCurSel();
  };
}


BEGIN_MESSAGE_MAP(EditPortrait, CDialog)
	//{{AFX_MSG_MAP(EditPortrait)
	ON_BN_CLICKED(IDC_DMcharacters, OnDMcharacters)
	ON_BN_CLICKED(IDC_External, OnExternal)
	ON_WM_HELPINFO()
	ON_BN_CLICKED(IDC_CSBcharacters, OnDMcharacters)
	ON_BN_CLICKED(IDC_Advanced, OnAdvanced)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditPortrait message handlers

int ACTUATOR_Portrait::Edit(LOCATIONREL locr, RN RNportrait)
{
  EditPortrait dlg(locr, RNportrait);
  dlg.m_pPortrait = this;
  return dlg.DoModal();
}


void EditPortrait::FillWallDecorationsComboBox()
{
  int i;
  WALLTYPE wt;
  m_graphic.ResetContent();
  m_graphic.AddString("none");
  for (i=0; i<pDoc->NumWallDecorations(m_locr.l()); i++)
  {
    wt = pDoc->WallDecorationType(m_locr.l(),i);
    m_graphic.AddString(wt.Name());
  };
}

static int tips[] =
{
  IDC_Advanced             , IDT_EditPortrait_Advanced,
  IDC_DisableMessageNone   , IDT_EditPortrait_Obituary,
  IDC_DisableMessageSet    , IDT_EditPortrait_Obituary,
  IDC_DisableMessageClear  , IDT_EditPortrait_Obituary,
  IDC_DisableMessageToggle , IDT_EditPortrait_Obituary,
  IDC_TargetPos            , IDT_EditPortrait_TargetPos,
  IDC_HidePortrait         , IDT_EditPortrait_HidePortrait,
  IDC_Inactive             , IDT_EditPortrait_Inactive,
  IDC_HideGraphic          , IDT_EditPortrait_HideGraphic
};

BOOL EditPortrait::OnInitDialog() 
{
  CWnd *pWnd;
  ExtractDataFromRecord();
  m_toolTips.Create(this, 0);
  for (int i=0; i<sizeof(tips)/sizeof(int); i+=2)
  {
    pWnd = GetDlgItem(tips[i]);
    if (pWnd == NULL) continue;
    m_toolTips.AddTool(pWnd,tips[i+1]);
  };
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditPortrait::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

void EditPortrait::OnDMcharacters() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  UpdateData(false);
}

void EditPortrait::OnExternal() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  UpdateData(false);
}

BOOL EditPortrait::OnHelpInfo(HELPINFO* pHelpInfo) 
{
	CString s;
  int id;
  switch (pHelpInfo->iCtrlId)
  {
  case IDC_DMcharacters: 
  case IDC_CSBcharacters:
    id = IDH_EditPortrait_DMcharacters; 
    break;
  case IDC_PortraitNumber:
  case IDC_External:
    id = IDH_EditPortrait_PortraitNumber;
    break;
  case IDC_CharacterName:
    id = IDH_EditPortrait_Character;
    break;
  case IDC_Graphic:
    id = IDH_EditPortrait_Graphic;
    break;
  case IDC_DisableMessageNone:
  case IDC_DisableMessageSet:
  case IDC_DisableMessageClear:
  case IDC_DisableMessageToggle:
  //case IDC_DisableMessageTitle:
  case IDC_DisableMessagePosTitle:
  case IDC_TargetPos:
    id = IDH_EditPortrait_DisableMessage;
    break;
  case IDC_HidePortrait:
    id = IDH_EditPortrait_HidePortrait;
    break;
  case IDC_Advanced:
    id = IDH_EditPortrait_Advanced;
    break;
  case IDC_Inactive:
    id = IDH_EditPortrait_Inactive;
    break;
  default: id = IDH_Context_None;
  };
	s.LoadString(id);
  HelpMsg(s);
  return true;
	
}

BOOL EditPortrait::PreTranslateMessage(MSG* pMsg) 
{
	// TODO: Add your specialized code here and/or call the base class
	m_toolTips.RelayEvent(pMsg);
	return CDialog::PreTranslateMessage(pMsg);
}

void EditPortrait::OnAdvanced() 
{
	// TODO: Add your control notification handler code here
  UpdateData(true);
  if (!m_advanced)
  {
    if (AnyAdvancedFeatures())
    {
      m_advanced = true;
    };
  };
  UpdateData(false);
}
