// EditTarget.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "CSBuildView.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CCSBuildView *pMainView;


/////////////////////////////////////////////////////////////////////////////
// EditTarget dialog


EditTarget::EditTarget(LOCATIONREL locr, 
                       char *title,
                       LOCATIONABS locaSource,
                       bool multiLevel,
                       CWnd* pParent /*=NULL*/)
	: CDialog(EditTarget::IDD, pParent)
{
  m_locr = locr;
  m_currentX = locr.x();
  m_currentY = locr.y();
  m_locaSource = locaSource;
  m_multiLevel = multiLevel;
  m_title = title;
	//{{AFX_DATA_INIT(EditTarget)
	m_currentLevel = locr.l();
	//}}AFX_DATA_INIT
  if (  (locr.l() >= pDoc->NumLevel())
      ||(locr.l() < 0) )
  {
    m_currentX = 0;
    m_currentY = 0;
    m_currentLevel = 0;
  };
}

void EditTarget::ShowControl(int ID, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(ID);
  if (pWnd != NULL) pWnd->ShowWindow(show);
}

void EditTarget::ReadOnlyControl(int ID, bool readOnly)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(ID);
  if (pWnd != NULL) pWnd->EnableWindow(!readOnly);
}
void EditTarget::DoDataExchange(CDataExchange* pDX)
{
  ShowControl(IDC_UpLevel, m_multiLevel);
  ShowControl(IDC_DownLevel, m_multiLevel);
  ReadOnlyControl(IDC_Level, !m_multiLevel);
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditTarget)
	DDX_Text(pDX, IDC_Level, m_currentLevel);
	DDV_MinMaxUInt(pDX, m_currentLevel, 0, pDoc->NumLevel()-1);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(EditTarget, CDialog)
	//{{AFX_MSG_MAP(EditTarget)
	ON_WM_DRAWITEM()
	ON_WM_PAINT()
	ON_BN_CLICKED(IDC_DownLevel, OnDownLevel)
	ON_BN_CLICKED(IDC_UpLevel, OnUpLevel)
	ON_EN_CHANGE(IDC_Level, OnChangeLevel)
	ON_WM_LBUTTONDOWN()
	ON_WM_KEYDOWN()
	ON_WM_CHAR()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditTarget message handlers

BOOL EditTarget::OnInitDialog() 
{
	CDialog::OnInitDialog();
	CString title;
	// TODO: Add extra initialization here
//	MoveWindow(0,0,16*32 + 130, 16*32 + 24);
  title.Format("Edit Target of %s at %d(%02d,%02d)",
               m_title,
               m_locaSource.l,
               m_locaSource.x,
               m_locaSource.y);
  SetWindowText(title);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditTarget::OnDrawItem(int nIDCtl, LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
	// TODO: Add your message handler code here and/or call default
	
	CDialog::OnDrawItem(nIDCtl, lpDrawItemStruct);
}

void EditTarget::DrawLevel(CPaintDC *pDC, 
                           int level, 
                           bool drawTarget)
{
  CDC myDC;
  int row, col;
  int top, bottom, left, right;
  CBitmap *origBitmap;
  CPen pen1, pen2, *pOldPen;
  myDC.CreateCompatibleDC(pDC);
  origBitmap = myDC.SelectObject(&pMainView->m_tileBitmap);
  for (row=0; row<=pDoc->LastRow(level); row++)
  {
    for (col=0; col<=pDoc->LastColumn(level); col++)
    {
      pMainView->DrawCell(pDC, &myDC, level, col, row); 
    };
  };
  if (drawTarget)
  {
    top = m_currentY*16;
    left = m_currentX*16;
    bottom = top+16;
    right = left + 16;
    pen1.CreatePen( PS_SOLID, 2, RGB(0,255,255) );
    pOldPen = pDC->SelectObject(&pen1);
    top -= 1;
    left -= 1;
    bottom += 1;
    right += 1;
    pDC->MoveTo(left,top);
    pDC->LineTo(right,top);
    pDC->LineTo(right,bottom);
    pDC->LineTo(left,bottom);
    pDC->LineTo(left,top);
    top -= 3;
    left -= 3;
    bottom += 2;
    right += 2;
    pen2.CreatePen(PS_SOLID, 3, RGB(255,0,0));
    pDC->SelectObject(&pen2);
    pDC->MoveTo(left,top);
    pDC->LineTo(right,top);
    pDC->LineTo(right,bottom);
    pDC->LineTo(left,bottom);
    pDC->LineTo(left,top);
    pDC->SelectObject(pOldPen);
    top -= 2;
    left -= 2;
    bottom += 3;
    right += 3;
    pDC->SelectObject(&pen1);
    pDC->MoveTo(left,top);
    pDC->LineTo(right,top);
    pDC->LineTo(right,bottom);
    pDC->LineTo(left,bottom);
    pDC->LineTo(left,top);
    pDC->MoveTo(0,0);
    pDC->LineTo(left,top);
    pDC->MoveTo(16*32,0);
    pDC->LineTo(right,top);
    pDC->MoveTo(16*32, 16*32);
    pDC->LineTo(right,bottom);
    pDC->MoveTo(0,16*32);
    pDC->LineTo(left,bottom);
    pDC->SelectObject(pOldPen);
  };
}

void EditTarget::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	// TODO: Add your message handler code here
	// Do not call CDialog::OnPaint() for painting messages
  DrawLevel(&dc, m_currentLevel, m_currentLevel==(unsigned)m_locr.l());
}

void EditTarget::OnDownLevel() 
{
	// TODO: Add your control notification handler code here
  UpdateData(true);
	if (m_currentLevel >= (unsigned)pDoc->NumLevel() - 1) return;
  m_currentLevel++;
  UpdateData(false);
  Invalidate();
}

void EditTarget::OnUpLevel() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  if (m_currentLevel == 0) return;
  m_currentLevel--;
  UpdateData(false);
  Invalidate();
}

void EditTarget::OnChangeLevel() 
{
	
	// TODO: Add your control notification handler code here
  UpdateData(true);
  if (m_currentLevel < 0) m_currentLevel = 0;
  if (m_currentLevel >= (unsigned)pDoc->NumLevel()-1) 
            m_currentLevel = pDoc->NumLevel()-1;
  UpdateData(false);
  Invalidate();
}

void EditTarget::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	int row, col;
  UpdateData(true);
  row = point.y/16;
  col = point.x/16;
  if (   (row >= 0)
      && (col >= 0)
      && (row <= pDoc->LastRow(m_currentLevel))
      && (col <= pDoc->LastColumn(m_currentLevel)))
  {
    m_currentX = col;
    m_currentY = row;
    m_locr.x(col);
    m_locr.y(row);
    m_locr.l(m_currentLevel);
    UpdateData(false);
    Invalidate();
  };
	CDialog::OnLButtonDown(nFlags, point);
}

void EditTarget::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	if (nChar == 0x22) OnDownLevel();  //Page Down
  if (nChar == 0x21) OnUpLevel();    //Page Up
	
	CDialog::OnKeyDown(nChar, nRepCnt, nFlags);
}

void EditTarget::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	
	CDialog::OnChar(nChar, nRepCnt, nFlags);
}

BOOL EditTarget::PreTranslateMessage(MSG* pMsg) 
{
	// TODO: Add your specialized code here and/or call the base class
	if (pMsg->message == WM_KEYDOWN)
  {
    OnKeyDown(pMsg->wParam, 1, 0);
  };
	return CDialog::PreTranslateMessage(pMsg);
}
