#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "ViewLevelInfo.h"
#include "CSBuildDoc.h"
#include "Misc.h"
#include "Actuator.h"
#include "GetExportArea.h"


CString EXPORTLOC::Format(void) const
{
  CString result;
  result.Format("%+d(%+d,%+d)%d",
                 dl, dx, dy, pos);
  return result;
}


void UnImplemented(FILE *f, char *msg, bool INI)
{
  static int numMessage = 0;
  if (numMessage == 0)
  {
    MessageBox(NULL, "One of more un-implemented objects\nare being exported", "Warning",MB_OK);
    numMessage++;
  };
  if (!INI)
  {
    fprintf(f,"%s",msg);
  };
}

void ExportTarget(FILE *f, 
                  const EXPORTLOC& src,
                  const LOCATIONREL& dst)
{
  fprintf(f," target=%+d(%+d,%+d)%d",
             dst.l()-src.level-src.dl,
             dst.x()-src.left-src.dx,
             dst.y()-src.top-src.dy,
             dst.p());
};

void ExportSTONE(FILE *f, const EXPORTLOC& eloc, bool /*INI*/)
{
  CELLFLAG cf;
  fprintf(f, ";\n   Stone ");
  cf = pDoc->GetCellFlag(eloc);
  if (cf & 0x08) fprintf(f,"RandN ");
  if (cf & 0x04) fprintf(f,"RandE ");
  if (cf & 0x02) fprintf(f,"RandS ");
  if (cf & 0x01) fprintf(f,"RandW ");
  fprintf(f,";\n");
}

void ExportOPEN(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  fprintf(f, ";\n  Open");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 0x8)
  {
    fprintf(f," RandomDecorations");
  };
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f," FootPrints");
  };
  fprintf(f,";\n");
}

void ExportTELEPORTER(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  RN RNteleporter;
  fprintf(f, ";\n  Teleporter");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 4) fprintf(f," Visible");
  if (cf & 8) fprintf(f," Active");
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f," FootPrints");
  };
  fprintf(f,";\n");
}

void ExportSTAIRS(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  fprintf(f, ";\n   Stairs");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 0x4)
  {
    fprintf(f," Up");
  }
  else
  {
    fprintf(f," Down");
  };
  if (cf & 0x8)
  {
    fprintf(f," North/South");
  }
  else
  {
    fprintf(f," East/West");
  };
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f," FootPrints");
  };
  fprintf(f,";\n");
}


void ExportPIT(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  fprintf(f, ";\n   Pit");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 0x1)
  {
    fprintf(f," Visible");
  };
  if (cf & 0x8)
  {
    fprintf(f," Open");
  };
  if (cf & 0x4)
  {
    fprintf(f," RandomDecorations");
  };
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f," FootPrints");
  };
  fprintf(f,";\n");
}

void ExportFALSEWALL(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  fprintf(f, ";\n   FalseWall");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 0x1)
  {
    fprintf(f," Open");
  };
  if (cf & 0x4)
  {
    fprintf(f," Visible");
  };
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f," FootPrints");
  };
  fprintf(f,";\n");
}

void ExportDOOR(FILE *f, const LOCATIONREL& locr, bool /*INI*/)
{
  CELLFLAG cf;
  char *state;
  fprintf(f, ";\n  DOOR");
  cf = pDoc->GetCellFlag(locr);
  if (cf & 0x8)
  {
    fprintf(f," North/South");
  }
  else
  {
    fprintf(f," East/West");
  };
  switch (cf & 7)
  {
  case 0: state="Open";        break;
  case 1: state="AlmostOpen";  break;
  case 2: state="HalfOpen";    break;
  case 3: state="AlmostClosed";break;
  case 4: state="Closed";      break;
  case 5: state="Bashed";      break;
  default:state="UnknownState";break;
  };
  fprintf(f, "  %s", state);
  if (pDoc->GetExtendedCellFlag(locr,ECF_FOOTPRINTS))
  {
    fprintf(f,"  FootPrints");
  };
  fprintf(f,";\n");
}

void ExportRoomType(FILE *f, const EXPORTLOC& eloc, bool INI)
{
  ROOMTYPE rt;
  rt = RoomType(eloc);
  switch(rt)
  {
  //case roomFALSEWALL:
  case roomBLOCKING:
  case roomNONBLOCKING:
    ExportFALSEWALL(f,eloc, INI);   break;
  case roomPIT:
    ExportPIT(f,eloc, INI);         break;
  case roomSTONE:
    ExportSTONE(f, eloc, INI);      break;
  case roomOPEN:
    ExportOPEN(f, eloc, INI);       break;
  case roomSTAIRS:
    ExportSTAIRS(f, eloc, INI);     break;
  case roomDOOR:
    ExportDOOR(f, eloc, INI);       break;
  case roomTELEPORTER:
    ExportTELEPORTER(f, eloc, INI); break;
  default: NotImplemented();
  };
}

void ExportActuator(FILE *f, const EXPORTLOC& /*eloc*/, RN /*obj*/, bool INI)
{
  UnImplemented(f, "  Actuator not implemented\n", INI);
}

void ExportTeleporter(FILE *f, const EXPORTLOC& eloc, RN obj, bool /*INI*/)
{
  DB1 *pTeleporter;
  LOCATIONREL locr;
  pTeleporter = pDoc->GetRecordAddressDB1(obj);
  pTeleporter->Export(f, eloc);
}

void ExportText(FILE *f, const EXPORTLOC& /*eloc*/, RN /*obj*/, bool INI)
{
  UnImplemented(f, "  Text not implemented\n", INI);
}

void ExportWeapon(FILE *f, const EXPORTLOC& eloc, RN obj, bool /*INI*/)
{
  DB5 *pWeapon;
  pWeapon = pDoc->GetRecordAddressDB5(obj);
  pWeapon->Export(f, eloc);
}

void ExportDoor(FILE *f, const EXPORTLOC& /*eloc*/, RN /*obj*/, bool INI)
{
  UnImplemented(f, "  Door not implemented\n", INI);
}

void ExportMonster(FILE *f, const EXPORTLOC& eloc, RN obj, bool /*INI*/)
{
  DB4 *pMonster;
  pMonster = pDoc->GetRecordAddressDB4(obj);
  pMonster->Export(f, eloc);
}

void ExportChest(FILE *f, const EXPORTLOC& /*eloc*/, RN /*obj*/, bool INI)
{
  UnImplemented(f, "  Chest not implemented\n", INI);
}

void ExportMisc(FILE *f, const EXPORTLOC& /*eloc*/, RN obj, bool /*INI*/)
{
  DB10 *pDB10;
  MISCTYPE mt;
  fprintf(f, "  Misc ");
  pDB10 = pDoc->GetRecordAddressDB10(obj);
  mt = pDB10->miscType();
  fprintf(f, "object=%s", mt.Name());
  if (pDB10->important()) fprintf(f, " important");
  if (pDB10->cursed()) fprintf(f," cursed");
  fprintf(f," value=%d;\n", pDB10->value());
}

void ExportPotion(FILE *f, const EXPORTLOC& eloc, RN obj, bool /*INI*/)
{
  pDoc->GetRecordAddressDB8(obj)->Export(f, eloc);
}

void ExportClothing(FILE *f, const EXPORTLOC& /*eloc*/, RN /*obj*/, bool INI)
{
  UnImplemented(f, "  Clothing not implemented\n", INI);
}

void ExportCellContents(FILE *f, const EXPORTLOC& eloc, bool INI)
{
  RN obj;
  DBTYPE objType;
  for (obj=pDoc->FindFirstObject(eloc);
       obj!=RNeof;
       obj=pDoc->GetDBRecordLink(obj))
  {
    objType = obj.dbType();
    switch(objType)
    {
    case dbACTUATOR:   ExportActuator(f, eloc, obj, INI);  break;
    case dbTEXT:       ExportText(f, eloc, obj, INI);      break;
    case dbWEAPON:     ExportWeapon(f, eloc, obj, INI);    break;
    case dbDOOR:       ExportDoor(f, eloc, obj, INI);      break;
    case dbMONSTER:    ExportMonster(f, eloc, obj, INI);   break;
    case dbCHEST:      ExportChest(f, eloc, obj, INI);     break;
    case dbMISC:       ExportMisc(f, eloc, obj, INI);      break;
    case dbPOTION:     ExportPotion(f, eloc, obj, INI);    break;
    case dbCLOTHING:   ExportClothing(f, eloc, obj, INI);  break;
    case dbTELEPORTER: ExportTeleporter(f, eloc, obj, INI);break;
    default: NotImplemented();
    };
  };
}

void ExportOneCell(FILE *f, const EXPORTLOC& eloc, bool INI)
{
  fprintf(f,"Cell Address= %s",eloc.Format());
  ExportRoomType(f, eloc, INI);  //Room Type and CellFlags
  ExportCellContents(f, eloc, INI); //All room objects
  fprintf(f,"EndCell; %s\n#\n",eloc.Format());
}

void ViewLevelInfo::ExportLevel(FILE *f) 
{
  EXPORTLOC eloc;
  LOCATIONABS loca;
  LOCATIONREL locr;
  GetExportArea dlg;
  dlg.m_height = pDoc->LastRow(m_currentLevel)+1;
  dlg.m_width = pDoc->LastColumn(m_currentLevel)+1;
  dlg.m_left = pDoc->OffsetX(m_currentLevel);
  dlg.m_top = pDoc->OffsetY(m_currentLevel);
  dlg.DoModal();
  loca.l = m_currentLevel;
  loca.x = dlg.m_left;
  loca.y = dlg.m_top;
  loca.p = 0;
  locr = loca;
  eloc.level = m_currentLevel;
  eloc.left = locr.x();
  eloc.dl = 0;
  eloc.top = locr.y();
  eloc.pos = 0;
  fprintf(f,
          "Area Width =  %d Height=%d;\n",
          dlg.m_width,
          dlg.m_height);
  for (eloc.dx=0; eloc.dx<(signed)dlg.m_width; eloc.dx++)
  {
    for (eloc.dy=0; eloc.dy<(signed)dlg.m_height; eloc.dy++)
    {
      ExportOneCell(f, eloc, dlg.m_ignoreNonImplemented?true:false);
    };
  };
  fprintf(f,"EndArea;\n");
}
