// Hints1.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "CSBuildView.h"
#include "misc.h"
#include "Hints.h"
#include "WarningDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


extern CCSBuildView *pMainView;


static i32 recentHintNum = 0;
HINTS hintLib;



HINT::HINT(CString& title)
{
  m_text      = title;
  m_next      = NULL;
  m_cells     = NULL;
  m_firstHint = NULL;
}

HINT::~HINT(void)
{
  HINT *temp;
  while (m_next != NULL)
  {
    temp = m_next;
    m_next = m_next->m_next;
    temp->m_next = NULL;
    delete temp;
  };
  if (m_cells != NULL) free(m_cells);
  m_cells = NULL;
  if (m_firstHint != NULL) delete m_firstHint;
  m_firstHint = NULL;
}

int HINT::CountCells(char *buf)
{
  int i, len;
  len = strlen(buf);
  m_numCell = 0;
  for (i=0; i<len; i++)
  {
    if (buf[i] == '(') m_numCell++;
  };
  return 0;
}

int HINT::Decimal(char *buf, int& col, int& n)
{
  char c;
  n = 0;
  c = buf[col++];
  if ( (c<'0') || (c > '9')) return HINTERR_BadDigit;
  n = c - '0';
  c = buf[col];
  if ( (c<'0') || (c > '9')) return 0;
  col++;
  n = 10 * n + c - '0';
  c = buf[col];
  if ( (c<'0') || (c > '9')) return 0;
  col++;
  n = 10 * n + c - '0';
  return 0;
}


int HINT::AddHint(char *buf)
{
  HINT *text, *pHint;
  CString title(buf);
  text = new HINT(title);
  if (m_firstHint == NULL) m_firstHint = text;
  else
  {
    for (pHint = m_firstHint; pHint->m_next!=NULL; pHint=pHint->m_next);
    pHint->m_next = text;
  };
  return 0;
}

int HINT::ReadCells(char *buf)
{
  int i, col;
  int i1, i2, i3;
  col = 0;
  for (i=0; i<m_numCell; i++)
  {
    if (buf[col++] != '(') return HINTERR_NoOpenParen;
    if (Decimal(buf,col,i1)) return HINTERR_BadDigit;
    if (buf[col++] != ',') return HINTERR_NoComma;
    if (Decimal(buf,col,i2)) return HINTERR_BadDigit;
    if (buf[col++] != ',') return HINTERR_NoComma;
    if (Decimal(buf,col,i3)) return HINTERR_BadDigit;
    if (buf[col++] != ')') return HINTERR_NoCloseParen;
    m_cells[i] = (i1<<16) | (i2<<8) | i3;
  };
  for (i=0; i<m_numCell; i++)
  {
    int j;
    for (j=i+1; j<m_numCell; j++)
    {
      if (m_cells[i] == m_cells[j])
      {
        m_cells[j] = m_cells[--m_numCell];
      };
    };
  };
  return 0;
}

bool HINT::AnyHintsFor(int location)
{
  int i;
  for (i=0; i<m_numCell; i++)
  {
    if (m_cells[i] == (unsigned)location) return true;
  };
  return false;
}

bool HINT::AnyHintForLevel(int level)
{
  int i;
  for (i=0; i<m_numCell; i++)
  {
    if ((m_cells[i] >> 16) == (unsigned)level) return true;
  };
  return false;
}

void HINT::SetHint(unsigned int location)
{
  if (AnyHintsFor(location)) return;
  m_cells = (unsigned int *)realloc(m_cells,4*((m_numCell/5 + 1) * 5),MALLOC099);
  m_cells[m_numCell++] = location;
}

unsigned int HINT::GetCell(int cellNum)
{
  return m_cells[cellNum];
}

void HINT::RemoveHint(unsigned int location)
{
  int i;
  for (i=0; i<m_numCell; i++)
  {
    if (m_cells[i] == location)
    {
      m_numCell--;
      m_cells[i] = m_cells[m_numCell];
      return;
    };
  };
  MessageBox(NULL,"Asked to remove non-existant cell reference","Error",MB_OK);
}

int HINT::Read(FILE *f, char *buf, int& lineNum)
{
  int len, result;
  if (fgets(buf, 9999, f) == NULL)
  {
    return HINTERR_NoCells;
  };
  lineNum++;
  len = strlen(buf);
  if (len == 0)
  {
    return HINTERR_NoCells;
  };
  result = CountCells(buf);
  if (result != 0) return result;
  m_cells = (unsigned int *)malloc(m_numCell*4,MALLOC097);
  if (m_cells == NULL) return HINTERR_NoMemory;
  result = ReadCells(buf);
  if (result != 0) return result;
  while (fgets(buf,9999,f) != NULL)
  {
    lineNum++;
    if (strlen(buf) == 0) break;
    if (buf[0] == ' ') break;
    if (buf[0] == '\n') break;
    AddHint(buf);
  };
  if (m_firstHint == NULL) return HINTERR_NoHint;
  return 0;
}

int HINT::Write(FILE *f)
{
  int i, i1, i2, i3;
  HINT *pHint;
  fprintf(f,"%s\n",m_text);
  for (i=0; i<m_numCell; i++)
  {
    i1 = m_cells[i] >> 16;
    i2 = (m_cells[i] >> 8) & 0xff;
    i3 = m_cells[i] & 0xff;
    fprintf(f,"(%d,%d,%d)", i1, i2, i3);
  };
  fprintf(f, "\n");
  for (pHint=m_firstHint; pHint!=NULL; pHint=pHint->m_next)
  {
    fprintf(f,"%s",pHint->m_text);
  };
  fprintf(f,"\n");
  return 0;
}

HINTS::HINTS(void)
{
  m_hintArray = NULL;
  m_numHint   = 0;
}

HINTS::~HINTS(void)
{
  Clear();
}

void HINTS::Clear(void)
{
  int i;
  if (m_hintArray != NULL) 
  {
    for (i=0; i<m_numHint; i++)
    {
      delete m_hintArray[i];
    };
    free (m_hintArray);
    m_numHint = 0;
  };
  m_hintArray = NULL;
}

void HINTS::SetHint(int level, int col, int row, int hint)
{
  unsigned int location;
  location = (level<<16) + (col << 8) + row;
  m_hintArray[hint]->SetHint(location);
}

void HINTS::RemoveHint(int level, int col, int row, int hint)
{
  unsigned int location;
  location = (level<<16) + (col << 8) + row;
  m_hintArray[hint]->RemoveHint(location);
}


bool HINTS::HasHint(int location, int hint)
{
  int i;
  if (hint >= 0)
  {
    if (hint >= m_numHint) return false;
    return m_hintArray[hint]->AnyHintsFor(location);
  }
  else
  { // does the location have any hints at all?
    for (i=0; i<m_numHint; i++)
    {
      if (m_hintArray[i]->AnyHintsFor(location)) return true;
    };
    return false;
  };
}

bool HINTS::LevelHasHint(int level, int hint)
{
  return m_hintArray[hint]->AnyHintForLevel(level);
}

int HINTS::Read(FILE *f)
{
  int lineNum = 0;
  int result, len;
  char line[10000];
  CString title;
  HINT *newTitle;
  Clear();
  while (fgets(line,9999,f) != NULL)
  {
    lineNum++;
    if (strcmp("HTCConverter Hints File\n", line) == 0) continue;
    if (strlen(line) == 0) continue;
    if (line[0] == ' ') continue;
    if (line[0] == '\n') continue;
    // Must be a title
    len = strlen(line);
    if (line[len-1] == '\n') line[len-1] = 0;
    title = line;
    newTitle = new HINT(title);
//    if (m_hintArray == NULL)
//    {
//      m_hintArray = (HINT *)malloc(5*sizeof(&newTitle), MALLOC098);
//      m_numHint = 1;
//    };
    m_hintArray = (HINT **)realloc(m_hintArray, 4*((m_numHint/5 + 1)*5), MALLOC099);
    m_hintArray[m_numHint++] = newTitle;
    result = newTitle->Read(f, line, lineNum);
    if (result != 0) 
    {
      return (lineNum << 16) + result;
    };
  };
  return 0;
}

int HINTS::Write(FILE *f)
{
  int i;
  for (i=0; i<m_numHint; i++)
  {
    m_hintArray[i]->Write(f);
  };
  return 0;
}





/////////////////////////////////////////////////////////////////////////////
// Hints dialog


Hints::Hints(CWnd* pParent /*=NULL*/)
	: CDialog(Hints::IDD, pParent)
{
	//{{AFX_DATA_INIT(Hints)
	m_showAll = FALSE;
	m_currentLevel = 0;
	//}}AFX_DATA_INIT
}


void Hints::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(Hints)
	DDX_Control(pDX, IDC_HintList, m_hintList);
	DDX_Check(pDX, IDC_ShowAll, m_showAll);
	DDX_Text(pDX, IDC_Level, m_currentLevel);
	DDV_MinMaxUInt(pDX, m_currentLevel, 0, 63);
	//}}AFX_DATA_MAP
  if (pDX->m_bSaveAndValidate)
  {
    int n;
    n = m_hintList.GetCurSel();
    if (n < 0) 
    {
      if (m_hintList.GetCount() > 0)
      {
        n = recentHintNum;
      }
      else
      {
        n=-1;
      };
    }
    else
    {
      n = m_hintList.GetItemData(n);
    };
    recentHintNum = n;
    m_currentHint = n;
  };

}


BEGIN_MESSAGE_MAP(Hints, CDialog)
	//{{AFX_MSG_MAP(Hints)
	ON_LBN_SELCHANGE(IDC_HintList, OnSelchangeHintList)
	ON_WM_PAINT()
	ON_WM_LBUTTONDOWN()
	ON_NOTIFY(UDN_DELTAPOS, IDC_LevelSpin, OnDeltaposLevelSpin)
	ON_EN_CHANGE(IDC_Level, OnChangeLevel)
	ON_BN_CLICKED(IDC_ShowAll, OnShowAll)
	ON_WM_RBUTTONDOWN()
	ON_BN_CLICKED(IDC_Read, OnRead)
	ON_BN_CLICKED(IDC_Write, OnWrite)
	ON_BN_CLICKED(IDC_Find, OnFind)
	ON_BN_CLICKED(IDC_CountByCell, OnCountByCell)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// Hints message handlers

void Hints::FillHintList(void)
{
  int i;
  CString line;
  int indx;
  LOCATIONABS loca;
  m_currentHint = -1;
  m_hintList.ResetContent();
  for (i=0; i<hintLib.NumHint(); i++)
  {
    indx = m_hintList.AddString(hintLib.GetTitle(i));
    m_hintList.SetItemData(indx, i);
    if (m_currentHint < 0) m_currentHint = i;
  };
}



void Hints::DrawLevel(CPaintDC *pDC, 
                       int level)
{
  CDC myDC;
  bool cellHasAnyHint, cellHasCurrentHint;
  int row, col, location;
  CBitmap *origBitmap;
  CPen pen1, pen2;
  myDC.CreateCompatibleDC(pDC);
  origBitmap = myDC.SelectObject(&pMainView->m_tileBitmap);
  for (row=0; row<=pDoc->LastRow(level); row++)
  {
    for (col=0; col<=pDoc->LastColumn(level); col++)
    {
      location = (level<<16) + (col<<8) + row;
      cellHasCurrentHint = hintLib.HasHint(location, m_currentHint);
      if (m_showAll)
      {
        cellHasAnyHint = hintLib.HasHint(location, -1);
        pMainView->DrawCell(pDC, &myDC, level, col, row,
                             (cellHasAnyHint)?(cellHasCurrentHint?2:1):0); 
      }
      else
      {
        pMainView->DrawCell(pDC, &myDC, level, col, row,
                                     cellHasCurrentHint?1:0); 
      }
    };
  };
}

/////////////////////////////////////////////////////////////////////////////
// ShowSkins message handlers

BOOL Hints::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
  SetWindowText(m_title);
  FillHintList();
  if (m_hintList.GetCount() > 0)
  {
    m_hintList.SetCurSel(0);
    recentHintNum = m_hintList.GetItemData(0);
    m_currentHint = recentHintNum;
  }
  else
  {
    recentHintNum = -1;
    m_currentHint = -1;
  };
  OnSelchangeHintList();
  GotoDlgCtrl(GetDlgItem(IDC_HintList));
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void Hints::OnSelchangeHintList() 
{
	// TODO: Add your control notification handler code here
  int n;
  n = m_hintList.GetCurSel();
  if (n < 0) 
  {
    if (m_hintList.GetCount() > 0)
    {
      n = recentHintNum;
    };
  }
  else
  {
    n = m_hintList.GetItemData(n);
  };
  recentHintNum = n;
  m_currentHint = n;
  Invalidate();	
}

void Hints::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	// TODO: Add your message handler code here
  DrawLevel(&dc, m_currentL);
	
	// Do not call CDialog::OnPaint() for painting messages
}

void Hints::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	
  int row, col;
	CDialog::OnLButtonDown(nFlags, point);
  col = point.x / pMainView->m_tileSize;
  row = point.y / pMainView->m_tileSize;
  if (col > pDoc->LastColumn(m_currentL)) return;
  if (row > pDoc->LastRow(m_currentL)) return;
  if (m_showAll)
  {
    return;
  }
  else
  {
    if (m_currentHint >= 0)
    {
      hintLib.SetHint(m_currentL, col, row, (ui8)m_currentHint);
    };
  };
  Invalidate();
}

void Hints::OnChangeLevel() 
{
	// TODO: If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.
	
	// TODO: Add your control notification handler code here
  UpdateData(true);
  if (m_currentLevel >= (unsigned)pDoc->NumLevel()) return;
  if (m_currentLevel == (unsigned)m_currentL) return;
  m_currentL = m_currentLevel;
  Invalidate();

}

void Hints::OnShowAll() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  Invalidate();
}

void Hints::OnDeltaposLevelSpin(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	// TODO: Add your control notification handler code here
  UpdateData(true);
	if (pNMUpDown->iDelta == 1)
  {
    if (m_currentL < pDoc->NumLevel() - 1) m_currentL++;
  }
  else
  {
    if (m_currentL > 0) m_currentL--;
  };
  m_currentLevel = m_currentL;
  UpdateData(false);
  Invalidate();
	*pResult = 0;
}

void Hints::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
  int row, col;
	CDialog::OnLButtonDown(nFlags, point);
  col = point.x / pMainView->m_tileSize;
  row = point.y / pMainView->m_tileSize;
  if (col > pDoc->LastColumn(m_currentL)) return;
  if (row > pDoc->LastRow(m_currentL)) return;
  hintLib.RemoveHint(m_currentL, col, row, m_currentHint);

  Invalidate();
	
	CDialog::OnRButtonDown(nFlags, point);
}


void Hints::OnRead() 
{
	// TODO: Add your control notification handler code here
  FILE *f;
  int result;
	CFileDialog fdlg(true);
  if (fdlg.DoModal() != IDOK) return;
  f = fopen(fdlg.GetPathName(), "r");
  if (f == NULL)
  {
    MessageBox("Cannot open that file","Error",MB_OK);
    return;
  };
  result = hintLib.Read(f);
  if (result != 0)
  {
    CString msg;
    msg.Format("Error type %d at line number %d", result &0xffff, result>>16);
    MessageBox(msg, "Hint File Error", MB_OK);
    fclose(f);
    return;
  };
  FillHintList();
  fclose (f);
  Invalidate();
}

void Hints::OnWrite() 
{
	// TODO: Add your control notification handler code here
  FILE *f;
  int result;
  CFileDialog fdlg(false);
  if (fdlg.DoModal() != IDOK) return;
  f = fopen(fdlg.GetPathName(), "w");
  if (f == NULL)
  {
    MessageBox("Cannot open that file","Error",MB_OK);
    return;
  };
  result = hintLib.Write(f);
  if (result != 0)
  {
    CString msg;
    MessageBox(msg, "Hint File Error", MB_OK);
    fclose(f);
    return;
  };
  fclose (f);
}

void Hints::OnFind() 
{
	// TODO: Add your control notification handler code here
  int startLevel = m_currentLevel;
  UpdateData(true);
  if (m_currentHint >= 0)
  {
	  m_currentLevel++;
    if (m_currentLevel >= (unsigned)pDoc->NumLevel()) m_currentLevel = 0;
    while (m_currentLevel != (unsigned)startLevel)
    {
      if (hintLib.LevelHasHint(m_currentLevel, m_currentHint)) break;
      m_currentLevel++;
      if (m_currentLevel >= (unsigned)pDoc->NumLevel()) m_currentLevel = 0;
    };
    if (m_currentLevel == (unsigned)startLevel) return;
    m_currentL = m_currentLevel;
    UpdateData(false);
    Invalidate();
  };
}

HINT *HINTS::GetHint(int hintNum)
{
  return m_hintArray[hintNum];
}

void Hints::OnCountByCell() 
{
	// TODO: Add your control notification handler code here
  ui16 *counts;
  int cell;
  int i,j;
  int level, col, row, maxCount;
  CHKERRMSG msg;
  LOCATIONREL locr;
  CString line;
  HINT *pHint;
  WarningDialog dlg;
  dlg.m_title = "Number of Hints in each cell";
  counts = (ui16 *)malloc(64*32*32*sizeof(ui16), MALLOC100);
  memset(counts,0,64*32*32*sizeof(ui16));
  for (i=0; i<hintLib.NumHint(); i++)
  {
    pHint = hintLib.GetHint(i);
    for (j=0; j<pHint->NumCell(); j++)
    {
      cell = pHint->GetCell(j);
      level = (cell >> 16) & 63;
      col   = (cell >> 8) & 31;
      row   = cell & 31;
      counts[(level*32 + col)*32+row]++;
    };
  };
  maxCount = 0;
  for (i=0; i<64*32*32; i++)
  {
    if (counts[i] > maxCount) maxCount = counts[i];
  };
  while (maxCount > 0)
  {
    for (i=0; i<64*32*32; i++)
    {
      if (counts[i] == maxCount)
      {
        level = i/32/32;
        col = (i/32) & 31;
        row = i & 31;
        line.Format("%5d   %d(%d,%d)", maxCount, level, col, row);
        msg.NewError(line);
      };
    };
    maxCount--;
  }
  free(counts);
  dlg.m_chkErrMsg = &msg;
  if (msg.FirstError() != NULL) dlg.DoModal();
  else pDoc->Information("No Hints Found");
}
