#include "stdafx.h"
#include "CSBuild.h"
//#include "types.h"
#include "CSBuildDoc.h"
#include "Actuator.h"
#include "misc.h"
#include "WarningDialog.h"



extern char ExtendedFeaturesVersion;
extern int  ExtendedFeaturesSize;
extern bool RecordDesignOption;
extern bool DMRulesOption;
extern bool invisibleMonsters;
extern bool drawAsSize4Monsters;
extern bool extendedWallDecorations;
extern bool extendedPortraits;
extern char *gameInfo;
CString     encipheredDataFile = "";
ui32        expectedGraphicsSignature1 = 0;
ui32        expectedGraphicsSignature2 = 0;
ui32        expectedCSBgraphicsSignature1 = 0;
ui32        expectedCSBgraphicsSignature2 = 0;
char        hintKey[8];
CHKERRMSG   clonedObjectList;


EXTENDEDFEATURESBLOCK::EXTENDEDFEATURESBLOCK(void)
{
  memset(this,0,512);
  memcpy(sentinel," Extended Features ",20);
  dataMapLength = 1;
  version = 'B';
}

EXTENDEDFEATURESBLOCK::~EXTENDEDFEATURESBLOCK(void)
{
}

unsigned int FormChecksum(ui8 *buf, int num)
{
  unsigned int result = 0;
  int i;
  for (i=0; i<num; i++)
  {
    result = result * ui32(0xbb40e62d) + 11 + buf[i];
  };
  return result;
}

LOADSTATUS CCSBuildDoc::ReadDSAs(CFile *f, int numDSA)
{
  int i;
  int dsaNum;
  unsigned int computedChecksum, fileChecksum;
  pDoc->RCS(f); //Discard old checksum and restart.
  for (i=0; i<numDSA; i++)
  {
    pDoc->RCS(&dsaNum);
    if (m_DSAIndex[dsaNum] != NULL)
    {
      Error("Duplicate DSA");
      return LOAD_ERROR;
    };
    if (dsaNum > 255)
    {
      Error("DSA number greater than 255");
      return LOAD_ERROR;
    };
    m_DSAIndex[dsaNum] = new DSA;
    m_DSAIndex[dsaNum]->Read();
  };
  computedChecksum = pDoc->RCS(f);
  f->Read(&fileChecksum, 4);
  if (fileChecksum != computedChecksum)
  {
    Error("Improper DSA checksum");
    return LOAD_ERROR;
  };
  return LOAD_OK;
}

void CCSBuildDoc::ClearDSALevelIndex(void)
{
  int i,j;
  for (i=0; i<64; i++)
  {
    for (j=0; j<32; j++)
    {
      m_DSALevelIndex[i][j] = 0xffff;
    };
  };
}

void CCSBuildDoc::ReadDSALevelIndex(CFile *f)
{
  unsigned char triple[3];
  for (;;)
  {
    f->Read(triple, 3);
    if (triple[0] > 63) break;
    m_DSALevelIndex[triple[0]][triple[1]] = triple[2];
  };
}

void ReadGameInfo(CFile *f, int size)
{
  if (gameInfo != NULL) free(gameInfo);
  gameInfo = NULL;
  if (size == 0)
  {
    gameInfo = (char *)malloc(100, MALLOC077);
    strcpy(gameInfo, "No game information was provided");
    return;
  };
  gameInfo = (char *)malloc(size, MALLOC078);
  f->Read(gameInfo, size);
}


int CCSBuildDoc::ReadExtendedFeatures(CFile *f)
{
  EXTENDEDFEATURESBLOCK efb;
  int n0, n1, n2;
  int startPos, endPos;
  unsigned int savechecksum;
  ExtendedFeaturesVersion = '@';
  extendedPortraits = false;
  RecordDesignOption = false;
  DMRulesOption = false;
  invisibleMonsters = false;
  drawAsSize4Monsters = false;
  extendedWallDecorations = false;
  m_indirectText = false;
  m_bigActuators = false;
  expectedGraphicsSignature1 = 0;
  expectedGraphicsSignature2 = 0;
  expectedCSBgraphicsSignature1 = 0;
  expectedCSBgraphicsSignature2 = 0;
  memset(hintKey,0,8);
  CellFlagArraySize(0);
  m_dataMapLength = 0;
  if (m_dataIndexMap != NULL) free(m_dataIndexMap);
  m_dataIndexMap = NULL;
  if (m_dataTypeMap  != NULL) free(m_dataTypeMap);
  m_dataTypeMap = NULL;
  if (f != NULL)
  {
    startPos = f->Seek(0, CFile::current);
    n0 = f->Read(&efb, 512);
    if (n0 != 512) {f->Seek(startPos,CFile::begin); return 0;};
    if (strcmp(efb.sentinel, " Extended Features ") != 0)
    {
      f->Seek(startPos,CFile::begin);
      return 0;
    };
    savechecksum = efb.extendedFeaturesChecksum;
    efb.extendedFeaturesChecksum = 0;
    if (FormChecksum((ui8 *)&efb,512) != savechecksum)
    {
      f->Seek(startPos,CFile::begin);
      return 0;
    };
  }
  else
  { //The constructor made a nice EFB for us.
    startPos = 0;
  };

  m_dataTypeMap = (ui8 *)malloc(efb.dataMapLength,MALLOC032);
  m_dataIndexMap = (ui16 *)malloc(2*efb.dataMapLength,MALLOC033);
  if ( (m_dataTypeMap==NULL) || (m_dataIndexMap==NULL))
  {
    DIE("Cannot Allocate Memory");
  };
  ExtendedFeaturesVersion = efb.version;
  CellFlagArraySize(efb.cellFlagArraySize);
  if (f != NULL)
  {
    n1 = f->Read(m_dataTypeMap,efb.dataMapLength);
    n2 = f->Read(m_dataIndexMap,2*efb.dataMapLength);
    if (  (FormChecksum((ui8 *)m_dataTypeMap, efb.dataMapLength) != efb.dataTypeMapChecksum)
        ||(FormChecksum((ui8 *)m_dataIndexMap,2*efb.dataMapLength) != efb.dataIndexMapChecksum)
        ||(n1 != efb.dataMapLength)
        ||(n2 != 2*efb.dataMapLength))
    {
      Error("Extended features damaged");
      return LOAD_ERROR;
    };
  };
  m_dataMapLength = efb.dataMapLength;
  if (efb.numDSA != 0) 
  {
    if (ReadDSAs(f, efb.numDSA)!=LOAD_OK) return false;
  };
  ReadGameInfo(f, efb.gameInfoSize);
  ClearDSALevelIndex();
  if (efb.flags & EXTENDEDFEATURESBLOCK::LevelDSAInfoPresent)
  {
    ReadDSALevelIndex(f);
  };
  RecordDesignOption =
    (efb.flags & EXTENDEDFEATURESBLOCK::AutoRecord) != 0;
  DMRulesOption =
    (efb.flags & EXTENDEDFEATURESBLOCK::DMRules) != 0;
  invisibleMonsters =
    (efb.extendedFlags & EXTENDEDFEATURESBLOCK::InvisibleMonsters) != 0;
  drawAsSize4Monsters =
    (efb.extendedFlags & EXTENDEDFEATURESBLOCK::DrawAsSize4Monsters) != 0;
  extendedWallDecorations =
    (efb.extendedFlags & EXTENDEDFEATURESBLOCK::ExtendedWallDecorations) != 0;
  extendedPortraits =
    (efb.flags & EXTENDEDFEATURESBLOCK::ExpandedPortraits) != 0;
  m_bigActuators = 
    (efb.flags & EXTENDEDFEATURESBLOCK::BigActuators) != 0;
  expectedGraphicsSignature1 = efb.graphicsSignature1;
  expectedGraphicsSignature2 = efb.graphicsSignature2;
  expectedCSBgraphicsSignature1 = efb.CSBgraphicsSignature1;
  expectedCSBgraphicsSignature2 = efb.CSBgraphicsSignature2;
  memcpy(hintKey,efb.hintKey,8);
  if ((efb.spellFilterLocation & 0x80000000) == 0)
  {
    m_spellFilterLocation.l(-1);
  }
  else
  {
    m_spellFilterLocation.Integer(efb.spellFilterLocation);
  };
  if (efb.flags & EXTENDEDFEATURESBLOCK::EditingOptionsPresent)
  {
    EditingOptions(efb.editingOptions);
  }
  else
  {
    EditingOptions(EDIT_OPT_DefaultOptions);
  };
  m_indirectText = (efb.flags & EXTENDEDFEATURESBLOCK::IndirectText) != 0;
  if (f != NULL)
  {
    endPos = f->Seek(0, CFile::current);
  }
  else
  {
    endPos = 512;
  };
  return endPos-startPos;
};


ui16 CCSBuildDoc::Unscramble(ui16 *buf, ui16 initialHash, i32 numword)
{
  ui16 currentHash, result;
  int i;
  currentHash = initialHash;
  result = initialHash;
  if (m_littleEndian)
  {
    for (i=0; i<numword; i++)
    {
      result = (ui16)(result + (buf[i]));
      buf[i] ^= (currentHash);
      result = (ui16)(result + (buf[i]));
      currentHash = (ui16)(currentHash + numword - i);
    };
  }
  else
  {
    for (i=0; i<numword; i++)
    {
      result = (ui16)(result + LE16(buf[i]));
      buf[i] ^= LE16(currentHash);
      result = (ui16)(result + LE16(buf[i]));
      currentHash = (ui16)(currentHash + numword - i);
    };
  };
  return result;
}

//*********************************************************
//
//*********************************************************
char getBits(i32 bitsNeeded, 
            int& bitsRemaining, 
            int& currentWord, 
            ui16 *(&pSRC))
// internal subroutine for ExpandData
{ 
  int result = 0;
  if (bitsRemaining < bitsNeeded)
  {
    bitsNeeded -= bitsRemaining;
    result = currentWord & 0xffff;
    result <<= bitsRemaining;
    currentWord = LE16(*(pSRC++));
    bitsRemaining = 16;
  };
  bitsRemaining -= bitsNeeded;
  result = (result & 0xffff0000) | (currentWord & 0xffff);
  result <<= bitsNeeded;
  currentWord <<= bitsNeeded;
  return (char)((result >> 16)&0xff);
}


void CCSBuildDoc::ExpandData(char *src,char *dst,int dstSize)
{
  // Data file consists of variable length fields.
  // There are 20 bytes at the beginning that are
  // lookup tables---the first is 4 bytes long
  // and is used to decode data of the form
  // ( in binary ) of 0xx.  xx is the index into the
  // table.
  // The next table is 16 bytes long and is used to
  // decode data of the form 10yyyy where yyyy is used
  // as an index into the table.
  // The only remaining case is data of the form
  // 11zzzzzzzz where zzzzzzzz is a literal byte.
  // Summary:
  // 0xx        index into small table
  // 10yyyy     index into large table
  // 11zzzzzzzz literal data
  //;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  //dReg D0, D1, D7;
  //aReg A0, A2, A3;
  char *pDST, *pTable;
  ui16 *pSRC;
  int currentWord=0;//To keep compiler quiet
  int bytesNeeded, bitsRemaining;
  pDST = dst;
  bytesNeeded = dstSize;
  pTable = src;
  memcpy(m_saveExpansionTable, pTable, 20);
  pSRC = (ui16 *)(src + 20);
  bitsRemaining = 0;
  do
  {
    if (bitsRemaining == 0)
    {
      currentWord = LE16(*(pSRC++));
      bitsRemaining = 16; // # bits in currentWord
    };
    if ((currentWord & 0x8000) != 0)
    {
      if (getBits(2, bitsRemaining, currentWord, pSRC) == 2)
      { //case 10yyyy
        *(pDST++) = *(pTable + getBits(4, bitsRemaining, currentWord, pSRC) + 4);
      }
      else
      { // case 11zzzzzzzz
        *(pDST++) = getBits(8, bitsRemaining, currentWord, pSRC);
      };
    }
    else
    {  //case 0xx
      *(pDST++) = *(pTable + getBits(3, bitsRemaining, currentWord, pSRC));
    };
    bytesNeeded--;
  } while (bytesNeeded != 0);
  return;
}



LOADSTATUS CCSBuildDoc::UnscrambleStream(
                  CFile *f,
                  ui16 *dest,        //data destination
                  i32 size,          //# of words
                  ui16 initialHash,   
                  ui16 Checksum)     //expected checksum
{
  int num;
  ui16 finalHash;
  num = f->Read(dest,2*size);
  if (num != 2*size) return LOAD_ERROR;
  finalHash = Unscramble(dest, initialHash, size);
  return (finalHash == Checksum) ? LOAD_OK : LOAD_ERROR;
}



LOADSTATUS CCSBuildDoc::ReadAndUnscrambleBlockBigEndian(CFile *f, ui16 *buf)
{
  ui16 hash, checkSum;
  int i;
  UINT bytesRead;
  bytesRead = f->Read(buf, 512);
  if (bytesRead != 512) return LOAD_ERROR;
  // Form a 16-bit hash code from the first 256 bytes.
  for (i=0, hash=0; i<128; i+=4)
  {
    hash = (ui16)(hash + LE16(buf[i]));
    hash ^= LE16(buf[i+1]);
    hash = (ui16)(hash - LE16(buf[i+2]));
    hash ^= LE16(buf[i+3]);
  };
  Unscramble(buf+128, LE16(buf[29]), 128);
  checkSum = 0;
  for (i=128; i<256; i++)
  {
    checkSum = (ui16)(checkSum + LE16(buf[i]));
  };
  if (hash != checkSum)
  {
    Unscramble(buf+128, LE16(buf[29]), 128); // Put things back as they were
    Unscramble(buf+128, LE16(buf[10]), 128); // Try the AtariDM method.
    checkSum = 0;
    for (i=128; i<256; i++)
    {
      checkSum = (ui16)(checkSum + LE16(buf[i]));
    };
    if (hash == checkSum) 
    {
      m_PCFormattedBlock1 = true;
    };
  };
  return (hash == checkSum) ? LOAD_OK : LOAD_ERROR;
}

LOADSTATUS CCSBuildDoc::ReadAndUnscrambleBlockLittleEndian(CFile *f, ui16 *buf)
{
  ui16 hash, checkSum;
  int i;
  UINT bytesRead;
  bytesRead = f->Read(buf, 512);
  if (bytesRead != 512) return LOAD_ERROR;
  // Form a 16-bit hash code from the first 256 bytes.
  for (i=0, hash=0; i<128; i+=4)
  {
    hash = (ui16)(hash + (buf[i]));
    hash ^= (buf[i+1]);
    hash = (ui16)(hash - (buf[i+2]));
    hash ^= (buf[i+3]);
  };
  Unscramble(buf+128, (buf[10]), 128);
  checkSum = 0;
  for (i=128; i<256; i++)
  {
    checkSum = (ui16)(checkSum + (buf[i]));
  };
  return (hash == checkSum) ? LOAD_OK : LOAD_ERROR;
}

void CCSBuildDoc::extractBlock1Data(GAMEBLOCK1 *pBlock)
{
  m_totalMoveCount = pBlock->totalMoveCount;
  m_RandomGameID = pBlock->RandomGameID;
  memcpy(m_Byte22808, pBlock->Byte22808, 132);//All zeroes in my CSBGAME2.DAT
  if (m_littleEndian)
  {
    pBlock->Byte22596 = 2;
    pBlock->Word22594 = 1;
  };
}

void CCSBuildDoc::extractPCBlock1Data(PCGAMEBLOCK1 *pBlock)
{
  m_totalMoveCount = pBlock->totalMoveCount;
  m_RandomGameID = pBlock->RandomGameID;
  memcpy(m_Byte22808, pBlock->Byte22808, 132);//All zeroes in my CSBGAME2.DAT
  if (m_littleEndian)
  {
    pBlock->Byte22596 = 2;
    pBlock->Word22594 = 1;
  };
}

void CCSBuildDoc::initializeBlock1Data(void)
{
  m_totalMoveCount = 0;
  m_RandomGameID = 0;
  memcpy(m_Byte22808, m_block1.Byte22808, 132);//All zeroes in my CSBGAME2.DAT
}

void CCSBuildDoc::extractBlock2Data(void)
{
  m_Time = m_Long11732 = m_block2.Time;
  m_RandomNumber = m_block2.ranseed;
  m_NumCharacter = m_block2.numcharacter;
  m_partyX = m_block2.partyx;
  m_partyY = m_block2.partyy;
  m_partyFacing = m_block2.partyfacing;
  m_partyLevel = m_block2.partyLevel;
  m_HandChar = m_block2.handChar; // chIdx
  m_MagicCaster = m_block2.MagicCaster;
  m_numTimer = m_block2.NumTimer;
  m_NextAvailTimer = m_block2.NextAvailTimer;
  m_MaxTimers = m_block2.MaxTimers;
  m_ITEM16QueLen = m_block2.ITEM16QueLen;
  m_Long12950 = m_block2.Long12950;
  m_Long12954 = m_block2.Long12954;
  m_Word11710 = m_block2.Word11710;
  m_Word11712 = m_block2.Word11712;
  m_Word11714 = m_block2.Word11714;
  m_objectInHand.NonExtendedInteger(m_block2.ObjectInHand);
  m_MaxITEM16 = m_block2.MaxITEM16;
}

void CCSBuildDoc::initializeBlock2Data(void)
{
  m_Time = m_Long11732 = 0;
  m_RandomNumber = 0;
  m_NumCharacter = 0;
  m_partyX = 0;
  m_partyY = 0;
  m_partyFacing = 0;
  m_partyLevel = 0;
  m_HandChar = 0; // chIdx
  m_MagicCaster = 0;
  m_numTimer = 0;
  m_NextAvailTimer = 0;
  m_MaxTimers = 0;
  m_ITEM16QueLen = 0;
  m_Long12950 = 0;
  m_Long12954 = 0;
  m_Word11710 = 0;
  m_Word11712 = 0;
  m_Word11714 = 0;
  m_objectInHand = RNnul;;
  m_MaxITEM16 = 0;
}

void CCSBuildDoc::InitializeItem16(void)
{
  i32 i;
  if (m_dungeonFormat) m_MaxITEM16 = 60;
  m_Item16 = (ITEM16 *)malloc(16*m_MaxITEM16, MALLOC007);
  for (i=0; i < m_MaxITEM16; i++) m_Item16[i].word0 = -1;
}

void CCSBuildDoc::InitializeTimers(void)
{
  if (m_Timers != NULL) free(m_Timers);
  if (m_timerQue != NULL) free(m_timerQue);
  m_Timers = (TIMER *)malloc(m_MaxTimers*10, MALLOC008);
  memset(m_Timers,0,m_MaxTimers*10);
  m_timerQue = (ui16 *)malloc(m_MaxTimers*2, MALLOC009);
  memset(m_timerQue,0,m_MaxTimers*2);
  //if (m_PrisonFile != 0)
  //{
  //  D7L = 0;
  //  for (pA3=d.Timers; D7W<d.MaxTimers; D7W++, pA3++)
  //  {
  //    pA3->function = 0;
  //  };
  //  d.numTimer = 0;
  //  d.NextAvailTimer = 0;
    //SetWatchdogTimer();
  //};
}

void CCSBuildDoc::InitializeCharacters(void)
{
  int i;
  m_NumCharacter = 0;
  for (i=0; i<4; i++)
  {
    memset(&m_Characters[i], 0, sizeof (m_Characters[i]));
  };
}

LOADSTATUS CCSBuildDoc::StreamInput(
                     CFile *f, ui8 *buf, int numByte)
{
  int numRead;
  if (numByte == 0) return LOAD_OK;
  if (m_compressed)
  {
    if (numByte > m_inStreamLeft) return LOAD_ERROR;
    memcpy(buf, m_inStreamNext, numByte);
    m_inStreamNext += numByte;
    m_inStreamLeft -= numByte;
    return LOAD_OK;
  };
  numRead = f->Read(buf, numByte);
  return numRead==numByte ? LOAD_OK : LOAD_ERROR;
}



LOADSTATUS CCSBuildDoc::FetchDataBytes(
                  CFile *f, ui8 *buf,i16 *checksum, i32 numByte)
{
  LOADSTATUS status;
  status = StreamInput(f, buf, numByte);
  if (status != LOAD_OK) return status;
  while (numByte != 0)
  {
    numByte--;
    *checksum = (i16)(*checksum + *buf);
    buf++;
  };
  return LOAD_OK;
}

struct DBHEADER
{
  ui16   sentinel;
  ui32   fileSize; //Not counting these 8 bytes
  ui16   uWord6;
};

LOADSTATUS CCSBuildDoc::readOldScrollDatabase(
                               CFile *f, i16 *pCheckSum)
{
  int entrySize;
  int i;
  ui8 *pOld, *pNew;
  int numEnt;
  LOADSTATUS status;
  DBCOMMON *pDB;
  entrySize = m_DBEntrySize[dbSCROLL];
  numEnt = m_dungeonDatIndex->DBSize(dbSCROLL);
  db.Allocate(dbSCROLL, numEnt);
  pDB = db.GetCommonAddress(DBTYPE(dbSCROLL),0);
  status = FetchDataBytes(
                f,
                (ui8 *)pDB, 
                pCheckSum, 
                (ui16)4*numEnt);
  if (status != LOAD_OK) return status;
#ifdef THERON
  {
    int i;
    ui16 *pScroll;
    FILE *fs;
    fs = fopen("db2.txt","a");
    for (i=0; i<m_dungeonDatIndex->DBSize(dbSCROLL); i++)
    {
      pScroll = (ui16 *)pDB + 2 * i;
      fprintf(fs,"Scroll[%03x]= %04x %04x",i,LE16(pScroll[0]),LE16(pScroll[1]));
      fprintf(fs," index = %03x\n",LE16(pScroll[1])&0x3f);
    };
    fclose(fs);
  };
#endif
  //Now make the entries the correct size.
  for (i=numEnt-1; i>=0; i--)
  {
    pOld = ((ui8 *)pDB) + 4*i;
    if (m_littleEndian)
    {
      *(ui16 *)pOld = (*(ui16 *)pOld); //The link
      *(ui16 *)(pOld+2) = (*(ui16 *)(pOld+2)); //The text index and open flag.
    }
    else
    {
      *(ui16 *)pOld = LE16(*(ui16 *)pOld); //The link
      *(ui16 *)(pOld+2) = LE16(*(ui16 *)(pOld+2)); //The text index and open flag.
    };
    pNew = ((ui8 *)pDB) + entrySize*i;
    memmove(pNew, pOld, 4);
  };
  return LOAD_OK;
}

void CCSBuildDoc::convertScrolls()
{
  RN RNtext;
  int numEnt;
  int i, textIndex;
  int openFlag;
  DB7 *pScroll;
  DB2 *pText;
  ui16 *pWord;
  int indirectIndex;
  bool warningIssued = false;
  bool nonEOFlinkWarning = false;
  numEnt = m_dungeonDatIndex->DBSize(dbSCROLL);
  for (i=0; i<numEnt; i++)
  {
    pScroll = db.GetCommonAddress(dbSCROLL,i)->CastToDB7();
    //The link is OK.
    pWord = (ui16 *)pScroll;
    if (pWord[0] == RNnul) continue;
    textIndex = pWord[1] & 0x3ff;
    if (textIndex >= db.NumEntry(dbTEXT))
    {
      if (!warningIssued)
      {
        Warning("One or more scrolls has illegal Text Index.\n"
                "I am going to set them to RNnul for now.\n"
                "Later, I will go through the dungeon and\n"
                "delete these scrolls if they are actually being\n"
                "used.  At that time I can tell you where\n"
                "they are.");
        warningIssued = true;
      };
      pScroll->text(RNnul);
      pScroll->open(false);
      continue;;
    };
    indirectIndex = AllocateIndirectIndex(dbTEXT,pWord[1]&0x3ff);
    openFlag = pWord[1]>>10;
    RNtext.ConstructFromInteger(indirectIndex);
    pScroll->text(RNtext);
    pText = db.GetCommonAddress(dbTEXT, textIndex)->CastToDB2();
    if (pText->link() != RNeof)
    {
      if (!nonEOFlinkWarning)
      {
        Warning("One or more Scrolls has text with\n"
                "a forward pointer.  A Scroll should not\n"
                "have more than one text entry.\n"
                "I am going to clear all such pointers.");
        nonEOFlinkWarning = true;
      };
      pText->link(RNeof);
    };
    pScroll->open(openFlag==0);
  };
}


LOADSTATUS CCSBuildDoc::readNewScrollDatabase(
                               CFile *f, i16 *pCheckSum)
{
  int entrySize;
  DBCOMMON *pDB;
  LOADSTATUS status;
  entrySize = m_DBEntrySize[dbSCROLL];
  db.Allocate(dbSCROLL, m_dungeonDatIndex->DBSize(dbSCROLL));
  if (m_dungeonDatIndex->DBSize(dbSCROLL) != 0)
  {
    pDB = db.GetCommonAddress(DBTYPE(dbSCROLL),0);
    status = FetchDataBytes(
                  f,
                  (ui8 *)pDB, 
                  pCheckSum, 
                  (ui16)entrySize * m_dungeonDatIndex->DBSize(dbSCROLL));
    if (status != LOAD_OK) return status;
    if (!m_littleEndian) db.swap(dbSCROLL);// database
  };
  return LOAD_OK;
}

LOADSTATUS CCSBuildDoc::readScrollDatabase(
                              CFile *f, i16 *pCheckSum)
{
  //If extendedFeatures Version is less than 'B' then
  //we are going to read the old version of scrolls
  //database and convert to the new version.
  //The old version had a 10-bit index into the
  //text database.  The new version has an RN reference
  //to the text.
  if (m_dungeonDatIndex->DBSize(dbSCROLL) == 0) return LOAD_OK;
  if (ExtendedFeaturesVersion < 'B') 
       return readOldScrollDatabase(f, pCheckSum);
  else return readNewScrollDatabase(f, pCheckSum);
}

void CCSBuildDoc::InitializeDungeonDatIndex(void)
{
  int i;
  m_dungeonDatIndex->Sentinel(0);
  NumLevel(1);
  for (i=0; i<16; i++) m_dungeonDatIndex->DBSize(i,0);
  m_dungeonDatIndex->LegacyCellFlagArraySize(0);
  CellFlagArraySize(1); //Used to be in dungeonDatIndex.
  m_dungeonDatIndex->NumWordsInTextArray(0);
  m_dungeonDatIndex->StartingPartyX(0);
  m_dungeonDatIndex->StartingPartyY(0);
  m_dungeonDatIndex->StartingPartyFacing(0);
  m_dungeonDatIndex->ObjectListLength(0);
}

LOADSTATUS CCSBuildDoc::ConvertToIndirectText(ui16 *text)
{
  int i, len, oldIndex, newIndex;
  int indirectIndexLen;
  DB2 *pDB2;
  if (m_indirectTextIndex != NULL)
  {
    free(m_indirectTextIndex);
    m_indirectTextIndex = NULL;
  };
  indirectIndexLen = 0;
  if (m_compressedText != NULL)
  {
    free(m_compressedText);
    m_compressedText = NULL;
  };
  m_sizeOfCompressedText = 0;
  for (i=0; i<db.NumEntry(dbTEXT); i++)
  {
    pDB2 = GetRecordAddressDB2(i);
    if (pDB2->link() == RNnul) continue;
    oldIndex = pDB2->index();
    if (oldIndex > NumWordsInTextArray())
    {
      Strange("A Text Record has an illegal text index");
      pDB2->index(0);
      continue;
    };
    len = ComputeTextLength(text+oldIndex);
    //Add one entry to m_indirectTextIndex;
    newIndex = indirectIndexLen;
    indirectIndexLen++;
    m_indirectTextIndex 
         = (int *)realloc(m_indirectTextIndex,
                          4*((indirectIndexLen+99)/100 * 100),
                          MALLOC071);
    m_indirectTextIndex[newIndex] = m_sizeOfCompressedText;
    m_sizeOfCompressedText += len;
    m_compressedText
         = (ui16 *)realloc(m_compressedText,
                           2*((m_sizeOfCompressedText+199)/200 * 200),
                           MALLOC072);
    memcpy(m_compressedText+m_indirectTextIndex[newIndex],
           text + oldIndex,
           2*len);
    pDB2->index(newIndex);
  };
  m_dungeonDatIndex->NumWordsInTextArray(indirectIndexLen);
  return LOAD_OK;
}

class SAFEMEM
{
public:
  ui16 *p;
  SAFEMEM(void){p=NULL;};
  ~SAFEMEM(void){if (p!=NULL) free(p);};
};

void CCSBuildDoc::MakeBigActuators()
{
  char *source, *destination;
  int oldSize, newSize, numEnt;
  int i, j;
  DB3 *pActuator;
  numEnt = m_dungeonDatIndex->DBSize(dbACTUATOR);
  newSize = m_DBEntrySize[dbACTUATOR];
  oldSize = newSize - 2;
  source = destination = (char *)db.GetCommonAddress(dbACTUATOR,0);
  source += numEnt * oldSize;
  destination += numEnt * newSize;
  for (i=0; i<numEnt; i++)
  {
    //insert two zero bytes.
    *(--destination) = 0;
    *(--destination) = 0;
    //copy oldSize bytes;
    for (j=0; j<oldSize; j++)
    {
      *(--destination) = *(--source);
    };
  };
  for (i=0; i<numEnt; i++)
  {
    pActuator = (DB3 *)db.GetCommonAddress(dbACTUATOR,i);
    pActuator->MakeBig();
  };
}


LOADSTATUS CCSBuildDoc::ReadDatabases(CFile *f)
{
  i32 numLevel;
  int i, level;
  int dbNum;
  int origNumEnt;
  int origDungeonDatIndex5;
  int entrySize;
  int size10450;
  int offsetOfCell;
  int numRow;
  LEVELDESC *pA0;
  DBCOMMON *pDB;
  CELLFLAG *pdD0;
  CELLFLAG **ppdA2;
  //i8  LOCAL_20[8];
  //DBHEADER local20; //For compressd files
  //i32 filesize; // LOCAL_12 //For compressed files
  i16 CheckSum;
  i16 LOCAL_6;
  CELLFLAG *LOCAL_4;
  LOADSTATUS status;
  SAFEMEM textArray;
  CheckSum = 0;
  if (m_dungeonDatIndex != NULL) free(m_dungeonDatIndex);
  m_dungeonDatIndex = (DUNGEONDATINDEX *)malloc(44, MALLOC010);
  if (f != NULL)
  {
    status = FetchDataBytes(f, (ui8 *)m_dungeonDatIndex, &CheckSum, 44);
  }
  else
  {
    InitializeDungeonDatIndex();
    status = LOAD_OK;
  };
  if (status != LOAD_OK) return status;
  if (f != NULL)
  {
    swapDungeonDatIndex();
  };
  if (m_dungeonFormat)
  {
    m_partyX = m_dungeonDatIndex->StartingPartyX();
    m_partyY = m_dungeonDatIndex->StartingPartyY();
    m_partyFacing = m_dungeonDatIndex->StartingPartyFacing();
    m_partyLevel = 0;
  };
  numLevel = m_dungeonDatIndex->NumLevel();;
  if (CellFlagArraySize() == 0)
  {
    CellFlagArraySize(m_dungeonDatIndex->LegacyCellFlagArraySize());
  };
  if (numLevel > 64)
  {
    Error("Number of Levels > 64");
    return LOAD_ERROR;
  };
  if (m_pLevelDescriptors != NULL) free(m_pLevelDescriptors);
  m_pLevelDescriptors =
           (LEVELDESC *)malloc(16 * 64 /*numLevel*/, MALLOC011);
  if (f!= NULL)
  {
    status = FetchDataBytes(f, 
                            (ui8 *)m_pLevelDescriptors,
                            &CheckSum,
                            16 * numLevel);
    if (status != LOAD_OK) return status;
    // We promised to littleEndian words 0, 8, 10, 12, and 14 of
    // each of the 16-byte structures.
    if (!m_littleEndian) swapLevelDescriptors(numLevel);
  }
  else
  {
    m_pLevelDescriptors[0].Level(0);
    m_pLevelDescriptors[0].CellOffset(0);
    m_pLevelDescriptors[0].OffsetX(0);
    m_pLevelDescriptors[0].OffsetY(0);
    m_pLevelDescriptors[0].LastColumn(0);
    m_pLevelDescriptors[0].LastRow(0);
    m_pLevelDescriptors[0].NumWallDecoration(0);
    m_pLevelDescriptors[0].NumFloorDecorations(0);
    m_pLevelDescriptors[0].NumDoorDecorations(0);
    m_pLevelDescriptors[0].NumRandomFloorDecorations(0);
    m_pLevelDescriptors[0].NumRandomWallDecorations(0);
    m_pLevelDescriptors[0].numMonsterType(0);
    m_pLevelDescriptors[0].experienceMultiplier(0);
    m_pLevelDescriptors[0].Door1Type(0);
    m_pLevelDescriptors[0].Door2Type(1);
    m_pLevelDescriptors[0].backgroundGraphicSet(0);
    m_pLevelDescriptors[0].wallGraphicSet(0);
  };
  for (level=0; level<numLevel; level++)
  {
    if (m_pLevelDescriptors[level].Level() != level)
    {
      int result;
      CString msg;
      msg.Format("Level descriptor for level %d\n"
                 "says it is for level %d\n"
                 "Should We fix this?",
                 level,m_pLevelDescriptors[level].Level());
      result = MessageBox(NULL,msg,"Warning",MB_YESNO|MB_ICONWARNING);
      if (result == IDYES)
      {
        m_pLevelDescriptors[level].Level(level);
      };
    };
  };
  if (m_objectLevelIndex != NULL) free (m_objectLevelIndex);
  m_objectLevelIndex = (ui16 *)malloc(2*numLevel, MALLOC012);
  m_numColumnPointers = 0;
  for (i=0; i<numLevel; i++)
  {
    m_objectLevelIndex[i] = (ui16)m_numColumnPointers;
    m_numColumnPointers += m_pLevelDescriptors[i].LastColumn() + 1;
  };
  origDungeonDatIndex5 = m_dungeonDatIndex->ObjectListLength();
  if (m_dungeonFormat != 0)
  {
    //m_dungeonDatIndex->ObjectListLength(m_dungeonDatIndex->ObjectListLength() + 300);
  };
  if (m_objectColumnIndex != NULL) free(m_objectColumnIndex);
  if (m_objectList        != NULL) free(m_objectList);
  if (m_indirectTextIndex != NULL) free(m_indirectTextIndex);
  if (m_compressedText    != NULL) free(m_compressedText);
  m_objectColumnIndex = (ui16 *)malloc(m_numColumnPointers*2, MALLOC013);
  m_objectList        = (RN *)malloc(m_dungeonDatIndex->ObjectListLength()*2,MALLOC014);
  m_indirectTextIndex = (int *)malloc(m_dungeonDatIndex->NumWordsInTextArray()*4,MALLOC015);
  m_compressedText    = NULL;
  m_sizeOfCompressedText = 0;
  if (f!= NULL)
  {
    status = FetchDataBytes(
                  f, 
                  (ui8 *)m_objectColumnIndex,
                  &CheckSum,
                  m_numColumnPointers*2);
    if (status != LOAD_OK) return status;
    if (!m_littleEndian) swapPointer10454();
  }
  else
  {
    m_objectColumnIndex[0] = 0;
  };
  if (f != NULL)
  {
    status = FetchDataBytes(
                  f,
                  (ui8 *)m_objectList,
                  &CheckSum,
                  origDungeonDatIndex5*2);
    if (status != LOAD_OK) return status;
    if (!m_littleEndian) swapObjectList(origDungeonDatIndex5);
  }
  else
  {
  };
  if (f != NULL)
  {
    if (!m_indirectText)
    { 
      textArray.p = (ui16 *)malloc(NumWordsInTextArray()*2,MALLOC070);
      status = FetchDataBytes(
                    f,
                    (ui8 *)textArray.p, 
                    &CheckSum, 
                    NumWordsInTextArray()*2);
      if (status != LOAD_OK) return status;
      if (m_littleEndian) swapTextArray(textArray.p);
#ifdef THERON
      {
        int i, j, word, iword, ibyte;
        char text[80];
        FILE *txt;
        txt = fopen("db2.txt","w");
        for (i=0; i<NumWordsInTextArray()-3; i++)
        {
          j=0;
          for (iword=0; iword<3; iword++)
          {
            word = LE16(((ui16 *)(textArray.p))[i+iword]);
            for (ibyte=10; ibyte>=0; ibyte-=5)
            {
              text[j++] = (char)(((word>>ibyte)&0x1f) + 'A');
            };
          };
          text[j] = 0;
          fprintf(txt,"TextIndex=%03x  %s\n",
                      i, text);
        };
        fprintf(txt,"\n\n\n");
        fclose(txt);
      };
#endif
    }
    else //if m_indirectText
    {//nothing is swapped!
      //fetch indirecttext(NumWordsInTextArray);
      status = FetchDataBytes(
                    f,
                    (ui8 *)m_indirectTextIndex, 
                    &CheckSum, 
                    NumWordsInTextArray()*4);
      if (status != LOAD_OK) return status;
        //fetch sizeof compressedText;
      if (m_compressedText != NULL)
      {
        free (m_compressedText);
        m_compressedText = NULL;
        m_sizeOfCompressedText = 0;
      };
      status = FetchDataBytes(
                    f,
                    (ui8 *)&m_sizeOfCompressedText, 
                    &CheckSum, 
                    4);
      //fetch compressedText;
      m_compressedText = (ui16 *)malloc(m_sizeOfCompressedText*2,
                                        MALLOC073);
      status = FetchDataBytes(
                    f,
                    (ui8 *)m_compressedText, 
                    &CheckSum, 
                    m_sizeOfCompressedText*2);
    }
  }
  else //f==NULL
  {
    m_sizeOfCompressedText = 0;
  };
  if (m_dungeonFormat) 
  {
    //m_MaxTimers = 100;
  };
  for (dbNum=0; dbNum<16; dbNum++)
  {
    origNumEnt = m_dungeonDatIndex->DBSize(dbNum);  // number of entries
    if (m_dungeonFormat)
    {
      //D0W = origNumEnt;
      //A0 = d.Byte7302 + dbNum;
      //D3W = *((ui8 *)A0);
      //D0W = sw(D0W + D3W);
      //D0W = sw(Smaller(dbNum==15 ? 768 : 1024, D0W));
      //d.dungeonDatIndex->DBSize(dbNum, D0W);
    };
    if (dbNum == dbSCROLL)
    {
      status = readScrollDatabase(f, &CheckSum);
      if (status != LOAD_OK) return status;
      continue;
    };
    entrySize = m_DBEntrySize[dbNum];
    if ((dbNum == dbACTUATOR) && !m_bigActuators)
    {
      entrySize -= 2;
    };
    // Always allocate!!  If it is already allocated then
    // it will get released and reallocated to the
    // correct size!  This was a bad bug that made
    // some saved games unloadable.
    //if (d.Word22584 == 0)
    {
      //D0L = d.dungeonDatIndex[dbNum+6] * entrySize;
      //D0L &= 0xffff;
      //if (D0L != 0) A0 = allocateMemory(D0L,1);
      //else A0 = NULL;
      //d.misc1052eight[D6W] = (UNKNOWN *)A0;
      db.Allocate(dbNum, m_dungeonDatIndex->DBSize(dbNum));
      /*
      if (m_dungeonDatIndex->DBSize(dbNum) != 0)
      {
        m_misc10528[dbNum] = db.GetCommonAddress(RN(0,dbNum,0));
      }
      else
      {
        m_misc10528[dbNum] = NULL;
      };
      */
    };
    if (m_dungeonDatIndex->DBSize(dbNum) != 0)
    {
      pDB = db.GetCommonAddress(DBTYPE(dbNum),0);
      status = FetchDataBytes(
                    f,
                    (ui8 *)pDB, 
                    &CheckSum, 
                    (ui16)entrySize * m_dungeonDatIndex->DBSize(dbNum));
      if (status != LOAD_OK) return status;
#ifdef THERON
      if ( (dbNum == dbTEXT) && (textArray.p!=NULL))
      {
        int dbindex, j, iword, ibyte, word;
        int textIndex;
        char text[80];
        FILE *db2;
        db2 = fopen("db2.txt","a");
        for (dbindex=0; dbindex<m_dungeonDatIndex->DBSize(dbNum); dbindex++)
        {
          pDB = db.GetCommonAddress(DBTYPE(dbNum),dbindex);
          textIndex = LE16(((i16 *)pDB)[1]) >> 3;
          j=0;
          for (iword=0; iword<10; iword++)
          {
            word = LE16(((ui16 *)(textArray.p))[textIndex+iword]);
            for (ibyte=10; ibyte>=0; ibyte-=5)
            {
              text[j++] = (char)(((word>>ibyte)&0x1f) + 'A');
            };
          };
          text[j] = 0;
          fprintf(db2,"TextRecord[%02x] textIndex=%02x %s\n",
                      dbindex, textIndex, text);
        };
        fprintf(db2,"\n\n\n");
        fclose(db2);
      };
#endif
      if (  (dbNum == dbACTUATOR) && !m_bigActuators)
      {
        MakeBigActuators();
      };
      if (!m_littleEndian) db.swap(dbNum);// database
    };
    if (m_dungeonFormat != 0)
    {
      //if ((dbNum == dbMONSTER) || (dbNum >= dbMISSILE))
      //{
      //  m_MaxTimers = (i16)(m_MaxTimers + m_dungeonDatIndex->DBSize(dbNum));
      //};
      /*
      All the enclosed code is to clear the extra
      entries that we allocated but did not initialize
      by reading from the file.
      But the new db.Allocate clears all the entries.
      Therefore, this code is not needed.  And that is
      nice because it is quite ugly.
      D1W = entrySize >> 1;
      D4W = D1W;
      LOCAL_30 = D1W;
      D1L = D5UW * (ui16)LOCAL_30;
      //D1 <<= 1;
      pwA3 += D1L;
      D5W = (UI8)(m_Byte7302[dbNum]); //# additional entries
      while (D5W != 0)
      {
        pwA3[0] = -1;
  //      D0W = D5W;
        D5W--;
        D1W = D4W;
        D1H1 = 0;
        //D1 <<= 1;
        pwA3 += D1L;

      };
      */
    };
    if (dbNum == dbEXPOOL) expool.Setup();
  }; //for
  if (!m_indirectText)
  {
    ConvertToIndirectText(textArray.p);
  };
  if (m_cellFlagArray != NULL) free(m_cellFlagArray);
  m_cellFlagArray = (CELLFLAG *)malloc(CellFlagArraySize(),MALLOC016);
  if (f!= NULL)
  {
    status = FetchDataBytes(
                  f,
                  (ui8 *)m_cellFlagArray,
                  &CheckSum,
                  CellFlagArraySize());
    if (status != LOAD_OK) return status;
  }
  else
  {
    m_cellFlagArray[0] = roomOPEN<<5;
  };
  if (f!= NULL)
  {
    status = StreamInput(f, (ui8 *)&LOCAL_6, 2); // Read without updating checksum
    if (!m_littleEndian) LOCAL_6 = LE16(LOCAL_6);
    if ((status == LOAD_OK) && (LOCAL_6!=CheckSum)) return LOAD_ERROR;
  };

  size10450 = 4 *(m_numColumnPointers + numLevel);
  // D7W is number of index pointers at the
  //  front of d.10450.
  if (m_pppdPointer10450 != NULL) free(m_pppdPointer10450);
  m_pppdPointer10450 = (CELLFLAG ***)malloc(size10450,MALLOC017);
  ppdA2 = (CELLFLAG **)&m_pppdPointer10450[numLevel];
    // Need cast because the 10450 array is used for
    // two things.  See comments in data definitions.
  for (level=0; level < numLevel; level++)
  {
    m_pppdPointer10450[level] = ppdA2;
    //pntGear(A0) = A2;
    pdD0 = m_cellFlagArray;
                                      //pA0 = &m_pLevelDescriptors[level];
    offsetOfCell = CellOffset(level); //offsetOfCell = pA0->cellOffset;
    //D3H1 = 0;

    pdD0 = pdD0 + offsetOfCell; // Add Byte offset
    //pdD0 = (CELLFLAG *)((i32)pdD0 + D3L); // Add Byte offset
    LOCAL_4 = pdD0;
    *ppdA2 = pdD0;
    ppdA2++;
    for (i32 column=1;
         column<= m_pLevelDescriptors[level].LastColumn();
         column++)
    {
      pA0 = &m_pLevelDescriptors[level];
      numRow = pA0->LastRow()+1;
      pdD0 = LOCAL_4;
      pdD0 = pdD0 + numRow; //add byte offset
      //pdD0 = (DUDAD16 *)((i32)pdD0+D1L); //add byte offset
      LOCAL_4 += numRow;
      *(ppdA2++) = LOCAL_4;
    }; // for column
  };/// for level
  return LOAD_OK;
}


bool CCSBuildDoc::AddLevel(void)
{
  int level;
  int i;
  unsigned int oldCellArraySize, newCellArraySize;
  int offset;
  FLOORTYPE floor;
  ORNATETYPE ornate;
  WALLTYPE wall;
  MONSTERTYPE monster;
  CELLFLAG *newCellArray, *pntr, *pCell;
  CELLFLAG **pColumnPointer;
  if (NumLevel() == 64) return true;
  //for (level=NumLevel(); level < 16; level++)
  level = NumLevel();
  {
    // First
    // We are going to add some bytes to the 
    // end of the CELLFLAG array.
    // We wll add:
    // 2 monsters           total 2
    // 2 wall decorations   total 4
    // 2 floor decorations  total 6
    // 2 door decorations   total 8
    // 1 CELLFLAG           total 9
    // Now we know that this is not a lot of bytes.  But
    // we are going to have to do a lot of work
    // to add those bytes.
    //
    oldCellArraySize = CellFlagArraySize();
    newCellArraySize = oldCellArraySize + 9;
    newCellArray = (CELLFLAG *)malloc(newCellArraySize,MALLOC018);
    if (newCellArray == NULL)
    {
      Sorry("Cannot allocate memory for additional levels");
      return true;
    };
    offset = newCellArray-m_cellFlagArray;
    memcpy(newCellArray, m_cellFlagArray, oldCellArraySize);
    free(m_cellFlagArray);
    m_cellFlagArray = newCellArray;
    //
    // Now fill in the monsters, etc.
    //
    pntr = m_cellFlagArray + oldCellArraySize;
    *(pntr++) = 0x28;  //An open room with random floor
    //
    // Add two monsters
    //
    monster = "Screamer";
    *(pntr++) = monster.Integer();
    monster = "Rock_Pile";
    *(pntr++) = monster.Integer();
    //
    // Add two Wall Decorations
    //
    wall = "SLIME";
    *(pntr++) = wall.Integer();
    wall = "FOUNTAIN";
    *(pntr++) = wall.Integer();
    //
    // Add two Floor Decorations
    //
    floor = "WATER PUDDLE";
    *(pntr++) = floor.Integer();
    floor = "SQUARE DRAIN";
    *(pntr++) = floor.Integer();
    //
    // Add two Door decorations
    //
    ornate = "METAL BARS";
    *(pntr++) = ornate.Integer();
    ornate = "WOOD VEN";
    *(pntr++) = ornate.Integer();
    //
    // Set new Cell Flag Array Size
    //
    CellFlagArraySize(newCellArraySize);
    //
    // We fix the LEVELDESC (Level Descriptor)
    // for this new level.
    //
    m_pLevelDescriptors[level].LastColumn(0);
    m_pLevelDescriptors[level].LastRow(0);
    m_pLevelDescriptors[level].OffsetX(0);
    m_pLevelDescriptors[level].OffsetY(0);
    m_pLevelDescriptors[level].Level(level);
    m_pLevelDescriptors[level].NumWallDecoration(2);
    m_pLevelDescriptors[level].NumRandomWallDecorations(1);
    m_pLevelDescriptors[level].NumFloorDecorations(2);
    m_pLevelDescriptors[level].NumRandomFloorDecorations(1);
    m_pLevelDescriptors[level].NumDoorDecorations(2);
    m_pLevelDescriptors[level].numMonsterType(2);
    m_pLevelDescriptors[level].experienceMultiplier(1);
    m_pLevelDescriptors[level].CellOffset(oldCellArraySize);
    m_pLevelDescriptors[level].Door1Type(0);
    m_pLevelDescriptors[level].Door2Type(1);
    m_pLevelDescriptors[level].backgroundGraphicSet(0);
    m_pLevelDescriptors[level].wallGraphicSet(0);

    //
    //
    // We have to fix all the pointers that
    // used to point into m_cellFlagArray.
    //
    free (m_pppdPointer10450);
    m_numColumnPointers++; //One column in new level.
    m_pppdPointer10450 = (CELLFLAG ***)malloc(4*(m_numColumnPointers + level+1),MALLOC019);
    pColumnPointer = (CELLFLAG **)m_pppdPointer10450 + level + 1; 
                   //Where first column pointer goes
    //ppdA2 = (CELLFLAG **)&m_pppdPointer10450[numLevel];
    // Need cast because the 10450 array is used for
    // two things.  See comments in data definitions.
    for (i=0; i<=level; i++)
    {
      m_pppdPointer10450[i] = pColumnPointer;
      pCell = m_cellFlagArray + CellOffset(i); // First cell in dungeon
      for (i32 column=0;
           column<= m_pLevelDescriptors[i].LastColumn();
           column++)
      {
        *(pColumnPointer++) = pCell;
        pCell += LastRow(i) + 1; //add one column of cells
      }; // for column
    };/// for i=1 to level
    free (m_objectLevelIndex);
    m_objectLevelIndex = (ui16 *)malloc(2*(level+1),MALLOC020);
    if (m_objectLevelIndex == NULL)
    {
      DIE("Cannot allocate memory");
    };
    offset = 0;
    for (i=0; i<=level; i++)
    {
      m_objectLevelIndex[i] = (ui16)offset;
      offset += LastColumn(i) + 1;
    };
    m_objectColumnIndex = 
          (ui16 *)realloc(m_objectColumnIndex,2*m_numColumnPointers,MALLOC001);
    if (m_objectColumnIndex == NULL)
    {
      DIE("Cannot allocate memory");
    };
    m_objectColumnIndex[m_objectLevelIndex[level]] = 
           m_objectColumnIndex[m_objectLevelIndex[level-1]+LastColumn(level-1)];
    for (i=0; i<=LastRow(level-1); i++)
    {
      if (GetCellFlag(level-1,LastColumn(level-1),i) &0x10)
      {
        m_objectColumnIndex[m_objectLevelIndex[level]]++;
      };
    };
    //
    // Last, we set the new number of levels.
    //
    NumLevel(NumLevel() + 1);
  };
  SetModifiedFlag();
  return VerifyLevelData();
}

bool CCSBuildDoc::AddColumn(int level)
{
  int i;
  unsigned int oldCellArraySize, newCellArraySize;
  int newColumnOffset;
  int newOffset, oldOffset;
  ui16 *oldColIndex, *newColIndex;
  FLOORTYPE floor;
  CELLFLAG *newCellArray, *pCell;
  CELLFLAG **pColumnPointer;
  if (level >= NumLevel()) return true;
  if (LastColumn(level) == 31) return true;
  {
    // First
    // We are going to increase the size of the
    // cellflag array by the height of the current
    // level.
    oldCellArraySize = CellFlagArraySize();
    newCellArraySize = oldCellArraySize + LastRow(level) + 1;
    newCellArray = (CELLFLAG *)malloc(newCellArraySize,MALLOC028);
    if (newCellArray == NULL)
    {
      Sorry("Cannot allocate memory for additional column");
      return true;
    };
    newColumnOffset = CellOffset(level)
                      + (LastRow(level)+1) * (LastColumn(level)+1);
    memcpy(newCellArray, 
           m_cellFlagArray, 
           newColumnOffset);
    memset(newCellArray+newColumnOffset,
           0,
           LastRow(level)+1);
    memcpy(newCellArray+newColumnOffset+LastRow(level)+1,
           m_cellFlagArray+newColumnOffset,
           oldCellArraySize-newColumnOffset);
    free(m_cellFlagArray);
    m_cellFlagArray = newCellArray;
    CellFlagArraySize(newCellArraySize);
    //
    // We fix the LEVELDESC (Level Descriptor)
    // for this new level.
    //
    m_pLevelDescriptors[level].LastColumn(LastColumn(level) + 1);
    // All levels after this one need a new CellOffset
    for (i=level+1; i<NumLevel(); i++)
    {
      m_pLevelDescriptors[i].CellOffset(
              CellOffset(i)+LastRow(level)+1);
    };
    //
    //
    // We have to fix all the pointers that
    // used to point into m_cellFlagArray.
    //
    free (m_pppdPointer10450);
    m_numColumnPointers++; //One additional column in level.
    m_pppdPointer10450 = (CELLFLAG ***)malloc(4*(m_numColumnPointers + NumLevel()),MALLOC019);
    pColumnPointer = (CELLFLAG **)m_pppdPointer10450 + NumLevel(); 
                   //Where first column pointer goes
    //ppdA2 = (CELLFLAG **)&m_pppdPointer10450[numLevel];
    // Need cast because the 10450 array is used for
    // two things.  See comments in data definitions.
    for (i=0; i<NumLevel(); i++)
    {
      m_pppdPointer10450[i] = pColumnPointer;
      pCell = m_cellFlagArray + CellOffset(i);
      for (i32 column=0;
           column<= m_pLevelDescriptors[i].LastColumn();
           column++)
      {
        *(pColumnPointer++) = pCell;
        pCell += LastRow(i) + 1; //add one column of cells
      }; // for column
    };/// for i=1 to level
    free (m_objectLevelIndex);
    m_objectLevelIndex = (ui16 *)malloc(2*(NumLevel()),MALLOC020);
    if (m_objectLevelIndex == NULL)
    {
      DIE("Cannot allocate memory");
    };
    oldColIndex = m_objectColumnIndex;
    newColIndex = (ui16 *)malloc(2*m_numColumnPointers, MALLOC021);
    if (newColIndex == NULL)
    {
      DIE("Cannot allocate memory");
    };
    newOffset = 0;
    oldOffset = 0;
    for (i=0; i<level; i++) 
    { //Copy previous levels
      m_objectLevelIndex[i] = (ui16)newOffset;
      memcpy(newColIndex+newOffset, 
             oldColIndex+oldOffset,
             2*(LastColumn(i)+1));
      newOffset += LastColumn(i) + 1;
      oldOffset += LastColumn(i) + 1;
    };
    // Now copy this level.
    m_objectLevelIndex[level] = (ui16)newOffset;
    for (i=0; i<LastColumn(level); i++)
    { //Copy previous columns in this level
      newColIndex[newOffset++] = oldColIndex[oldOffset++];
    };
    //Now compute this column's index;
    // Start with previous column's index.
    newColIndex[newOffset] = oldColIndex[oldOffset-1];
    //For each cell in previous column increment our index
    for (i=0; i<=LastRow(level); i++)
    {
      if (GetCellFlag(level,LastColumn(level)-1,i) &0x10)
      {
        newColIndex[newOffset]++;
      };
    };
    newOffset++;
    for (i=level+1; i<NumLevel(); i++)
    { //Copy following levels
      m_objectLevelIndex[i] = (ui16)newOffset;
      memcpy(newColIndex+newOffset, 
             oldColIndex+oldOffset,
             2*(LastColumn(i)+1));
      newOffset += LastColumn(i) + 1;
      oldOffset += LastColumn(i) + 1;
    };
    free (m_objectColumnIndex);
    m_objectColumnIndex = newColIndex;
  };
  SetModifiedFlag();
  return VerifyLevelData();
}

bool CCSBuildDoc::AddRow(int level)
{
  int i;
  unsigned int oldCellArraySize, newCellArraySize;
  int newOffset, oldOffset;
  FLOORTYPE floor;
  CELLFLAG *newCellArray, *pCell;
  CELLFLAG **pColumnPointer;
  if (level >= NumLevel()) return true;
  if (LastRow(level) == 31) return true;
  {
    // First
    // We are going to increase the size of the
    // cellflag array by the width of the current
    // level.
    oldCellArraySize = CellFlagArraySize();
    newCellArraySize = oldCellArraySize + LastColumn(level) + 1;
    newCellArray = (CELLFLAG *)malloc(newCellArraySize,MALLOC028);
    if (newCellArray == NULL)
    {
      Sorry("Cannot allocate memory for additional row");
      return true;
    };
    newOffset = 0;
    oldOffset = 0;
    // Copy all levels up to this level.
    memcpy(newCellArray + newOffset, 
           m_cellFlagArray + oldOffset, 
           CellOffset(level));
    newOffset += CellOffset(level);
    oldOffset += CellOffset(level);
    //Now copy the column in this level.
    for (i=0; i<=LastColumn(level); i++)
    {
      //Copy one old column
      memcpy(newCellArray + newOffset,
             m_cellFlagArray + oldOffset,
             LastRow(level) + 1);
      newOffset += LastRow(level) + 1;
      oldOffset += LastRow(level) + 1;
      // Add a cell at the bottom of the current column.
      *(newCellArray + newOffset) = 0;
      newOffset++;
    };
    //Now copy the rest of the cellflags
    memcpy(newCellArray+newOffset,
           m_cellFlagArray+oldOffset,
           oldCellArraySize-oldOffset);
    free(m_cellFlagArray);
    m_cellFlagArray = newCellArray;
    CellFlagArraySize(newCellArraySize);
    //
    // We fix the LEVELDESC (Level Descriptor)
    // for this new level.
    //
    m_pLevelDescriptors[level].LastRow(LastRow(level) + 1);
    //
    // All levels after this one need a new CellOffset
    //
    for (i=level+1; i<NumLevel(); i++)
    {
      m_pLevelDescriptors[i].CellOffset(
              CellOffset(i)+LastColumn(level)+1);
    };
    //
    //
    // We have to fix all the pointers that
    // used to point into m_cellFlagArray.  We 
    // regenerate them from scratch.
    //
    free (m_pppdPointer10450);
    m_pppdPointer10450 = (CELLFLAG ***)malloc(4*(m_numColumnPointers + NumLevel()),MALLOC019);
    pColumnPointer = (CELLFLAG **)m_pppdPointer10450 + NumLevel(); 
                   //Where first column pointer goes
    //ppdA2 = (CELLFLAG **)&m_pppdPointer10450[numLevel];
    // Need cast because the 10450 array is used for
    // two things.  See comments in data definitions.
    for (i=0; i<NumLevel(); i++)
    {
      m_pppdPointer10450[i] = pColumnPointer;
      pCell = m_cellFlagArray + CellOffset(i);
      for (i32 column=0;
           column<= m_pLevelDescriptors[i].LastColumn();
           column++)
      {
        *(pColumnPointer++) = pCell;
        pCell += LastRow(i) + 1; //add one column of cells
      }; // for column
    };/// for i=1 to level
    //
    // Now we have to adjust the objectLevel and objectColumn
    // indexes.
    //

    // On second thought....there is nothing to adjust.
    // We added no columns and we added no objects
    // to the columns.  Everything stays the same.

  };
  SetModifiedFlag();
  return VerifyLevelData();
}

//*********************************************************
//
//*********************************************************
void CCSBuildDoc::ConvertListOfObjects(RN *pRN, bool DuplicateOK,
                                       int level, int x, int y)
{
  int dbNum, pos, idx;
  int i, newDML, duplicate;
  DBCOMMON *pDB;
  bool clone = false;
  while ( (*pRN != RNeof) && !clone)
  {
    if (    (*pRN == RNnul)
        // || (*pRN == RNempty)
         || pRN->IsMagicSpell())
    {
      CString msg;
      *pRN = RNeof;
      msg.Format("Illegal object at %d(%02d,%02d)\n"
                 "Object value = 0x%04x",
                 level,x,y,
                 pRN->ConvertToInteger());
      Error(msg);
      return;
    };
    dbNum = (pRN->ConvertToInteger() >> 10) & 15;
    pos   = (pRN->ConvertToInteger() >> 14) & 3;
    idx   = pRN->ConvertToInteger() & 0x3ff;
    if (idx >= m_dungeonDatIndex->DBSize(dbNum))
    {
      CString msg;
      *pRN = RNeof;
      msg.Format("Illegal object at %d(%02d,%02d)\n"
                 "dbNum = %d; pos = %d;\n"
                 "idx = %d; dbSize = %d",
                 level,x,y,
                 dbNum, pos,
                 idx, m_dungeonDatIndex->DBSize(dbNum));
      Error(msg);
      return;
    };
    pDB   = db.GetCommonAddress(DBTYPE(dbNum), idx);
    duplicate = -1;
    for (i=1; i<m_dataMapLength; i++)
    { //Search to see if it is already present
      if ((m_dataTypeMap[i] & 15) != dbNum) continue;
      if (m_dataIndexMap[i] != idx) continue;
      if (DuplicateOK)
      {
        duplicate = i;
        break;
      }
      else if ((pDoc->EditingOptions() & EDIT_OPT_AllowClonedObjects) == 0)
      {
        CString msg;
        if (clonedObjectList.FirstError() == NULL)
        {
          clonedObjectList.NewError(  "ONE OR MORE OBJECTS HAVE BEEN DELETED");
          clonedObjectList.NewError("FROM THE DUNGEON BECAUSE THEY ARE LOCATED");
          clonedObjectList.NewError("IN MORE THAN ONE PLACE IN THE DUNGEON AT");
          clonedObjectList.NewError("THE SAME TIME.  THIS IS NOT ONLY UNNATURAL");
          clonedObjectList.NewError("IT CAN CAUSE BIG PROBLEMS.");
          clonedObjectList.NewError("IF YOU REALLY WANT THIS TO HAPPEN, SELECT");
          clonedObjectList.NewError("Edit|Options|Allow Cloned Objects BEFORE");
          clonedObjectList.NewError("YOU LOAD YOUR DESIGN TO BE EDITED.");
          clonedObjectList.NewError(" ");
        };   
        msg.Format("Object %d at %d(%02d,%02d) is in more than one place.",i,level,x,y);
        *pRN = RNeof;
        clonedObjectList.NewError(msg);
        return;
      }
      else
      {
        //And we will create a new index entry for this cloned
        //object.  So there will be two index entries for the
        //same database entry.  That will make them easier to
        //find since we don't have to search the dungeon.  And
        //it will distinguish us from legally cloned objects
        //which have but one index entry for the multiple 
        //references.
        RN rn;
        rn.ConstructFromInteger(i);
//        rn.IsCloned(true);
        clone = true;
      };
    };
    if (duplicate >= 0)
    {
      pRN->ConstructFromInteger(duplicate);
//      pRN->IsCloned(true);
    }
    else
    {
      newDML = 32*(m_dataMapLength/32) + 33;
      m_dataTypeMap = (ui8 *)realloc(m_dataTypeMap, newDML,MALLOC034);
      m_dataIndexMap = (ui16 *)realloc(m_dataIndexMap, 2* newDML,MALLOC035);
      if (m_dataMapLength == 0)
      {
        // Don't use entry zero.
        m_dataTypeMap[0] = 0;
        m_dataIndexMap[0] = 0;
        m_dataMapLength = 1;
      };
      m_dataTypeMap[m_dataMapLength] = (ui8)((pos<<4) | dbNum);
      m_dataIndexMap[m_dataMapLength] = (ui16)idx;
      m_dataMapLength++;
      pRN->ConstructFromInteger(m_dataMapLength-1);
      if (clone)
      {
//        pRN->IsCloned(true);
      };
    };
    if ( (duplicate == -1) & !clone) //Don't convert objects twice.
    {
      switch (dbNum)
      {
      case dbCHEST:
        {
          DB9 *pChest;
          pChest = pDB->CastToDB9();
          if (pChest->contents() != RNeof)
          {
            ConvertListOfObjects(pChest->pContents(), false,
                                 level,x,y);
          };
        };
        break;
      case dbMONSTER:
        {
          DB4 *pMonster;
          pMonster = pDB->CastToDB4();
          if (pMonster->possession() != RNeof)
          {
            ConvertListOfObjects(pMonster->pPossession(), false,
                                 level,x,y);
          };
        };
        break;
      case dbMISSILE:
        {
          DB14 *pMissile;
          pMissile = pDB->CastToDB14();
          if (    (pMissile->flyingObject() != RNeof)
               && !pMissile->flyingObject().IsMagicSpell())
          {
            ConvertListOfObjects(pMissile->pFlyingObject(),true,
                                 level,x,y);
          };
        };
        break;
      case dbSCROLL:
        //Nothing to do.  It has index of Text Entry.
        break;
      };
    };
    pRN = pDB->pLink();
  };
  return;
}

void CCSBuildDoc::ConvertTimers(void)
{
  //UI_MessageBox("ConvertTimers NotImplemented","error",MB_OK);
  for (i32 i=0; i<pDoc->MaxTimers(); i++)
  {
    switch (m_Timers[i].timerFunction)
    {
    case 24: 
    case 25: ConvertListOfObjects(&m_Timers[i].timerObj8(), false,
                                  0,-1,0);
             break;
    case 48:
    case 49: ConvertListOfObjects(&m_Timers[i].timerObj6(), true,
                                  0,-2,0); // The flying object
             break;
    case 60:
    case 61: ConvertListOfObjects(&m_Timers[i].timerObj8(), false,
                                  0,-3,0);
             break;
    };
  };
}

void CCSBuildDoc::ConvertCharacters(void)
{
  int i, j;
  for (i=0; i<m_NumCharacter; i++)
  {
    for (j=0; j<30; j++)
    {
      if (m_Characters[i].possessions[j] != RNnul)
      {
        ConvertListOfObjects(&m_Characters[i].possessions[j], false,
                             0,-4,0);
      };
    };
  };
}

//****************************************************
//
// We need to expand the database to the 'linear'
// model.
// We need to search the dungeon and convert all
// objects of class RN to the new model.
// Should be fun.
void CCSBuildDoc::ConvertDataBases(void)
{
  int level, x, y;
  CELLFLAG cf, *columnPointerCF;
  RN *columnPointerRN;
  for (level=0; level<m_dungeonDatIndex->NumLevel(); level++)
  {
    for (x=0; x<=m_pLevelDescriptors[level].LastColumn(); x++)
    {
      columnPointerCF = m_pppdPointer10450[level][x];
      columnPointerRN = m_objectList + m_objectColumnIndex[m_objectLevelIndex[level]+x];
      for (y=0; y<=m_pLevelDescriptors[level].LastRow(); y++)
      {
        cf = m_pppdPointer10450[level][x][y];
        if ((cf & 0x10) == 0) continue;
        ConvertListOfObjects(columnPointerRN, false,
                             level,
                             x + OffsetX(level),
                             y + OffsetY(level));
          //And sub-lists, too. Chest/Monster/scroll/missile
        columnPointerRN++;
      };
    };
  };
  if (m_Timers != NULL) ConvertTimers();
  ConvertCharacters();
  //if (RememberToPutObjectInHand != -1)
  //{
  //  InternalError("Object in hand");
  //};
  //  if (RememberToPutObjectInHand == RNnul) d.objectInHand = RNnul;
  //  else 
  //  {
  //    d.objectInHand.ConstructFromInteger(RememberToPutObjectInHand);
      if ( (ExtendedFeaturesVersion == '@')&&(m_objectInHand!=RNnul))
      {
        ConvertListOfObjects(&m_objectInHand, false,
                             0,-5,0);
      };
  //  };
  //  RememberToPutObjectInHand = -1;
  //};
}

void CCSBuildDoc::translatePCcharacters(void)
{
  int i, j;
  ui8 *pSrc, *pDst, *pC;
  for (i=3; i>=0; i--)
  {
    pSrc = ((ui8 *)(m_Characters)) + 319*i;
    pDst = ((ui8 *)(m_Characters)) + 800*i;
    memmove(pDst, pSrc, 319);
  };
  for (i=0; i<4; i++)
  {
    pC = (ui8 *)(&m_Characters[i]);
    memmove(pC+92, pC+91, 319-92);
    m_Characters[i].portrait[0] = (i8)pC[276];
    if (i >= m_NumCharacter)
    {
      for (j=0; j<30; j++)
      {
        m_Characters[i].possessions[j] = RNnul;
      };
    };
  };
  memmove(m_Characters+4,(ui8 *)m_Characters+319*4, 128);
}


void CCSBuildDoc::VerifyListOfObjects(
                         RN *pRN,
                         int level, int x, int y,
                         int depth)
{
  DBTYPE dbType;
  DBCOMMON *pDB;
  RN obj;
  DB7 *pScroll;
  int dbIndex;
  if (depth == 30)
  {
    LOCATIONREL locr;
    CString msg = "We have reached our limit of recursion while\n"
                  "examining the objects in this dungeon.\n"
                  "This is probably due to an object containing itself.\n";
    locr.l(level);
    locr.x(x);
    locr.y(y);
    msg += locr.Format();
    pDoc->Sorry(msg);
    return;
  };
  while (*pRN != RNeof)
  {
    dbType = pRN->dbType();
    dbIndex = pRN->dbIndex();
    if (dbIndex >= db.NumEntry(dbType))
    {
      CString msg;   
      msg.Format("At %02d(%02d,%02d) we found a %s\n"
                 "with an Object Indirect Pointer greater\n"
                 "than the number of objects in its database.\n"
                 "We have deleted it.",
                 level, x, y, db.DBName(dbType));
      Warning(msg);
      *pRN = RNeof;
      return;
    };
    pDB = GetCommonAddress(*pRN);
    switch (dbType)
    {
    case dbSCROLL:
      pScroll = pDB->CastToDB7();
      if (pScroll->text() == RNnul)
      { //We marked this during the conversion process.
        //We promised to delete it now that we know where it is.
        CString msg;
        msg.Format("One of the Scrolls with an illegal\n"
                   "text index is at %d(%02d,%02d)"
                   "We are deleting it.",
                   level, x, y);
        Warning(msg);
        *pRN = pDB->link();
        break;
      };
      pRN = pDB->pLink();
      break;
    case dbDOOR:
    case dbTELEPORTER:
    case dbCLOTHING:
    case dbMISC:
    case dbTEXT:
    case dbPOTION:
    case dbACTUATOR:
    case dbWEAPON:
    case dbCLOUD:
      pRN = pDB->pLink();
      break;
    case dbCHEST:
      VerifyListOfObjects(
              pDB->CastToDB9()->pContents(),
              level,x,y,
              depth+1);
      pRN = pDB->pLink();
      break;
    case dbMONSTER:
      VerifyListOfObjects(
              pDB->CastToDB4()->pPossession(),
              level,x,y,
              depth+1);
      pRN = pDB->pLink();
      break;
    case dbMISSILE:
      pRN = pDB->pLink();
      obj = pDB->CastToDB14()->flyingObject();
      if (obj.IsMagicSpell()) break;
      VerifyListOfObjects(
              pDB->CastToDB14()->pFlyingObject(),
              level, x, y,
              depth+1);
      break;
    default:
      NotImplemented();
      pRN = pDB->pLink();
      break;
    };
  };
}

void CCSBuildDoc::VerifyDatabases(void)
{
  int i, j, level, x, y;
  LOCATIONREL locr;
  CELLFLAG cf, *columnPointerCF;
  RN *columnPointerRN;
  for (level=0; level<m_dungeonDatIndex->NumLevel(); level++)
  {
    for (x=0; x<=m_pLevelDescriptors[level].LastColumn(); x++)
    {
      columnPointerCF = m_pppdPointer10450[level][x];
      columnPointerRN = m_objectList + m_objectColumnIndex[m_objectLevelIndex[level]+x];
      for (y=0; y<=m_pLevelDescriptors[level].LastRow(); y++)
      {
        cf = m_pppdPointer10450[level][x][y];
        if ((cf & 0x10) == 0) continue;
        VerifyListOfObjects(columnPointerRN,
                            level,
                            x + OffsetX(level),
                            y + OffsetY(level),
                            0);
          //And sub-lists, too. Chest/Monster/scroll/missile
        columnPointerRN++;
      };
    };
  };
  for (i=0; i<pDoc->NumCharacter(); i++)
  {
    for (j=0; j<30; j++)
    {
      if (pDoc->CharacterPossession(i,j) != RNnul)      
      {
         VerifyListOfObjects(
             pDoc->pCharacterPossession(i,j),
             -1, 0, 0,
             0);
      };
    };
  };
  if (pDoc->ObjectInHand() != RNnul)
  {
    VerifyListOfObjects(pDoc->pObjectInHand(),
                     -1, 0, 0,
                     0);
  };
}

void CCSBuildDoc::ExtendPortraits(void)
{
  int i;
  int numEnt;
  DB3 *pActuator;
  ACTUATOR_Portrait *pPortrait;
  ACTUATORCLASS ac;
  numEnt = db.NumEntry(dbACTUATOR);
  for (i=0; i<numEnt; i++)
  {
    pActuator = GetRecordAddressDB3(i);
    if (pActuator->link() == RNnul) continue;
    ac = pActuator->actuatorClass(roomSTONE);
    if (ac == AC_PORTRAIT)
    {
      pPortrait = pActuator->Cast2AC_Portrait(roomSTONE);
      pPortrait->ClearResponses(0);
      pPortrait->SetResponses(0);
      pPortrait->ToggleResponses(0);
      pPortrait->ClearUnusedPor();
      pPortrait->State(0);
      pPortrait->BlockDisables(false);
      pPortrait->DisableMessage(0);
    };
  };
  extendedPortraits = true;
}


void PrepareExtendedWallDecorations(void)
{
  if (!extendedWallDecorations)
  {
    LOCATIONREL locr;
    for (locr.First(); locr.l()>=0; locr.Next())
    {
      RN obj;
      for ( obj = pDoc->FindFirstObject(locr); obj!=RNeof; obj = pDoc->GetDBRecordLink(obj))
      {
        int dsaLevelIndex, absIndex;
        DB3 *pDB3;
        ACTUATOR_DSA *pDSA;
        if (obj.dbNum() != dbACTUATOR) continue;
        pDB3 = pDoc->GetRecordAddressDB3(obj);
        if (pDB3->actuatorClass(roomSTONE) != AC_DSA) continue;
        pDSA = pDB3->Cast2AC_DSA(roomSTONE);
        dsaLevelIndex = pDSA->DSAselector();
        absIndex = pDoc->getDSA(locr.l(), dsaLevelIndex);
        if (absIndex != 255) continue;
        pDSA->ParameterA(0);
      };
    };
  };
  extendedWallDecorations = true;
}

bool CCSBuildDoc::LoadDungeon(CArchive *archive)
{
  LOADSTATUS status;
  CFile *f;
  int i;
  m_PCFormattedBlock1 = false;
  if (archive != NULL)
  {
    archive->Flush();
    f = archive->GetFile();
    f->Seek(0,SEEK_SET);
  }
  else
  {
    f=NULL;
  };
  if (encipheredDataFile != "")
  {
    Sorry("You cannot decode an encrypted file");
    encipheredDataFile = "";
    return false;
  };
  ExtendedFeaturesSize = ReadExtendedFeatures(f);
  if (f != NULL)
  {
    if (m_littleEndian)
    {
      status = ReadAndUnscrambleBlockLittleEndian(f, (ui16 *)&m_block1);
    }
    else
    {
      status = ReadAndUnscrambleBlockBigEndian(f,(ui16 *)&m_block1);
    };
  }
  else
  {
    status = LOAD_OK;
  };
  if (f != NULL)
  {
    m_dungeonFormat = false;
  };
  m_compressed = false;
  if (status != LOAD_OK)
  {
    f->Seek(ExtendedFeaturesSize, CFile::begin);
    int length = f->GetLength();
    char *fileBuffer;
    fileBuffer = (char *)malloc(length-ExtendedFeaturesSize,MALLOC022);
    f->Read(fileBuffer, length-ExtendedFeaturesSize);
    if (*(ui16 *)fileBuffer != 0x0481)
    {
      // So perhaps it is uncompress dungeon.dat.
      m_dungeonFormat = true;
      f->Seek(ExtendedFeaturesSize, CFile::begin);
    }
    else
    {
      m_dungeonFormat = true;
      ui32 expandedLength = LE32(*((ui32 *)(fileBuffer+2)));
      if (expandedLength > 500000)
      {
        Error("Expanded file length > 500000");
        return false;
      };
      m_expandedBuffer = (char *)realloc(m_expandedBuffer,expandedLength,MALLOC023);
      ExpandData(fileBuffer+8, m_expandedBuffer, expandedLength);
      m_inStreamLeft = expandedLength;
      m_inStreamNext = m_expandedBuffer;
      m_dungeonFormat = true;
      m_compressed = true;
    };
    free (fileBuffer);
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    if (m_PCFormattedBlock1)
    {
      swapPCBlock1((PCGAMEBLOCK1 *)(&m_block1));
      extractPCBlock1Data((PCGAMEBLOCK1 *)(&m_block1));
    }
    else
    {
      swapBlock1(&m_block1);
      extractBlock1Data(&m_block1);
    };
  }
  else
  {
    initializeBlock1Data();
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    if (m_littleEndian | m_PCFormattedBlock1)
    {
      status = UnscrambleStream(
                  f,
                  (ui16 *)&m_block2,       // buffer
                  64,                      // #words
                  ((PCGAMEBLOCK1 *)(&m_block1))->Block2Hash,     // initial hash
                  ((PCGAMEBLOCK1 *)(&m_block1))->Block2Checksum);// expected checksum
    }
    else
    {
      status = UnscrambleStream(
                  f,
                  (ui16 *)&m_block2,       // buffer
                  64,                      // #words
                  m_block1.Block2Hash,     // initial hash
                  m_block1.Block2Checksum);// expected checksum
    };
    if (status != LOAD_OK)
    {
      Error("Second block of game file is damaged");
      return false;
    };
    swapBlock2(&m_block2);
    extractBlock2Data();
  }
  else
  {
    initializeBlock2Data();
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    InitializeItem16();
    if (m_littleEndian | m_PCFormattedBlock1)
    {
      status = UnscrambleStream(
                    f,
                    (ui16 *)m_Item16,     // buffer
                    16/2 * m_MaxITEM16,   // #words
                    ((PCGAMEBLOCK1 *)(&m_block1))->ITEM16Hash,
                    ((PCGAMEBLOCK1 *)(&m_block1))->ITEM16Checksum);
    }
    else
    {
      status = UnscrambleStream(
                    f,
                    (ui16 *)m_Item16,     // buffer
                    16/2 * m_MaxITEM16,   // #words
                    m_block1.ITEM16Hash,
                    m_block1.ITEM16Checksum);
    };
    if (status != LOAD_OK)
    {
      Error("The monsters on the active level appear damaged");
      return false;
    };
    if (!m_littleEndian) swapITEM16s();
  }
  else
  {
    InitializeItem16();
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    if (m_littleEndian )
    {
      status = UnscrambleStream(
                  f,
                  (ui16 *)m_Characters,   // buffer
                  1404/2,                 // # words
                  ((PCGAMEBLOCK1 *)(&m_block1))->CharacterHash,
                  ((PCGAMEBLOCK1 *)(&m_block1))->CharacterChecksum);
      //TranslateDMCharacters();
    }
    else
    {
      if (m_PCFormattedBlock1)
      {
        status = UnscrambleStream(
                    f,
                    (ui16 *)m_Characters,   // buffer
                    3328/2,                 // # words
                    ((PCGAMEBLOCK1 *)(&m_block1))->CharacterHash,
                    ((PCGAMEBLOCK1 *)(&m_block1))->CharacterChecksum);
      }
      else
      {
        status = UnscrambleStream(
                    f,
                    (ui16 *)m_Characters,   // buffer
                    3328/2,                 // # words
                    m_block1.CharacterHash,
                    m_block1.CharacterChecksum);
      };
    };
    if (status != LOAD_OK)
    {
      Error("The characters in the savegame file appear damaged");
      return false;
    };
    if (m_littleEndian) 
    {
      translatePCcharacters();
    }
    else
    {
      swapCharacterData();
    };
  }
  else
  {
    InitializeCharacters();
    brightness = 0;
    SeeThruWalls = 0;
    Byte13279 = 0;   
    Word13278 = 0;       //Party shield
    Word13276 = 0;       //Spell effect
    Word13274 = 0;       //Spell effect
    NumHistEnt = 0;      // #entries in 13268???
    uByte13271 = 0;      //Life frozen time
    Byte13270 = 0;       //index of entry in 13268
    Byte13269 = 0;       //index of entry in 13268
    for (i=0; i<24; i++)
    {
      Word13268[i] = 0;  //history of moves???
                         // bits 0-4 = mapX
                         // bits 5_9 = mapY
                         // bits 10_15 = loaded level
      Byte13220[i] = 0;  //parallels 13268??  
                         // Is this a once-only kind of thing?
    };
    Invisible = 0;       // 13196;
    memset (Byte13195,0,sizeof(Byte13195));
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    InitializeTimers();
    if (m_littleEndian | m_PCFormattedBlock1)
    {
      status = UnscrambleStream(
                    f,
                    (ui16 *)m_Timers,        // buffer
                    (ui16)m_MaxTimers * 10/2,// # words
                    ((PCGAMEBLOCK1 *)(&m_block1))->TimersHash,        // initial hash
                    ((PCGAMEBLOCK1 *)(&m_block1))->TimersChecksum);   // expected checksum
    }
    else
    {
      status = UnscrambleStream(
                    f,
                    (ui16 *)m_Timers,        // buffer
                    (ui16)m_MaxTimers * 10/2,// # words
                    m_block1.TimersHash,        // initial hash
                    m_block1.TimersChecksum);   // expected checksum
    };
    if (status != LOAD_OK)
    {
      Error("TIMER data in savegame file appear damaged");
      return false;
    };
    if (!m_littleEndian) swapTimers();
  }
  else
  {
    m_numTimer = 0;
    m_MaxTimers = 10;
    m_NextAvailTimer = 0;
    InitializeTimers();
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    if (m_littleEndian | m_PCFormattedBlock1)
    {
      status = UnscrambleStream(
                   f,
                  (ui16 *)m_timerQue,     // buffer
                  (ui16)m_MaxTimers,      // # words
                  ((PCGAMEBLOCK1 *)(&m_block1))->TimerQueHash,     // initial hash
                  ((PCGAMEBLOCK1 *)(&m_block1))->TimerQueChecksum);// expected checksum
    }
    else
    {
      status = UnscrambleStream(
                   f,
                  (ui16 *)m_timerQue,     // buffer
                  (ui16)m_MaxTimers,      // # words
                  m_block1.TimerQueHash,     // initial hash
                  m_block1.TimerQueChecksum);// expected checksum
    };
    if (status != LOAD_OK) 
    {
      Error("TIMER queue in savegame file appears damaged");
      return false;
    };
    if (!m_littleEndian) swapTimerQue();
  };
  if ((!m_dungeonFormat) && (f!=NULL))
  {
    if (m_littleEndian)
    { //Read the characters' portraits
      ui8 *buf;
      ui16 *pPortrait;
      int i, g, b, p;
      ui8 c;
      buf = (ui8 *)malloc(8,MALLOC057);
      for (i=0; i<4; i++)
      {
        //Now we have 29 rows of 32 pixels.
        //Or 58 groups of 16 pixels.
        pPortrait = (ui16 *)m_Characters[i].portrait;
        for (g=0; g<58; g++)
        {
          f->Read(buf, 8);
          for (b=0; b<16; b++)
          {
            c = buf[b/2];
            if ((b&1)==0) c >>= 4;
            for (p=0; p<4; p++)
            {
              pPortrait[p] = (ui16)(((pPortrait[p])<<1) | ((c&1)<<0));
              c >>= 1;
            };
          };
          pPortrait[0] = LE16(pPortrait[0]);
          pPortrait[1] = LE16(pPortrait[1]);
          pPortrait[2] = LE16(pPortrait[2]);
          pPortrait[3] = LE16(pPortrait[3]);
          pPortrait += 4;
        };
      };
    };
  };
  status = ReadDatabases(f);
  if (status != LOAD_OK)
  {
    Error("The databases in the savegame file appear damaged");
    return false;
  };
  if (m_compressed) {free (m_expandedBuffer);m_expandedBuffer=NULL;};
  if (ExtendedFeaturesVersion == '@')
  {
    LOCATIONREL locr;
    CELLFLAG cf;
    ROOMTYPE rt;
    RN obj;
    DB3 *pActuator;
    ACTUATORCLASS ac;
    //-----
    ConvertDataBases();
    if (clonedObjectList.FirstError() != NULL)
    {
      WarningDialog dlg;
      dlg.m_chkErrMsg = &clonedObjectList;
      dlg.DoModal();
      clonedObjectList.Clear();
    };
    for (locr.First(); locr.l()!=-1; locr.Next())
    {
      cf = pDoc->GetCellFlag(locr);
      rt = RoomType(cf);
      if (rt != roomSTONE) continue;
      for (obj=pDoc->FindFirstActuator(locr);
           obj!=RNeof;
           obj = pDoc->FindNextActuator(obj))
      {
        pActuator = pDoc->GetRecordAddressDB3(obj);
        ac = pActuator->actuatorClass(rt);
        if (ac != AC_ANDOR) continue;
        pActuator->Cast2AC_AndOr(rt)->Edge(0);
      };
    };
    ExtendedFeaturesVersion = 'A';
  };
  if (ExtendedFeaturesVersion < 'B') 
  {
    convertScrolls();
    ExtendedFeaturesVersion = 'B';
  };
  VerifyDatabases();
  if (!extendedPortraits) ExtendPortraits();
  DeleteItem16Entries();
  PrepareExtendedWallDecorations();
  //Force16Levels();
//  FILE *ch = fopen("ch.bin","wb");
//  fwrite(this->m_Characters,1,3200,ch);
//  fclose (ch);
  return true;
}


//*********************************************************
//
//*********************************************************
//   TAG00b152
void CCSBuildDoc::DetachItem16(i32 P1)
{
  ITEM16 *pi16A3;
  DB4 *pDB4A2;
//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
//
//
  if ( (P1 > m_MaxITEM16) || (m_Item16[P1].word0 < 0) ) return;
//
  ASSERT(P1 < m_MaxITEM16);
  pi16A3 = &m_Item16[P1];
  pDB4A2 = (DB4 *)db.GetCommonAddress(dbMONSTER,pi16A3->word0);
  //pDB4A2 = &((DB4 *)d.misc1052eight[4])[pi16A3->word0];
  m_ITEM16QueLen--;
  pDB4A2->groupIndex(pi16A3->positions());
  // 30OCT02SETWBITS8_9(pDB4A2->word14, pi16A3->facings());
  pDB4A2->facing(pi16A3->facings());
  // 30OCT02if (BITS0_3(pDB4A2->word14) >=4)
  if (pDB4A2->fear() >= 4)
  {
    //30OCT02SETWBITS0_3(pDB4A2->word14, 0);
    pDB4A2->fear(0);
  };
  pi16A3->word0 = -1;
}

void CCSBuildDoc::DeleteItem16Entries(void)
{
  i32 i=0;
  while (m_ITEM16QueLen != 0)
  {
    ASSERT(i < m_MaxITEM16);
    if ( m_Item16[i].word0 >= 0)
    {
      DetachItem16(i); //(item16Index)
    };
    i++;
  };
}

