#include "stdafx.h"
#include "CSBuild.h"
//#include "types.h"
#include "misc.h"
#include "Graphics.h"
#include "CSBuildDoc.h"
#include "CSbuildView.h"
#include "Actuator.h"
#include "ShowLocations.h"

int messageBoxActive = 0;


bool MessageBoxActive(void)
{
  return messageBoxActive!=0;
}

void DIE(const char *msg)
{
  MessageBox(NULL, msg, "Fatal Error", MB_OK|MB_TOPMOST);
  pDoc->MarkInvalid();
  throw(0x12345);
}

void NoSelection(void)
{
  MessageBox(NULL,"You have made no selection","Error", MB_OK);
}

void InternalError(void)
{
  InternalError("Internal Error");
}

void InternalError(char *msg)
{
  int result = MessageBox(NULL,msg,"Sorry - InternalError/n"
    "Should We Continue?", MB_YESNO);
  if (result == IDNO) DIE("Goodbye");
}

int AreYouSure(const char *msg)
{
  int result = MessageBox(NULL,msg,"Are You Sure???",
                          MB_YESNO|MB_ICONEXCLAMATION);
  return result;
}

int ShouldWe(const char *msg)
{
  int result = MessageBox(NULL,msg,"What Should We Do???",
                          MB_YESNO|MB_ICONQUESTION);
  return result;
}

void DungeonError(void)
{
  Error("Before starting the Enlarge operation we\n"
        "found inconsistencies in this dungeon's structure\n"
        "Enlarging can only make it worse.\n"
        "Send this dungeon to CSBuild's author for analysis.");
}

void EnlargeError(void)
{
  Error("Enlarging has caused inconsistencies in the\n"
        "dungeon's structure.  Please let CSBuild's\n"
        "author know that this has happened.  And quit\n"
        "the program without saving so as not to\n"
        "cause permanent damage to your dungeon.");
}


void CCSBuildDoc::Sorry(const char *msg)
{
  messageBoxActive++;
  //POSITION pos = GetFirstViewPosition();
  //while (pos != NULL)
  //{
    //CView* pView = GetNextView(pos);
    MessageBox(NULL,msg,"Sorry", MB_TASKMODAL|MB_TOPMOST);
    //break;
  //};
  messageBoxActive--;
}

bool CCSBuildDoc::DoYouReallyWantTo(const char *msg)
{
  bool result=false;
  messageBoxActive++;
  POSITION pos = GetFirstViewPosition();
  while (pos != NULL)
  {
    CView* pView = GetNextView(pos);
    CString m;
    m.Format("%s\n\nDo you really want to do that?",msg);
    result = pView->MessageBox(m,"Warning",
              MB_YESNO|MB_TOPMOST|MB_APPLMODAL) == IDYES;
  };
  messageBoxActive--;
  return result;
}

void CCSBuildDoc::Warning(const char *msg)
{
  messageBoxActive++;
  POSITION pos = GetFirstViewPosition();
  while (pos != NULL)
  {
    CView* pView = GetNextView(pos);
    pView->MessageBox(msg,"Warning", MB_APPLMODAL|MB_TOPMOST);
  };
  messageBoxActive--;
}

void CCSBuildDoc::Information(const char *msg)
{
  messageBoxActive++;
  POSITION pos = GetFirstViewPosition();
  while (pos != NULL)
  {
    CView* pView = GetNextView(pos);
    pView->MessageBox(msg,"Information", MB_APPLMODAL|MB_TOPMOST);
  };
  messageBoxActive--;
}

void Strange(const char *msg)
{
  MessageBox(NULL,msg,"Strange but not fatal", NULL);
}

void Error(const char *msg)
{
  MessageBox(NULL,msg,"Error", NULL);
}

void HelpMsg(const char *msg)
{
  MessageBox(NULL,msg,"Help", NULL);
}

void NotImplemented(void)
{
  MessageBox(NULL,"Not Implemented", "Sorry", MB_TASKMODAL|MB_OK);
}

LOCATIONREL::LOCATIONREL(const EXPORTLOC& eloc)
{
  m_l = (char)(eloc.level + eloc.dl);
  m_x = (char)(eloc.left + eloc.dx);
  m_y = (char)(eloc.top + eloc.dy);
  m_p = (char)eloc.pos;
}


LOCATIONREL& LOCATIONREL::operator =(const LOCATIONREL &locr)
{
  m_l = locr.m_l;
  m_x = locr.m_x;
  m_y = locr.m_y;
  m_p = locr.m_p;
  return *this;
}

LOCATIONREL& LOCATIONREL::operator =(const LOCATIONABS &loca)
{
  m_l = (char)loca.l;
  m_x = (char)loca.x;
  m_y = (char)loca.y;
  m_p = (char)loca.p;
  if ((m_l >= 0) && (m_l<pDoc->NumLevel()))
  {
    m_x = (char)(m_x-pDoc->OffsetX(m_l));
    m_y = (char)(m_y-pDoc->OffsetY(m_l));
  };
//  if (  (m_x > pDoc->LastColumn(m_l)) || (m_x < 0)
//      ||(m_y > pDoc->LastRow(m_l))    || (m_y < 0) ) m_l=-1;
  return *this;
}

LOCATIONREL LOCATIONREL::operator +(const LOCATIONREL& locr)
{
  return LOCATIONREL(m_l+locr.m_l, m_x+locr.m_x, m_y+locr.m_y);
}

bool LOCATIONREL::operator ==(const LOCATIONREL& locr) const
{
  if (m_l != locr.m_l) return false;
  if (m_x != locr.m_x) return false;
  if (m_y != locr.m_y) return false;
  if (m_p != locr.m_p) return false;
  return true;
}

LOCATIONABS& LOCATIONABS::operator =(const LOCATIONABS &loca)
{
  l = loca.l;
  x = loca.x;
  y = loca.y;
  p = loca.p;
  return *this;
}

LOCATIONABS& LOCATIONABS::operator =(const LOCATIONREL &locr)
{
  l = locr.l();
  x = locr.x();
  y = locr.y();
  p = locr.p();
  if ((l >= 0) && (l<pDoc->NumLevel()))
  {
//    if (  (x > pDoc->LastColumn(l)) || (x < 0)
//        ||(y > pDoc->LastRow(l))    || (y < 0) )
//    {
//      l = -1;
//    }
//    else
    {
      x += pDoc->OffsetX(l);
      y += pDoc->OffsetY(l);
    };
  };
  return *this;
}

CString LOCATIONREL::Format() const
{
  LOCATIONABS loca;
  if (m_l >= 0)
  {
    loca = *this;
  }
  else
  {
    loca.l = -1;
    loca.x = 0;
    loca.y = 0;
  };
  CString result;
  result.Format("%02d(%02d,%02d)", loca.l, loca.x, loca.y);
  return result;
}


ui32 LOCATIONREL::Integer(void) const
{
  ASSERT(x < 32);
  ASSERT(y < 32);
  ASSERT(l < 63);
  ASSERT(p < 4);
  return ((m_p&3)<<16)|(m_l<<10)|(m_x<<5)|m_y;
}

void LOCATIONREL::Integer(int t)
{
  m_p = (char)((t >> 16) & 3);
  m_l = (char)((t >> 10) & 63);
  m_x = (char)((t >> 5) & 31);
  m_y = (char)(t & 31);
}

void LOCATIONREL::Vertical(int num)
{
  // Set l (level) to -1 if illegal
  LOCATIONABS loca;
  loca = *this;  //Get absolute location
  loca.l += num;
  if (  (loca.l < 0) || (loca.l >= pDoc->NumLevel()))
  {
    m_l = -1;
  }
  else
  {
    *this = loca; //Convert back to relative location
  };
}

bool LOCATIONREL::Next(void)
{
  if (m_l < 0) {m_l = 0;m_x = 0;m_y = 0;}
  else m_y++;
  for (;;)
  {
    if (m_y > pDoc->LastRow(m_l)) {m_y=0;m_x++;continue;}; 
    if (m_x > pDoc->LastColumn(m_l)) {m_x=0;m_y=0;m_l++;continue;};
    if (m_l >= pDoc->NumLevel()){m_l = -1;};
    return m_l>=0;
  };
}

bool LOCATIONREL::First(void)
{
  m_l = -1;
  return Next();
}

bool LOCATIONREL::AdjacentCell(int direction)
{
  static char dx[4] = {0,1,0,-1};
  static char dy[4] = {-1,0,1,0};
  m_x = (char)(m_x + dx[direction&3]);
  m_y = (char)(m_y + dy[direction&3]);
  if (    (m_y > pDoc->LastRow(m_l))
       || (m_x > pDoc->LastColumn(m_l))
       || (m_x < 0)
       || (m_y < 0) ) 
  {
    m_l = -1;
    return false;
  };
  return true;
}


bool LOCATIONABS::Next(void)
{
  LOCATIONREL locr;
  bool result;
  locr = *this;
  result = locr.Next();
  *this = locr;
  return result;
}

LEVELINFO& LEVELINFO::operator =(LOCATIONREL& locr)
{
  level  = locr.l();
  top    = pDoc->OffsetY(level);
  bottom = pDoc->LastRow(level)+pDoc->OffsetY(level);
  left   = pDoc->OffsetX(level);
  right  = pDoc->LastColumn(level)+pDoc->OffsetX(level);
  return *this;
}

ROOMTYPE RoomType(CELLFLAG cf)
{
  i32 n;
  n = ((cf>>5) & 7);
  if (n == 6) // A trick wall
  {
    if ((cf & 5) == 0) return roomBLOCKING;
    return roomNONBLOCKING;
  };
  return (ROOMTYPE)n;
}

ROOMTYPE RoomType(LOCATIONREL locr)
{
  return RoomType(pDoc->GetCellFlag(locr));
}

ROOMTYPE RoomType(int level, int col, int row)
{
  if (col < 0) return roomSTONE;
  if (row < 0) return roomSTONE;
  if (col > pDoc->LastColumn(level)) return roomSTONE;
  if (row > pDoc->LastRow(level)) return roomSTONE;
  return RoomType(pDoc->GetCellFlag(level, col, row));
}



CString DoorTypeName(int type)
{
  switch (type)
  {
  case 0: return CString("Iron Bars");
  case 1: return CString("Wood");
  case 2: return CString("Solid Iron");
  case 3: return CString("Ra");
  default:
    {
      CString result;
      result.Format("Door type %d",type);
      return result;
    };
  };
}

void FillTargetListBox(CListBox *lb, LOCATIONREL dest,
                       OBJECTLIST *objectList)
{
  LOCATIONREL locr;
  LOCATIONREL target;
  int n, index;
  RN obj, o;
  DB3 *pActuator;
  DB1 *pTeleporter;
  CString s;
  LOCATIONABS loca;
  int x=0, y=0, l=0;
  ROOMTYPE rt;
  lb->ResetContent();
  for (locr.First(); locr.l()>=0; locr.Next())
  {
    rt = RoomType(locr);
    for (o = pDoc->FindFirstObject(locr);
         o != RNeof;
         o = pDoc->GetDBRecordLink(o))
    {
      if (o.dbType() == dbACTUATOR)
      {
        obj = o;
      }
      else if (o.dbType() == dbMONSTER)
      {
        DB4 *pDB4;
        pDB4 = pDoc->GetRecordAddressDB4(o);
        for (obj=pDB4->possession(); obj!=RNeof; obj=pDoc->GetDBRecordLink(obj))
        {
          if (obj.dbType() == dbACTUATOR) break;
        };
        if (obj == RNeof) continue;
      }
      else if (o.dbType() == dbTELEPORTER)
      {
        obj = o;
      }
      else obj = o;
      if (obj.dbType() == dbACTUATOR)
      {
        pActuator = pDoc->GetRecordAddressDB3(obj);
        for (n=0, target=pActuator->GetNthTarget(locr,n);
             target.l() != -1;
             n++, target=pActuator->GetNthTarget(locr,n))
        {
          if (target.x() != dest.x()) continue;
          if (target.y() != dest.y()) continue;
          if (target.l() != dest.l()) continue;
          loca = locr;
          s.Format("%02d(%02d,%02d) Actuator",
                   loca.l,loca.x,loca.y);
          index = lb->AddString(s);
          lb->SetItemData(index, objectList->Insert(obj,locr));
        };
      }
      else if (obj.dbType() == dbTELEPORTER)
      {
        pTeleporter = pDoc->GetRecordAddressDB1(obj);
        x = pTeleporter->destX();
        y = pTeleporter->destY();
        l = pTeleporter->destLevel();
        if (x != dest.x()) continue;
        if (y != dest.y()) continue;
        if (l != dest.l()) continue;
        loca = locr;
        s.Format("%02d(%02d,%02d) Teleporter",
                 loca.l,loca.x,loca.y);
        index = lb->AddString(s);
        lb->SetItemData(index, objectList->Insert(obj,locr));
      };
    };
  };
}

#undef malloc
#undef realloc
#undef free
void *DebugMalloc(int size, MALLOCS m)
{
  int *result;
  ASSERT (_heapchk() == _HEAPOK);
  result = (int *)malloc(size+4);
  if (result == NULL)
  {
    DIE("Cannot allocate memory");
  };
  *result = m;
  return (void *)(result+1);
}

void *DebugRealloc(void *a, int size, MALLOCS m)
{
  int *result;
  ASSERT (_heapchk() == _HEAPOK);
  if (a == NULL)
  {
    result = (int *)malloc(size+4);
  }
  else
  {
    result = (int *)realloc((int *)a-1, size+4);
  };
  if (result == NULL)
  {
    DIE("Cannot allocate memory");
  };
  *result = m;
  return (void *)(result+1);
}

void DebugFree(void *a)
{
  ASSERT (_heapchk() == _HEAPOK);
  free((char *)a - 4);
}
#define malloc DebugMalloc
#define realloc DebugRealloc
#define free DebugFree

CString NicelyDecapitalize(CString& str)
{
  int len = str.GetLength();
  int i;
  bool blank;
  char *newstr = (char *)alloca(len + 1);
  memcpy(newstr, (LPCTSTR)str, len);
  newstr[len] = 0;
  blank = true;
  for (i=0; i<len; i++)
  {
    if (!blank)
    {
      if ((newstr[i]>='A') && (newstr[i]<='Z'))
        newstr[i] += 'a'-'A';
    };
    blank =    (newstr[i] == ' ') 
            || (newstr[i] == '(')
            || (newstr[i] == ')')
            || (newstr[i] == '/');
  };
  return CString(newstr);
}

CString DecodeText(const CString& str, TEXTTYPE flags)
{
  int i;
  int len;
  unsigned char ch;
  CString text;
  if (flags == TEXTTYPE_SCROLL)
  {
    text = "";
    len = str.GetLength();
    for (i=0; i<len; i++)
    {
      ch = str.GetAt(i);
      if (ch == 10) ch = '\n';
      text += ch;
    };
  }
  else if (flags == TEXTTYPE_WALL)
  {
    text = "";
    len = str.GetLength();
    for (i=0; i<len; i++)
    {
      ch = str.GetAt(i);
      ch += 'A';
      if (ch == 0x5b) text += ' ';
      else if (ch == 0x5c) text += '.';
      else if (ch == 0xc1) text+='\n';
      else text += ch;
    };
  };
  return text;
}

CHKERRMSG::CHKERRMSG(void)
{
  m_next = NULL;
  m_prev = NULL;
  m_root = true;
  m_errorType = CEMT_UNKNOWN;
  m_parentIndirectIndex = -1;
  m_cloneIndirectIndex = -1;
  m_location.l(-1);
  m_repaired = false;
}

CHKERRMSG::~CHKERRMSG(void)
{
  if (m_root) Clear();
  ASSERT(m_next == NULL);
  ASSERT(m_prev == NULL);
}

void CHKERRMSG::Clear(void)
{
  CHKERRMSG *temp;
  ASSERT(m_root);
  while (m_next != NULL)
  {
    temp = m_next;
    m_next = temp->m_next;
    if (m_next == NULL) m_prev = NULL;
    temp->m_next = NULL;
    temp->m_prev = NULL;
    delete temp;
  };
}

void CHKERRMSG::operator +=(const CString &msg)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_message += msg;
}

CString CHKERRMSG::ErrorMessage(void) const
{
  return m_message;
}

void CHKERRMSG::NewError(const CString& msg)
{
  CHKERRMSG *temp;
  ASSERT(m_root);
  if (msg == "") return;
  temp = new CHKERRMSG;
  temp->m_root = false;
  if (m_prev != NULL) m_prev->m_next = temp;
  temp->m_prev = m_prev;
  m_prev = temp;
  if (m_next == NULL) m_next = temp;
  temp->m_message = msg;
}

void CHKERRMSG::NewError(const CHKERRMSG *pcem)
{
  NewError(pcem->m_message);
  m_prev->m_errorType           = pcem->m_errorType;
  m_prev->m_parentIndirectIndex = pcem->m_parentIndirectIndex;
  m_prev->m_cloneIndirectIndex  = pcem->m_cloneIndirectIndex;
  m_prev->m_location            = pcem->m_location;
}

CHKERRMSG *CHKERRMSG::NextError(const CHKERRMSG *pcem)const
{
  ASSERT(m_root);
  ASSERT(pcem != NULL);
  if (pcem == NULL) return NULL;
  return pcem->m_next;
}

CHKERRMSG *CHKERRMSG::Next(void)const
{
  ASSERT(!m_root);
  return m_next;
}

CHKERRMSG *CHKERRMSG::Prev(void)const
{
  ASSERT(!m_root);
  return m_prev;
}

CHKERRMSG *CHKERRMSG::FirstError(void) const
{
  ASSERT(m_root);
  return m_next;
}

CHKERRMSG *CHKERRMSG::LastError(void) const
{
  ASSERT(m_root);
  return m_prev;
}

CEMT CHKERRMSG::ErrorType(void) const
{
  ASSERT(!m_root);
  return m_errorType;
};

void CHKERRMSG::ErrorType(CEMT errorType)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_errorType = errorType;
};

LOCATIONREL CHKERRMSG::Location(void) const
{
  ASSERT(!m_root);
  return m_location;
};

void CHKERRMSG::Location(LOCATIONREL locr)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_location = locr;
};

void CHKERRMSG::NewLocation(LOCATIONREL locr)
{
  ASSERT(!m_root);
  m_location = locr;
};

int CHKERRMSG::CloneIndirectIndex(void) const
{
  ASSERT(!m_root);
  return m_cloneIndirectIndex;
};

void CHKERRMSG::CloneIndirectIndex(int indirectIndex)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_cloneIndirectIndex = indirectIndex;
};

int CHKERRMSG::ParentIndirectIndex(void) const
{
  ASSERT(!m_root);
  return m_parentIndirectIndex;
};

void CHKERRMSG::ParentIndirectIndex(int indirectIndex)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_parentIndirectIndex = indirectIndex;
};

void CHKERRMSG::Repaired(bool repaired)
{
  ASSERT(!m_root);
  m_repaired = repaired;
}

bool CHKERRMSG::Repaired(void) const
{
  ASSERT(!m_root);
  return m_repaired;
}

bool CHKERRMSG::AnyRepairs(void) const
{
  ASSERT(m_root);
  CHKERRMSG *pcem;
  for (pcem=m_next; pcem!=NULL; pcem=pcem->m_next)
  {
    if (pcem->m_repaired) return true;
  };
  return false;
}

DBCOMMON *CHKERRMSG::DBCommonAddress(void) const
{
  ASSERT(!m_root);
  return m_dbCommonAddress;
}

void CHKERRMSG::DBCommonAddress(DBCOMMON *pDBCommon)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_dbCommonAddress = pDBCommon;
}

DBTYPE CHKERRMSG::DBType(void) const
{
  ASSERT(!m_root);
  return m_dbType;
}

void CHKERRMSG::DBType(DBTYPE dbType)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_dbType = dbType;
}

int CHKERRMSG::Index1(void) const
{
  ASSERT(!m_root);
  return m_index1;
}

void CHKERRMSG::Index1(int index)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_index1 = index;
}

int CHKERRMSG::Index2(void) const
{
  ASSERT(!m_root);
  return m_index2;
}

void CHKERRMSG::Index2(int index)
{
  ASSERT(m_root);
  ASSERT(m_prev != NULL);
  if (m_prev == NULL) return;
  m_prev->m_index2 = index;
}

static bool SearchForRecord(DBTYPE recType, 
                            int dbIndex, 
                            RN obj)
{
  DB4 *pDB4;
  DB7 *pDB7;
  DB9 *pDB9;
  for (;obj!=RNeof; obj=pDoc->GetDBRecordLink(obj))
  {
    DBTYPE dbType;
    dbType = obj.dbType();
    if ( (dbType==recType) && (obj.dbIndex()==dbIndex) )
    {
      return true;
    };
    switch (dbType)
    {
    case dbCLOTHING:
    case dbPOTION:
    case dbDOOR:
    case dbWEAPON:
    case dbMISC:
    case dbACTUATOR:
    case dbTELEPORTER:
    case dbTEXT:
      break;  //ignore...no links
    case dbMONSTER:
      pDB4 = pDoc->GetRecordAddressDB4(obj);
      if (SearchForRecord(recType, dbIndex, pDB4->possession()))
      {
        return true;
      };
      break;
    case dbSCROLL:
      pDB7 = pDoc->GetRecordAddressDB7(obj);
      if (SearchForRecord(recType, dbIndex, pDB7->text()))
      {
        return true;
      };
      break;
    case dbCHEST:
      pDB9 = pDoc->GetRecordAddressDB9(obj);
      if (SearchForRecord(recType, dbIndex, pDB9->contents()))
      {
        return true;
      };
      break;
    default: NotImplemented();
    };
  };
  return false;
}

LOCATIONREL SearchForRecord(DBTYPE dbType, int dbIndex)
{
  LOCATIONREL locr;
  int i;
  int j;
  for (locr.First(); locr.l()>=0; locr.Next())
  {
    if (SearchForRecord(dbType, 
                        dbIndex, 
                        pDoc->FindFirstObject(locr)))
    {
      return locr;
    };
  };
  for (i=0; i<pDoc->NumCharacter(); i++)
  {
    for (j=0; j<30; j++)
    {
      if (pDoc->CharacterPossession(i,j) == RNnul) continue;
      if (SearchForRecord(dbType,dbIndex,pDoc->CharacterPossession(i,j)))      
      {
         locr.l(16+30*i+j);
         return locr;
      };
    };
  };
  if (pDoc->ObjectInHand() != RNnul)
  {
    if (SearchForRecord(dbType, dbIndex, pDoc->ObjectInHand()))
    {
      locr.l(16+30*4);
      return locr;
    };
  };
  for (i=0; i<pDoc->NumTimers(); i++)
  {
    locr.l(16+30*4+1);
    switch (pDoc->Timer(i)->timerFunction)
    {
    case 24: 
    case 25: if (SearchForRecord(
                          dbType,
                          dbIndex,
                          pDoc->Timer(i)->timerObj8())) 
             { 
               return locr;  
             };
    case 48:
    case 49: if (SearchForRecord(
                          dbType,
                          dbIndex,
                          pDoc->Timer(i)->timerObj8()))
             { 
               return locr;  
             };
    case 60:
    case 61: if (SearchForRecord(
                          dbType,
                          dbIndex,
                          pDoc->Timer(i)->timerObj8()))
             { 
               return locr;  
             };
    };
  };
  locr.l(-1);
  return locr;
}

CROSSREF::CROSSREF(void)
{
  int s,d;
  for (s=0; s<32; s++) for (d=0; d<32; d++)
  {
    m_ref[s][d]       = NULL;
    m_size[s][d]      = 0;
    m_allocated[s][d] = 0;
  };
  m_initialized = false;
}

CROSSREF::~CROSSREF(void)
{
  Clear();
}

void CROSSREF::Clear(void)
{
  int s,d;
  for (s=0; s<32; s++) for (d=0; d<32; d++)
  {
    if (m_ref[s][d] != NULL)
    {
      free (m_ref[s][d]);
      m_ref[s][d] = NULL;
    };
    m_size[s][d] = 0;
    m_allocated[s][d] = 0;
  };
  m_initialized = false;
}

#define CROSSREFslBIT 22
#define CROSSREFsyBIT 17
#define CROSSREFspBIT 15
#define CROSSREFdlBIT  9
#define CROSSREFdyBIT  4
#define CROSSREFdpBIT  2
#define CROSSREFtBIT   0
#define CROSSREFdstMASK (0x7FF<<CROSSREFdyBIT)
#define CROSSREFsrcMASK (0x7FF<<CROSSREFsyBIT)
#define CROSSREFsy(v) (((v)>>CROSSREFsyBIT)&0x1f)
#define CROSSREFsl(v) (((v)>>CROSSREFslBIT)&0x3f)
#define CROSSREFsp(v) (((v)>>CROSSREFspBIT)&0x03)
#define CROSSREFtMASK 1

LOCATIONREL CROSSREF::NextTrigger(void)
{
  LOCATIONREL locr;
  int len;
  ui32 *pC, v; 
  ASSERT( (m_dx >= 0) && (m_sx >= 0));
  for (;m_sx < 32; m_sx++, m_index=-1)
  {
    len = m_size[m_sx][m_dx];
    pC = m_ref[m_sx][m_dx];
    for (m_index++; m_index<len; m_index++)
    {
      v = pC[m_index];
      if (((v ^ m_destCell) & (CROSSREFdstMASK|CROSSREFtMASK)) == 0)
      {
        locr.x(m_sx);
        locr.y(CROSSREFsy(v));
        locr.l(CROSSREFsl(v));
        locr.p(CROSSREFsp(v));
        return locr;
      };
    };
  };
  locr.l(-1);
  m_sx = -1;
  return locr;
}

LOCATIONREL CROSSREF::FirstTrigger(LOCATIONREL locr, int type)
{
  if (!m_initialized) Initialize();
  m_sx = 0;
  m_dx = locr.x();
  m_destCell =  (locr.l() << CROSSREFdlBIT) 
              | (locr.y() << CROSSREFdyBIT)
              | (type     << CROSSREFtBIT);
  m_index = -1;
  return NextTrigger();
}

void CROSSREF::AddTrigger(LOCATIONREL locrS, LOCATIONREL locrD, int type)
{
  ui32 v;
  int sx, dx;
  if (    (locrD.l() == 0)
       && (locrD.x() == 0)
       && (locrD.y() == 0) ) return;
  v =   (locrS.l() << CROSSREFslBIT)
      | (locrS.y() << CROSSREFsyBIT)
      | (locrS.p() << CROSSREFspBIT)
      | (locrD.l() << CROSSREFdlBIT)
      | (locrD.y() << CROSSREFdyBIT)
      | (locrD.p() << CROSSREFdpBIT)
      | (type << CROSSREFtBIT);
  sx = locrS.x();
  dx = locrD.x();
  if (m_size[sx][dx] >= m_allocated[sx][dx])
  {
    m_allocated[sx][dx] = m_allocated[sx][dx]+10;
    m_ref[sx][dx] = (ui32 *)realloc(m_ref[sx][dx],
                                    m_allocated[sx][dx]*sizeof(ui32),
                                    MALLOC080);
  };
  m_ref[sx][dx][m_size[sx][dx]++] = v;
}

void CROSSREF::AddTriggersForCell(LOCATIONREL locrS)
{
  int n;
  RN obj;
  RN poss;
  LOCATIONREL locrD;
  DB1 *pTeleporter;
  DB3 *pDB3;
  DB4 *pDB4;
  for (obj = pDoc->FindFirstObject(locrS);
       obj != RNeof;
       obj = pDoc->FindNextActuator(obj))
  {
    switch (obj.dbType())
    {
    case dbTELEPORTER:
      locrS.p(obj.pos());
      pTeleporter = pDoc->GetRecordAddressDB1(obj);
      locrD.l(pTeleporter->destLevel());
      locrD.x(pTeleporter->destX());
      locrD.y(pTeleporter->destY());
      locrD.p(pTeleporter->rotation());
      AddTrigger(locrS,locrD,0);
      break;
    case dbACTUATOR:
      pDB3 = pDoc->GetRecordAddressDB3(obj);
      locrS.p(obj.pos());
      for (n=0; locrD=pDB3->GetNthTarget(locrS,n), locrD.l()>=0 ;n++)
      {
        AddTrigger(locrS, locrD, 1);
      };
      break;
    case dbMONSTER:
      pDB4 = pDoc->GetRecordAddressDB4(obj);
      for (poss = pDB4->possession(); poss!=RNeof; poss = pDoc->GetDBRecordLink(poss))
      {
        switch(poss.dbType())
        {
          case dbACTUATOR:
          pDB3 = pDoc->GetRecordAddressDB3(poss);
          locrS.p(poss.pos());
          for (n=0; locrD=pDB3->GetNthTarget(locrS,n), locrD.l()>=0 ;n++)
          {
            AddTrigger(locrS, locrD, 1);
          };
          break;
        };
      };
      break;
    };
  };
}

void CROSSREF::RebuildTriggers(LOCATIONREL locr)
{
  ui32 srcCell;
  ui32 *pC;
  int index, len;
  if (!m_initialized) Initialize();
  srcCell =  (locr.y() << CROSSREFsyBIT)
            |(locr.l() << CROSSREFslBIT);
  //Remove all entries with this cell as source.
  m_sx = locr.x();
  for (m_dx=0; m_dx<32; m_dx++)
  {
    pC = m_ref[m_sx][m_dx];
    if (pC != NULL)
    {
      len = m_size[m_sx][m_dx];
      for (index=0; index<len; index++)
      {
        if ( ((pC[index]^srcCell) & CROSSREFsrcMASK) == 0)
        { //Remove this entry.
          pC[index] = pC[--len];
        };
      }
    };
  };
  AddTriggersForCell(locr);
  m_sx = -1;
}

void CROSSREF::Initialize(void)
{
  LOCATIONREL locr;
  Clear();
  for (locr.First(); locr.l()>=0; locr.Next())
  {
    AddTriggersForCell(locr);
  };
  m_initialized = true;
}

OBJECTLIST::OBJECTLIST(void)
{
  nument = 0;
  list = NULL;
}

OBJECTLIST::~OBJECTLIST(void)
{
  if (list != NULL) free (list);
  nument = 0;
  list = NULL;
}

int OBJECTLIST::NumEnt(void)
{
  return nument;
}

int OBJECTLIST::Insert(RN obj, LOCATIONREL locr)
{
  nument++;
  list = (OBJENT *)realloc(list, 
                           (20*((nument+19)/20))*sizeof(OBJENT),
                           MALLOC081);
  list[nument-1].locr = locr;
  list[nument-1].obj = obj;
  return nument-1;
}

LOCATIONREL OBJECTLIST::GetLocation(int index)
{
  return list[index].locr;
}

EXTRACTED_LEVEL::EXTRACTED_LEVEL(void)
{
  m_origLevel = -1;
  m_pCellflags = NULL;
  m_pContents = NULL;
  m_cellflagSize = -1;
  m_contentsSize = -1;
}

EXTRACTED_LEVEL::~EXTRACTED_LEVEL(void)
{
  m_origLevel = -1;
  if (m_pCellflags != NULL) free(m_pCellflags);
  m_pCellflags = NULL;
  if (m_pContents != NULL) free(m_pContents);
  m_pContents = NULL;
  m_cellflagSize = -1;
  m_contentsSize = -1;
}

void CCSBuildView::SwapAllReferences(int level1, int level2)
{
  LOCATIONREL locr;
  RN obj;
  DBTYPE objType;
  for (locr.First(); locr.l()>=0; locr.Next())
  {
    for (obj = pDoc->FindFirstObject(locr);
         obj != RNeof;
         obj = pDoc->GetDBRecordLink(obj))
    {
      objType = obj.dbType();
      switch (objType)
      {
      case dbTELEPORTER:
        {
          DB1 *pTeleporter;
          pTeleporter = pDoc->GetRecordAddressDB1(obj);
          pTeleporter->SwapDestLevel(level1, level2);
          break;
        };
      case dbACTUATOR:
        {
          DB3 *pActuator;
          pActuator = pDoc->GetRecordAddressDB3(obj);
          pActuator->SwapTargetLevel(level1, level2, &locr);
          break;
        };
      default: break;
      };
    };
  };
}

