// ShowSkins.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "CSBuildView.h"
#include "misc.h"
#include "ShowSkins.h"


extern CCSBuildView *pMainView;

static i32 recentSkinNum = 0;


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// ShowSkins dialog


void ShowSkins::FillSkinList(void)
{
  int i;
  CString line;
  LOCATIONABS loca;
  for (i=0; i<256; i++)
  {
    line.Format("%03d", i);
    m_skinList.AddString(line);
  };
}



void ShowSkins::DrawLevel(CPaintDC *pDC, 
                              int level)
{
  CDC myDC;
  int row, col, n;
  CBitmap *origBitmap;
  CPen pen1, pen2;
  myDC.CreateCompatibleDC(pDC);
  origBitmap = myDC.SelectObject(&pMainView->m_tileBitmap);
  for (row=0; row<=pDoc->LastRow(level); row++)
  {
    for (col=0; col<=pDoc->LastColumn(level); col++)
    {
      if (m_showAll)
      {
        n = skinCache.GetSkin(level, col, row);
        pMainView->DrawCell(pDC, &myDC, level, col, row,
                       ((n==0)?0:(n==m_currentSkin)?2:1)); 
      }
      else
      {
        pMainView->DrawCell(pDC, &myDC, level, col, row,
                            (m_currentSkin != 0) &&
                            (skinCache.GetSkin(level, col, row) == m_currentSkin)); 
      }
    };
  };
}

ShowSkins::ShowSkins(CWnd* pParent /*=NULL*/)
	: CDialog(ShowSkins::IDD, pParent)
{
	//{{AFX_DATA_INIT(ShowSkins)
	m_showAll = FALSE;
	m_currentLevel = 0;
	//}}AFX_DATA_INIT

  m_clickRow = -1;
}


void ShowSkins::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(ShowSkins)
	DDX_Control(pDX, IDC_LevelSpin, m_levelSpin);
	DDX_Control(pDX, IDC_SkinList, m_skinList);
	DDX_Check(pDX, IDC_ShowAll, m_showAll);
	DDX_Text(pDX, IDC_Level, m_currentLevel);
	DDV_MinMaxUInt(pDX, m_currentLevel, 0, 63);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(ShowSkins, CDialog)
	//{{AFX_MSG_MAP(ShowSkins)
	ON_LBN_SELCHANGE(IDC_SkinList, OnSelchangeSkinList)
	ON_WM_PAINT()
	ON_WM_LBUTTONDOWN()
	ON_EN_CHANGE(IDC_Level, OnChangeLevel)
	ON_BN_CLICKED(IDC_ShowAll, OnShowAll)
	ON_NOTIFY(UDN_DELTAPOS, IDC_LevelSpin, OnDeltaposLevelSpin)
	ON_WM_RBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// ShowSkins message handlers

BOOL ShowSkins::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
  SetWindowText(m_title);
  FillSkinList();
  m_skinList.SetCurSel(0);
  OnSelchangeSkinList();
  GotoDlgCtrl(GetDlgItem(IDC_SkinList));
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void ShowSkins::OnSelchangeSkinList() 
{
	// TODO: Add your control notification handler code here
  int n;
  n = m_skinList.GetCurSel();
  if (n < 0) n = recentSkinNum;
  recentSkinNum = n;
  m_currentSkin = n;
  Invalidate();	
}

void ShowSkins::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	// TODO: Add your message handler code here
  DrawLevel(&dc, m_currentL);
	
	// Do not call CDialog::OnPaint() for painting messages
}

void ShowSkins::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	
  int row, col, n;
	CDialog::OnLButtonDown(nFlags, point);
  col = point.x / pMainView->m_tileSize;
  row = point.y / pMainView->m_tileSize;
  if ( (col == m_clickCol) && (row == m_clickRow) ) return;
  m_clickCol = col;
  m_clickRow = row;
  if (    (col > pDoc->LastColumn(m_currentL))
       || (row > pDoc->LastRow(m_currentL)) 
     )
  {
    m_clickRow = -1;
    return;
  };
  if (m_showAll)
  {
    n = skinCache.GetSkin(m_currentL, col, row);
    m_skinList.SetCurSel(n);
    m_currentSkin = n;
  }
  else
  {
    skinCache.SetSkin(m_currentL, col, row, (i8)m_currentSkin);
  };
  Invalidate();
}

void ShowSkins::OnMouseMove(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default	
	CDialog::OnMouseMove(nFlags, point);
  if (m_clickRow >= 0) OnLButtonDown(nFlags, point);
}

void ShowSkins::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	
	CDialog::OnLButtonUp(nFlags, point);
  m_clickRow = -1;
}


void ShowSkins::OnChangeLevel() 
{
	// TODO: If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.
	
	// TODO: Add your control notification handler code here
  UpdateData(true);
  if (m_currentLevel >= (unsigned)pDoc->NumLevel()) return;
  if (m_currentLevel == (unsigned)m_currentL) return;
  m_currentL = m_currentLevel;
  Invalidate();

}

void ShowSkins::OnShowAll() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  Invalidate();
}

void ShowSkins::OnDeltaposLevelSpin(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	// TODO: Add your control notification handler code here
  UpdateData(true);
	if (pNMUpDown->iDelta == 1)
  {
    if (m_currentL < pDoc->NumLevel() - 1) m_currentL++;
  }
  else
  {
    if (m_currentL > 0) m_currentL--;
  };
  m_currentLevel = m_currentL;
  UpdateData(false);
  Invalidate();
	*pResult = 0;
}

void ShowSkins::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
  int row, col;
	CDialog::OnLButtonDown(nFlags, point);
  col = point.x / pMainView->m_tileSize;
  row = point.y / pMainView->m_tileSize;
  if (col > pDoc->LastColumn(m_currentL)) return;
  if (row > pDoc->LastRow(m_currentL)) return;
  skinCache.SetSkin(m_currentL, col, row, 0);

  Invalidate();
	
	CDialog::OnRButtonDown(nFlags, point);
}

