#include <stdafx.h>
//#include "types.h"
#include "CSBuildDoc.h"

bool bigEndian = true;

ui16 LE16(ui32 w)
{
  if (bigEndian)
  {
    return (ui16)(((w>>8) & 0xff) | ((w << 8) & 0xff00));
  }
  else
  {
    return (i16)(w & 0xffff);
  };
}

ui32 LE32(ui32 i)
{
  if (bigEndian)
  {
    return   ((i>>24) & 0xff)
           | ((i>>8)  & 0xff00)
           | ((i<<8)  & 0xff0000)
           | ((i<<24)  & 0xff000000);
  }
  else
  {
    return i;
  };
}



void BO(ui16& word)
{
  if (bigEndian) word = LE16(word);
}

void BO(i16& word)
{
  if (bigEndian) word = LE16(word);
}

void BO(i32& longword)
{
  if (bigEndian) longword = LE32(longword);
}


void CCSBuildDoc::swapBlock1(GAMEBLOCK1 *b)
{ // Swap and reverse byte order in first 512-byte
  // block of savegame file.
  if (m_littleEndian) return;
  BO(b->RandomGameID);
  BO(b->Block2Hash);       //428
  BO(b->ITEM16Hash);       //426
  BO(b->CharacterHash);    //424
  BO(b->TimersHash);       //422
  BO(b->TimerQueHash);     //420

  BO(b->Block2Checksum);   //396
  BO(b->ITEM16Checksum);   //394
  BO(b->CharacterChecksum);//392
  BO(b->TimersChecksum);   //390
  BO(b->TimerQueChecksum); //388

  BO(b->Word22594);        //364
  BO(b->Word22592);        //362
}

void CCSBuildDoc::swapPCBlock1(PCGAMEBLOCK1 *b)
{ // Swap and reverse byte order in first 512-byte
  // block of savegame file.
  if (m_littleEndian) return;
  BO(b->RandomGameID);
  BO(b->Block2Hash);       //428
  BO(b->ITEM16Hash);       //426
  BO(b->CharacterHash);    //424
  BO(b->TimersHash);       //422
  BO(b->TimerQueHash);     //420

  BO(b->Block2Checksum);   //396
  BO(b->ITEM16Checksum);   //394
  BO(b->CharacterChecksum);//392
  BO(b->TimersChecksum);   //390
  BO(b->TimerQueChecksum); //388

  BO(b->Word22594);        //364
  BO(b->Word22592);        //362
}

void CCSBuildDoc::swapBlock2(GAMEBLOCK2 *b)
{ // Swap and reverse byte order in second 128-byte
  // block of savegame file.
  if (m_littleEndian) return;
  BO(b->Time);          //228
  BO(b->ranseed);       //224
  BO(b->ObjectInHand);  //220
  BO(b->numcharacter);  //218
  BO(b->partyx);        //216
  BO(b->partyy);        //214
  BO(b->partyfacing);   //212
  BO(b->partyLevel);    //210
  BO(b->handChar);      //208
  BO(b->MagicCaster);   //206
  BO(b->NumTimer);      //204
  BO(b->NextAvailTimer);//202
  BO(b->MaxTimers);     //200
  BO(b->ITEM16QueLen);  //198
  BO(b->Long12950);     //196
  BO(b->Long12954);     //192
  BO(b->Word11710);     //188
  BO(b->Word11712);     //186
  BO(b->Word11714);     //184
  BO(b->MaxITEM16);     //182
}

void CCSBuildDoc::swapTimerQue(void)
{
  int i;
  for (i=0; i<m_MaxTimers; i++)
    m_timerQue[i]=LE16(m_timerQue[i]);
}

void CCSBuildDoc::swapTimers(void)
{
  int i;
  for (i=0; i<m_MaxTimers; i++)
  {
    m_Timers[i].time = LE32(m_Timers[i].time);
    switch (m_Timers[i].timerFunction)
    {
    case 24: m_Timers[i].swapObj8();
             break;
    case 25: m_Timers[i].swapObj8();
             break;
    case 48:
    case 49: m_Timers[i].swapObj6(); // The flying object
             m_Timers[i].swapWord8();// x, y, and direction
             break;
    case 60:
    case 61: m_Timers[i].swapObj8();
             break;
    case 70: m_Timers[i].swapWord6();
             break;
    case 72: m_Timers[i].swapWord6();
             break;
    case 75: m_Timers[i].swapWord6();
             break;
    };
  };
}

void CCSBuildDoc::swapCharacter(int i)
{
  int j;
  CHARDESC *pc;
  i16 *pp;
  pc = &m_Characters[i];
  pc->busyTimer = LE16(pc->busyTimer);
  pc->timerIndex = LE16(pc->timerIndex);
  pc->word48 = LE16(pc->word48);
  pc->ouches = LE16(pc->ouches);
  pc->HP = LE16(pc->HP);
  pc->maxHP = LE16(pc->maxHP);
  pc->stamina = LE16(pc->stamina);
  pc->maxStamina = LE16(pc->maxStamina);
  pc->mana = LE16(pc->mana);
  pc->maxMana = LE16(pc->maxMana);
  pc->word64 = LE16(pc->word64);
  pc->food = LE16(pc->food);
  pc->water = LE16(pc->water);
  for (j=0; j<20; j++)
  {
    pc->skills92[j].word0 = LE16(pc->skills92[j].word0);
    pc->skills92[j].Long2 = LE32(pc->skills92[j].Long2);
  };
  for (j=0; j<30; j++)
  {
    pp = (i16 *)&pc->possessions[j];
    *pp = LE16(*pp);
  };
  pc->load = LE16(pc->load);
  pc->shieldStrength = LE16(pc->shieldStrength);
}


void CCSBuildDoc::swapCharacterData(void)
{
  int i;
  for (i=0; i<4; i++) swapCharacter(i);
  //m_Brightness = LE16(m_Brightness);
  brightness = LE16(brightness);
  for (i=0; i<24; i++)
    m_Word13268[i]=LE16(m_Word13268[i]);
}

void CCSBuildDoc::swapTextArray(ui16 *text)
{
  int i;
  ui16 *pWord;
  for (i=0; i<m_dungeonDatIndex->NumWordsInTextArray(); i++) 
  {
    pWord = text + i;
    *pWord = LE16(*pWord);
  };
}



void CCSBuildDoc::swapITEM16s(void)
{ // Fix byte order in ITEM16 entries
  int i;
  for (i=0; i<m_MaxITEM16; i++)
  {
    m_Item16[i].word0 = LE16(m_Item16[i].word0);
  };
}


void TIMER::swapWord6(void)
{
  *(i16 *)(&uByte6) = LE16(*(i16 *)(&uByte6));
}


void TIMER::swapObj8(void)
{
  *(i16 *)(&uByte8) = LE16(*(i16 *)(&uByte8));
}

void TIMER::swapObj6(void)
{
  *(i16 *)(&uByte6) = LE16(*(i16 *)(&uByte6));
}


void TIMER::swapWord8(void)
{
  *(i16 *)(&uByte8) = LE16(*(i16 *)(&uByte8));
}

void CCSBuildDoc::swapObjectList(i32 num)
{
  i32 i;
  for (i=0; i<num; i++) m_objectList[i].LittleEndian();
}

void CCSBuildDoc::swapPointer10454(void)
{
  for (i32 i=0; i<m_numColumnPointers; i++)
    m_objectColumnIndex[i] = LE16(m_objectColumnIndex[i]);
}

void CCSBuildDoc::swapLevelDescriptors(int numLevel)
{
  for (i32 j=0; j<numLevel; j++)
  {
    m_pLevelDescriptors[j].swap();
  };
}

void LEVELDESC::swap(void)
{
  m_cellOffset = LE16(m_cellOffset);
  word8 = LE16(word8);
  word10= LE16(word10);
  word12= LE16(word12);
  word14= LE16(word14);
}

void CCSBuildDoc::swapDungeonDatIndex(void)
{
  m_dungeonDatIndex->Swap();
}

