// WarningDialog.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "WarningDialog.h"
#include "FixCloneDlg.h"
#include "Misc.h"

bool FixSharedText(CHKERRMSG *pcem);
bool FixClonedIndirectTextEntry(CHKERRMSG *pcem);

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// WarningDialog dialog


WarningDialog::WarningDialog(CWnd* pParent /*=NULL*/)
	: CDialog(WarningDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(WarningDialog)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}

CString WarningDialog::GetGraphicName(const CHKERRMSG *pcem)
{
  RN obj;
  DB3 *pDB3;
  int indirectIndex;
  indirectIndex = pcem->CloneIndirectIndex();
  for (pcem=pcem->Next(); pcem!=NULL; pcem=pcem->Next())
  {
    switch(pcem->ErrorType())
    {
    case CEMT_CLONE: return CString("{unknown}");
    case CEMT_PARENTCELL:
    case CEMT_CLONEDPARENTLIST:
    case CEMT_UNCLONEDPARENTLIST:
      ASSERT(pcem->CloneIndirectIndex() == indirectIndex);
      obj.ConstructFromInteger(indirectIndex);
      pDB3 = pDoc->GetRecordAddressDB3(obj);
      return CString("{") + pDB3->GraphicName(pcem->Location()) + "}";
    };
  };
  return CString("{unknown}");
}

void WarningDialog::DoDataExchange(CDataExchange* pDX)
{
  CString line;
  int numLine, newIndex, oldIndex, i;
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(WarningDialog)
	DDX_Control(pDX, IDC_List, m_list);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    const CHKERRMSG *pcem;
    int index;
    oldIndex = m_list.GetCurSel();
    if (oldIndex < 0) oldIndex = 0;
    m_list.ResetContent();
    for (pcem = m_chkErrMsg->FirstError();
         pcem != NULL;
         pcem = m_chkErrMsg->NextError(pcem))
    {
      if (pcem->Repaired())
      {
        line = "Repaired ";
      }
      else
      {
        line = "";
      };
      line += pcem->ErrorMessage();
      if (pcem->ErrorType() == CEMT_CLONE)
      {
        if (pDoc->DataType(pcem->CloneIndirectIndex()) 
                        == dbACTUATOR)
        {
          line += GetGraphicName(pcem);
        };
      };
      index = m_list.AddString(line);
      m_list.SetItemDataPtr(index, (void *)pcem);
    };
    numLine = m_list.GetCount();
    newIndex = -1;
    for (i=oldIndex; i<numLine; i++)
    {
      pcem = (CHKERRMSG *)m_list.GetItemDataPtr(i);
      if (pcem->Repaired()) continue;
      newIndex = i;
      break;
    };
    if (newIndex == -1)
    {
      for (i=oldIndex; i>=0; i--)
      {
        pcem = (CHKERRMSG *)m_list.GetItemDataPtr(i);
        if (pcem->Repaired()) continue;
        newIndex = i;
        break;
      };
    };
    m_list.SetCurSel(newIndex);
  };
}


BEGIN_MESSAGE_MAP(WarningDialog, CDialog)
	//{{AFX_MSG_MAP(WarningDialog)
	ON_BN_CLICKED(IDC_WriteToFile, OnWriteToFile)
	ON_BN_CLICKED(IDC_Repair, OnRepair)
	ON_LBN_DBLCLK(IDC_List, OnDblclkList)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// WarningDialog message handlers

BOOL WarningDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	// TODO: Add extra initialization here
  m_repaired = false;
	if (m_title != "")
  {
    SetWindowText(m_title);
  };
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void WarningDialog::OnWriteToFile() 
{
	// TODO: Add your control notification handler code here
  FILE *f;
  int i;
	CFileDialog dlg(false,"txt","Warnings.txt");
  if (dlg.DoModal() != IDOK) return;
  f = fopen(dlg.GetPathName(),"w");
  if (f == NULL)
  {
    Error("I cannot open that file for writing");
    return;
  };
  for (i=0; i<m_list.GetCount(); i++)
  {
    CString t;
    m_list.GetText(i,t);
    fprintf(f,"%s\n",t);
  };
  fclose(f);
}


bool FixBadlyPositionedMonster(CHKERRMSG *pcem)
{
  RN objMon;
  DB4 *pMonster;
  objMon.ConstructFromInteger(pcem->Index1());
  pMonster = pDoc->GetRecordAddressDB4(objMon);
  pMonster->positions(0xd8); // 0, 2, 1, 3
  return true;
}

void WarningDialog::OnRepair() 
{
	// TODO: Add your control notification handler code here
	int index;
  CHKERRMSG *pcem;
  UpdateData(true);
  index = m_list.GetCurSel();
  if (index < 0)
  {
    pDoc->Sorry("No Item Has Been Selected");
    return;
  };
  pcem = (CHKERRMSG *)m_list.GetItemDataPtr(index);
  if (pcem == NULL)
  {
    pDoc->Sorry("No Repair Implemented");
    return;
  };
  switch (pcem->ErrorType())
  {
  case CEMT_TEXTPLUSRANDOMDECORATION:
    {
      LOCATIONREL locr;
      CELLFLAG cf;
      locr = pcem->Location();
      cf = pDoc->GetCellFlag(locr);
      cf &= ~(8 >> locr.p());
      pDoc->SetCellFlag(locr, cf);
      pcem->Repaired(true);
      m_repaired = true;
      UpdateData(false);
    };
    break;
  case CEMT_RANDOMDECORATIONONSKINNEDWALL:
    {
      LOCATIONREL locr;
      CELLFLAG cf;
      CString str;
      locr = pcem->Location();
      cf = pDoc->GetCellFlag(locr);
      cf &= ~15;
      pDoc->SetCellFlag(locr, cf);
      pcem->Repaired(true);
      m_repaired = true;
      m_list.GetText(index, str);
      m_list.DeleteString(index);
      m_list.InsertString(index, CString("Repaired ") + str);
      m_list.SetItemDataPtr(index, pcem);
      m_list.SetCurSel(index+1);
      //UpdateData(false);
    };
    break;
  case CEMT_RANDOMDECORATIONONSKINNEDFLOOR:
    {
      LOCATIONREL locr;
      CELLFLAG cf;
      CString str;
      locr = pcem->Location();
      cf = pDoc->GetCellFlag(locr);
      cf &= ~8;
      pDoc->SetCellFlag(locr, cf);
      pcem->Repaired(true);
      m_repaired = true;
      m_list.GetText(index, str);
      m_list.DeleteString(index);
      m_list.InsertString(index, CString("Repaired ") + str);
      m_list.SetItemDataPtr(index, pcem);
      m_list.SetCurSel(index+1);
      //UpdateData(false);
    };
    break;
  case CEMT_RANDOMDECORATIONONDOORFRAME:
    {
      LOCATIONREL locr;
      CELLFLAG cf;
      locr = pcem->Location();
      cf = pDoc->GetCellFlag(locr);
      cf &= ~pcem->Index1();
      pDoc->SetCellFlag(locr, cf);
      pcem->Repaired(true);
      m_repaired = true;
      UpdateData(false);
    };
    break;
   case CEMT_CHARPOSSESSIONISLIST:
    {
      RN firstPossession, secondPossession;
      DBCOMMON *pFirstDB;
      LOCATIONREL locr;
      firstPossession = 
           pDoc->CharacterPossession(pcem->Index1(),pcem->Index2());
      ASSERT(firstPossession != RNnul);
      if (firstPossession == RNnul) break;
      secondPossession = pDoc->GetDBRecordLink(firstPossession);
      ASSERT(secondPossession != RNeof);
      if (secondPossession == RNeof) break;
      pFirstDB = pDoc->GetCommonAddress(firstPossession);
      pFirstDB->link(RNeof);
      m_repaired = true;
      pcem->Repaired(true);
      if (pDoc->DungeonFormat())
      {
        locr.x(pDoc->StartingX());
        locr.l(0);
        locr.y(pDoc->StartingY());
      }
      else
      {
        locr.x(pDoc->PartyX());
        locr.l(pDoc->PartyLevel());
        locr.y(pDoc->PartyY());
      };
      locr.p(0);
      for (firstPossession = secondPossession;
           firstPossession!=RNeof;
           firstPossession=secondPossession)
      {
        pFirstDB = pDoc->GetCommonAddress(firstPossession);
        secondPossession = pDoc->GetDBRecordLink(firstPossession);
        pFirstDB->link(RNeof);
        pDoc->AddObjectToRoom(firstPossession, locr);
      };     
      UpdateData(false);
    };
    break;
  case CEMT_INDIRECTNOTREFERENCED:
    {
      RN obj;
      LOCATIONREL locr;
      locr.l(-1);
      obj.ConstructFromInteger(pcem->Index1());
      if (EditOrphan(obj) == 0) break;
    };
    pDoc->ReleaseObjectIndirectIndex(pcem->Index1());
    pcem->Repaired(true);
    m_repaired = true;
    UpdateData(false);
    break;
  case CEMT_CLONE:
    {
      FixCloneDlg dlg;
      dlg.m_pcem = pcem;
      dlg.DoModal();
      if (dlg.m_modified) 
      {
        m_repaired = true;
        OnOK();
        return;
      };
    };
    break;
  case CEMT_SCROLLNONTEXT:
    if (pDoc->RepairScrollNonText(pcem))
    {
      m_repaired = true;
      OnOK();
      return;
    };
    break;
  case CEMT_ORPHANEDOBJECT:
    if (pDoc->RemoveOrphan(pcem))
    {
      pcem->Repaired(true);
      m_repaired = true;
      UpdateData(false);
    };
    break;
  case CEMT_OVERLAPPEDTEXT:
    {
      NotImplemented();
      //FixOverlappedText dlg;
      //dlg.m_pcem = pcem;
      //dlg.DoModal();
      //if (dlg.m_modified) 
      //{
      //  pcem->Repaired(true);
      //  m_repaired = true;
      //  UpdateData(false);
      //};
    };
    break;
  case CEMT_UNKNOWNCELL:
    pDoc->SetRoomType(pcem->Location(),roomSTONE);
    UpdateData(false);
    pcem->Repaired(true);
    m_repaired = true;
    break;
  case CEMT_CLONEDTEXT1:
    {
      if (FixSharedText(pcem))
      {
        m_repaired = true;
        for (;;)
        {
          pcem->Repaired(true);
          pcem->Next();
          if (pcem == NULL) break;
          if (pcem->ErrorType() != CEMT_CLONEDTEXT2) break;
        };
        UpdateData(false);
      };
    };
    break;
  case CEMT_SINGLESHAREDTEXT:
    {
      int textindex, len;
      DB2 *pDB2;
      ui16 *text;
      pDB2 = pDoc->GetRecordAddressDB2(pcem->Index1());
      textindex = pDB2->index();
      len = pDoc->ComputeTextLength(textindex);
      text = pDoc->FindRawText(textindex);
      pDB2->index(pDoc->AddText(text, len));
      pcem->Repaired(true);
      m_repaired = true;
      UpdateData(false);
    };
    break;
  case CEMT_CLONEDINDIRECTTEXTENTRY:
    {
      if (FixClonedIndirectTextEntry(pcem))
      {
        m_repaired = true;
        do
        {
          pcem->Repaired(true);
          pcem = pcem->Next();
        } while (
                    (pcem != NULL)
                 && (pcem->ErrorType() 
                         == CEMT_CLONEDINDIRECTTEXTINDEXPARENT)
                );
        UpdateData(false);
      };
    };
    break;
  case CEMT_MONSTERPOSITIONEDBADLY:
    {
      if (FixBadlyPositionedMonster(pcem))
      {
        pcem->Repaired(true);
        m_repaired = true;
        UpdateData(false);
      };
    };
    break;
  default:
    pDoc->Sorry("No Repair Implemented");
    break;
  };
}

void WarningDialog::OnDblclkList() 
{
	// TODO: Add your control notification handler code here
	OnRepair();
}
