// myClistCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "myClistCtrl.h"
#include "misc.h"
//#include "wingdi.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// myClistCtrl

myClistCtrl::myClistCtrl()
{
  m_editMode = false;
}

myClistCtrl::~myClistCtrl()
{
}


BEGIN_MESSAGE_MAP(myClistCtrl, CListCtrl)
	//{{AFX_MSG_MAP(myClistCtrl)
	ON_WM_LBUTTONDOWN()
	ON_WM_CREATE()
	ON_WM_CTLCOLOR()
	ON_WM_DRAWITEM()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// myClistCtrl message handlers

void myClistCtrl::SetDSA(DSA *pDSA)
{
  int state, relState, messageIndex, row, column;
  CString pgm;
  char text[20];
  for (relState=0; relState<16; relState++)
  {
    row = relState + 2 + relState/4;
    state = pDSA->FirstDisplayedState() + relState;
    text[0] = 0;
    sprintf(text, "%d", state);
    SetItemText(row, 0, text);
    for (messageIndex=0; messageIndex<12; messageIndex++)
    {
      column = messageIndex + 1;
      pgm       = pDSA->Pgm(state, messageIndex);
      SetItemText(row,column,pgm);
    };
  };
}

void myClistCtrl::GetDSA(DSA *pDSA)
{
  int state, relState, msgIndex;
  CString text;
  for (relState=0; relState<16; relState++)
  {
    state = pDSA->FirstDisplayedState() + relState;
    for (msgIndex=0; msgIndex<12; msgIndex++)
    {
      text = GetStateEntry(relState, msgIndex);      
      if (!pDSA->Pgm(state, msgIndex, text))
      {
        //Strange ("Illegal state entry.\nCell not updated.");
      };
    };
  };
}


RECT myClistCtrl::GetRectangle(int state, int msgIndex)
{
  RECT result;
  int row, column;
  row = state+2+state/4;
  column = msgIndex;
  result.top = row*14 + 1;
  result.bottom = result.top + 12;
  result.left = 34 + 62 * column;
  result.right = result.left + 60;
  return result;
}



bool myClistCtrl::SelectEditCell(int state, int messageIndex)
{
  RECT rect;
  TRACE3("SelectEditCell %d %d %d\n",
         m_editMode, state, messageIndex);
  if (m_editMode)
  {
    if ( (state==m_editState) && (messageIndex==m_editMessageIndex) )
    {
      return true;
    };
    if (!DeselectEditCell()) return false;
  };
  TRACE0("SelectEditCell - Activating cell\n");
  rect = GetRectangle(state, messageIndex);
  m_editCell.SetWindowPos(&wndTop,
                          rect.left,
                          rect.top,
                          rect.right-rect.left+20,
                          rect.bottom-rect.top+1,
                          0);
  m_editCell.EnableWindow(true);
  m_editCell.SetFocus();
  m_editCell.ShowWindow(true);
  m_editMode = true;
  m_editState = state;
  m_editMessageIndex = messageIndex;
  m_editCellText = GetStateEntry(state, messageIndex);
  m_editCell.SetWindowText(m_editCellText);
  Invalidate(true);
  return false;
}

bool myClistCtrl::EditMode(bool enableEdit, DSA *pDSA)
{
  if (enableEdit == m_editMode) return true;
  if (enableEdit)
  {
    SelectEditCell(m_editState, m_editMessageIndex);
    return true;
  }
  else
  {
    DSAAction action;
    char *text;
    int len;
    TRACE3("DisableEdit %d %d %d\n",
          m_editMode, m_editState, m_editMessageIndex);
    TRACE0("DisableEdit   Set m_editMode false\n");
    m_editMode = false;
    UpdateData(true);
    len = m_editCell.LineLength(0);
    text = (char *)malloc(len+20, MALLOC041);
    if (text == NULL)
    {
      Error("Cannot allocate memory");
      return false;
    };
    m_editCell.GetLine(0,text,90);
    m_editCellText = text;
    free (text);
    TRACE0("DisableEdit   Return from UpdateData(true)\n");
    if (!SetStateEntry(m_editState,
                      m_editMessageIndex,
                      m_editCellText))
    {
      TRACE0("DisableEdit -Bad data\n");
      pDoc->Sorry("Illegal State Machine Entry\n"
                  "Data not updated");
      m_editMode = true;
      return false;
    };
    pDSA->Pgm(pDSA->FirstDisplayedState() + m_editState, 
              m_editMessageIndex, 
              m_editCellText);
    return true;
  };
}

bool myClistCtrl::DeselectEditCell(void)
{
  DSAAction action;
  char *text;
  int len;
  TRACE3("DeselectEditCell %d %d %d\n",
         m_editMode, m_editState, m_editMessageIndex);
  if (!m_editMode) return true;
  TRACE0("DeselectEditCell   Set m_editMode false\n");
  m_editMode = false;
  UpdateData(true);
  len = m_editCell.LineLength(0);
  text = (char *)malloc(len+20, MALLOC041);
  if (text == NULL)
  {
    Error("Cannot allocate memory");
    return false;
  };
  m_editCell.GetLine(0,text,len);
  m_editCellText = text;
  free (text);
  TRACE0("DeselectEditCell   Return from UpdateData(true)\n");
  if (!SetStateEntry(m_editState,
                     m_editMessageIndex,
                     m_editCellText))
  {
    TRACE0("DeselectEditCell -Bad data\n");
    pDoc->Sorry("Illegal State Machine Entry\n"
               "Data not updated");
    m_editMode = true;
    return false;
  };
  return true;
}



void myClistCtrl::OnLButtonDown(UINT nFlags, CPoint point)
{
	LVHITTESTINFO hitTestInfo;
  int state;
  //if (MessageBoxActive()) return;
  TRACE2("MyCListCtrl::OnLButtonDown %d %d\n",point.x,point.y);
  hitTestInfo.pt = point;
  SubItemHitTest(&hitTestInfo);
  if (hitTestInfo.flags == LVHT_ONITEMLABEL)
  {
    if (hitTestInfo.iSubItem == 0) return;
    if (hitTestInfo.iItem < 2) return;
    state = hitTestInfo.iItem-2;
    if (state % 5 == 4) return; //Empty line.
    state -= state/5;
    if (SelectEditCell(state,hitTestInfo.iSubItem-1))
    {
      CListCtrl::OnLButtonDown(nFlags, point);
      return;
    };
  }
  else
  {
    CListCtrl::OnLButtonDown(nFlags, point);
    TRACE0("Pass LButton to CListCtrl");
  };
}


int myClistCtrl::SetStateEntry(int state,
                                int messageIndex,
                                CString& str)
{
  DSAAction action;
  int row, col;
  const char *pC;
  TRACE3("Setting State Entry %d %d %s\n",
         state, messageIndex, str);
  col = messageIndex + 1;
  row = state + 2 + state/4;
  SetItem(row,
          col,
          LVIF_TEXT,
          str,
          0,0,0,0);
  pC = (LPCTSTR)str;
  return action.DecodeFromASCII(pC, state);
}

CString myClistCtrl::GetStateEntry(int state,
                                   int messageIndex)
{
  int row, col;
  char text[10000];
  col = messageIndex + 1;
  row = state + 2 + state/4;
  GetItemText(row,
              col,
              text,
              9990);
  return CString(text);
}



int myClistCtrl::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CListCtrl::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
  {
	  RECT rect = {0,0,160,17};
    m_editCell.Create(WS_CHILD|ES_AUTOHSCROLL|ES_UPPERCASE,rect,this,IDC_EditCell);
    m_editCell.ShowWindow(false);
  };
  CFont *pFont;
  pFont = GetFont();
  LOGFONT lfont;
  pFont->GetLogFont(&lfont);
  m_fnt.CreateFontIndirect(&lfont);
  m_editCell.SetFont(&m_fnt);

	return 0;
}

void myClistCtrl::OnOK(void)
{
  DeselectEditCell();
}

CBrush brush(RGB(255,255,128));

HBRUSH myClistCtrl::OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor) 
{
	HBRUSH hbr = CListCtrl::OnCtlColor(pDC, pWnd, nCtlColor);
	
	// TODO: Change any attributes of the DC here
	hbr = (HBRUSH)GetStockObject(GRAY_BRUSH);
  pDC->SetBkColor(RGB(255,255,128));
	// TODO: Return a different brush if the default is not desired
	return brush;
}

void myClistCtrl::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
  int row, column;
  int x, y;
  char text[20];
  row = lpDrawItemStruct->itemID;
  y = row*14+1;
  if (row%5 == 1) 
  {
    Rectangle(lpDrawItemStruct->hDC,0,row*14+1,777,row*14+14);
  }
  else
  {
    for (column=0; column<=12; column++)
    {
      x=1;
      if (column>0)
      {
        x=34 + (column-1)*62;
      };
      GetItemText(row,column,text,19);
      TextOut(lpDrawItemStruct->hDC,x,y,text,strlen(text));
    };
  };
}

void myClistCtrl::OnDrawItem(int nIDCtl, LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
	// TODO: Add your message handler code here and/or call default
	
	CListCtrl::OnDrawItem(nIDCtl, lpDrawItemStruct);
}
