#pragma warning (disable:4243)

#define TILEROWS 13
#define TILECOLUMNS 3

#define MAXDBSIZE 8000


typedef          _int8     i8;
typedef unsigned _int8    ui8;
typedef          _int16   i16;
typedef unsigned _int16  ui16;
typedef          _int32   i32;
typedef unsigned _int32  ui32;


#define MAXHP 999
#define MAXSTAMINA 9999
#define MAXMANA 900

class LOCATIONABS;
class LEVELINFO;
class CLIPBOARD;

class EXPORTLOC
{
public:
  int left; //= locr.x
  int top;  //= locr.y
  int level;
  int dl;
  int dx;   //relative to left
  int dy;   //relative to top
  int pos;
  CString Format() const;
};

class LOCATIONREL
{
  char m_l; //level
  char m_x;  //column
  char m_y;  //row
  char m_p;  //position
public:
  LOCATIONREL(void){m_l=-1;};
  LOCATIONREL(int l, int x, int y)
  {
    m_l=(char)l;m_x=(char)x;m_y=(char)y;m_p=0;
  };
  LOCATIONREL(const EXPORTLOC& eloc);
  LOCATIONREL& operator =(const LOCATIONREL &);
  LOCATIONREL& operator =(const LOCATIONABS &);
  LOCATIONREL operator +(const LOCATIONREL& locr);
  bool operator ==(const LOCATIONREL &)const;
  int  l(void)const{return m_l;};
  int  x(void)const{return m_x;};
  int  y(void)const{return m_y;};
  int  p(void)const{return m_p;};
  void l(int l){m_l=(char)l;};
  void x(int x){m_x=(char)x;};
  void y(int y){m_y=(char)y;};
  void p(int p){m_p=(char)p;};
  void Vertical(int n);
  bool First(void);
  bool Next(void); //Scan the dungeon. l = -1 at start and end;
  ui32 Integer(void) const; //Pack in a single 32-bit integer.
  void Integer(int i); //Unpack from 32-bit integer
  CString Format(void) const;
  bool AdjacentCell(int direction);
};

class LOCATIONABS
{
public:
  int l; //level
  int x; //column
  int y; //row
  int p; //position
  LOCATIONABS(void){l=-1;};
  LOCATIONABS& operator =(const LOCATIONABS &);
  LOCATIONABS& operator =(const LOCATIONREL &);
  void Vertical(int n);
  bool Next(void); //Scan the dungeon. l = -1 at start and end;
};

class FLOORTYPE
{
//public:
  enum m_FLOORTYPE
  {
    floor_2small        =-1,
    floor_SquareDrain   = 0,
    floor_SquarePad     = 1,
    floor_PoisonPuddle  = 2,
    floor_RoundGrid     = 3,
    floor_HexagonalPad  = 4,
    floor_FirePot       = 5,
    floor_GreyMark      = 6,
    floor_TinyPad       = 7,
    floor_WaterPuddle   = 8,
    floor_2large        = 9 
  } m_floorType;
public:
  FLOORTYPE(m_FLOORTYPE floorType = floor_2small);
  FLOORTYPE(int floorTypeIndex);
  CString Name(void);
  ui8 Integer(void){return (ui8)m_floorType;};
  void Integer(int i){m_floorType=(m_FLOORTYPE)i;};
  FLOORTYPE& operator =(const char *name);
  bool operator ==(FLOORTYPE& ft){return m_floorType==ft.m_floorType;};
  void FirstType(void);
  bool IsLegal(void);
  bool IsLegal(int floorDecoration);
  void NextType(void);
};

class ORNATETYPE
{
  enum m_ORNATETYPE
  {
    ornate_2small          = -1,
    ornate_SquareGrid      =  0,
    ornate_MetalBars       =  1,
    ornate_Jewels          =  2,
    ornate_WoodVen         =  3,
    ornate_ArchedGrid      =  4,
    ornate_BlockLock       =  5,
    ornate_CornerMetal     =  6,
    ornate_Ebony           =  7,
    ornate_RedTriangleLock =  8,
    ornate_TriangleLock    =  9,  // never used in DM nor CSB
    ornate_BlueField       = 10,
    ornate_DamagedMetal    = 11,
    ornate_2large          = 12
  } m_ornateType;
public:
  ORNATETYPE(int ornateTypeIndex);
  ORNATETYPE(m_ORNATETYPE ornateType = (m_ORNATETYPE)(ornate_2small+1));
  CString Name(void);
  ui8 Integer(void){return (ui8)m_ornateType;};
  void Integer(int i){m_ornateType=(m_ORNATETYPE)i;};
  ORNATETYPE& operator =(const char *name);
  void FirstType(void);
  bool IsLegal(void);
  void NextType(void);
};


//..............................................................................

// 60 types for wall graphics

class WALLTYPE
{
  enum m_WALLTYPE
  {
    wall_2small         = -1,
    wall_GarbledText    = 0x00,
    wall_SquareAlcove   = 0x01,
    wall_ViAltar        = 0x02,
    wall_OrnateAlcove   = 0x03,
    wall_Hook           = 0x04,
    wall_BlockLock      = 0x05,
    wall_WoodRing       = 0x06,
    wall_SmallSwitch    = 0x07,
    wall_Dent1          = 0x08,
    wall_Dent2          = 0x09,
    wall_IronRing       = 0x0a,
    wall_Crack          = 0x0b,
    wall_SlimeOutlet    = 0x0c,
    wall_Point          = 0x0d,
    wall_TinySwitch     = 0x0e,
    wall_GreenButtonOut = 0x0f,
    wall_BlueButtonOut  = 0x10,
    wall_CoinSlot       = 0x11,
    wall_DoubleLock     = 0x12,
    wall_SquareLock     = 0x13,
    wall_WingedLock     = 0x14,
    wall_OnyxLock       = 0x15,
    wall_StoneLock      = 0x16,
    wall_CrossLock      = 0x17,
    wall_JeweledLock    = 0x18,
    wall_SkeletonLock   = 0x19,
    wall_GoldLock       = 0x1a,
    wall_TurquoiseLock  = 0x1b,
    wall_EmeraldLock    = 0x1c,
    wall_RubyLock       = 0x1d,
    wall_RaLock         = 0x1e,
    wall_MasterLock     = 0x1f,
    wall_GemHole        = 0x20,
    wall_Slime          = 0x21,
    wall_Grate          = 0x22,
    wall_Fountain       = 0x23,
    wall_Menacles       = 0x24,
    wall_DemonFace      = 0x25,
    wall_TorchHolderEmpty = 0x26,
    wall_DragonClaws    = 0x27,
    wall_PoisonHoles    = 0x28,
    wall_FireballHoles  = 0x29,
    wall_DaggerHoles    = 0x2a,
    wall_HeroMirror     = 0x2b,
    wall_LeverUp        = 0x2c,
    wall_LeverDown      = 0x2d,
    wall_TorchHolderFull= 0x2e,
    wall_RedButtonOut   = 0x2f,
    wall_EyeSwitch      = 0x30,
    wall_BigSwitchOut   = 0x31,
    wall_CrackSwitchOut = 0x32,
    wall_GreenButtonIn  = 0x33,
    wall_BlueButtonIn   = 0x34,
    wall_RedButtonIn    = 0x35,
    wall_BigSwitchIn    = 0x36,
    wall_CrackSwitchIn  = 0x37,
    wall_Amalgam        = 0x38,
    wall_PoweredAmalgam = 0x39,
    wall_BurntAmalgam   = 0x3a,
    wall_EvilOutside    = 0x3b,
    wall_2large         = 0x3c
  } m_wallType;
public:
  WALLTYPE(m_WALLTYPE wallType = wall_2small);
  WALLTYPE(int wallTypeIndex);
  CString Name(void);
  ui8 Integer(void){return (ui8)m_wallType;};
  void Integer(int i){m_wallType=(m_WALLTYPE)i;};
  WALLTYPE& operator =(const char *name);
  bool operator ==(WALLTYPE& wt){return m_wallType==wt.m_wallType;};
  bool IsAlcove(void);
  void FirstType(void);
  bool IsLegal(void);
  bool IsLegal(int wallDecoration);
  void NextType(void);
};

typedef ui8 CELLFLAG;
    // For roomtype = roomSTONE
    //    bit 0 = west random decoration
    //    bit 1 = south random decoration
    //    bit 2 = east random decoration
    //    bit 3 = north random decoration
    // For roomtype = roomDOOR
    //    bits 0-2 = door state (0, 1, or 5 == allow passage)
    //                  (I have yet to see the value of 1!)
    //                  (0, 4, and 5 are all I have seen).      
    //              00 = open 
    //              01 = almost open
    //              02 = half open
    //              03 = almost closed
    //              04 = closed
    //              05 = door has been bashed open.
    //    bit 3 = N/S door (direction of travel through door)
    // For roomtype = roomSTAIRS
    //    bit 2 (0x04) = stairs up
    //    bit 3 (0x08) = Traverse North/South
    // For roomtype = roomPIT
    //    bit 0 means False Pit.  Don't fall even when open.
    //    bit 2 means Obscure (not obvious)
    //    if (0x08) then pit is open (you will fall through)
    // For roomType = roomOPEN
    //    bit 3 means random decoration
    // For roomtype = roomTELEPORTER
    //    Bit 2 (0x04) means visible
    //    bit 3 (0x08) means teleporter is active
    // For roomtype = roomFALSEWALL
    //    bit 0 (0x1) means wall is open to party
    //    bit 2 (0x4) means wall is invisible and open to party 
    // 0x08   active  (like teleporter)
    // bits 5-7 room type (see enum ROOMTYPE)

enum ROOMTYPE
{
  roomSTONE      = 0,
  roomOPEN       = 1,
  roomPIT        = 2,
  roomSTAIRS     = 3,
  roomDOOR       = 4,
  roomTELEPORTER = 5,
  //roomFALSEWALL  = 6,
  roomDOOREDGE   = 16, //A door seen on edge
  roomDOORFACING = 17, //A door seen on flat side
  roomType18     = 18,
  roomType19     = 19,
  roomBLOCKING   = 20,
  roomNONBLOCKING= 21,
  roomILLEGAL    = 7,
};

ROOMTYPE RoomType(CELLFLAG cf);
ROOMTYPE RoomType(LOCATIONREL locr);
ROOMTYPE RoomType(int level, int col, int row);

void setmodifiedflag(void);
#define MODIFY setmodifiedflag();

#define BITS0_0(x)   (ui8)((((x)    )&0x0001))
#define BITS0_1(x)   (ui8)((((x)    )&0x0003))
#define BITS0_2(x)   (ui8)((((x)    )&0x0007))
#define BITS0_3(x)   (ui8)((((x)    )&0x000f))
#define BITS0_4(x)   (ui8)((((x)    )&0x001f))
#define BITS0_5(x)   (ui8)((((x)    )&0x003f))
#define BITS0_6(x)   (ui8)((((x)    )&0x007f))
#define BITS0_7(x)   (ui8)((((x)    )&0x00ff))
#define BITS0_9(x)   (((x)    )&0x03ff)
#define BITS0_8(x)   (((x)    )&0x00ff)
#define BITS1_2(x)   (((x)>> 1)&0x0003)
#define BITS1_4(x)   (((x)>> 1)&0x000f)
#define BITS2_2(x)   (((x)>> 2)&0x0001)
#define BITS3_4(x)   (((x)>> 3)&0x0003)
#define BITS3_15(x)  (((x)>> 3)&0x1fff)
#define BITS4_5(x)   (ui8) ((((x)>> 4)&0x0003))
#define BITS4_7(x)   (ui8) ((((x)>> 4)&0x000f))
#define BITS4_9(x)   (ui8) ((((x)>> 4)&0x003f))
#define BITS4_11(x)  (ui8) ((((x)>> 4)&0x00ff))
#define BITS4_15(x)  (ui16)((((x)>> 4)&0x0fff))
#define BITS5_5(x)   (ui8) ((((x)>> 5)&0x0001))
#define BITS5_6(x)   (ui8) ((((x)>> 5)&0x0003))
#define BITS5_9(x)   (ui8) ((((x)>> 5)&0x001f))
#define BITS5_15(x)  (ui16)((((x)>> 5)&0x07ff))
#define BITS6_6(x)   (ui8) ((((x)>> 6)&0x0001))
#define BITS6_7(x)   (ui8) ((((x)>> 6)&0x0003))
#define BITS6_10(x)  (ui8) ((((x)>> 6)&0x001f))
#define BITS6_11(x)  (ui8) ((((x)>> 6)&0x003f))
#define BITS7_7(x)   (ui8) ((((x)>> 7)&0x0001))
#define BITS7_8(x)   (ui8) ((((x)>> 7)&0x0003))
#define BITS7_10(x)  (ui8) ((((x)>> 7)&0x000f))
#define BITS7_11(x)  (ui8) ((((x)>> 7)&0x001f))
#define BITS7_15(x)  (ui16)((((x)>> 7)&0x01ff))
#define BITS8_8(x)   (ui8) ((((x)>> 8)&0x0001))
#define BITS8_11(x)  (ui8) ((((x)>> 8)&0x000f))
#define BITS8_9(x)   (ui8) ((((x)>> 8)&0x0003))
#define BITS8_12(x)  (ui8) ((((x)>> 8)&0x001f))
#define BITS8_14(x)  (ui8) ((((x)>> 8)&0x007f))
#define BITS8_15(x)  (ui8) ((((x)>> 8)&0x00ff))
#define BITS9_9(x)   (ui8) ((((x)>> 9)&0x0001))
#define BITS9_11(x)  (ui8) ((((x)>> 9)&0x007))
#define BITS9_12(x)  (ui8) ((((x)>> 9)&0x000f))
#define BITS9_15(x)  (ui8) ((((x)>> 9)&0x007f))
#define BITS10_10(x) (ui8) ((((x)>>10)&0x0001))
#define BITS10_13(x) (ui8) ((((x)>>10)&0x000f))
#define BITS10_14(x) (ui8) ((((x)>>10)&0x001f))
#define BITS10_11(x) (ui8) ((((x)>>10)&0x0003))
#define BITS10_15(x) (ui8) ((((x)>>10)&0x003f))
#define BITS11_11(x) (ui8) ((((x)>>11)&0x0001))
#define BITS11_14(x) (ui8) ((((x)>>11)&0x000f))
#define BITS11_15(x) (ui8) ((((x)>>11)&0x001f))
#define BITS12_12(x) (ui8) ((((x)>>12)&0x0001))
#define BITS12_13(x) (ui8) ((((x)>>12)&0x0003))
#define BITS12_15(x) (ui8) ((((x)>>12)&0x000f))
#define BITS13_13(x) (ui8) ((((x)>>13)&0x0001))
#define BITS13_14(x) (ui8) ((((x)>>13)&0x0003))
#define BITS13_15(x) (ui8) ((((x)>>13)&0x0007))
#define BITS14_14(x) (ui8) ((((x)>>14)&0x0001))
#define BITS14_15(x) (ui8) ((((x)>>14)&0x0003))
#define BITS15_15(x) (ui8) ((((x)>>15)&0x0001))

#define SETBBITS0_0(d,s)   (d) = (I8)(((d)&0xfffe)|(((s)&0x0001)    ))
#define SETBBITS0_2(d,s)   (d) = (I8)(((d)&0xfff8)|(((s)&0x0007)    ))

#define SETWBITS0_0(d,s)   (d) = (i16)(((d)&0xfffe)|(((s)&0x0001)    ))
#define SETWBITS0_1(d,s)   (d) = (i16)(((d)&0xfffc)|(((s)&0x0003)    ))
#define SETWBITS0_2(d,s)   (d) = (i16)(((d)&0xfff8)|(((s)&0x0007)    ))
#define SETWBITS0_3(d,s)   (d) = (i16)(((d)&0xfff0)|(((s)&0x000f)    ))
#define SETWBITS0_4(d,s)   (d) = (i16)(((d)&0xffe0)|(((s)&0x001f)    ))
#define SETWBITS0_5(d,s)   (d) = (i16)(((d)&0xffc0)|(((s)&0x003f)    ))
#define SETWBITS0_6(d,s)   (d) = (i16)(((d)&0xff80)|(((s)&0x007f)    ))
#define SETWBITS0_7(d,s)   (d) = (i16)(((d)&0xff00)|(((s)&0x00ff)    ))
#define SETWBITS1_4(d,s)   (d) = (i16)(((d)&0xffe1)|(((s)&0x000f)<< 1))
#define SETWBITS2_2(d,s)   (d) = (i16)(((d)&0xfffb)|(((s)&0x0001)<< 2))
#define SETWBITS3_4(d,s)   (d) = (i16)(((d)&0xffe7)|(((s)&0x0003)<< 3))
#define SETWBITS3_15(d,s)  (d) = (i16)(((d)&0x0007)|(((s)&0x1fff)<< 3))
#define SETWBITS4_5(d,s)   (d) = (i16)(((d)&0xffcf)|(((s)&0x0003)<< 4))
#define SETWBITS4_7(d,s)   (d) = (i16)(((d)&0xff0f)|(((s)&0x000f)<< 4))
#define SETWBITS4_11(d,s)  (d) = (i16)(((d)&0xf00f)|(((s)&0x00ff)<< 4))
#define SETWBITS4_15(d,s)  (d) = (i16)(((d)&0x000f)|(((s)&0x0fff)<< 4))
#define SETWBITS5_5(d,s)   (d) = (i16)(((d)&0xffdf)|(((s)&0x0001)<< 5))
#define SETWBITS5_6(d,s)   (d) = (i16)(((d)&0xff9f)|(((s)&0x0003)<< 5))
#define SETWBITS5_9(d,s)   (d) = (i16)(((d)&0xfc1f)|(((s)&0x001f)<< 5))
#define SETWBITS6_6(d,s)   (d) = (i16)(((d)&0xffbf)|(((s)&0x0001)<< 6))
#define SETWBITS6_7(d,s)   (d) = (i16)(((d)&0xff3f)|(((s)&0x0003)<< 6))
#define SETWBITS6_10(d,s)  (d) = (i16)(((d)&0xf83f)|(((s)&0x001f)<< 6))
#define SETWBITS6_11(d,s)  (d) = (i16)(((d)&0xf03f)|(((s)&0x003f)<< 6))
#define SETWBITS7_7(d,s)   (d) = (i16)(((d)&0xff7f)|(((s)&0x0001)<< 7))
#define SETWBITS7_10(d,s)  (d) = (i16)(((d)&0xf87f)|(((s)&0x000f)<< 7))
#define SETWBITS7_11(d,s)  (d) = (i16)(((d)&0xf07f)|(((s)&0x001f)<< 7))
#define SETWBITS7_15(d,s)  (d) = (i16)(((d)&0x007f)|(((s)&0x01ff)<< 7))
#define SETWBITS8_8(d,s)   (d) = (i16)(((d)&0xfeff)|(((s)&0x0001)<< 8))
#define SETWBITS8_9(d,s)   (d) = (i16)(((d)&0xfcff)|(((s)&0x0003)<< 8))
#define SETWBITS8_11(d,s)  (d) = (i16)(((d)&0xf0ff)|(((s)&0x000f)<< 8))
#define SETWBITS8_14(d,s)  (d) = (i16)(((d)&0x80ff)|(((s)&0x007f)<< 8))
#define SETWBITS8_15(d,s)  (d) = (i16)(((d)&0x00ff)|(((s)&0x00ff)<< 8))
#define SETWBITS9_9(d,s)   (d) = (i16)(((d)&0xfdff)|(((s)&0x0001)<< 9))
#define SETWBITS9_11(d,s)  (d) = (i16)(((d)&0xf1ff)|(((s)&0x0007)<< 9))
#define SETWBITS9_12(d,s)  (d) = (i16)(((d)&0xe1ff)|(((s)&0x000f)<< 9))
#define SETWBITS10_10(d,s) (d) = (i16)(((d)&0xfbff)|(((s)&0x0001)<<10))
#define SETWBITS10_11(d,s) (d) = (i16)(((d)&0xf3ff)|(((s)&0x0003)<<10))
#define SETWBITS10_13(d,s) (d) = (i16)(((d)&0xc3ff)|(((s)&0x000f)<<10))
#define SETWBITS10_15(d,s) (d) = (i16)(((d)&0x03ff)|(((s)&0x003f)<<10))
#define SETWBITS11_11(d,s) (d) = (i16)(((d)&0xf7ff)|(((s)&0x0001)<<11))
#define SETWBITS11_14(d,s) (d) = (i16)(((d)&0x87ff)|(((s)&0x000f)<<11))
#define SETWBITS11_15(d,s) (d) = (i16)(((d)&0x07ff)|(((s)&0x001f)<<11))
#define SETWBITS12_12(d,s) (d) = (i16)(((d)&0xefff)|(((s)&0x0001)<<12))
#define SETWBITS12_13(d,s) (d) = (i16)(((d)&0xcfff)|(((s)&0x0003)<<12))
#define SETWBITS12_15(d,s) (d) = (i16)(((d)&0x0fff)|(((s)&0x000f)<<12))
#define SETWBITS13_13(d,s) (d) = (i16)(((d)&0xdfff)|(((s)&0x0001)<<13))
#define SETWBITS13_14(d,s) (d) = (i16)(((d)&0x9fff)|(((s)&0x0003)<<13))
#define SETWBITS14_14(d,s) (d) = (i16)(((d)&0xbfff)|(((s)&0x0001)<<14))
#define SETWBITS14_15(d,s) (d) = (i16)(((d)&0x3fff)|(((s)&0x0003)<<14))
#define SETWBITS15_15(d,s) (d) = (i16)(((d)&0x7fff)|(((s)&0x0001)<<15))
#define SETIBITS24_35(d,s) (d) = (i32)(((d)&0xffffff)|(((s)&0xff)<<24))

class LEVELINFO
{
public:
  LEVELINFO& operator =(LOCATIONREL& locr);
  int level;
  int top;
  int bottom;
  int left;
  int right;
};

enum DBTYPE
{
 dbDOOR      =  0,
 dbTELEPORTER=  1,
 dbTEXT      =  2,
 dbACTUATOR  =  3,
 dbMONSTER   =  4,
 dbWEAPON    =  5,
 dbCLOTHING  =  6,
 dbSCROLL    =  7,
 dbPOTION    =  8,
 dbCHEST     =  9,
 dbMISC      = 10,
 dbEXPOOL    = 11,
 db12        = 12,
 db13        = 13,
 dbMISSILE   = 14,
 dbCLOUD     = 15,
 dbILLEGAL   = 16
};


enum LOADSTATUS
{
  LOAD_ERROR,
  LOAD_OK
};

enum STORESTATUS
{
  STORE_ERROR,
  STORE_OK
};

ui16 LE16(ui32);
ui32 LE32(ui32);

void DIE(const char *msg);

enum RNVAL
{
  RNeof             =0xfffe,
  RNnul             =0xffff,
  RNpit             =0xfffd,
  RNFireball        =0xff80, //fireball
  RNff81            =0xff81,
  RNff82            =0xff82,
  RNDispellMissile  =0xff83, //dispell - Damage non-Material
  RNZoSpell         =0xff84, //open door?
  RNff86            =0xff86,
  RNff87            =0xff87, //Poison Cloud
  RNffa8            =0xffa8,
  RNffe4            =0xffe4,
  RNempty           =0x0000
};

class RN // Record Name
{
  friend class DBCOMMON;
  friend class DB4;
  friend class DB9;
  friend class DB14;
  void swap(void) {indirectIndex=LE16(indirectIndex);};
private:
  ui16 indirectIndex;
public:
  RN(void) {indirectIndex = 0xffff;};
  RN(DBTYPE dbType, int dbIndex);
  //RN(ui32 pos, ui32 db, ui32 idx)
  //{
  //  ASSERT((pos<4) && (db<16) && (idx<1024));
  //  d.bf.m_pos=pos;
  //  d.bf.m_db=db;
  //  d.bf.m_idx=idx;
  //};
  RN(RNVAL i);
  void ConstructFromInteger(i32 i);
  bool CopyToClipboard(CLIPBOARD *pClipboard);
  void NonExtendedInteger(i32 i);
  void LittleEndian(void) {indirectIndex = LE16(indirectIndex);};
  DBTYPE dbType(void) const;
  ui8 dbNum(void) const;
  ui8 pos(void) const;
  void pos(ui32 p);
  ui16 dbIndex(void) const;
  //bool operator == (RN rn) const;
  //bool operator != (RN rn) const;
  void Delete(LOCATIONREL *plocr=NULL); //Also contents(if chest) 
                                 //and possesions(monster).
                                 //and timer for disabled generator
  bool operator == (RNVAL i) const;
  //bool operator >= (RNVAL i) const;
  //bool operator <  (RNVAL i) const;
  bool operator != (RNVAL i) const;
  ui16 ConvertToInteger(void) const; //USE SPARINGLY!
  bool IsIndirectIndexOK(void) const;
  bool IsDBIndexOK(void) const;
  bool IsMagicSpell(void) const;
  bool IsSameObjectAs(RN obj) const;
  bool IsSpell(void) const;
//  bool IsCloned(void) const;
//  void IsCloned(bool b);
  CString Name(ROOMTYPE rt) const;
};

enum CEMT
{
  CEMT_UNKNOWN,
  CEMT_CLONE,      //A cloned object error.
  CEMT_PARENTCELL, 
         //This is a dungeon cell that points to
         //a cloned object.
         //Location() is location in dungeon
  CEMT_UNCLONEDPARENTLIST, 
         //This is an uncloned object that has a cloned
         //object in its 'list link'.
         //IndirectIndex() is this 'parent' object.
         //Location() is this parent's location.
  CEMT_CLONEDPARENTLIST,
         //This is a cloned object that has a cloned
         //object in its 'list link'.
         //IndirectIndex() is this 'parent' object.
         //Location() is this parent's location.
  CEMT_MONSTERPOSSESSION,
         //A monster possesses a cloned object.
  CEMT_CHESTCONTENTS,
         //A chest contains a cloned object.
  CEMT_SCROLLTEXT,
         //The text of a scroll is a cloned object.
  CEMT_CHARACTERPOSSESSION,
         //A character possesses a cloned object.
         //Location.x() is the character index.
         //Location.y() is the possession index
         //If character index == -1 then object in hand.
  CEMT_SCROLLNONTEXT,
         //A scroll contains something other than text.
         //DBCommonAddress() is address of DB7 record
         //of the scroll.
  CEMT_ORPHANEDOBJECT,
         //DBCommonAddress() and DBType() describe the
         //object that is an orphan.
  CEMT_OVERLAPPEDTEXT,
         //DBIndex1() and DBIndex2() are the dbTEXT
         //entries that overlap;
  CEMT_CLONEDTEXT1,
         //Index1() = First of dbTEXT entry that shares 
         //compressed text.  CEMT_CLONEDTEXT2 entries follow.
  CEMT_CLONEDTEXT2,
         //Index1() = Second and subsequent dbTEXT entries 
         //that share compressed text.  See CEMT_CLONEDTEXT1
  CEMT_SINGLESHAREDTEXT,
         //DB2[Index1()] shares compressed text with some
         //other DB2 entries.
  CEMT_UNKNOWNCELL,
         //At Location()
  CEMT_INDIRECTNOTREFERENCED,
         //Index1() = objectIndirectIndex
  CEMT_CLONEDINDIRECTTEXTENTRY,
         //Index1() = Indirect Text Index
         //Index2() = dbIndex of first DB2 entry parent
         //  Following CEMT_CLONEDINDIRETTEXTINDEDXPARENT 
         //  records are parent records
  CEMT_CLONEDINDIRECTTEXTINDEXPARENT,
         //Index1() = DB2 index of parent record
         // See CEMT_CLONEDINDIRECTTEXTENTRY
  CEMT_TEXTPLUSRANDOMDECORATION,
         //location is place problem occurred
  CEMT_CHARPOSSESSIONISLIST,
         //One of the possessions of a character is 
         //not a single object but rather a list of objects
         //Index1 = character index.
         //Index2 = possession index.
  CEMT_MONSTERPOSITIONEDBADLY,
         //Index1() = Monster object;
  CEMT_RANDOMDECORATIONONDOORFRAME,
         //Location of Stone Cell
         //Index1 - Cellflag mask 8,4,2,or 1
  CEMT_RANDOMDECORATIONONSKINNEDWALL,
         //Location of Stone Cell
  CEMT_RANDOMDECORATIONONSKINNEDFLOOR
         //Location of Open Cell
};

class CHKERRMSG
{
  //An instance of this class serves as the root.
  //The root contains only the pointers to the
  //first and last entries in the list.
  //The member function 'NewError' adds another entry
  //to the end of the list.  Other functions
  //reference the last entry in the list.
  CHKERRMSG  *m_next;
  CHKERRMSG  *m_prev;
  bool        m_root;
  CString     m_message;
  CEMT        m_errorType;
  int         m_parentIndirectIndex;
  int         m_cloneIndirectIndex;
  int         m_index1;
  int         m_index2;
  LOCATIONREL m_location;
  bool        m_repaired;
  DBCOMMON   *m_dbCommonAddress;
  DBTYPE      m_dbType;
public:
  CHKERRMSG(void);
  ~CHKERRMSG(void);
  void Clear(void);
  void NewError(const CString& msg); //Add another entry to list.
  void NewError(const CHKERRMSG *pcem); //Add another entry to list.
  void operator += (const CString& moreMsg); //Extend message.
  CHKERRMSG  *FirstError(void) const;
  CHKERRMSG  *LastError(void) const;
  CHKERRMSG  *NextError(const CHKERRMSG *pcem) const;
  CHKERRMSG  *Next(void)const;
  CHKERRMSG  *Prev(void)const;
  CString     ErrorMessage(void)const;
  CEMT        ErrorType(void) const;
  void        ErrorType(CEMT errorType);
  int         ParentIndirectIndex(void) const;
  void        ParentIndirectIndex(int indirectIndex);
  int         CloneIndirectIndex(void) const;
  void        CloneIndirectIndex(int indirectIndex);
  LOCATIONREL Location(void) const;
  void        Location(LOCATIONREL locr);
  void        NewLocation(LOCATIONREL locr);
  void        Repaired(bool repaired);//Set repaired flag.
  bool        Repaired(void) const;
  bool        AnyRepairs(void) const;
  DBCOMMON   *DBCommonAddress(void) const;
  void        DBCommonAddress(DBCOMMON *pDBCommon);
  DBTYPE      DBType(void) const;
  void        DBType(DBTYPE dbType);
  int         Index1(void) const;
  void        Index1(int index);
  int         Index2(void) const;
  void        Index2(int index);
};



class MODIFIEDOBJECTTYPE;

class MISCTYPE
{
  enum m_MISCTYPE
  {
    misc_2small         =-1,
    misc_Compass        = 0,
    misc_Waterskin      = 1,
    misc_JewelSymal     = 2,
    misc_Illumulet      = 3,
    misc_Ashes          = 4,
    misc_BonesOfHero    = 5, //value = character index
    misc_SarCoin        = 6,
    misc_SilverCoin     = 7,
    misc_GorCoin        = 8,
    misc_IronKey        = 9,
    misc_KeyOfB         = 10,
    misc_SolidKey       = 11,
    misc_SquareKey      = 12,
    misc_TourquoiseKey  = 13,
    misc_CrossKey       = 14,
    misc_OnyxKey        = 15,
    misc_SkeletonKey    = 16,
    misc_GoldKey        = 17,
    misc_WingedKey      = 18,
    misc_TopazKey       = 19,
    misc_SapphireKey    = 20,
    misc_EmeraldKey     = 21,
    misc_RubyKey        = 22,
    misc_RaKey          = 23,
    misc_MasterKey      = 24,
    misc_Boulder        = 25,
    misc_BlueGem        = 26,
    misc_OrangeGem      = 27,
    misc_GreenGem       = 28,
    misc_Apple          = 29,
    misc_Corn           = 30,
    misc_Bread          = 31,
    misc_Cheese         = 32,
    misc_ScreamerSlice  = 33,
    misc_WormRound      = 34,
    misc_Shank          = 35,
    misc_DragonSteak    = 36,
    misc_GemOfAges      = 37,
    misc_EkkhardCross   = 38,
    misc_Moonstone      = 39,
    misc_TheHellion     = 40,
    misc_PendantFeral   = 41,
    misc_MagicalBoxA    = 42,
    misc_MagicalBoxB    = 43,
    misc_MirrorOfDawn   = 44,
    misc_Rope           = 45,
    misc_RabbitsFoot    = 46,
    misc_Corbum         = 47,
    misc_Choker         = 48,
    misc_LockPicks      = 49,
    misc_Magnifier      = 50,
    misc_ZokathraSpell  = 51,
    misc_Bones          = 52,
    misc_2large         = 53
  } m_miscType;
  void CheckValid(void);
public:
  MISCTYPE(int n = misc_2small);
  ui8 Integer(void){return (ui8)m_miscType;};
  bool IsValid(void);
  MISCTYPE& operator =(MISCTYPE pt);
  MISCTYPE& operator =(const char *name);
  bool operator ==(MISCTYPE& pt){return pt.m_miscType==m_miscType;};
  CString Name(void);
  int FillComboBox(CComboBox *pBox);
};

enum CHESTTYPE
{
  chest_0 = 0
};

class POTIONTYPE
{
  enum m_POTIONTYPE
  {
    potion_2small     = -1,
    potion_MonPotionA =  0,
    potion_UmPotion   =  1,
    potion_DesPotion  =  2,
    potion_VenPotion  =  3,
    potion_SarPotion  =  4,
    potion_ZoPotion   =  5,
    potion_RosPotion  =  6,
    potion_KuPotion   =  7,
    potion_DanePotion =  8,
    potion_NetaPotion =  9,
    potion_AntiVenin  = 10,
    potion_MonPotionB = 11,
    potion_YaPotion   = 12,
    potion_EEPotion   = 13,
    potion_ViPotion   = 14,
    potion_WaterFlask = 15,
    potion_KathBomb   = 16,
    potion_PewBomb    = 17,
    potion_RaBomb     = 18,
    potion_FulBomb    = 19,
    potion_EmptyFlask = 20,
    potion_2large     = 21
  } m_potionType;
  void CheckValid(void);
public:
  POTIONTYPE(int n = potion_2small);
  void Integer(int i){m_potionType=(m_POTIONTYPE)i;};
  ui8 Integer(void){return (ui8)m_potionType;};
  bool IsValid(void);
  POTIONTYPE& operator =(POTIONTYPE pt);
  POTIONTYPE& operator =(const char *name);
  bool operator ==(POTIONTYPE& pt){return pt.m_potionType==m_potionType;};
  CString Name(void);
  int FillComboBox(CComboBox *pBox);
};

class CLOTHINGTYPE
{
  enum m_CLOTHINGTYPE
  {
    clothing_2small        = -1,
    clothing_Cape          =  0,
    clothing_CloakOfNight  =  1,
    clothing_TatteredPants =  2,
    clothing_Sandals       =  3,
    clothing_LeatherBoots  =  4,
    clothing_TatteredShirt =  5,
    clothing_Robe          =  6,
    clothing_FineRobeA     =  7,
    clothing_FineRobeB     =  8,
    clothing_Kirtle        =  9,
    clothing_SilkShirt     = 10,
    clothing_Tabard        = 11,
    clothing_Gunna         = 12,
    clothing_ElvenDoublet  = 13,
    clothing_ElvenHuke     = 14,
    clothing_ElvenBoots    = 15,
    clothing_LeatherJerkin = 16,
    clothing_LeatherPants  = 17,
    clothing_SuedeBoots    = 18,
    clothing_BluePants     = 19,
    clothing_Tunic         = 20,
    clothing_Ghi           = 21,
    clothing_GhiTrousers   = 22,
    clothing_Calista       = 23,
    clothing_CrownOfNerra  = 24,
    clothing_BezerkerHelm  = 25,
    clothing_Helmet        = 26,
    clothing_Basinet       = 27,
    clothing_NetaShield    = 28,
    clothing_CrystalShield = 29,
    clothing_WoodenShield  = 30,
    clothing_SmallShield   = 31,
    clothing_MailAketon    = 32,
    clothing_LegMail       = 33,
    clothing_MithralAketon = 34,
    clothing_MithralMail   = 35,
    clothing_CasqueNCoif   = 36,
    clothing_Hosen         = 37,
    clothing_Armet         = 38,
    clothing_TorsoPlate    = 39,
    clothing_LegPlate      = 40,
    clothing_FootPlate     = 41,
    clothing_SarShield     = 42,
    clothing_HelmOfRa      = 43,
    clothing_PlateOfRa     = 44,
    clothing_PoleynOfRa    = 45,
    clothing_GreaveOfRa    = 46,
    clothing_ShieldOfRa    = 47,
    clothing_DragonHelm    = 48,
    clothing_DragonPlate   = 49,
    clothing_DragonPoleyn  = 50,
    clothing_DragonGreave  = 51,
    clothing_DragonShield  = 52,
    clothing_Dexhelm       = 53,
    clothing_Flamebain     = 54,
    clothing_PowerTowers   = 55,
    clothing_BootsOfSpeed  = 56,
    clothing_Halter        = 57,
    clothing_2large        = 58
  } m_clothingType;
  void CheckValid(void);
public:
  CLOTHINGTYPE(int n = clothing_2small);
  ui8 Integer(void){return (ui8)m_clothingType;};
  bool IsValid(void);
  CLOTHINGTYPE& operator =(CLOTHINGTYPE wt);
  CLOTHINGTYPE& operator =(MODIFIEDOBJECTTYPE mot);
  CLOTHINGTYPE& operator =(const char *name);
  bool operator ==(CLOTHINGTYPE& wt){return wt.m_clothingType==m_clothingType;};
  CString Name(void);
  int FillComboBox(CComboBox *pBox);
};

class WEAPONTYPE
{
  enum m_WEAPONTYPE
  {
    weapon_2small       = -1,
    weapon_EyeOfTime    =  0,
    weapon_StormRing    =  1,
    weapon_Torch        =  2,
    weapon_Flamitt      =  3,
    weapon_StaffOfClaws =  4,
    weapon_Storm        =  5,
    weapon_RaBlade      =  6,
    weapon_TheFirestaffA=  7,
    weapon_Dagger       =  8,
    weapon_Falchion     =  9,
    weapon_Sword        = 10,
    weapon_Rapier       = 11,
    weapon_Biter        = 12,
    weapon_SamuraiSword = 13,
    weapon_SideSplitter = 14,
    weapon_DiamondEdge  = 15,
    weapon_VorpalBlade  = 16,
    weapon_DragonFang   = 17,
    weapon_Axe          = 18,
    weapon_Executioner  = 19,
    weapon_Mace         = 20,
    weapon_MaceOfOrder  = 21,
    weapon_Morningstar  = 22,
    weapon_Club         = 23,
    weapon_StoneClub    = 24,
    weapon_ClawBow      = 25,
    weapon_Crossbow     = 26,
    weapon_Arrow        = 27,
    weapon_Slayer       = 28,
    weapon_Sling        = 29,
    weapon_Rock         = 30,
    weapon_PoisonDart   = 31,
    weapon_ThrowingStar = 32,
    weapon_Stick        = 33,
    weapon_Staff        = 34,
    weapon_Wand         = 35,
    weapon_TeoWand      = 36,
    weapon_YewStaff     = 37,
    weapon_StaffOfIrra  = 38,
    weapon_CrossOfNeta  = 39,
    weapon_SerpentStaff = 40,
    weapon_DragonSpit   = 41,
    weapon_SceptreOfLyf = 42,
    weapon_HornOfFear   = 43,
    weapon_Speedbow     = 44,
    weapon_TheFirestaffB= 45,
    weapon_2large       = 46
  } m_weaponType;
  void CheckValid(void);
public:
  WEAPONTYPE(int n = weapon_2small);
  ui8 Integer(void){return (ui8)m_weaponType;};
  bool IsValid(void);
  WEAPONTYPE& operator =(WEAPONTYPE wt);
  WEAPONTYPE& operator =(const char *name);
  bool operator ==(WEAPONTYPE& wt){return wt.m_weaponType==m_weaponType;};
  CString Name(void);
  int FillComboBox(CComboBox *pBox);
};

class MONSTERTYPE
{
  enum m_MONSTERTYPE
  {
    mon_2small         =-1,
    mon_Scorpion       = 0,   //0
    mon_SlimeDevil     = 1,   //1
    mon_Giggler        = 2,   //2
    mon_FlyingEye      = 3,   //3
    mon_Hellhound      = 4,   //4
    mon_5              = 5,   //5
    mon_Screamer       = 6,   //6
    mon_RockPile       = 7,   //7
    mon_Rive           = 8,   //8
    mon_StoneGolem     = 9,   //9
    mon_Mummy          = 0x0a,//10
    mon_BlackFlame     = 0x0b,//11
    mon_Skeleton       = 0x0c,//12
    mon_Couatl         = 0x0d,//13
    mon_Vexirk         = 0x0e,//14
    mon_Worm           = 0x0f,//15
    mon_AntMan         = 0x10,//16
    mon_Muncher        = 0x11,//17
    mon_DethKnight     = 0x12,//18
    mon_Zytaz          = 0x13,//19
    mon_WaterElemental = 0x14,//20
    mon_Oitu           = 0x15,//21
    mon_Demon          = 0x16,//22
    mon_LordChaos      = 0x17,//23
    mon_Dragon         = 0x18,//24
    mon_LordOrder      = 0x19,//25
    mon_GreyLord       = 0x1a,//26
    mon_2large         = 0x1b
  } m_monsterType;
  void CheckValid(void);
public:
  MONSTERTYPE(int n = mon_2small);
  ui8 Integer(void){return (ui8)m_monsterType;};
  void Integer(int i){m_monsterType=(m_MONSTERTYPE)i;};
  bool IsValid(void);
  MONSTERTYPE& operator =(MONSTERTYPE wt);
  MONSTERTYPE& operator =(const char *name);
  bool operator ==(MONSTERTYPE& wt){return wt.m_monsterType==m_monsterType;};
  CString Name(void);
  CString Name(int graphic);
  int FillComboBox(CComboBox *pBox, int level=-1);
  int FillListBox(CListBox *pBox, int level=-1);
};

#define CHEST_TYPE_OFFSET 1
#define POTION_TYPE_OFFSET 2
#define WEAPON_TYPE_OFFSET 23
#define CLOTHING_TYPE_OFFSET 69
#define MISC_TYPE_OFFSET 127


class MODIFIEDOBJECTTYPE
{
  enum m_MODIFIEDOBJECTTYPE
  {
    obj_FIRSTEXTRA = -9,
    obj_ERROR      = -9,
    obj_ANY        = -8,
    obj_BEING      = -7,
    obj_STEP       = -6,
    obj_MOVE       = -5,
    obj_MONSTER    = -4,
    obj_PARTY      = -3,
    obj_NONE       = -2,
    obj_LASTEXTRA  = -2,
    obj_2small     = -1,
    obj_Compass_N = 0,  //0  0
      obj_FirstModifiableObject = obj_Compass_N,
    obj_Compass_E,      //1  1
    obj_Compass_S,      //2  2
    obj_Compass_W,      //3  3
    obj_Torch_a,        //4  4
    obj_Torch_b,        //5  5
    obj_Torch_c,        //6  6
    obj_Torch_d,        //7  7
    obj_Waterskin,      //8  8
    obj_Water,          //9  9
    obj_JewelSymal_a,   //0a  10
    obj_JewelSymal_b,   //0b  11
    obj_Illumulet_a,    //0c  12
    obj_Illumulet_b,    //0d
    obj_Flamitt_a,      //0e
    obj_Flamitt_b,      //0f
    obj_EyeOfTime_a,    //10
    obj_EyeOfTime_b,    //11
    obj_StormRing_a,    //12
    obj_StormRing_b,    //13
    obj_StaffOfClaws_a, //14  20
    obj_StaffOfClaws_b, //15  21
    obj_StaffOfClaws_c, //16  22
    obj_Storm_a,        //17
    obj_Storm_b,        //18
    obj_RABlade_a,      //19  25
    obj_RABlade_b,      //1a  26
    obj_TheFirestaff_a, //1b
    obj_TheFirestaff_b, //1c
    obj_TheFirestaff_c, //1d
    obj_OpenScroll,     //1e
    obj_Scroll,         //1f
      obj_LastModifiableObject = obj_Scroll,
    obj_Dagger,         //20  32
    obj_Falchion,       //21  33
    obj_Sword,          //22  34
    obj_Rapier,         //23  35
    obj_Biter,          //24  36
    obj_SamuraiSword,   //25  37
    obj_SideSplitter,   //26  38
    obj_DiamondEdge,    //27  39
    obj_VorpalBlade,    //28  40
    obj_DragonFang,     //29  41
    obj_Axe,            //2a
    obj_Executioner,    //2b
    obj_Mace,           //2c
    obj_MaceOfOrder,    //2d  45
    obj_Morningstar,    //2e
    obj_Club,           //2f
    obj_StoneClub,      //30
    obj_ClawBow,        //31
    obj_Crossbow,       //32
    obj_Arrow,          //33
    obj_Slayer,         //34
    obj_Sling,          //35
    obj_Rock,           //36
    obj_PoisonDart,     //37
    obj_ThrowingStar,   //38
    obj_Stick,          //39
    obj_Staff,          //3a  58
    obj_Wand,           //3b  59
    obj_TeoWand,        //3c  60
    obj_YewStaff,       //3d  61
    obj_StaffOfIrra,    //3e  62
    obj_CrossOfNeta,    //3f  63
    obj_SerpentStaff,   //40  64
    obj_DragonSpit,     //41  65
    obj_SceptreOfLyf,   //42  66
    obj_TatteredShirt,  //43
    obj_FineRobe_a,     //44
    obj_Kirtle,         //45
    obj_SilkShirt,      //46
    obj_ElvenDoublet,   //47
    obj_LeatherJerkin,  //48
    obj_Tunic,          //49
    obj_Ghi,            //4a
    obj_MailAketon,     //4b
    obj_MithralAketon,  //4c
    obj_TorsoPlate,     //4d
    obj_PlateOfRa,      //4e
    obj_DragonPlate,    //4f
    obj_Cape,           //50 80
    obj_CloakOfNight,   //51 81
    obj_TatteredPants,  //52 82
    obj_Robe,           //53 83
    obj_FineRobe_b,     //54 84
    obj_Tabard,         //55 85
    obj_Gunna,          //56
    obj_ElvenHuke,      //57
    obj_LeatherPants,   //58
    obj_BluePants,      //59
    obj_GhiTrousers,    //5a
    obj_LegMail,        //5b
    obj_MithralMail,    //5c
    obj_LegPlate,       //5d
    obj_PoleynOfRa,     //5e
    obj_DragonPoleyn,   //5f
    obj_BezerkerHelm,   //60
    obj_Helmet,         //61
    obj_Basinet,        //62
    obj_CasqueNCoif,    //63
    obj_Armet,          //64 100
    obj_HelmOfRa,       //65 101
    obj_DragonHelm,     //66 102
    obj_Calista,        //67 103
    obj_CrownOfNerra,   //68 104
    obj_NetaShield,     //69
    obj_CrystalShield,  //6a
    obj_SmallShield,    //6b
    obj_WoodenShield,   //6c
    obj_SarShield,      //6d
    obj_ShieldOfRa,     //6e
    obj_DragonShield,   //6f
    obj_Sandals,        //70
    obj_SuedeBoots,     //71
    obj_LeatherBoots,   //72
    obj_Hosen,          //73
    obj_FootPlate,      //74
    obj_GreaveOfRa,     //75
    obj_DragonGreave,   //76
    obj_ElvenBoots,     //77
    obj_GemOfAges,      //78
    obj_EkkhardCross,   //79
    obj_Moonstone,      //7a 122
    obj_TheHellion,     //7b 123
    obj_PendantFeral,   //7c 124
    obj_SarCoin,        //7d 125
    obj_SilverCoin,     //7e 126
    obj_GorCoin,        //7f 127
    obj_Boulder,        //80 128
    obj_BlueGem,        //81 129
    obj_OrangeGem,      //82 130
    obj_GreenGem,       //83 131
    obj_MagicalBox_a,   //84 132
    obj_MagicalBox_b,   //85 133
    obj_MirrorOfDawn,   //86 134
    obj_HornOfFear,     //87 135
    obj_Rope,           //88 136
    obj_RabbitsFoot,    //89 137
    obj_Corbum,         //8a 138
    obj_Choker,         //8b 139
    obj_Dexhelm,        //8c 140
    obj_Flamebain,      //8d 141
    obj_Powertowers,    //8e 142 
    obj_Speedbow,       //8f 143
    obj_Chest,          //90 144
    obj_OpenChest,      //91 145
    obj_Ashes,          //92
    obj_Bones_a,        //93
    obj_MonPotion_a,    //94
      obj_FirstFullFlask = obj_MonPotion_a,
    obj_UmPotion,       //95
    obj_DesPotion,      //96
    obj_VenPotion,      //97
    obj_SarPotion,      //98
    obj_ZoPotion,       //99
    obj_RosPotion,      //9a
    obj_KuPotion,       //9b
    obj_DanePotion,     //9c
    obj_NetaPotion,     //9d
    obj_AntiVenin,      //9e
    obj_MonPotion_b,    //9f
    obj_YaPotion,       //a0 160
    obj_EePotion,       //a1
    obj_ViPotion,       //a2
    obj_WaterFlask,     //a3
      obj_LastFullFlask = obj_WaterFlask,
    obj_KathBomb,       //a4
    obj_PewBomb,        //a5
    obj_RaBomb,         //a6
    obj_FulBomb,        //a7
    obj_Apple,          //a8  168
    obj_Corn,           //a9  169
    obj_Bread,          //aa  170
    obj_Cheese,         //ab  171
    obj_ScreamerSlice,  //ac  172
    obj_WormRound,      //ad  173
    obj_Shank,          //ae  174
    obj_DragonSteak,    //af  175
    obj_IronKey,        //b0  176
      obj_FirstKey = obj_IronKey,
    obj_KeyOfB,         //b1  177  
    obj_SolidKey,       //b2  178
    obj_SquareKey,      //b3  179
    obj_TourquoiseKey,  //b4
    obj_CrossKey,       //b5
    obj_OnyxKey,        //b6
    obj_SkeletonKey,    //b7
    obj_GoldKey,        //b8
    obj_WingedKey,      //b9
    obj_TopazKey,       //ba
    obj_SapphireKey,    //bb
    obj_EmeraldKey,     //bc
    obj_RubyKey,        //bd
    obj_RaKey,          //be
    obj_MasterKey,      //bf
      obj_LastKey = obj_MasterKey,
    obj_LockPicks,      //c0 192
    obj_Magnifier,      //c1
    obj_BootsOfSpeed,   //c2
    obj_EmptyFlask,     //c3
    obj_Halter,         //c4
    obj_ZokathraSpell,  //c5
    obj_Bones_b,        //c6 198
    obj_Special_a,      //c7
    obj_Special_b,      //c8
    obj_Special_c,      //c9
    obj_Special_d,      //ca 202
    obj_Special_e,      //cb 203 
    obj_Special_f,      //cc 204
    obj_Special_g,      //cd 205
    obj_Special_h,      //ce 206
    obj_Special_i,      //cf 207
    obj_Special_j,      //d0 208
    obj_Special_k,      //d1
    obj_Special_l,      //d2
    obj_Special_m,      //d3
    obj_Special_n,      //d4
    obj_2large          //d5
  } m_modifiedObjectType;
  void CheckValid(bool allowExtras=false);
  int GetObjectIndex();
public:
  DBTYPE DatabaseType(void);
  CLOTHINGTYPE clothingType(void);
  MODIFIEDOBJECTTYPE(int n = obj_2small);
  MODIFIEDOBJECTTYPE(RN obj);
  MODIFIEDOBJECTTYPE(WEAPONTYPE wt);
  MODIFIEDOBJECTTYPE(CLOTHINGTYPE ct);
  MODIFIEDOBJECTTYPE(POTIONTYPE pt);
  MODIFIEDOBJECTTYPE(MISCTYPE mt);
  ui8 Integer(void){return (ui8)m_modifiedObjectType;};
  CString Name(void);
  int CarryLocations(void);
  bool IsValid(void);
  bool IsValid(int n);
  bool operator ==(MODIFIEDOBJECTTYPE& mot);
  MODIFIEDOBJECTTYPE operator =(int n);
  MODIFIEDOBJECTTYPE operator =(const char *name);
  MODIFIEDOBJECTTYPE operator =(WEAPONTYPE wt);
  MODIFIEDOBJECTTYPE operator =(CLOTHINGTYPE ct);
  MODIFIEDOBJECTTYPE operator =(POTIONTYPE pt);
  MODIFIEDOBJECTTYPE operator =(MISCTYPE mt);
  int FillComboBox(CComboBox *pBox);
  bool IsBasicObject(void);
};

#define tmrAct_SET    0
#define tmrAct_CLEAR  1
#define tmrAct_TOGGLE 2

enum TIMERFUNCTION
{
  timerFunctionM3 = -3,
  timerFunctionM2 = -2,
  timerFunctionM1 = -1,
  timerFunction0  = 0,
  timerFunction1  = 1,
  timerFunction2  = 2,
  timerFunction5  = 5,
  timerFunction6  = 6,
  timerFunction7  = 7,
  timerFunction8  = 8,
  timerFunction9  = 9,
  timerFunction10 = 10,
  timerFunction11 = 11,
  timerFunction12 = 12,
  timerFunction13 = 13,
  timerFunction20 = 20,
  timerFunction22 = 22,
  timerFunction24 = 24,
  timerFunction25 = 25,
  timerFunction28 = 28,
  timerFunction29 = 29,
  timerFunction30 = 30,
  timerFunction31 = 31,
  timerFunction32 = 32,
  timerFunction33 = 33,
  timerFunction37 = 37,
  timerFunction38 = 38,
  timerFunction41 = 41,
  timerFunction48 = 48,
  timerFunction49 = 49,
  timerFunction53 = 53,
  timerFunction60 = 60,
  timerFunction61 = 61,
  timerFunction65 = 65,
  timerFunction70 = 70,
  timerFunction71 = 71,
  timerFunction72 = 72,
  timerFunction73 = 73,
  timerFunction74 = 74,
  timerFunction75 = 75,
  timerFunction77 = 77,
  timerFunction78 = 78,
  timerFunction79 = 79,
};


class TIMER
{ //  We read them from
  // DUNGEON.DAT when game reloaded.
public:
  TIMER(void) {uByte5 = uByte6 = uByte7 = uByte8 = uByte9 = 0xcd;};
  ui32 time; // Top 8 bits=partyLevel, bottom 24 bits = time
  ui8  timerFunction;   
    //  1
    //    Happened when I pressed open door switch. 
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 =
    //  2 Delayed door bashing.
    //         byte 5 = 0
    //         byte 6 = mapX
    //         byte 7 = mapY
    //  5 
    //    
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 = function
    //              for DB2 fiddle bit 0 of word 2
    //                = 0 means set
    //                = 1 means clear
    //                = 2 means toggle bit 0 of word 2
    //              for DB3 (Actuator) 
    //  6 Fiddle with DB2 and DB3
    //    Appears to toggle things.
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 8 = position when sent to object
    //                    of DB type 2 (text??)
    //                  For DB type 3 (Actuators) the
    //                  position of the actuator is
    //                  ignored.  All actuators in the
    //                  target cell receive the timer.
    //         byte 9 = 0 means set bit
    //                = 1 means clear
    //                = 2 means toggle
    //    When sent to actuator type 14 ...(see actuator type 14)
    //  7 Change bit 0x04 in CELLFLAG 
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 = 0 = Set Unconditionally
    //                = 1 = If party at X,Y requeue
    //                      If material monster at X,Y requeue
    //                      else clear
    //                = 2 = Toggle (same rules as set/clear)
    //  8 Change bit 0x08 in CELLFLAG.
    //    If the final state of the bit is a one then the 
    //    function 'ProcessNewlyActiveRoom' is called.
    //    The timer goes away without restarting anything.
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 = tmrAct
    //                = 0 = tmrAct_SET
    //                = 1 = tmrAct_CLEAR
    //                = 2 = tmrAct_TOGGLE
    //  9
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 =
    // 10
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 9 = 
    // 11 Disable action after casting spell
    //         byte 5 = chIdx
    //         byte 6 = 0
    // 12 When party moves. Byte 5=chIdx.  Removes timerIndex from chIdx
    // 13 Happened when I put bones in Altar of Vi.
    //         byte 5 = character index
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 8 = position
    //         uByte9 = 0, 1, or 2;
    //         case 2: TAG00dea8(RNffe4) and requeue type 1 in 5 ticks.
    // 22
    //         Processing for function 22 disabled.
    // 24      
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         obj  8 = object name
    //          
    // 25 
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         object name in word 8
    // 29 to 41 have everything to do with monster movement
    //         This is the heart of the monster AI.  Each monster
    //         gets to do its thing when its timer expires.
    // 32 Causes monster action     
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         byte 8 = additional time until function changed to 37
    // 37 When character dies and bones
    //       put on floor???
    //         byte5 comes from Item9660.byte2[0]
    //         byte6 = mapX
    //         byte7 = mapY
    //         bz`z yte 8 = 0
    // 48 and 49TAG01826c  Missile????
    //         byte5 = 0
    //         word6 objectID
    //         word8 bits 10-11 direction
    //               bits  5- 9 y
    //               bits  0- 4 x
    // 53 = Watchdog Timer
    // 60 and 61 = ???
    //    After processing this is reset to time+5
    //    This appears to be a pending monster generator.
    //    A monster was generated where a monster already
    //    existed.  Wait a while and try again.
    //         byte6 = mapX
    //         byte7 = mapY
    //         obj8  = object to move 
    // 65
    //         byte 6 = mapX
    //         byte 7 = mapY
    //         All actuators at (X,Y) with actuatorType==0
    //           get set to actuatorType 6.  Used to
    //           reactivate monster generators and such.
    // 70 Adjust Brightness (light level)
    //    Timer is reset to d.Time+4 if index has
    //    not been reduced to zero.
    //         word6 = (+/-)index to d.Byte1074
    //                 Add or subtract from d.13282
    //     
    // 71 Set by 3-2-6 spell (OH EW SAR) invisibility     
    // 72 Character has used shield potion (Ya potion)
    //    It is time to remove the shield.
    //         byte 5 = character index
    //         word 6 = shield increment to be removed
    // 73 set by 3-2-5 spell (OH EW RA See thru walls)
    // 74 Set by protection spell
    // 75 Character hit by missile??
    ///   Character Poison effect??
    //         byte 5 = character index
    //         word 6 = damage
    //         Byte 42 of the character has the timer index.
    // 77 Some sort of spell will wear off
    //    d.Word13274 was incremented by word6
    //         byte 6 = value to subtract from d.Word13274
    // 78 Some sort of spell will wear off
    //    d.Word13276 was incremented by word6
    //         byte 6 = value to subtract from d.Word13276
    // 79 Result of 1-1-5-2 spell ya-bro-ros (magic footprints)
  ui8  uByte5;
  ui8  uByte6;
  ui8  uByte7;
  ui8  uByte8;
  ui8  uByte9;
  RN&  timerObj8(void) 
  {
    ASSERT(   (timerFunction==24)
           || (timerFunction==25)
           || (timerFunction==60)
           || (timerFunction==61));
    return *(RN *)(&uByte8);
  };
  void swapWord6(void);
  void swapObj6(void);
  void swapWord8(void);
  void swapObj8(void);
  RN&  timerObj6(void)
  {
    ASSERT((timerFunction==48) || (timerFunction==49));
    return *(RN *)(&uByte6);
  };
  i16& word6(void)
  {
    return *(i16 *)(&uByte6);
  };
  ui16& uWord6(void)
  {
    return *(ui16 *)(&uByte6);
  };
  i16& word8(void)
  {
    return *(i16 *)(&uByte8);
  };
  ui8 missileX(void){return BITS0_4(word8());};
  ui8 missileY(void){return BITS5_9(word8());};
  ui8 missilePos(void){return BITS10_11(word8());};
};


struct GAMEBLOCK1
{
  // Data below this line is 512 bytes long.  This
  // block is the first read from the game save file.
  i8  Byte740[300];  //000
  i8  Byte22598;     //300
  ui8 Byte22596;     //301
  i8  FILL438[4];    //302
  i16 SaveOption;    //306
  i32 RandomGameID;  //308 reversed
  i16 Block2Hash;    //312 swapped
  i16 ITEM16Hash;    //314 swapped
  i16 CharacterHash; //316 swapped
  i16 TimersHash;    //318 swapped
  i16 TimerQueHash;  //320 swapped
  i32 totalMoveCount;//322
  i16 Hash326;       //326
  i16 Hash328;       //328
  i16 Hash330;       //330
  i16 Hash332;       //332
  i16 Hash334;       //334
  i16 Hash336;       //336
  i16 Hash338;       //338
  i16 Hash340;       //340
  i16 Hash342;       //342
  i16 Block2Checksum;//344swapped
  i16 ITEM16Checksum;   //346 swapped
  i16 CharacterChecksum;//348 swapped
  i16 TimersChecksum;   //350 swapped
  i16 TimerQueChecksum; //352 swapped
  i16 Checksum354;      //354
  i16 Checksum356;      //356
  i16 Checksum358;      //358
  i16 Checksum360;      //360
  i16 Checksum362;      //362
  i16 Checksum364;      //364
  i16 Checksum366;      //366
  i16 Checksum368;      //368
  i16 Checksum370;      //370
  i16 Checksum372;      //372
  i16 Checksum374;      //374
  i16 Word22594;        //376 swapped
  i16 Word22592;        //378 swapped
  i8  Byte22808[132];   //380 moved as a unit
};

struct PCGAMEBLOCK1
{
  // Data below this line is 512 bytes long.  This
  // block is the first read from the game save file.
  i8  Byte740[300];  //000
  i8  Byte22598;     //300
  ui8 Byte22596;     //301
  i8  FILL438[4];    //302
  i16 SaveOption;    //306
  i16 RandomGameID;  //308 //i32 RandomGameID;  //308 reversed
  i16 Block2Hash;    //310 swapped
  i16 ITEM16Hash;    //312 swapped
  i16 CharacterHash; //314 swapped
  i16 TimersHash;    //316 swapped
  i16 TimerQueHash;  //318 swapped
  i32 totalMoveCount;//320
  i16 Hash326;       //324
  i16 Hash328;       //326
  i16 Hash330;       //328
  i16 Hash332;       //330
  i16 Hash334;       //332
  i16 Hash336;       //334
  i16 Hash338;       //335
  i16 Hash340;       //338
  i16 Hash342;       //340
  i16 Block2Checksum;//342swapped
  i16 ITEM16Checksum;   //344 swapped
  i16 CharacterChecksum;//346 swapped
  i16 TimersChecksum;   //348 swapped
  i16 TimerQueChecksum; //350 swapped
  i16 Checksum354;      //352
  i16 Checksum356;      //354
  i16 Checksum358;      //356
  i16 Checksum360;      //358
  i16 Checksum362;      //360
  i16 Checksum364;      //362
  i16 Checksum366;      //364
  i16 Checksum368;      //366
  i16 Checksum370;      //368
  i16 Checksum372;      //370
  i16 Checksum374;      //372
  i16 Word22594;        //374 swapped
  i16 Word22592;        //376 swapped
  i8  Byte22808[134];   //378 moved as a unit
};

struct GAMEBLOCK2
{
  // The following 128 bytes are the second thing
  // read from the game file.
  i32 Time;          //000; reversed
  i32 ranseed;       //004; reversed
  ui16 ObjectInHand; //008; swapped
  i16 numcharacter;  //010; swapped
  i16 partyx;        //012; swapped
  i16 partyy;        //014; swapped
  i16 partyfacing;   //016; swapped
  i16 partyLevel;    //018; swapped
  i16 handChar;      //020; swapped Character index
  i16 MagicCaster;   //022; swapped
  i16 NumTimer;      //024; swapped
  i16 NextAvailTimer;//026; swapped
  i16 MaxTimers;     //028; swapped
  i16 ITEM16QueLen;  //030; swapped
  i32 Long12950;     //032; reversed
  i32 Long12954;     //036; A timestamp. reversed
  i16 Word11710;     //040; swapped
  i16 Word11712;     //042; swapped
  i16 Word11714;     //044; swapped
  i16 MaxITEM16;     //046; swapped
  i8  FILL180[80];   //048;
};

struct ITEM16
{
  i16 word0;    // monster (DB4) index //swapped when read
private:
  ui8 facing_2; // Four 2-bit fields.
                // Each member of group is in 2 bits starting from bottom.
  ui8 pos_3;    // Four 2-bit fields
                // The position within the room of each
                // member of group.  If the monster occupies 2
                // corners of the room then its position
                // is either 0 or 2.
public:
  ui8 uByte4;   // Least significant byte of d.Time
                // set to d.Time-127
  ui8 uByte5;   // Direction facing??? Giggler got this set to 
                // Random(64) + 20.  Time to pause????
  ui8 uByte6;   // mapX square being approached
  ui8 uByte7;   // mapY square being approached
  ui8 uByte8;   // compared to new mapX
  ui8 uByte9;   // compared to new mapY
  ui8 uByte10;  // mapX
  ui8 uByte11;  // mapY
  ui8 uByte12[4];// Treated as array by TAG00ecca(ProcessTimer25)
  inline ui8& positions(void) {return pos_3;};
  inline ui8& facings(void) {return facing_2;};
};

class ATTRIBUTE // 3-byte structure in character
{
private:
  // Let me tell you why I hid these variables.....
  // From what I can tell, these are treated as unsigned
  // bytes.  But at least one of them can go negative.
  // The Luck-Minimum starts at 10 and is decreased by
  // three for each cursed object.  Gather four such
  // objects and you have set the minimum to a very
  // large number.  So I want to handle them exactly the
  // same except that when you ask for a number and it 
  // has gone negative, I will return a zero.  But its
  // value will export to the Atari perfectly.
  ui8 ubMaximum;
  ui8 ubCurrent;
  ui8 ubMinimum;
public:
  ui8 Minimum(void);
  void Minimum(i32);
  void IncrMinimum(i32 increment); // Minus to decrement
  ui8 Maximum(void);
  void Maximum(i32);
  void IncrMaximum(i32 increment); // Minus to decrement
  ui8 Current(void);
  void Current(i32);
  void IncrCurrent(i32 increment); // Minus to decrement
};

struct SKILL // 6-byte structure in character
{
  i16 word0;
  i32 Long2;  // Experience in this skill
};

enum skillNames {
  SKILL_Fighter  = 0,
  SKILL_Ninja    = 1,
  SKILL_Priest   = 2,
  SKILL_Wizard   = 3
};

struct CHARDESC // character info???
{ i8  name[8];  // 00
  i8  title[16];// 08 // size??
  i16 wordx24; // Not swapped because I don't see it used.
  ui8 FILL26[28-26];
  ui8 facing;  //28
  ui8 position; //29
  ui8 byte30;
  ui8 byte31;
   i8 attackType; //32Signed so check for -1 works.
   i8 byte33;
   i8 incantation[4];//34; 
        // [0] = power 96 through 111 --> 1 through 6
        // [1] =      102 through 107 
        // [2] =      108 through 113
        // [3] =      114 through 119
  ui8 FILL38[40-38];
  ui8 facing3; //40
  ui8 uByte41;
  ui8 uByte42; //Poison count.  A timer for each one??
  ui8 uByte43;
  i16 busyTimer;// 44 // Timer Index. Signed so -1 works.
  i16 timerIndex;//46;
  i16 word48; // ORed with 0x280
              // 0x8000 while selecting attack option
  i16 ouches;     //50;// mask of damaged body parts.
  i16 HP;         //52
  i16 maxHP;      //54
  i16 stamina;    //56;
  i16 maxStamina; //58;
  i16 mana;       //60;
  i16 maxMana;    //62;
  i16 word64;
  i16 food;       //66;
  i16 water;      //68;
  ATTRIBUTE Attributes[7]; // 70 (maximum, current, minimum)
                           // 70 = [0] Luck??
                           // 73 = [1] Strength
                           // 76 = [2] Dexterity
                           // 79 = [3] Wisdom    
                           // 82 = [4] Vitality  
                           // 85 = [5] AntiMagic 
                           // 88 = [6] AntiFire  
  i8  FILL91;
  SKILL skills92[20]; // 92 //0 and 4-7  =Fighter
                            //1 and 8-11 =Ninja
                            //2 and 12-15=Priest
                            //3 and 16-19=Wizard
  RN  possessions[30]; //212
  ui16 load;//272; In 10ths of KG
  ui16 shieldStrength; //274;
  ui8 FILL276[336-276];
  i8  portrait[464]; //336
  //ui8 FILL337[800-337];
};

struct LEVELDESC
{ // Describes a level.  We read them from
  // DUNGEON.DAT
  friend class CCSBuildDoc;
private:
  ui16 m_cellOffset;// littleEndian when read
                    //Offset of first cell in CELLFLAG array
                    //This is only the bottom 16 bits.
  i8  FILL2[4];
  ui8 offsetX; //6;
  ui8 offsetY; //7;
  i16 word8;  // bits  0-5  = level number
              // bits  6-10 = # columns -1
              // bits 11-15 = # rows - 1
  i16 word10; // bits 0-3 = #wall decorations used on level
              // 4-7,  
              // 8-11,    = #of floor decorations
              // 12-15

  i16 word12; // bits 0-3,  = number of Door decorations
              // bits 4-7   = number monsters types
              // bits 12-15 = experience multiplier
  i16 word14; // bits 0-3
              // bits 4-7
              // bits 8-11, //First door type this level
              // bits 12-15 //Second door type this level
  int CellOffset(void){return m_cellOffset;};
  void CellOffset(int n){MODIFY m_cellOffset = (ui16)n;};
  int OffsetX(void){return offsetX;};
  void OffsetX(int n)
  {
    if (OffsetX() == n) return;
    MODIFY 
    offsetX=(ui8)n;
  };
  int OffsetY(void){return offsetY;};
  void OffsetY(int n)
  { 
    if (OffsetY() == n) return;
    MODIFY 
    offsetY=(ui8)n;
  };
  void swap(void);
  ui8 LastColumn(void) { return BITS6_10(word8);};
  void LastColumn(int c) {MODIFY SETWBITS6_10(word8, c);};
  ui8 LastRow(void) { return BITS11_15(word8);};
  void LastRow(int r) {MODIFY SETWBITS11_15(word8, r);};
  ui8 Level(void) { return BITS0_5(word8);};
  void Level(int n) {MODIFY SETWBITS0_5(word8,n);};
  ui8 NumWallDecoration(void) {return BITS0_3(word10);};
  void NumWallDecoration(int n) {MODIFY SETWBITS0_3(word10,n);};
  ui8 NumRandomWallDecorations(void) {return BITS4_7(word10);};
  void NumRandomWallDecorations(int n) {MODIFY SETWBITS4_7(word10,n);};
  ui8 NumFloorDecorations(void) {return BITS8_11(word10);};
  void NumFloorDecorations(int n) {MODIFY SETWBITS8_11(word10, n);};
  ui8 NumRandomFloorDecorations(void) {return BITS12_15(word10);};
  void NumRandomFloorDecorations(int n) {MODIFY SETWBITS12_15(word10,n);};
  ui8 NumDoorDecorations(void) {return BITS0_3(word12);};
  void NumDoorDecorations(int n) {MODIFY SETWBITS0_3(word12,n);};
  ui8 numMonsterType(void) {return BITS4_7(word12);};
  void numMonsterType(int n) {MODIFY SETWBITS4_7(word12,n);};
         //The monster types are listed just after the cellflags.
         //This is also the offset of the wall decoration
         //table in the information following the cellflags.
  ui8 experienceMultiplier(void) {return BITS12_15(word12);};
  void experienceMultiplier(int n) {MODIFY SETWBITS12_15(word12,n);};
  ui8 backgroundGraphicSet(void) {return BITS0_3(word14);};
  void backgroundGraphicSet(int g) {MODIFY SETWBITS0_3(word14,g);};
  ui8 wallGraphicSet(void) {return BITS4_7(word14);};
  void wallGraphicSet(int g) {MODIFY SETWBITS4_7(word14,g);};
  ui8 Door1Type(void) {return BITS8_11(word14);};
  void Door1Type(int d) 
  {
    if (Door1Type() == d) return;
    MODIFY 
    SETWBITS8_11(word14,d);
  };
  ui8 Door2Type(void){return BITS12_15(word14);};
  void Door2Type(int d) 
  {
    if (Door2Type() == d) return;
    MODIFY 
    SETWBITS12_15(word14,d);
  };
};

struct DUNGEONDATINDEX
{
private:
  ui16  sentinel;
  ui16  word2;
  ui16  word4;
  ui16  word6;
  ui16  word8;
  ui16  word10;
  ui16  dbSizes[16];
public:
  ui8  NumLevel(void) {return BITS8_15(word4);};
  void NumLevel(int n) {MODIFY SETWBITS8_15(word4,n);};
  //ui16 Sentinel(void) {return sentinel;};
  void Sentinel(ui16 s){sentinel=s;};
  ui16 DBSize(int dbNum) {return dbSizes[dbNum];};
  void DBSize(int dbNum, ui16 size) {MODIFY dbSizes[dbNum] = size;};
  ui16 LegacyCellFlagArraySize(void) {return word2;};
  //This has been moved into ExtendedFeaturesBlock because
  //16 bits are not sufficient for 64 level dungeons.
  //The cellflag array can be about 68000 bytes when you
  //include the decoration lists.
  void LegacyCellFlagArraySize(int n) {MODIFY word2 = (ui16)n;};
  ui16 NumWordsInTextArray(void) {return word6;};
  void NumWordsInTextArray(int n){MODIFY word6 = (ui16)n;};
  ui8  StartingPartyX(void) {return BITS0_4(word8);};
  void  StartingPartyX(int x) {MODIFY SETWBITS0_4(word8,x);};
  ui8  StartingPartyY(void) {return BITS5_9(word8);};
  void  StartingPartyY(int y) {MODIFY SETWBITS5_9(word8,y);};
  ui8  StartingPartyFacing(void) {return BITS10_11(word8);};
  void  StartingPartyFacing(int f) {MODIFY SETWBITS10_11(word8,f);};
  ui16 ObjectListLength(void) {return word10;};
  void ObjectListLength(ui16 v) {MODIFY word10 = v;};
  void Swap(void);
};


class DBCOMMON
{
  friend class DB0;
  friend class DB1;
  friend class DB2;
  friend class DB3;
  friend class DB4;
  friend class DB5;
  friend class DB6;
  friend class DB7;
  friend class DB8;
  friend class DB9;
  friend class DB10;
  friend class DB11;
  friend class DB14;
  friend class DB15;
//public:
  RN   m_link;
  void swapLink(void);
public:
  const RN link(void) const;
  RN *pLink(void){return &m_link;};
  void link(RN rn);
  bool IsDBType(DBTYPE);
  class DB0 *CastToDB0(void);
  class DB1 *CastToDB1(void);
  class DB2 *CastToDB2(void);
  class DB3 *CastToDB3(void);
  class DB4 *CastToDB4(void);
  class DB5 *CastToDB5(void);
  class DB6 *CastToDB6(void);
  class DB7 *CastToDB7(void);
  class DB8 *CastToDB8(void);
  class DB9 *CastToDB9(void);
  class DB10 *CastToDB10(void);
  class DB11 *CastToDB11(void);
  class DB14 *CastToDB14(void);
  class DB15 *CastToDB15(void);
};

class DB0:public DBCOMMON
{ //Door
  friend class DATABASES;
  i16 m_word2; //bit 0 = door Type
               //bits 1-4 = ornateness type
               //bit 5 = Mode of operation (0=sideways, 1=up)
               //bit 6 = door switch
               //bit 7 = Can open with fireball
               //bit 8 = Can open with axe
  void swap(void);
public:
  DB0(void);
  bool operator != (DB0& db0); //Undefined!!!
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  DB0& operator = (DB0& db0);
  const ui8 doorType(void)   {return (ui8)(m_word2 & 1);}; 
  void  doorType(int t) {SETWBITS0_0(m_word2,t);};
    // Each level can have two door types.
  const i32 ornateness(void){return (m_word2>>1) & 0xf;};
  void ornateness(int o){SETWBITS1_4(m_word2,o);};
  const ui8 mode(void)   {return (ui8)((m_word2>>5) & 1);};
  void mode(int m){SETWBITS5_5(m_word2,m);};
  const bool canOpenWithFireball(void)    {return (m_word2>>7) & 1;};
  void canOpenWithFireball(BOOL b){SETWBITS7_7(m_word2,b);};
  const bool canOpenWithAxe(void) {return (m_word2>>8) & 1;};
  void canOpenWithAxe(BOOL b) {SETWBITS8_8(m_word2,b);};
  const ui8 doorSwitch(void) {return (ui8)((m_word2>>6) & 1);};
  void doorSwitch(BOOL b) {SETWBITS6_6(m_word2, b);};
  void Clear(void);
  CString Name(void);
  void Dump(FILE *f, int index, bool title=false);
  void Export(FILE *f, const LOCATIONREL& locr);
};



class DB1:public DBCOMMON
{ //Teleporter
  friend class DATABASES;
  friend class CCSBuildDoc;
  void swap(void);
private:
  i16 m_word2;  // bits 0-4   = mapX
                // bits 5-9   = mapY
                // bits 10-11 = new facing or rotation
                // bit 12     = 1 = absolute facing; 0=rotate relative
                // bits 13-14 = 0 = objects only
                //              1 = monsters only
                //              2 = party or objects only
                //              3 = anything
                // bit 15     = audible buzz
  i16 m_word4;  // bits 8-15  = level
public:
  DB1(void);
  bool operator != (DB1& db1);
  DB1& operator = (DB1& db1); //Undefined!!!
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  i32  destX(void) {return BITS0_4(m_word2);};
  i32  destY(void) {return BITS5_9(m_word2);};
  ui8  destLevel(void) {return BITS8_15(m_word4);};
  void SwapDestLevel(int L1, int L2);
  void destX(int x) {SETWBITS0_4(m_word2,x);};
  void destY(int y) {SETWBITS5_9(m_word2,y);};
  void destLevel(int l) {SETWBITS8_15(m_word4,l);};
  LOCATIONREL dest(void);
  ui8  rotation(void) {return BITS10_11(m_word2);};
  void rotation(int r){SETWBITS10_11(m_word2,r);};
  bool facingMode(void) {return BITS12_12(m_word2);};
  void facingMode(bool b){SETWBITS12_12(m_word2,b);};
  ui8  what(void) {return BITS13_14(m_word2);};
  void what(int w){SETWBITS13_14(m_word2,w);};
  bool audible(void) {return BITS15_15(m_word2);};
  void audible(bool b){SETWBITS15_15(m_word2,b);};
  void dest(LOCATIONREL locr);
  void Clear(void);
  CString Name(void);  
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const EXPORTLOC& eloc);
};

enum TEXTTYPE
{ //The flags used in ExpandText.
  TEXTTYPE_WALL = 0,         //Wall text
  TEXTTYPE_MIRROR = 0x8002,  //On floor adjacent to portrait
  TEXTTYPE_ENDGAME = 0x8001, //Hidden in nearby floor
  TEXTTYPE_SCROLL = 0x8002
};

class DB2:public DBCOMMON  // Text
{
  friend class DATABASES;
  i16 m_word2;  // bit 0 = ??
                // bits 3-15 = text (ui16 *) array index
                // 0x1fff = No text.
  void swap(void);
public:
  DB2(void);
  ui8 show(void) {return BITS0_0(m_word2);};
  void show(i32 val) {SETWBITS0_0(m_word2,val);};
  i32 index(void) {return BITS3_15(m_word2);};
  void index(int i){SETWBITS3_15(m_word2,i);};
  bool ValidText(void){return index() != 0x1fff;};
  void Clear(void);
  void Verify(LOCATIONREL locr, 
              CHKERRMSG *result, 
              int containerMAsk);
  DB2& operator =(DB2& db2); //Undefined!!!
  bool IsLegalPortratText(void);
  void Dump(FILE *f, int index, bool title=false);
  void Export(FILE *f, const LOCATIONREL& locr);
};

// Action field in DB3 record
enum ACTUATORACTION
{
  AA_SET     = 0,
  AA_CLEAR   = 1,
  AA_TOGGLE  = 2,
  AA_HOLD    = 3
};




enum ACTUATORCLASS
{
  AC_UNKNOWN = 1,
  AC_PRESSUREPAD,
  AC_DISABLED,
  AC_MONSTERGENERATOR,
  AC_PUSHBUTTON,
  AC_ANDOR,
  AC_COUNTER,
//  AC_NICHESWITCH,
  AC_ENDGAME,
  AC_LAUNCHER,
  AC_PORTRAIT,
  AC_OBJECTHOLDER,   //Torch Holder, eg.
  AC_SWAPOBJECT,     //Swap ZoKathRa for Firestaff, eg.
  AC_DSA
};

bool IsActuatorClassAllowed(ACTUATORCLASS ac, LOCATIONREL locr);

class ACTUATOR_PressurePad;
class ACTUATOR_MonsterGenerator;
class ACTUATOR_PushButton;
class ACTUATOR_Launcher;
class ACTUATOR_AndOr;
class ACTUATOR_SkillIncreaser;
class ACTUATOR_Rotate;
class ACTUATOR_Counter;
//class ACTUATOR_NicheSwitch;
class ACTUATOR_ObjectHolder;
class ACTUATOR_EndGame;
class ACTUATOR_SwapObject;
class ACTUATOR_Disabled;
class ACTUATOR_Portrait;
class ACTUATOR_DSA;

enum 
{
  //Bit flags in State
  PORTRAIT_InActive       = 1,
  PORTRAIT_HidePortrait   = 2,
  PORTRAIT_HideGraphic    = 4,
  // How to respond to messsages
  PORTRAIT_ResponseNone    = 0, //Do nothing
  PORTRAIT_ResponseSet     = 1, //Set Active or Show
  PORTRAIT_ResponseClear   = 2, //Set Inactive or Hide
  PORTRAIT_ResponseToggle  = 3, //Toggle Active or Show
  // Where to find 2-bit Response actions in 'responses' fields
  PORTRAIT_ActiveResponse   = 0,
  PORTRAIT_PortraitResponse = 2,
  PORTRAIT_GraphicResponse  = 4
};


class DB3:public DBCOMMON  // An actuator (pressure pad, keyhole, etc.)
{
  friend class DATABASES;
//  friend class ACTUATOR_PressurePad;
  friend class CCSBuildDoc;
  void swap(void);
private:
  i16 m_word2;// bits 0-6 = Actuator Type
              //    0 = Do nothing //but local action still works
              //    1 = A simple wall switch that
              //        can be pressed with anything.
              //    2 = ?
              //    3 = party??
              //    4 = In an open room
              //    5 = 
              //        Byte 8 (position) of Timer6 is bit number.
              //        Bits 7-15 are value.  Bits in this value
              //        are Set/Cleared/Toggled by Timer6.
              //        "position" of timer6 tells which bit to
              //        fiddle with.
              //        Actuator is 'pressed' when bits 7-10 are
              //        equal to bits 11-14.  (The first 4 bits of
              //        value equals second 4 bits).  
              //    6 = 
              //        In a closed (stone) room
              //           Bits 7-15 are a 9-bit counter.  A SET
              //           function increments and anything else
              //           decrements the counter.  You cannot change
              //           a counter that equals zero.  (Why?)
              //           The actuator is considered 'pressed'
              //           if the count is non-zero.
              //           NOTE: New counters are type 48.
              //        In an open room this is a monster generator
              //        and is actuated by a timer event type 5.
              //           word 6 bits 8-15 non-zero causes once only.
              //                            If >= 128 then a type 65 timer
              //                            is queued to go off at
              //                            64 * ((word 6 bits 8-15)-126)
              //                            (I presume to reactivate generator!).
              //    7 = ?
              //    8 = In open room:
              //        Operated only by party.  Value is
              //        an OBJECTTYPE that we search for in the
              //        party's possessions.  Actuated if party
              //        possesses such an object.
              //
              //        In Stone: A missile launcher.
              //          Value = 0  Fireballs
              //    9 = ?
              //   13 = In a wall (local action)
              //          Value is a MODIFIEDOBJECTTYPE.
              //          If the hand is empty and a object of the
              //          correct type is stored at the proper
              //          poition in the wall then the object is
              //          moved to the hand.  And vise versa.
              //          If an object is moved then:
              //           if (localAction)
              //           {
              //            switch (word6 bits4-15)
              //            {
              //            case 1:
              //            case 2:
              //              Rotate actuators;
              //              break;
              //            case 10:
              //              Increase skill
              //              break
              //            default:
              //              Do nothing
              //            };      
              //           }
              //           else
              //           {
              //            Always send message. 
              //            Ignore normally closed.
              //            Hold sends set if object placed
              //           };
              //   14 = 
              //        When a timer of type 6 is sent to 
              //        an actuator of type 14 then the first 
              //        object at the proper position in the same
              //        room as the actuator is removed 
              //        from the room and launched as a missile 
              //        in the adjacent room.
              //   15 = 
              //        When a timer of type 6 is sent to 
              //        an actuator of type 15 then the first two 
              //        objects at the proper position in the same
              //        room as the actuator are removed 
              //        from the room and launched as missiles 
              //        in the adjacent room.
              //   47 = Designer-specified Actuator,
              //   48 = Counter.  See type 6 (above).  We added
              //        this type so that counters will not share
              //        a type number with monster generators.
              //        You can put a counter anywhere.
              // bits 7-15 = value
private:
  i16 m_word4;// bit 2 = causes bits 1-6 of word 2 to be cleared. "Once only"???
              // bits 3-4 Action when switch changes state
              //           0=set when closed
              //           1=clear when closed
              //           2=toggle when closed
              //           3=set when closed, clear when opened
              // bit 5 = 0=normally open; 1=normally closed
              // bit 6 = make click sound
              // bits 7-10 = delta time (or number of monsters+1)
              //   For a monster generator, a value with bit 3
              //   set means to generate a random number modulo
              //   bits 0-2.
              // bit 11 = Local action only.  Like increase
              //          skill.
              // bits 12-15 = wall graphic index+1
private:
  i16 m_word6;// (if word4 & 0x800)
                 // bits 4-15 
              // (if ! word4 & 0x800)
                 // bits 4-5    position????
                 // bits 6_10   =mapX
                 // bits 11_15  =mapY
              // bits 4-7  hitPoint multiplier (Monster Generator)
              // bits 4-11 energy remaning in missile
              // bits 8_15 disable time (Monster Generator)
              // bits 12-15
private:
  i16 m_word8;//Always stored in big-endian format.
              // DSA bits 0_1 Parameter A position
              //     bits 2_3 Parameter B position
              //     bit 4 Parameter A is integer
              //     bit 5 Parameter B is integer
protected:
  DB3(void);
  CString StandardCheckTarget(LOCATIONREL locr, 
                              char t[16][32][32],
                              const char *name);
  bool StandardCheckTrigger(LOCATIONREL locr, char t[16][32][32]);
  i16  delay(void) {return (i16)BITS7_10(m_word4);};
  void delay(int d) {SETWBITS7_10(m_word4,d);};
  i16  targetX(void) {ASSERT((m_word4&0x800)==0);return (i16)BITS6_10(m_word6);};
  void targetX(int x){SETWBITS6_10(m_word6,x);};
  i16  targetY(void) {ASSERT((m_word4&0x800)==0);return (i16)BITS11_15(m_word6);};
  void targetY(int y){SETWBITS11_15(m_word6,y);};
  i16  targetPos(void) {ASSERT((m_word4&0x800)==0);return (i16)BITS4_5(m_word6);};
  void targetPos(int p){SETWBITS4_5(m_word6,p);};
  i16 w6_4_15(void) {ASSERT((m_word4&0x800)!=0);return (i16)BITS4_15(m_word6);};
  void w6_4_15(int i) {ASSERT((m_word4&0x800)!=0);SETWBITS4_15(m_word6,i);};
  bool audible(void) {return (i16)BITS6_6(m_word4)!=0;};
  void audible(bool b){SETWBITS6_6(m_word4,b);};
  ACTUATORACTION action(void) {return ACTUATORACTION(BITS3_4(m_word4));};
  void action(ACTUATORACTION a) {SETWBITS3_4(m_word4,a);};
  bool onceOnly(void) {return (i16)BITS2_2(m_word4)!=0;};
  void onceOnly(bool b){SETWBITS2_2(m_word4,b);};
  bool normallyClosed(void) {return BITS5_5(m_word4)!=0;};
  void normallyClosed(bool b){SETWBITS5_5(m_word4,b);};
  bool localAction(void) {return (actuatorType()!=47)&&(BITS11_11(m_word4)!=0);};
  void localAction(bool b){SETWBITS11_11(m_word4,b);};
  ui16 value(void) {return BITS7_15(m_word2);};
  void value(i32 val) {SETWBITS7_15(m_word2,val);};
  ui8  initialValue(void) {return BITS7_10(m_word2);};
  void initialValue(i32 val) {SETWBITS7_10(m_word2,val);};
  ui8  finalValue(void) {return BITS11_14(m_word2);};
  void finalValue(i32 val) {SETWBITS11_14(m_word2,val);};
  ui16 actuatorType(void) {return BITS0_6(m_word2);};
  void actuatorType(i32 type){SETWBITS0_6(m_word2,type);};
  void actuatorTypeOR(i32 bits){m_word2 |= bits;};
  i16  graphicOrdinal(void) {return (i16)BITS12_15(m_word4);};
  void graphicOrdinal(int w) {SETWBITS12_15(m_word4,w);};
  i32 disableTime(void) {return BITS8_15(m_word6);}; //Monster Generator
  void disableTime(int t) {SETWBITS8_15(m_word6,t);};
  i32 hitPointMultiplier(void) {return BITS4_7(m_word6);};//Monster Generator
           //If zero then level experience multiplier is used
  void hitPointMultiplier(int h) {SETWBITS4_7(m_word6,h);};
  i32 energyRemaining(void) {return BITS4_11(m_word6);};
  void energyRemaining(int e) {SETWBITS4_11(m_word6,e);};
  ui8 andOrEdge(void){return BITS0_2(m_word6);};
  void andOrEdge(int e){SETWBITS0_2(m_word6,e);};
  bool drawAsSize4(void){return BITS0_0(m_word6);};
  void drawAsSize4(bool b){SETWBITS0_0(m_word6,b?1:0);};
  ui8 DSAstate(void){return BITS12_15(m_word2);};
  void DSAstate(ui8 s){SETWBITS12_15(m_word2,s);};
  ui8 DSAselector(void){return BITS7_11(m_word2);};
  void DSAselector(ui8 s){SETWBITS7_11(m_word2,s);};
  ui32 ParameterA(void){return (ui16)m_word4+((m_word8&3)<<16);};
  void ParameterA(ui32 a){m_word4=(ui16)a;m_word8=(ui16)(m_word8&0xfffc|((a>>16)&0x03));};
  bool AisInteger(void){return (m_word8&0x10)!=0;};
  void AisInteger(bool b){m_word8=(ui16)(b?(m_word8|0x10):(m_word8&0xffef));};
  ui32 ParameterB(void){return (ui16)m_word6+((m_word8&0x0c)<<14);};
  void ParameterB(ui32 b){m_word6=(ui16)b;m_word8=(ui16)(m_word8&0xfff3|((b>>14)&0x0c));};
  bool BisInteger(void){return (m_word8&0x20)!=0;};
  void BisInteger(bool b){m_word8=(ui16)(b?(m_word8|0x20):(m_word8&0xffdf));};
public:
  void operator = (const ACTUATOR_PressurePad& p);
  void operator = (const ACTUATOR_Counter& p);
  void operator = (const ACTUATOR_AndOr& p);
  void operator = (const ACTUATOR_PushButton& p);
  void operator = (const ACTUATOR_ObjectHolder& p);
  void operator = (const ACTUATOR_Disabled& p);
  void operator = (const ACTUATOR_Launcher& p);
  void operator = (const ACTUATOR_MonsterGenerator& p);
  void operator = (const ACTUATOR_Portrait& p);
  bool operator != (DB3& db3);
  DB3& operator = (const DB3& db3); //Undefined!!!
  void Copy(const DB3& db3);
  void MakeBig(void);
  ACTUATORCLASS actuatorClass(ROOMTYPE rt);
//  ACTUATORCLASS actuatorClass(LOCATIONREL locr);
  ACTUATOR_PressurePad      *Cast2AC_PressurePad(ROOMTYPE rt);
  ACTUATOR_MonsterGenerator *Cast2AC_MonsterGenerator(ROOMTYPE rt);
  ACTUATOR_PushButton       *Cast2AC_PushButton(ROOMTYPE rt);
  ACTUATOR_Launcher         *Cast2AC_Launcher(ROOMTYPE rt);
  ACTUATOR_AndOr            *Cast2AC_AndOr(ROOMTYPE rt);
  ACTUATOR_SkillIncreaser   *Cast2AC_SkillIncreaser(ROOMTYPE rt);
  ACTUATOR_Rotate           *Cast2AC_Rotate(ROOMTYPE rt);
  ACTUATOR_Counter          *Cast2AC_Counter(ROOMTYPE rt);
  //ACTUATOR_NicheSwitch      *Cast2AC_NicheSwitch(ROOMTYPE rt);
  ACTUATOR_ObjectHolder     *Cast2AC_ObjectHolder(ROOMTYPE rt);
  ACTUATOR_EndGame          *Cast2AC_EndGame(ROOMTYPE rt);
  ACTUATOR_SwapObject       *Cast2AC_SwapObject(ROOMTYPE rt);
  ACTUATOR_Disabled         *Cast2AC_Disabled(ROOMTYPE rt);
  ACTUATOR_Portrait         *Cast2AC_Portrait(ROOMTYPE rt);
  ACTUATOR_DSA              *Cast2AC_DSA(ROOMTYPE rt);
  int Edit(LOCATIONREL locr, RN RNactuator); //Return IDOK if OK
  bool HasTarget(LOCATIONREL locr);
  LOCATIONREL GetNthTarget(LOCATIONREL locr, int n);
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  i16 GraphicOrdinal(LOCATIONREL locr);
  CString Name(ROOMTYPE rt);
  CString ShortName(ROOMTYPE rt);
  CString GraphicName(LOCATIONREL locr);
  CString StandardGraphicName(int graphicOrdinal, LOCATIONREL locr);
  MODIFIEDOBJECTTYPE TriggerObject(LOCATIONREL locr);
  ui16 BlockDisables(void){return BITS8_8(m_word4);};
  void BlockDisables(ui16 v){SETWBITS8_8(m_word4,v);};
  ui16 SetResponses(void){return BITS6_11(m_word6);};
  void SetResponses(ui16 v){SETWBITS6_11(m_word6,v);};
  ui16 ClearResponses(void){return BITS0_5(m_word6);};
  void ClearResponses(ui16 v){SETWBITS0_5(m_word6,v);};
  ui16 ToggleResponses(void){return BITS0_5(m_word4);};
  void ToggleResponses(ui16 v){SETWBITS0_5(m_word4,v);};
  ui16 DisableMessage(void){return BITS6_7(m_word4);};
  void DisableMessage(ui16 v){SETWBITS6_7(m_word4,v);};
  ui16 State(void){return BITS9_11(m_word4);};
  void State(ui16 v){SETWBITS9_11(m_word4,v);};
  ui16 Position(void){return BITS12_13(m_word6);};
  void Position(ui16 v){SETWBITS12_13(m_word6,v);};
  void ClearUnusedPor(void){SETWBITS14_15(m_word6,0);};
  void Dump(FILE *f, int index, bool title, ROOMTYPE rt);
  void Export(FILE *f, const LOCATIONREL& locr);
  bool OnceOnly(ROOMTYPE rt);
  void SwapTargetLevel(int L1, int L2, LOCATIONREL *pLocr);
};           



class DB4:public DBCOMMON // Monsters??
{
  friend class DATABASES;
  void swap(void);
  RN   possession2; // Swapped when read
               // //I think this is a list of possessions.
private:
  ui8  uByte4; // Index into d.MonsterDescriptor (ITEM26)

  ui8  uByte5; //ITEM16 index when the monster group is on 
               // the same level as the party.  Else it is
               // four 2-bit fields with positions of each
               // group member.
public:
  DB4(void);
  ui16  hitPoints6[4]; //Swapped when read
                // Hit Points for each monster in group.
  i16  word14;  // Swapped when read
                // Bits 0-3 = 6 = 
                //  We set these bits to a 5 when a giggler stole
                //  some of our articles.
                // Bits 5-6 = index into uByte12 of Item9664 (ITEM26)
                //            Number of monsters in group-1
                //            Comes from delay field of Timer5 entry
                //             when monster created
                // Bits 8-9 set random when created
                //  When not on same level as party then
                //   this is direction facing.
                // Bit 10 set to zero when created
                // Bit 11 makes monster invisible
                // Bit 12 draws monster as size 4
                // Bit 13 means the monster ID is unique.  This was installed
                //        because the designer wanted to track a special monster
                //        by its ID but when the monster was killed another 
                //        monster might inherit its ID.  We clear this flag
                //        when we generate a monster.
                // Bit 14 means poisoned.  Only DSAs use this.
  MONSTERTYPE monsterType(void) {return (MONSTERTYPE)(uByte4&0x1f);};
  void monsterType(MONSTERTYPE mt) {uByte4=(ui8)((mt.Integer()&0x1f)|(uByte4&0xe0));};
  i32 alternateGraphics(void){return (uByte4>>5)&0x7;};
  void alternateGraphics(i32 a){uByte4=(ui8)((uByte4&0x1f)|((a&0x7)<<5));};
  int numMonster(void){return BITS5_6(word14)+1;};
  void numMonster(int n){SETWBITS5_6(word14,n-1);};
  RN possession(void) {return possession2;};
  RN *pPossession(void) {return &possession2;};
  void possession(RN object) {possession2 = object;};
  int hitPoints(int index){return hitPoints6[index];};
  void hitPoints(int index, int h){hitPoints6[index]=(ui16)h;};
  ui8 groupIndex(void) {return uByte5;};
  void groupIndex(ui32 i){uByte5 = (ui8)i;};
  ui8 positions(void){return uByte5;};
  void positions(i32 i){uByte5=(ui8)i;};
  int fear(void){return BITS0_3(word14);};
  void fear(int f){SETWBITS0_3(word14,f);};
  void Clear(void);
  void w14b10(int n){SETWBITS10_10(word14,n);};
  bool Invisible(void){return BITS11_11(word14)!=0;};
  void Invisible(bool b){SETWBITS11_11(word14,b?1:0);};
  bool DrawAsSize4(void){return BITS12_12(word14)!=0;};
  void DrawAsSize4(bool b){SETWBITS12_12(word14,b?1:0);};
  bool Unique(void){return BITS13_13(word14)!=0;};
  void Unique(bool b){SETWBITS13_13(word14,b?1:0);};
  bool Poisoned(void){return BITS14_14(word14)!=0;};
  void Poisoned(bool b){SETWBITS14_14(word14,b?1:0);};
  ui8  facing(void){return BITS8_9(word14);};
  void facing(int n){SETWBITS8_9(word14,n);};
  bool operator !=(DB4& db4);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  int Edit(LOCATIONREL locr, RN obj);
  DB4& operator =(DB4& db4); //Undefinded!!!
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB5:public DBCOMMON
{
  friend class DATABASES;
  void swap(void);
  i16 word2; //<bits 0-6> = weapon Type
             //             + 23 = Object Index
             // bit 7 = ??? (important item???)
             // bit 8 = Cursed
             // bit 9 = Poisoned
             //<bits 10-13> ;charges remaining
             // bit 14 = Broken
             // bit 15 = ???
public:
  DB5(void);
  bool operator != (DB5& db5);
  WEAPONTYPE  weaponType(void){return WEAPONTYPE(BITS0_6(word2));};
  void weaponType(WEAPONTYPE w){SETWBITS0_6(word2,w.Integer());};
  i32 cursed(void) {return BITS8_8(word2);};
  void cursed(i32 c) {SETWBITS8_8(word2,c);};
  bool poisoned(void) {return BITS9_9(word2)!=0;};
  void poisoned(bool b) {SETWBITS9_9(word2,b?1:0);};
  ui8 charges(void){return BITS10_13(word2);};
  void charges(i32 n) {SETWBITS10_13(word2,n);};
  i32 broken(void) {return BITS14_14(word2);};
  void broken(bool b) {SETWBITS14_14(word2,b?1:0);};
  bool important(void){return BITS7_7(word2)!=0;};
  void important(bool i) {SETWBITS7_7(word2,i?1:0);};
  i32 b15(void) {return BITS15_15(word2);};
  void b15(i32 n) {SETWBITS15_15(word2, n);};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  int Edit(LOCATIONREL locr, RN RNweapon); // return IDOK if OK
  DB5& operator =(DB5& db5); //Undefined!!!
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB6:public DBCOMMON  //Sandals, robe, shield
{
  friend class DATABASES;
  void swap(void);
  i16 word2; //<bits 0-6>  = clothing type
             // bit 7 = ??? (important item???)
             // bit 8 = cursed
             //<bits 9-12> = charges remaining
             // bit 13 = broken
public:
  DB6(void);
  int Edit(LOCATIONREL locr, RN RNclothing); // return IDOK if OK
  CLOTHINGTYPE clothingType(void) {return BITS0_6(word2);};
  void clothingType(CLOTHINGTYPE c){SETWBITS0_6(word2,c.Integer());};
  bool important(void) {return (word2 & 0x80) != 0;};
  void important(bool b) {SETWBITS7_7(word2,b?1:0);};
  i32 charges(void) {return BITS9_12(word2);};
  void charges(i32 n) {SETWBITS9_12(word2, n);};
  i32 cursed(void) {return BITS8_8(word2);};
  void cursed(i32 c) {SETWBITS8_8(word2, c);};
  i32 broken(void) {return BITS13_13(word2);};
  void broken(i32 c) {SETWBITS13_13(word2,c);};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  DB6& operator =(DB6& db6); //Undefined!!!
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB7:public DBCOMMON
{ // Scroll?
  friend class DATABASES;
// old Scroll  i16 m_word2; //swapped when read
// old Scroll             // bits 0-9 = text index
// old Scroll             // bits 10-15 = ????
  void swap(void);
  RN m_RNtext;
  ui16 m_word4; // Bit 0 = open
                // Bits 1-15 unused
public:
  DB7(void);
  void Verify(LOCATIONREL locr, 
              CHKERRMSG *result, 
              int containerMask);
// old Scroll  i32 textIndex(void) {return BITS0_9(m_word2);};
// old Scroll  i32 b10_15(void) {return BITS10_15(m_word2);};
// old Scroll  void b10_15(i32 n) {SETWBITS10_15(m_word2,n);};
  RN text(void){return m_RNtext;};
  void text(RN text){m_RNtext = text;};
  RN  *pText(void){return &m_RNtext;};
  bool open(void){return BITS0_0(m_word4)!=0;};
  void open(bool o){SETWBITS0_0(m_word4,o?1:0);};
  void Clear(void);
  DB7& operator =(DB7& db7); //Undefined!!!
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB8:public DBCOMMON
{ // potion??
  friend class DATABASES;
  void swap(void);
  i16 m_word2;// <bits 0-7> Strength of potion
            // <bits 8-14> = potion type
            // bit 15 = ??? (important item??)
public:
  DB8(void);
  ui8 strength(void) {return BITS0_7(m_word2);};
  void strength(i32 s) {SETWBITS0_7(m_word2,s);};
  bool important(void) {return (m_word2&0x8000)!=0;};
  void important(bool b) {SETWBITS15_15(m_word2,b?1:0);};
  POTIONTYPE potionType(void) {return (POTIONTYPE)BITS8_14(m_word2);};
  void potionType(POTIONTYPE p){SETWBITS8_14(m_word2,p.Integer());};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  DB8& operator =(DB8& db8); //Undefined!!!
  int Edit(LOCATIONREL locr, RN RNpotion);
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB9:public DBCOMMON // A chest?? 
{
  friend class DATABASES;
  RN  m_contents;  
  i16 m_word4;    // <Bits 1-2> + 1 = object index
  ui8 m_uByte6[2];
  void swap(void);
public:
  DB9(void);
  ~DB9(void);
  int chestType(void) {return (CHESTTYPE)BITS1_2(m_word4);};
  const RN contents(void) {return m_contents;};
  RN *pContents(void){return &m_contents;};
  void contents(RN rn) {m_contents = rn;};
  const i32 word4(void) {return m_word4;};
  const ui32 uByte6(i32 n) {return m_uByte6[n];};
  bool operator != (DB9& db9);
  DB9& operator = (DB9& db9); //Undefined!!!
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB10:public DBCOMMON  //corn, shank, necklace, miscellaneous
{
  friend class DATABASES;
  void swap(void);
public:
  i16 m_word2; // For Water.Waterskin
             //     <Bits 0-6> = misc type
             // bit 7 = ??? (important item???)
             // bit 8 = cursed???
             //     Bits 14-15 = how full; 0=empty 3 = full
             //                = charges remaining
             //                = character index (for bones)
public:
  DB10(void);
  bool operator != (DB10& db10);
  DB10& operator =(DB10& db10); //Undefined!!!
  int Edit(LOCATIONREL locr, RN RNmisc); // return IDOK if OK
  MISCTYPE miscType(void) {return (MISCTYPE)BITS0_6(m_word2);};
  void miscType(MISCTYPE t) {SETWBITS0_6(m_word2, t.Integer());};
  bool important(void) {return BITS7_7(m_word2) != 0;};
  void important(bool b) {SETWBITS7_7(m_word2,b?1:0);};
  bool cursed(void) {return BITS8_8(m_word2);};
  void cursed(bool c) {SETWBITS8_8(m_word2,c?1:0);};
  ui8 value(void) {return BITS14_15(m_word2);};
  void value(i32 n) {SETWBITS14_15(m_word2, n);};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  void Dump(FILE *f, int index, bool title);
  void Export(FILE *f, const LOCATIONREL& locr);
};

class DB11:public DBCOMMON //Expool
{
  friend class EXPOOL;
  friend class DATABASES;
  ui16 size;
  ui32 d[63];
  void swap(void);
  void Size(int newSize){size=(ui16)newSize;};
};

class DB12:public DBCOMMON //Not used
{
};

class DB13:public DBCOMMON //Not used
{
};

class DB14:public DBCOMMON // Fireball.  Other missiles.  Thrown objects.
{
  friend class DATABASES;
  RN  m_obj2;            //2 missile object dagger, fireball, etc 
                       // 
  ui8 m_energyRemaining; //4
  ui8 m_energyRemaining2;//5
  ui16 m_timerIndex;     //6
  void swap(void);
public:
  DB14(void);
  RN  flyingObject(void) {return m_obj2;};
  void flyingObject(RN rn) {m_obj2 = rn;};
  RN *pFlyingObject(void){return &m_obj2;};
  ui8 energyRemaining(void) {return m_energyRemaining;};
  void energyRemaining(i32 e) {m_energyRemaining=(ui8)e;};
  ui8 energyRemaining2(void) {return m_energyRemaining2;};
  void energyRemaining2(i32 e) {m_energyRemaining2=(ui8)e;};
  ui16 timerIndex(void) {return m_timerIndex;};
  void timerIndex(i32 i) {m_timerIndex = (ui16)i;};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  DB14& operator =(DB14& db14); //Undefinded!!!
  void Dump(FILE *f, int index, bool title);
};

class DB15:public DBCOMMON
{ // Puff of smoke when monster expires??
  friend class DATABASES;
  i16 m_word2; // Bits 0-6 = type (50 is special to monsters???)
             // Bit  7 = ???
             // Bits 8-15 = value.  Like Damage.
  void swap(void);
public:           //   
  DB15(void);
  ui8 type(void) {return BITS0_6(m_word2);};
  void type(i32 t){SETWBITS0_6(m_word2,t);};
  i32 b7(void) {return BITS7_7(m_word2);};
  void b7(i32 n){SETWBITS7_7(m_word2,n);};
  i32 value(void) {return BITS8_15(m_word2);};
  void value(i32 n) {SETWBITS8_15(m_word2,n);};
  void Clear(void);
  void Verify(LOCATIONREL locr, CHKERRMSG *result);
  DB15& operator =(DB15& db15); //Undefined!!!
  void Dump(FILE *f, int index, bool title);
};

class DATABASES
{
  friend FindEmptyDBEntry(const i32,const DBTYPE,const bool);
  friend class DBCOMMON;
  friend class CCSBuildDoc;
  friend class EXPOOL;
  DBCOMMON *m_Address[16];
  i32       m_numEnt[16];
  i32       m_noEnlarge;//A bit for each DB.  Set when the
                        //player says NOT to enlarge the #entries.
  void swapDOOR(int dbNum);
  void swapTELEPORTER(int dbNum);
  void swapTEXT(int dbNum);
  void swapACTUATOR(int dbNum);
  void swapMONSTER(int dbNum);
  void swapWEAPON(int dbNum);
  void swapCLOTHING(int dbNum);
  void swapSCROLL(int dbNum);
  void swapPOTION(int dbNum);
  void swapCHEST(int dbNum);
  void swapMISC(int dbNum);
  void swapEXPOOL(int dbNum);
  void swapMISSILE(int dbNum);
  void swapCLOUD(int dbNum);
  RN FindEmptyDBEntry(DBTYPE dbType);
  void Allocate(i32 dbNum, i32 numEnt);
  void DeAllocate(i32 dbNum);
  RN Enlarge(int dbNum);//Returns RNnul if impossible.
  RN AllocateSegment(int segment, DBTYPE dbType);
  void swap (int dbNum);
  bool IsDBIndexOK(DBTYPE dbType, int dbIndex);;
  DBCOMMON *GetCommonAddress(RN obj);
  DBCOMMON *GetCommonAddress(DBTYPE dbType, int dbIndex);
  bool IsDBType(DBCOMMON *, DBTYPE);
  void Dump(FILE *f);
  void SetAddress(DBTYPE dbType, DBCOMMON *addr);
  void NumEntry(int dbNum, int numEntry);
public:
  DATABASES(void);
  ~DATABASES(void);
  i32 NumEntry(int dbNum);
  i32 NumUsedEntry(DBTYPE dbType);
  CString DBName(int dbNum);
  DB3 *GetNextActuator(int *prev);
};

enum DSACOMMAND
{
  DSACMD_MESSAGE          = 1, //deprecated. This has 16-bit target
  DSACMD_OVERRIDE         = 2,
  DSACMD_NOOP             = 3,
  DSACMD_COPYTELEPORTER   = 4, //deprecated. This has 16-bit target
  DSACMD_GOSUB            = 5,
  DSACMD_LOAD             = 6,
  DSACMD_FETCH            = 7,
  DSACMD_EQUAL            = 8,
  DSACMD_QUESTION         = 9,
  DSACMD_RANDOM           =10,
  DSACMD_AMPERSAND        =11,
  DSACMD_JUMP             =12,
  DSACMD_STORE            =13,
  DSACMD_COPYTELEPORTER32 =14, //Absolute location is 32-bits
  DSACMD_MESSAGE32        =15, //Absolute location is 32-bits
  DSACMD_CASE             =16,
  DSACMD_VARIABLEFETCH    =17,
  DSACMD_VARIABLESTORE    =18,
  DSACMD_GLOBALFETCH      =19,
  DSACMD_GLOBALSTORE      =20,
};

enum MESSAGETYPE
{
  MSG_NULL   =  0,
  MSG_SET    =  1,
  MSG_CLEAR  =  2,
  MSG_TOGGLE =  3
};

enum TARGETTYPE
{
  TARGET_A    = 0,
  TARGET_B    = 1,
  TARGET_ABS  = 2,
  TARGET_GEAR = 3  //Target on stack
};

struct DSAgeneralCmd
{
private:
  unsigned int cmdCode:6; //DSACOMMAND
  int          unused:10;
public:
  DSACOMMAND CmdCode(void){return (DSACOMMAND)cmdCode;};
};

struct DSAmessageCmd
{
private:
  unsigned int cmdCode  :6; //DSACOMMAND
  unsigned int msgType  :2; //MESSAGETYPE
  unsigned int delay    :2; //0=0/1=X/2=Y/3=explicit
  unsigned int target   :2; //TARGETTYPE
  signed int   nextState:4; //relative to current state
  enum
  {
    MAXSTATE     = 8,
    MAXDELAY     = 0,
    COMMANDCODE  = DSACMD_MESSAGE
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  unsigned int Delay(void){return delay;};
  void Delay(unsigned int d){delay=d;};
  MESSAGETYPE MsgType(void){return (MESSAGETYPE)msgType;};
  void MsgType(MESSAGETYPE m){msgType=m;};
  TARGETTYPE Target(void){return (TARGETTYPE)target;};
  void Target(TARGETTYPE t){target=t;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAmessageCmd32
{
private:
  unsigned int cmdCode  :6; //DSACOMMAND
  unsigned int msgType  :2; //MESSAGETYPE
  unsigned int delay    :2; //0=0/1=X/2=Y/3=explicit
  unsigned int target   :2; //TARGETTYPE
  signed int   nextState:4; //relative to current state
  enum
  {
    MAXSTATE     = 8,
    MAXDELAY     = 0,
    COMMANDCODE  = DSACMD_MESSAGE32
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  unsigned int Delay(void){return delay;};
  void Delay(unsigned int d){delay=d;};
  MESSAGETYPE MsgType(void){return (MESSAGETYPE)msgType;};
  void MsgType(MESSAGETYPE m){msgType=m;};
  TARGETTYPE Target(void){return (TARGETTYPE)target;};
  void Target(TARGETTYPE t){target=t;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAcopyTeleporterCmd
{
private:
  unsigned int cmdCode    :6; //DSACOMAND
  unsigned int targetfrom :2; //TARGETTYPE
  unsigned int targetto   :2; //TARGETTYPE
  signed   int nextState  :6; //relative to curent state
  enum
  {
    MAXSTATE      = 32,
    COMMANDCODE   = DSACMD_COPYTELEPORTER
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  TARGETTYPE TargetFrom(void){return (TARGETTYPE)targetfrom;};
  void TargetFrom(TARGETTYPE t){targetfrom=t;};
  TARGETTYPE TargetTo(void){return (TARGETTYPE)targetto;};
  void TargetTo(TARGETTYPE t){targetto=t;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAcopyTeleporterCmd32
{
private:
  unsigned int cmdCode    :6; //DSACOMAND
  unsigned int targetfrom :2; //TARGETTYPE
  unsigned int targetto   :2; //TARGETTYPE
  signed   int nextState  :6; //relative to curent state
  enum
  {
    MAXSTATE      = 32,
    COMMANDCODE   = DSACMD_COPYTELEPORTER32
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  TARGETTYPE TargetFrom(void){return (TARGETTYPE)targetfrom;};
  void TargetFrom(TARGETTYPE t){targetfrom=t;};
  TARGETTYPE TargetTo(void){return (TARGETTYPE)targetto;};
  void TargetTo(TARGETTYPE t){targetto=t;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

enum LOADTYPE
{
  LOAD_INTEGER32 = 30,
  LOAD_ABS32     = 29,
  LOAD_$         = 28,
  LOAD_ABS16     = 27,
  LOAD_INTEGER   = 26
};

struct DSAloadCmd
{
private:
  unsigned int cmdCode  :6; //DSACOMMAND
  unsigned int where    :5; //LOADTYPE
  signed int   nextState:5;
enum
{
  MAXSTATE      = 16,
  COMMANDCODE   = DSACMD_LOAD
};
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Where(void){return where;};
  void Where(unsigned int w){where=w;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAstoreCmd
{
private:
  unsigned int cmdCode  :6; //DSACOMMAND
  unsigned int where    :5; //0=A...25=Z
  signed int   nextState:5;
enum
{
  MAXSTATE      = 16,
  COMMANDCODE   = DSACMD_STORE
};
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Where(void){return where;};
  void Where(unsigned int w){where=w;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAoverrideCmd
{
private:
  unsigned int cmdCode  :6; //DSACOMMAND
  unsigned int what     :3;
  unsigned int value    :3; // means use next word
  signed int   nextState:4;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_OVERRIDE
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  unsigned int What(void){return what;};
  void What(unsigned int w){what=w;};
  unsigned int Value(void){return value;};
  void Value(unsigned int v){value=v;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAnoopCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  signed int   nextState:10;
  enum
  {
    MAXSTATE    = 512,
    COMMANDCODE = DSACMD_NOOP
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSArandomCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int value    :5;  //0 to 27;
                             //28=next word;29=A;30=B;31=*;
  signed int   nextState:5;
  enum
  {
    MAXSTATE    = 16,
    COMMANDCODE = DSACMD_RANDOM
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Value(void){return value;};
  void Value(unsigned int v){value=v;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAequalCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  signed int   nextState:10;
  enum
  {
    MAXSTATE    = 512,
    COMMANDCODE = DSACMD_EQUAL
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

enum STKOP
{
    STKOP_Noop                        =  0,
    STKOP_Plus                        =  1,
    STKOP_Roll                        =  2,
    STKOP_Pick                        =  3,
    STKOP_Neg                         =  4,
    STKOP_Equal                       =  5,
    STKOP_Poke                        =  6,
    STKOP_Drop                        =  7,
    STKOP_Type                        =  8,
    STKOP_Del                         =  9,
    STKOP_Add                         = 10,
    STKOP_Dup                         = 11,
    STKOP_1Minus                      = 12,
    STKOP_Swap                        = 13,
    STKOP_1Plus                       = 14,
    STKOP_Over                        = 15,
    STKOP_MinusRoll                   = 16,
    STKOP_Rot                         = 17,
    STKOP_MinusRot                    = 18,
    STKOP_And                         = 19,
    STKOP_Shift                       = 20,
    STKOP_Comp                        = 21,
    STKOP_Or                          = 22,
    STKOP_2Dup                        = 23,
    STKOP_Slash                       = 24,
    STKOP_Percent                     = 25,
    STKOP_SetNewState                 = 26,
    STKOP_Less                        = 27,
    STKOP_2Drop                       = 28,
    STKOP_NotEqual                    = 29,
    STKOP_2Pick                       = 30,
    STKOP_RShift                      = 31,
    STKOP_Not                         = 32,
    STKOP_FalsePit                    = 33,
    STKOP_GeneratorDelayStore         = 34,
    STKOP_Overlay                     = 35,
    STKOP_DisableSaves                = 36,
    STKOP_Gear                        = 37,
    STKOP_Fetch                       = 38,
    STKOP_Store                       = 39,
    STKOP_ParamFetch                  = 40,
    STKOP_ParamStore                  = 41,
    STKOP_Message                     = 42,
    STKOP_Cast                        = 43,
    STKOP_FetchExCellFlg              = 44,
    STKOP_StoreExCellFlg              = 45,
    STKOP_ChPoss                      = 46,
    STKOP_Say                         = 47,
    STKOP_Loc2AbsCoord                = 48,
    STKOP_Mastery                     = 49,
    STKOP_ULess                       = 50,
    STKOP_GetCurse                    = 51,
    STKOP_SetCurse                    = 52,
    STKOP_GetCharges                  = 53,
    STKOP_SetCharges                  = 54,
    STKOP_GetBroken                   = 55,
    STKOP_SetBroken                   = 56,
    STKOP_CellFetch                   = 57,
    STKOP_CellStore                   = 58,
    STKOP_GlobalFetch                 = 59,
    STKOP_MonsterFetch                = 60,
    STKOP_Throw                       = 61,
    STKOP_TeleportParty               = 62,
    STKOP_MonsterStore                = 63,
    STKOP_PartyFetch                  = 64,
    STKOP_CharFetch                   = 65,
    STKOP_CharStore                   = 66,
    STKOP_Random                      = 67,
    STKOP_CreateCloud                 = 68,
    STKOP_Sound                       = 69,
    STKOP_Xor                         = 70,
    STKOP_MonBlk                      = 71,
    STKOP_Describe                    = 72,
    STKOP_Move                        = 73,
    STKOP_GetPoisoned                 = 74,
    STKOP_SetPoisoned                 = 75,
    STKOP_Copy                        = 76,
    STKOP_MonPoss                     = 77,
    STKOP_Palette                     = 78,
    STKOP_I_Del                       = 79,
    STKOP_I_Add                       = 80,
    STKOP_I_CreateCloud               = 81,
    STKOP_I_Cast                      = 82,
    STKOP_I_TeleportParty             = 83,
    STKOP_I_MonsterStore              = 84,
    STKOP_I_CharStore                 = 85,
    STKOP_I_Move                      = 86,
    STKOP_I_Copy                      = 87,
    STKOP_I_CellStore                 = 88,
    STKOP_I_Throw                     = 89,
    STKOP_I_Indirect                  = 90,
    STKOP_I_Delay                     = 91,
    STKOP_SetAdjustSkillsParameters   = 92,
    STKOP_I_DelMon                    = 93,
    STKOP_I_InsMon                    = 94,
    STKOP_DelMon                      = 95,
    STKOP_InsMon                      = 96,
    STKOP_TimeFetch                   = 97,
    STKOP_JumpGear                    = 98,
    STKOP_GosubGear                   = 99,
    STKOP_GeneratorDelayFetch         = 100,
    STKOP_MissileInfoFetch            = 101,
    STKOP_MissileInfoStore            = 102,
    STKOP_TextFetch                   = 103,
    STKOP_TextSay                     = 104,
    STKOP_MonLandD                    = 105,
    STKOP_CountInjury                 = 106,
    STKOP_Neighbors                   = 107,
    STKOP_BitCount                    = 108,
    STKOP_ThisCell                    = 109,
    STKOP_CausePoison                 = 110,
    STKOP_I_CausePoison               = 111,
    STKOP_MultiplierFetch             = 112,
    STKOP_ExperiencePlus              = 113,
    STKOP_IsCarried                   = 114,
    STKOP_DiscardText                 = 115,
    STKOP_WhereIsChar                 = 116,
    STKOP_WhoHasTalent                = 117,
    STKOP_SwapCharacter               = 118,
    STKOP_FilteredCast                = 119,
    STKOP_I_FilteredCast              = 120,
    STKOP_GlobalTextStore             = 121,
    STKOP_CharNameFetch               = 122,
    STKOP_ObjectID                    = 123,

};


struct DSAampersandCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int subCode  :7;
  signed int   nextState:3;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_AMPERSAND
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  STKOP SubCode(void){return (STKOP)subCode;};
  void SubCode(STKOP n){subCode=n;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAvariableFetchCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int index    :7;  // 0-99 the variable to fetch
  signed int   nextState:3;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_VARIABLEFETCH
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Index(void){return index;};
  void Index(unsigned int n){index=n;};
  void CmdCode(DSACOMMAND cmd){cmdCode=cmd;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAvariableStoreCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int index    :7;  // 0-99 the variable to store
  signed int   nextState:3;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_VARIABLESTORE
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Index(void){return index;};
  void Index(unsigned int n){index=n;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAglobalFetchCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int index    :7;  // 0-99 the variable to fetch
  signed int   nextState:3;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_GLOBALFETCH
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Index(void){return index;};
  void Index(unsigned int n){index=n;};
  void CmdCode(DSACOMMAND cmd){cmdCode=cmd;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAglobalStoreCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  unsigned int index    :7;  // 0-99 the variable to store
  signed int   nextState:3;
  enum
  {
    MAXSTATE    = 8,
    COMMANDCODE = DSACMD_GLOBALSTORE
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  unsigned int Index(void){return index;};
  void Index(unsigned int n){index=n;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAfetchCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  signed int   nextState:10;
  enum
  {
    MAXSTATE    = 512,
    COMMANDCODE = DSACMD_FETCH
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAcaseCmd
{
private:
  unsigned int cmdCode  :6;  //DSACOMMAND
  signed int   nextState:10;
  enum
  {
    MAXSTATE    = 512,
    COMMANDCODE = DSACMD_CASE
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAgosubCmd
{
private:
  unsigned int cmdCode  :6;
  unsigned int row      :6;
  unsigned int column   :1;
  signed   int nextState:3;
  enum
  {
    MAXSTATE    = 4,
    COMMANDCODE = DSACMD_GOSUB
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  int Row(void){return row;};
  void Row(int r){row=r;};
  int Column(void){return column;};
  void Column(int c){column=c;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAjumpCmd
{
private:
  unsigned int cmdCode  :6;
  unsigned int row      :6;
  unsigned int column   :1;
  signed   int nextState:3;
  enum
  {
    MAXSTATE    = 4,
    COMMANDCODE = DSACMD_JUMP
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  int Row(void){return row;};
  void Row(int r){row=r;};
  int Column(void){return column;};
  void Column(int c){column=c;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

struct DSAquestionCmd
{
private:
  unsigned int cmdCode   :6;
  signed   int nextState :4;
  unsigned int ifColumn  :1;
  unsigned int ifCMD     :2; //0=None,1=Jump,2=GoSub
  unsigned int elseColumn:1;
  unsigned int elseCMD   :2; //0=None,1=Jump,2=GoSub
  enum
  {
    MAXSTATE    = 2,
    COMMANDCODE = DSACMD_QUESTION
  };
public:
  int NextState(void){return nextState;};
  void NextState(int n){nextState=n;};
  bool IfColumn(void){return ifColumn!=0;};
  void IfColumn(bool b){ifColumn=b?1:0;};
  int IfCmd(void){return ifCMD;};
  void IfCmd(int n){ifCMD=n;};
  bool ElseColumn(void){return elseColumn!=0;};
  void ElseColumn(bool b){elseColumn=b?1:0;};
  int ElseCmd(void){return elseCMD;};
  void ElseCmd(int n){elseCMD=n;};
  bool IsMaxState(void){return nextState==-MAXSTATE;};
  bool StateFits(int r){return abs(r)<MAXSTATE;};
  void SetMaxState(void){nextState=MAXSTATE;};
  DSACOMMAND CmdCode(void){return (DSACOMMAND)(cmdCode=COMMANDCODE);};
};

union DSAcommand
{
  DSAmessageCmd          msgCmd;
  DSAgeneralCmd          generalCmd;
  DSAoverrideCmd         overrideCmd;
  DSAnoopCmd             noopCmd;
  DSAcopyTeleporterCmd   copyTeleporterCmd;
  DSAgosubCmd            gosubCmd;
  DSAfetchCmd            fetchCmd;
  DSAequalCmd            equalCmd;
  DSArandomCmd           reandomCmd;
  DSAampersandCmd        ampersandCmd;
  DSAjumpCmd             jumpCmd;
  DSAvariableFetchCmd    variableFetchCmd;
  DSAvariableStoreCmd    variableStroeDmd;
  DSAglobalFetchCmd      globalFetchCmd;
  DSAglobalStoreCmd      globalStoreCmd;
};

class DSAAction
{
  friend class DSAState;
  int         m_column;
  int         m_numPgmWords;
  ui16       *m_program;
  void compress(void);
  bool DecodeEqual (const char* &pC, int relstate);
  bool DecodeQuestion (const char* &pC, int relstate, int curstate);
  bool DecodeCase(const char*  &pC ,int relstate, int curstate);
  bool DecodeAmpersand(const char*  & pC,int relstate);
//  bool DecodeVariableFetch(const char*  & pC,int relstate);
//  bool DecodeVariableStore(const char*  & pC,int relstate);
  bool DecodeF (const char* &pC, int relstate);
  bool DecodeG (const char* &pC, int relstate, int curstate);
  bool DecodeJ (const char* &pC, int relstate, int curstate);
  bool DecodeM (const char* &pC, int relstate);
  bool DecodeL (const char* &pC, int relstate);
  bool DecodeO (const char* &pC, int relstate);
  bool DecodeN (const char* &pC, int relstate);
  bool DecodeR (const char* &pC, int relstate);
  bool DecodeS (const char* &pC, int relstate);
  bool DecodeV (const char* &pC, int relstate);
  bool DecodeGV(const char* &pC, int relstate);
  bool DecodeCT(const char* &pC, int relstate);
public:
  DSAAction(void);
  ~DSAAction(void);
  DSAAction& operator =(DSAAction&);
  bool operator != (DSAAction&);
  int Column(void){return m_column;};
  void Column(int column){m_column = column;};
  CString EncodeToASCII      (int curState, bool explicitState=false);
  CString EncodeEqual        (int &index, int curState);
  CString EncodeQuestion     (int &index, int curState, bool explicitState=false);
  CString EncodeCase         (int &index, int curstate, bool explicitState=false);
  CString EncodeAmpersand    (int &index, int curState);
  CString EncodeVariableFetch(int &index, int curState);
  CString EncodeVariableStore(int &index, int curState);
  CString EncodeGlobalFetch  (int &index, int curState);
  CString EncodeGlobalStore  (int &index, int curState);
  CString EncodeF            (int &index, int curState);
  CString EncodeG            (int &index, int curState, bool explicitState=false);
  CString EncodeJ            (int &index, int curState, bool explicitState=false);
  CString EncodeL            (int &index, int curState);
  CString EncodeM            (int &index, int curState);
  CString EncodeM32          (int &index, int curState);
  CString EncodeO            (int &index, int curState);
  CString EncodeN            (int &index, int curState);
  CString EncodeR            (int &index, int curState);
  CString EncodeS            (int &index, int curState);
  CString EncodeCT           (int &index, int curState);
  CString EncodeCT32         (int &index, int curState);
  int DecodeFromASCII(const CString& pgm, 
                      int curState);
  void Write(void);
  void Read(void);
  void Export(FILE *f, int state);
  bool Import(FILE *f, int state);
};

class DSAState
{
  int         m_numAction;
  DSAAction **m_pActions;
  void        AddAction(int column);
  void        DeleteAction(int column);
public:
  DSAState(void);
  ~DSAState(void);
  DSAState& operator =(DSAState&);
  bool operator !=(DSAState&);
  CString  Pgm(int msgIdx, int curState);
  int      Pgm(int msgIdx, const CString& pgm, int state);
  void compress(void);
  void Write(void);
  void Read(void);
  void Export(FILE *f, int state);
  bool Import(FILE *f, int state);
};

class DSA
{
  char       m_description[80];
  ui16       m_state;
  char       m_StateLocation;//0 state is in 4-bit field
                             //    in actuator record
                             //1 state is kept globally in
                             //    the DSA structure
                             //2 state is kept locally in
                             //    the actuators 16-bit
                             //    ParameterB field
                             //3 this is a 'SLAVE' - state
                             //    is kept in 'MASTER'
  ui8        m_groupID;
  int        m_numState;
  int        m_firstDisplayedState;
  DSAState **m_pStates;
  void       Expand(int state);
public:
  DSA(void);
  ~DSA(void);
  DSA& operator =(DSA&);
  bool operator !=(DSA&);
  CString Description(void){return CString(m_description);};
  void Description(CString& desc);
  ui8 GroupID(void){return m_groupID;};
  int FirstDisplayedState(void){return m_firstDisplayedState;};
  void FirstDisplayedState(int s){m_firstDisplayedState=s;};
  void GroupID(int g){m_groupID=(ui8)g;};
  int StateLocation(void){return m_StateLocation;};
  void StateLocation(int loc){m_StateLocation=(char)loc;};
  CString Pgm(int state, int msgIdx);
  int Pgm(int state, int msgIdx, const CString& pgm);
  int State(void){return m_state;};
  void State(int s){m_state = (ui16)s;};
  void compress(void); //return true if anything remains.
  void Write(void);
  void Read(void);
  void Export(void);
  void Import(void);
};

enum EXPOOL_DATA_TYPE
{
  EDT_MessageParameters = 1,
  EDT_ExtendedCellFlags = 2, //one 32-bit word per flag
       ECF_FOOTPRINTS = 0,
  EDT_SpecialLocations  = 3,
       ESL_ILLEGAL                = -1,
       ESL_MONSTERATTACKFILTER    = 0,
       ESL_FEEDINGFILTER          = 1,
       ESL_MONSTERMOVEFILTER      = 2,
       ESL_VIEWINGFILTER          = 3,
       ESL_PARTYATTACKFILTER      = 4,
       ESL_INDIRECTACTION         = 5,
       ESL_ADJUSTSKILLSFILTER     = 6,
       ESL_DELETEMONSTERFILTER    = 7,
       ESL_SOUNDFILTER            = 8,
       ESL_CURSORFILTER           = 9,
       ESL_ACTIONNAMEFILTER       = 10,
       ESL_CHARDEATHFILTER        = 11,
  EDT_Skins             = 4,
  EDT_Database          = 5,
       EDBT_ObjectWeights         = 0,
       EDBT_DeleteDuplicateTimers = 1,
       EDBT_RuntimeFileSignature  = 2,
       EDBT_Debuging              = 3,
       EDBT_GlobalVariables       = 4,
       EDBT_DisableSaves          = 5,
       EDBT_MonsterNames          = 6,
       EDBT_DSAtraces             = 7,
       EDBT_AltMonGraphics        = 8,
};


class EXPOOL
{
  friend class CCSBuildDoc;
  ui32 *m;
  void Setup(void);//After loading game.
  void enlarge(ui32 size);
public:
  EXPOOL(void);
  ~EXPOOL(void);
  void Write(ui32 key, ui32 *record, ui32 size);
  // Read() and Locate() return the size of the
  // record (or -1 if no record exists).
  i32 Read(ui32 key, ui32 *record, ui32 max); //Deletes it, too.
  i32 Locate(ui32 key, ui32 **record);
};

extern EXPOOL expool;

class PALETTE
{
  ui32 level;       //dungeon level to which this applies
  ui16 colors[16];
};

extern DATABASES db;
void FillComboBox(CComboBox *pCCombobox);
int  ObjectEditor(RN obj, LOCATIONREL locr);
int  EditOrphan(RN obj);